package datavisualizer;

import javax.swing.JTable;
import javax.swing.JPanel;
import java.awt.Color;
import javawrapper.ModelException;

/**
 * Produces a table of state variable results.  This requires the detailed
 * output file to have data saved from the "State Variables" grid.
 *
 * Copyright: Copyright (c) Charles D. Canham 2010
 * Company: Institute of Ecosystem Studies
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>January 10, 2010: Created (LEM)
 */

public class StateVariableTableDataRequest extends DataRequest {

  /**The precipitation for this timestep*/
  private float m_fPrecip;
  
  /**The temperature for this timestep*/
  private float m_fTemp;
    
  /**The code for the precipitation value.*/
  private int m_iPrecipCode;
  
  /**The code for the temperature value.*/
  private int m_iTempCode;
  
  /**Table column headers.  Putting them here makes them consistent in all
   * possible display methods.*/
  private String[] mp_sColumnNames = {"Temperature C", "Precipitation mm"};

  /**Name of the grid from which the data will be collected.*/
  private String m_sThisGrid;

  /**Number of columns in the table*/
  private int m_iNumCols = 2;

  /**
   * Constructor
   * @param sChartName Name of the chart
   * @param oManager Detailed output file manager
   * @throws ModelException not really, but I have to declare this.
   */
  public StateVariableTableDataRequest(String sChartName, DetailedOutputFileManager oManager) throws
      ModelException {
    super(sChartName, oManager);

    m_sThisGrid = "State Variables";

  }

  /**
   * Writes the table's data to tab-delimited text.
   * @param jOut java.io.FileWriter The file to write to.
   * @throws java.io.IOException if there's a problem writing the file.
   */
  protected void WriteChartDataToFile(java.io.FileWriter jOut) throws java.io.
      IOException {
      
      int iCol;

      //Write header row
      jOut.write(mp_sColumnNames[0]);
      for (iCol = 1; iCol < m_iNumCols; iCol++) {
        jOut.write("\t" + mp_sColumnNames[iCol]);
      }
      jOut.write("\n");

      //Write out data
      jOut.write(String.valueOf(m_fPrecip) + "\t" +
                   String.valueOf(m_fTemp) + "\n\n");       
    }

  /**
   * This accepts float data. 
   * @param sGridName Name of the grid for this data
   * @param iX X number of the cell from which this value came
   * @param iY Y number of the cell from which this value came
   * @param iCode Data member code of this value.
   * @param fVal Value.
   */
  public void AddGridFloatData(String sGridName, int iX, int iY, int iCode,
                               float fVal) {
    if (sGridName.equals(m_sThisGrid)) {
      if (iCode == m_iPrecipCode) {
        m_fPrecip = fVal;
      }
      else if (iCode == m_iTempCode) {
        m_fTemp = fVal;
      }        
    }
  }

  /**
   * This wants floats.
   * @return True.
   */
  public boolean WantAnyGridFloats() {
    return true;
  }

  /**
   * Accepts a grid float data member code for future reference when passed
   * float data members.
   * @param sGridName Name of the grid
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   */
  public void AddGridFloatDataMemberCode(String sGridName, String sLabel,
                                         int iCode) {
    if (sGridName.equals(m_sThisGrid)) {
      
      if (sLabel.equalsIgnoreCase("Temp.C")) {
      	m_iTempCode = iCode;
      }
      else if (sLabel.equalsIgnoreCase("Precip.mm")) {
      	m_iPrecipCode = iCode;
      }      
    }
  }

  /**
   * Creates a table of the results that have been collected.
   * @param oLegend The species legend.
   * @param sChartTitle The chart title.
   * @return A JInternalFrame with the table present in it.
   * @throws datavisualizer.ModelException
   */
  ModelInternalFrame DrawChart(Legend oLegend, String sChartTitle) throws
      ModelException {

    m_oChartFrame = new ModelInternalFrame(sChartTitle, oLegend);
    m_oChartFrame.setContentPane(CreateTable(oLegend));
    m_oChartFrame = DataGrapher.AddFileMenu(m_oChartFrame, this, m_bShowOneTimestep);
    return m_oChartFrame;
  }

  void UpdateChart(Legend oLegend) throws ModelException {
    m_oChartFrame.setContentPane(CreateTable(oLegend));
  }

  /**
   * Creates the results table.
   * @param oLegend Legend for this chart.
   * @return Panel containing the table.
   * @throws ModelException If anything goes wrong with the drawing.
   */
  private JPanel CreateTable(Legend oLegend) throws ModelException {
    //I need to be able to format a range of numbers
    java.text.NumberFormat jFormat = java.text.NumberFormat.getInstance();
    jFormat.setMaximumFractionDigits(3);
         
    //Create our table data array
    Object[][] p_oData = new Object[1][];
    p_oData[0] = new Object[m_iNumCols];

    p_oData[0][0] = jFormat.format(m_fPrecip);
    p_oData[0][1] = jFormat.format(m_fTemp);
        
    //Now our column names
    JTable jTable = new JTable(p_oData, mp_sColumnNames);
    jTable.setName("ResultsTable");
    jTable.setGridColor(Color.WHITE);
    jTable.setBackground(Color.WHITE);
    jTable.setFont(new uihelpers.SortieFont());
    jTable.getTableHeader().setFont(new uihelpers.SortieFont());

    //Create the panel and add the table to it.
    JPanel jPanel = new JPanel(new java.awt.BorderLayout());
    jPanel.add(jTable.getTableHeader(), java.awt.BorderLayout.NORTH);
    jPanel.add(jTable, java.awt.BorderLayout.CENTER);
    jPanel.setBackground(Color.WHITE);
    jPanel.add(jTable);
    return jPanel;
  }

  /**
   * Clears out existing data.
   * @throws ModelException Doesn't throw an exception.
   */
  void ClearData() throws ModelException {;}
}
