package datavisualizer;

import java.awt.Dimension;
import java.awt.FlowLayout;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;

import org.xml.sax.SAXException;

import uihelpers.SortieFont;
import uihelpers.SortieLabel;
import uihelpers.TextFileFilter;

import javawrapper.DataMember;
import javawrapper.DetailedTreeSettings;
import javawrapper.ErrorGUI;
import javawrapper.ModelException;
import javawrapper.ModelFileChooser;
import javawrapper.OutputBehaviors;
import javawrapper.TreePopulation;

/**
 * This class will write a complete tree list for any desired time step.
 * 
 * @author Lora E. Murphy
 * @version 1.0
 * 
 * Edit history: <br>
 * ----------------- <br>
 * March 24, 2011: Created (LEM) <br>
 * 
 */
public class TreeListWriter extends DataRequest {
  
  /** Field for output filename. */
  private JTextField m_jOutputFileName = new JTextField();
  /** Field for timestep to save. */
  private JTextField m_jTimestep = new JTextField();
  
  /**Text output file*/
  private java.io.FileWriter m_jOut;
  
  private String[] mp_sTypeNames = new String[TreePopulation.GetNumberOfTypes()];
  /** Make our own copy because the one in DetailedOutputFileManager omits
   * not dead*/
  private String[] mp_sDeadCodeNames = new String[OutputBehaviors.NUMCODES];
  
  /**
   * Column positions for tree integer data members. First index is species, 
   * second is type. Vector position equals the data member position in the map 
   * being read, and vector value is the position in the column of the file
   * being written.
   */
  private Vector<Vector<Vector<Integer>>> mp_iTreeIntTransforms,
  /**
   * Column positions for tree float data members. First index is species, 
   * second is type. Vector position equals the data member position in the map 
   * being read, and vector value is the position in the column of the file
   * being written.
   */
  mp_iTreeFloatTransforms,
  /**
   * Column positions for tree char data members. First index is species, 
   * second is type. Vector position equals the data member position in the map 
   * being read, and vector value is the position in the column of the file
   * being written.
   */
  mp_iTreeCharTransforms,
  /**
   * Column positions for tree bool data members. First index is species, 
   * second is type. Vector position equals the data member position in the map 
   * being read, and vector value is the position in the column of the file
   * being written.
   */
  mp_iTreeBoolTransforms;
  
  /**
   * The current timestep. This helps process an entire detailed output file. It
   * also lets this class know when to ignore offered data - whenever this value
   * is set to -1, some other chart event triggered the parse.
   */
  private int m_iCurrentTimestep = -1;

  /**
   * The columns in our output text file. The columns will depend on the data 
   * members available in the tree map of the output file. This is a unique list
   *  of the code names of all available tree data members.
   */
  private String[] mp_sColumnsByCode;

  /**
   * One tree's worth of data. One space for each column in the output text 
   * file, even if they will not all be used.
   */
  private String[] mp_sFileRow;
  
  private int m_iSpeciesIndex = 0;
  private int m_iTypeIndex = 1;
  private int m_iDeadCodeIndex = 2;
  private int m_iTreeSpecies;
  private int m_iTreeType;
  private int m_iTreeDeadCode;

  public TreeListWriter(String chartName, DetailedOutputFileManager manager) {
    super(chartName, manager);
    mp_sTypeNames[TreePopulation.SEED] = "Seed";
    mp_sTypeNames[TreePopulation.SEEDLING] = "Seedling";
    mp_sTypeNames[TreePopulation.SAPLING] = "Sapling";
    mp_sTypeNames[TreePopulation.ADULT] = "Adult";
    mp_sTypeNames[TreePopulation.STUMP] = "Stump";
    mp_sTypeNames[TreePopulation.SNAG] = "Snag";
    mp_sTypeNames[TreePopulation.WOODY_DEBRIS] = "Woody debris";
    
    mp_sDeadCodeNames[OutputBehaviors.NOTDEAD] = "Alive";
    for (int i = OutputBehaviors.NOTDEAD +1; i < mp_sDeadCodeNames.length; i++)
      mp_sDeadCodeNames[i] = m_oManager.mp_sDeadCodeNames[i];
  }

  /** Does nothing */
  void ClearData() throws ModelException {
    ;
  }

  /**
   * Respons to the button clicks for this class's chart window.
   * 
   * @param oEvent ActionEvent Event to process.
   */
  public void actionPerformed(java.awt.event.ActionEvent oEvent) {

    if (oEvent.getActionCommand().equals("Browse")) {
      ModelFileChooser jChooser = new ModelFileChooser();

      // If there's a value for the detailed output file, navigate to that
      // directory
      File oDir = new File(m_jOutputFileName.getText());
      if (oDir.getParentFile() != null && oDir.getParentFile().exists()) {
        jChooser.setCurrentDirectory(oDir.getParentFile());
      }

      jChooser.setSelectedFile(new File(oDir.getName()));
      jChooser.setFileFilter(new TextFileFilter());

      int returnVal = jChooser.showSaveDialog(m_oChartFrame);
      if (returnVal == JFileChooser.APPROVE_OPTION) {
        // User chose a file - does it already exist?
        File oFile = jChooser.getSelectedFile();

        // Does this file already exist?
        if (oFile.exists()) {
          returnVal = JOptionPane.showConfirmDialog(m_oChartFrame,
              "This file already exists.  Do you wish to overwrite it?",
              "Model", JOptionPane.YES_NO_OPTION);
          if (returnVal == JOptionPane.YES_OPTION) {
            String sFileName = oFile.getAbsolutePath();
            m_jOutputFileName.setText(sFileName);
          }
        } else {
          // File didn't exist
          String sFileName = oFile.getAbsolutePath();
          m_jOutputFileName.setText(sFileName);
        }
      }

    } else if (oEvent.getActionCommand().equals("WriteTrees")) {
      int iTimestep = 0;
      try {
        Integer oTS = new Integer(m_jTimestep.getText());
        iTimestep = oTS.intValue();
      } catch (java.lang.NumberFormatException oErr) {
        JOptionPane.showMessageDialog(m_oChartFrame,
            "The value for timestep is not a recognized number.");
        return;
      }
      if (iTimestep < 0 || iTimestep > m_oManager.GetNumberOfActualTimesteps()) {
        JOptionPane.showMessageDialog(m_oChartFrame,
            "The value for timestep must be between 0 and "
                + m_oManager.GetNumberOfActualTimesteps() + ".");
        return;
      }
      WriteTrees(iTimestep);
    }
  }
  
  /**
   * Writes the trees for a time step to file. This will set up the 
   * infrastructure for the file, and parse the output file. 
   */
  private void WriteTrees(int iTimestep) {
    
    try {
      m_oChartFrame.setCursor(java.awt.Cursor
          .getPredefinedCursor(java.awt.Cursor.WAIT_CURSOR));

      //This serves as a flag to show that parsing was caused by this
      m_iCurrentTimestep = iTimestep;

      int i, j, m;

      //********************************************
      // Assemble a list of the columns in the text output file
      // Make a vector of all column headers - display names and code names
      Vector<String> p_oColumnHeader = new Vector<String>(3), 
                     p_oColumnCodes = new Vector<String>(3);

      // Add species and type
      p_oColumnHeader.add(m_iSpeciesIndex, "Species");
      p_oColumnHeader.add(m_iTypeIndex, "Type");
      p_oColumnHeader.add(m_iDeadCodeIndex, "Dead Code");
      p_oColumnCodes.add(m_iSpeciesIndex, "Species");
      p_oColumnCodes.add(m_iTypeIndex, "Type");
      p_oColumnCodes.add(m_iDeadCodeIndex, "Dead Code");            

      // Now add each data member uniquely
      String sTemp, sTemp2;
      boolean bFound;

      DetailedTreeSettings oSetting;
      for (i = 0; i < m_oManager.GetNumberTreeSettings(); i++) {
        oSetting = m_oManager.GetTreeSetting(i);
        for (j = 0; j < oSetting.GetNumberOfFloats(); j++) {
          sTemp2 = oSetting.GetFloat(j).GetDisplayName();
          bFound = false;
          for (m = 0; m < p_oColumnHeader.size(); m++) {
            sTemp = (String) p_oColumnHeader.elementAt(m);
            if (sTemp.equals(sTemp2)) {
              bFound = true;
              break;
            }
          }
          if (!bFound) {
            p_oColumnHeader.add(sTemp2);
            p_oColumnCodes.add(oSetting.GetFloat(j).GetCodeName());
          }
        }

        for (j = 0; j < oSetting.GetNumberOfInts(); j++) {
          sTemp2 = oSetting.GetInt(j).GetDisplayName();
          bFound = false;
          for (m = 0; m < p_oColumnHeader.size(); m++) {
            sTemp = (String) p_oColumnHeader.elementAt(m);
            if (sTemp.equals(sTemp2)) {
              bFound = true;
              break;
            }
          }
          if (!bFound) {
            p_oColumnHeader.add(sTemp2);
            p_oColumnCodes.add(oSetting.GetInt(j).GetCodeName());
          }
        }

        for (j = 0; j < oSetting.GetNumberOfBools(); j++) {
          sTemp2 = oSetting.GetBool(j).GetDisplayName();
          bFound = false;
          for (m = 0; m < p_oColumnHeader.size(); m++) {
            sTemp = (String) p_oColumnHeader.elementAt(m);
            if (sTemp.equals(sTemp2)) {
              bFound = true;
              break;
            }
          }
          if (!bFound) {
            p_oColumnHeader.add(sTemp2);
            p_oColumnCodes.add(oSetting.GetBool(j).GetCodeName());
          }
        }

        for (j = 0; j < oSetting.GetNumberOfChars(); j++) {
          sTemp2 = oSetting.GetChar(j).GetDisplayName();
          bFound = false;
          for (m = 0; m < p_oColumnHeader.size(); m++) {
            sTemp = (String) p_oColumnHeader.elementAt(m);
            if (sTemp.equals(sTemp2)) {
              bFound = true;
              break;
            }
          }
          if (!bFound) {
            p_oColumnHeader.add(sTemp2);
            p_oColumnCodes.add(oSetting.GetChar(j).GetCodeName());
          }
        }
      }
      
      //Turn these into arrays that we will use for writing our file
      mp_sColumnsByCode = new String[p_oColumnCodes.size()];
      mp_sFileRow = new String[p_oColumnCodes.size()];
      for (i = 0; i < mp_sColumnsByCode.length; i++) {
        mp_sColumnsByCode[i] = p_oColumnCodes.elementAt(i);
        mp_sFileRow[i] = "";
      }
      
      // Null the arrays to reset them before they are redeclared
      mp_iTreeBoolTransforms = null;
      mp_iTreeIntTransforms = null;
      mp_iTreeCharTransforms = null;
      mp_iTreeFloatTransforms = null;

      // Declare the tree data member arrays
      int iNumSpecies = m_oManager.GetNumberOfSpecies();
      int iNumTypes = m_oManager.GetNumberOfTypes();
      mp_iTreeBoolTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);
      mp_iTreeIntTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);
      mp_iTreeCharTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);
      mp_iTreeFloatTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);

      for (i = 0; i < iNumSpecies; i++) {
        mp_iTreeBoolTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
        mp_iTreeIntTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
        mp_iTreeCharTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
        mp_iTreeFloatTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
        for (j = 0; j < iNumTypes; j++) {
          mp_iTreeBoolTransforms.elementAt(i).add(j, new Vector<Integer>(0));
          mp_iTreeIntTransforms.elementAt(i).add(j, new Vector<Integer>(0));
          mp_iTreeCharTransforms.elementAt(i).add(j, new Vector<Integer>(0));
          mp_iTreeFloatTransforms.elementAt(i).add(j, new Vector<Integer>(0));
        }
      }

      for (i = 0; i < mp_sFileRow.length; i++) {
        mp_sFileRow[i] = "--";
      }
      m_jOut = new java.io.FileWriter(m_jOutputFileName.getText());
      m_jOut.write("Trees for timestep " + m_jTimestep.getText() + 
          " of output file " + m_oManager.GetFileName() + "\n");
      m_jOut.write(p_oColumnHeader.elementAt(0));
      for (i = 1; i < p_oColumnHeader.size(); i++) 
        m_jOut.write("\t" + p_oColumnHeader.elementAt(i));
      m_jOut.write("\n");

      // Make the file manager parse this file
      m_oManager.ReadFile(iTimestep);
      
      JOptionPane.showMessageDialog(m_oChartFrame, "File has been saved.");

    } catch (IOException e) {
      ModelException oErr = new ModelException(ErrorGUI.BAD_FILE, 
          "TreeListWriter.WriteTrees", e.getMessage());
      javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_oChartFrame);
      oHandler.WriteErrorMessage(oErr);
    } catch (ModelException oErr) {
      javawrapper.ErrorGUI oHandler = new javawrapper.ErrorGUI(m_oChartFrame);
      oHandler.WriteErrorMessage(oErr);
    }  finally {
      m_oChartFrame.setCursor(java.awt.Cursor
          .getPredefinedCursor(java.awt.Cursor.DEFAULT_CURSOR));
      m_iCurrentTimestep = -1;
      try {
      m_jOut.close();
      } catch (IOException e) {;}
    }
  }
  

  /**
   * Writes out the accumulated tree data and resets it.
   */
  public void EndTree() throws SAXException {
    if (-1 == m_iCurrentTimestep) return;
    try {
    int i;
    mp_sFileRow[m_iSpeciesIndex] = m_oManager.GetSpeciesNameFromCode(m_iTreeSpecies);
    mp_sFileRow[m_iTypeIndex] = mp_sTypeNames[m_iTreeType];
    mp_sFileRow[m_iDeadCodeIndex] = mp_sDeadCodeNames[m_iTreeDeadCode];

    try {      
      m_jOut.write(mp_sFileRow[0]);
      for (i = 1; i < mp_sFileRow.length; i++) {
        m_jOut.write("\t" + mp_sFileRow[i]);
      }
      m_jOut.write("\n");
    } catch (IOException e) {
      throw(new org.xml.sax.SAXException(e));
    }
    for (i = 0; i < mp_sFileRow.length; i++) {
      mp_sFileRow[i] = "--";
    }
    } catch (ModelException e) {
      throw(new SAXException(e.getMessage()));
    }
  }

  /**
   * Draws the panel allowing the user to select a timestep to save.
   * 
   * @param oLegend Legend Legend for this chart.
   * @param sChartTitle String Chart name for the window title.
   * @throws ModelException Passed through from other called methods - this
   * method doesn't throw it.
   * @return JInternalFrame The finished chart window.
   */
  ModelInternalFrame DrawChart(Legend oLegend, String sChartTitle)
      throws javawrapper.ModelException {
    ModelInternalFrame jFrame = new ModelInternalFrame(sChartTitle, oLegend);

    JPanel jMainPanel = new JPanel();
    jMainPanel.setLayout(new BoxLayout(jMainPanel, BoxLayout.PAGE_AXIS));

    // Label at the top
    SortieLabel jTempLabel = new SortieLabel(
        "Write a tab delimited text file of every tree saved for a timestep",
        SwingConstants.LEFT, java.awt.Font.BOLD, 0);
    JPanel jTempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    jTempPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempLabel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel.add(jTempLabel);
    jMainPanel.add(jTempPanel);

    jTempLabel = new SortieLabel("Timestep to write:", SwingConstants.LEFT);
    m_jTimestep.setFont(new SortieFont());
    m_jTimestep.setText("0");
    m_jTimestep.setPreferredSize(new Dimension(30, (int) m_jTimestep
        .getPreferredSize().getHeight()));
    m_jTimestep.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    jTempPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel.add(jTempLabel);
    jTempPanel.add(m_jTimestep);
    jMainPanel.add(jTempPanel);

    // Filename and browse button
    jTempLabel = new SortieLabel("Output File Name:", SwingConstants.LEFT);
    jTempLabel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    jTempPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel.add(jTempLabel);
    jMainPanel.add(jTempPanel);

    m_jOutputFileName.setFont(new SortieFont());
    m_jOutputFileName.setText("");
    m_jOutputFileName.setPreferredSize(new Dimension(300,
        (int) m_jOutputFileName.getPreferredSize().getHeight()));
    m_jOutputFileName.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    JButton jButton = new JButton("Browse");
    jButton.setFont(new SortieFont());
    jButton.addActionListener(this);
    jButton.setActionCommand("Browse");
    jButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    jTempPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel.add(m_jOutputFileName);
    jTempPanel.add(jButton);
    jMainPanel.add(jTempPanel);

    jButton = new JButton("Write Trees");
    jButton.setFont(new SortieFont());
    jButton.addActionListener(this);
    jButton.setActionCommand("WriteTrees");
    jTempPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
    jTempPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTempPanel.add(jButton);
    jMainPanel.add(jTempPanel);
    jMainPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

    // Recreate the content pane with the controls and the chart
    JPanel jContentPanel = new JPanel(new java.awt.BorderLayout());
    jContentPanel.setLayout(new java.awt.BorderLayout());
    jContentPanel.add(jMainPanel, java.awt.BorderLayout.NORTH);
    m_oChartFrame = jFrame;
    m_oChartFrame.setContentPane(jContentPanel);

    return m_oChartFrame;
  }

  /** Does nothing */
  void UpdateChart(Legend legend) throws ModelException {
    ;
  }

  /** Does nothing */
  protected void WriteChartDataToFile(FileWriter out) throws IOException,
      ModelException {
    ;
  }

  /**
   * This object always wants tree float data members.
   * 
   * @return True
   */
  public boolean WantAnyTreeFloats() {
    return true;
  }

  /**
   * This object always wants tree int data members.
   * 
   * @return True
   */
  public boolean WantAnyTreeInts() {
    return true;
  }

  /**
   * This object always wants tree char data members.
   * 
   * @return True
   */
  public boolean WantAnyTreeChars() {
    return true;
  }

  /**
   * This object always wants tree bool data members.
   * 
   * @return True
   */
  public boolean WantAnyTreeBools() {
    return true;
  }

  /**
   * This object always wants dead tree float data members.
   * 
   * @return True.
   */
  public boolean WantAnyDeadTreeFloats() {
    return true;
  }

  /**
   * This object always wants dead tree int data members.
   * 
   * @return True.
   */
  public boolean WantAnyDeadTreeInts() {
    return true;
  }

  /**
   * This object always wants dead tree char data members.
   * 
   * @return True.
   */
  public boolean WantAnyDeadTreeChars() {
    return true;
  }

  /**
   * This object always wants dead tree bool data members.
   * 
   * @return True.
   */
  public boolean WantAnyDeadTreeBools() {
    return true;
  }
  
  /**
   * Adds a piece of tree data to the file row currently being built.
   * @param iSpecies Species number.
   * @param iType Type number.
   * @param iDeadCode Dead code reason number.
   * @param iPos Column position number.
   * @param sVal Value for the column.
   */
  private void AddDataToFileRow(int iSpecies, int iType, int iDeadCode, 
      int iPos, String sVal) {
    m_iTreeSpecies = iSpecies;
    m_iTreeType = iType;
    m_iTreeDeadCode = iDeadCode;
    mp_sFileRow[iPos] = sVal; 
  }

  /**
   * Accepts a piece of tree float data from the parser.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param fVal Value.
   */
  public void AddTreeFloatData(int iSpecies, int iType, int iCode, float fVal) {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeFloatTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, OutputBehaviors.NOTDEAD, 
        iIndex.intValue(), String.valueOf(fVal));      
  }

  /**
   * Accepts a piece of tree int data from the parser.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param iVal Value.
   */
  public void AddTreeIntData(int iSpecies, int iType, int iCode, int iVal) {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeIntTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, OutputBehaviors.NOTDEAD, 
        iIndex.intValue(), String.valueOf(iVal));
  }

  /**
   * Accepts a piece of tree char data from the parser.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param sVal Value.
   */
  public void AddTreeCharData(int iSpecies, int iType, int iCode, String sVal) {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeCharTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, OutputBehaviors.NOTDEAD, 
        iIndex.intValue(), sVal);
  }

  /**
   * Accepts a piece of tree bool data from the parser.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param bVal Value.
   */
  public void AddTreeBoolData(int iSpecies, int iType, int iCode, boolean bVal) {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeBoolTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, OutputBehaviors.NOTDEAD, 
        iIndex.intValue(), String.valueOf(bVal));
  }

  /**
   * Accepts a piece of dead tree float data from the parser. If this particular
   * piece is not wanted, do nothing.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param iDeadCode Dead code for this tree.
   * @param fVal Value.
   */
  public void AddDeadTreeFloatData(int iSpecies, int iType, int iCode,
      int iDeadCode, float fVal) {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeFloatTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, iDeadCode,
        iIndex.intValue(), String.valueOf(fVal));
  }

  /**
   * Accepts a piece of dead tree int data from the parser. If this particular
   * piece is not wanted, do nothing.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param iDeadCode Dead code for this tree.
   * @param iVal Value.
   */
  public void AddDeadTreeIntData(int iSpecies, int iType, int iCode,
      int iDeadCode, int iVal) {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeIntTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, iDeadCode, 
        iIndex.intValue(), String.valueOf(iVal));
  }

  /**
   * Accepts a piece of dead tree char data from the parser. If this particular
   * piece is not wanted, do nothing.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param iDeadCode Dead code for this tree.
   * @param sVal Value.
   */
  public void AddDeadTreeCharData(int iSpecies, int iType, int iCode,
      int iDeadCode, String sVal) {
 // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeCharTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, iDeadCode, 
        iIndex.intValue(), sVal);
  }

  /**
   * Accepts a piece of dead tree bool data from the parser.
   * 
   * @param iSpecies Species of the tree from which this value came.
   * @param iType Type of the tree from which this value came.
   * @param iCode Data member code of this value.
   * @param iDeadCode Dead code for this tree.
   * @param bVal Value.
   */
  public void AddDeadTreeBoolData(int iSpecies, int iType, int iCode,
      int iDeadCode, boolean bVal) {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }    
    Integer iIndex = mp_iTreeBoolTransforms.elementAt(iSpecies).elementAt(
        iType).elementAt(iCode);
    AddDataToFileRow(iSpecies, iType, iDeadCode, 
        iIndex.intValue(), String.valueOf(bVal));
  }

  /**
   * Accepts a tree float data member code for future reference when passed
   * float data members.
   * 
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   */
  public void AddTreeFloatDataMemberCode(int iSpecies, int iType,
      String sLabel, int iCode) throws SAXException {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }
    int iOfficialIndex = GetCodeForDataMember(DataMember.FLOAT, iSpecies,
        iType, sLabel);

    if (mp_iTreeFloatTransforms.elementAt(iSpecies).elementAt(iType).size() <= iCode) {
      mp_iTreeFloatTransforms.elementAt(iSpecies).elementAt(iType).setSize(
          iCode + 1);
    }
    mp_iTreeFloatTransforms.elementAt(iSpecies).elementAt(iType).remove(iCode);
    mp_iTreeFloatTransforms.elementAt(iSpecies).elementAt(iType).add(iCode,
        new Integer(iOfficialIndex));
  }

  /**
   * Accepts a tree int data member code for future reference when passed int
   * data members.
   * 
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   */
  public void AddTreeIntDataMemberCode(int iSpecies, int iType, String sLabel,
      int iCode) throws SAXException {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }
    int iOfficialIndex = GetCodeForDataMember(DataMember.INTEGER, iSpecies,
        iType, sLabel);

    if (mp_iTreeIntTransforms.elementAt(iSpecies).elementAt(iType).size() <= iCode) {
      mp_iTreeIntTransforms.elementAt(iSpecies).elementAt(iType).setSize(
          iCode + 1);
    }
    mp_iTreeIntTransforms.elementAt(iSpecies).elementAt(iType).remove(iCode);
    mp_iTreeIntTransforms.elementAt(iSpecies).elementAt(iType).add(iCode,
        new Integer(iOfficialIndex));
  }

  /**
   * Accepts a tree char data member code for future reference when passed char
   * data members.
   * 
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   */
  public void AddTreeCharDataMemberCode(int iSpecies, int iType, String sLabel,
      int iCode) throws SAXException {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }
    int iOfficialIndex = GetCodeForDataMember(DataMember.CHAR, iSpecies, iType,
        sLabel);

    if (mp_iTreeCharTransforms.elementAt(iSpecies).elementAt(iType).size() <= iCode) {
      mp_iTreeCharTransforms.elementAt(iSpecies).elementAt(iType).setSize(
          iCode + 1);
    }
    mp_iTreeCharTransforms.elementAt(iSpecies).elementAt(iType).remove(iCode);
    mp_iTreeCharTransforms.elementAt(iSpecies).elementAt(iType).add(iCode,
        new Integer(iOfficialIndex));
  }

  /**
   * Accepts a tree bool data member code for future reference when passed bool
   * data members.
   * 
   * @param iSpecies The species for which this is a data member.
   * @param iType The tree type for which this is a data member.
   * @param sLabel The label of the data member.
   * @param iCode The data member code.
   */
  public void AddTreeBoolDataMemberCode(int iSpecies, int iType, String sLabel,
      int iCode) throws SAXException {
    // If this was triggered by some other event, ignore
    if (-1 == m_iCurrentTimestep) {
      return;
    }
    int iOfficialIndex = GetCodeForDataMember(DataMember.BOOLEAN, iSpecies,
        iType, sLabel);

    if (mp_iTreeBoolTransforms.elementAt(iSpecies).elementAt(iType).size() <= iCode) {
      mp_iTreeBoolTransforms.elementAt(iSpecies).elementAt(iType).setSize(
          iCode + 1);
    }
    mp_iTreeBoolTransforms.elementAt(iSpecies).elementAt(iType).remove(iCode);
    mp_iTreeBoolTransforms.elementAt(iSpecies).elementAt(iType).add(iCode,
        new Integer(iOfficialIndex));
  }

  /**
   * Get the code for a particular data member label.
   * 
   * @param iDataType One of the data type statics from DataMember
   * @param iSpecies Tree species
   * @param iType Tree type
   * @param sLabel Tree data member label
   * @return Data member code
   * @throws ModelException If settings for the tree type and species cannot be
   * found, or if the data type is invalid
   */
  private int GetCodeForDataMember(int iDataType, int iSpecies, int iType,
      String sLabel) throws SAXException {
    int i;
    // Find the column position of this data member    
    for (i = 0; i < mp_sColumnsByCode.length; i++) {
      if (sLabel.equals(mp_sColumnsByCode[i])) {
        return i;
      }
    }
    throw(new SAXException("TreeListWriter: Couldn't find column space for \"" 
        + sLabel + "\"."));  
  }
}
