package javawrapper;

import javax.swing.*;

import uihelpers.SortieFont;

/**
 * Displays an edit window for editing the tree species/type combos to
 * which a single behavior applies.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>February 4, 2008: Updated to conform to Java 6 compliance (LEM)
 */
public class DisplayBehaviorComboEdit
    extends JDialog
    implements java.awt.event.ActionListener {
  
  

  /**The list displaying the assigned species/type combos for this
   * behavior*/
  protected JList m_jAssignedCombosList;

  /** Tree population object */
  protected TreePopulation m_oPop;

  /** Behavior object to edit */
  protected Behavior m_oBehavior;

  /**The list of species*/
  protected JList m_jSpecies;

  /**The list of types*/
  protected JList m_jTypes;

  /**The assigned species/type combos for this behavior*/
  protected DefaultListModel m_jAssignedCombosListModel;

  /**The help topic ID for this window*/
  private String m_sHelpID = "windows.behavior_assignments";

  /**
   * Constructor.
   * @param jParent Parent dialog in which to display this dialog.
   * @param oBehavior Behavior to edit.
   * @param oPop Tree population.
   * @param oHelpBroker The application's help broker.
   * @throws ModelException if there is a problem constructing the window.
   */
  public DisplayBehaviorComboEdit(JDialog jParent, Behavior oBehavior,
                                  TreePopulation oPop,
                                  javax.help.HelpBroker oHelpBroker) throws
      ModelException {
    super(jParent, "Tree assignments for " + oBehavior.GetDescriptor(), true);
    m_oPop = oPop;
    m_oBehavior = oBehavior;

    //Help ID
    oHelpBroker.enableHelpKey(this.getRootPane(), m_sHelpID, null);

    /********************************************************************
     * Leftmost panel - two lists with species and type for choosing
     ********************************************************************/
    JPanel jNewComboPicker = new JPanel();
    int i;
    //Species choices
    String[] p_sSpeciesOptions = new String[m_oPop.GetNumberOfSpecies()];
    for (i = 0; i < p_sSpeciesOptions.length; i++) {
      p_sSpeciesOptions[i] = m_oPop.GetSpeciesNameFromCode(i).replace('_', ' ');
    }

    //Type choices
    String[] p_sTypeOptions = new String[TreePopulation.GetNumberOfTypes()];
    for (i = 0; i < p_sTypeOptions.length; i++) {
      p_sTypeOptions[i] = TreePopulation.GetTypeNameFromCode(i);
    }

    m_jSpecies = new JList(p_sSpeciesOptions);
    m_jTypes = new JList(p_sTypeOptions);
    m_jSpecies.setFont(new SortieFont());
    m_jTypes.setFont(new SortieFont());
    m_jSpecies.setVisibleRowCount(6);
    m_jTypes.setVisibleRowCount(6);
    JScrollPane jSpeciesScroller = new JScrollPane(m_jSpecies);
    JScrollPane jTypeScroller = new JScrollPane(m_jTypes);

    //Add button
    JButton jAddButton = new JButton(new ModelIcon(15, 15,
        ModelIcon.RIGHT_TRIANGLE));
    jAddButton.setFont(new SortieFont());
    jAddButton.setToolTipText(
        "Add this tree species and type to assigned list");
    jAddButton.setActionCommand("Add");
    jAddButton.addActionListener(this);

    //Assemble the panel
    jNewComboPicker.setLayout(new BoxLayout(jNewComboPicker,
                                            BoxLayout.PAGE_AXIS));
    JLabel jLabel = new JLabel("Please choose the species:");
    jLabel.setFont(new SortieFont());
    jNewComboPicker.add(jLabel);
    jNewComboPicker.add(jSpeciesScroller);
    jLabel = new JLabel("Please choose the tree types:");
    jLabel.setFont(new SortieFont());
    jNewComboPicker.add(jLabel);
    jNewComboPicker.add(jTypeScroller);

    /********************************************************************
     * List showing existing species/type combos
     ********************************************************************/
    JPanel jExisting = new JPanel(new java.awt.BorderLayout());
    m_jAssignedCombosListModel = new DefaultListModel();
    m_jAssignedCombosList = new JList(m_jAssignedCombosListModel);
    m_jAssignedCombosList.setFont(new SortieFont());
    JScrollPane jScroller = new JScrollPane(m_jAssignedCombosList);
    jScroller.setPreferredSize(new java.awt.Dimension(200,
                                             (int) jScroller.getPreferredSize().
                                             getHeight()));

    //Load the combos
    for (i = 0; i < oBehavior.GetNumberOfCombos(); i++) {
      m_jAssignedCombosListModel.addElement(
          new ComboDisplay(m_oPop, oBehavior.GetSpeciesTypeCombo(i).GetSpecies(),
                           oBehavior.GetSpeciesTypeCombo(i).GetType()));
    }
    jLabel = new JLabel("Behavior currently assigned to:");
    jLabel.setFont(new SortieFont());
    jExisting.add(jLabel, java.awt.BorderLayout.NORTH);
    jExisting.add(jScroller, java.awt.BorderLayout.CENTER);

    /********************************************************************
     * Remove button for combos
     ********************************************************************/
    JButton jRemoveButton = new JButton("Remove");
    jRemoveButton.setFont(new SortieFont());
    jRemoveButton.setToolTipText(
        "Remove this tree species / type from this behavior's list");
    jRemoveButton.setActionCommand("Remove");
    jRemoveButton.addActionListener(this);

    //Assemble the entire window
    //Package everything but the buttons in a center panel
    JPanel jCenterPanel = new JPanel();
    jCenterPanel.setLayout(new java.awt.FlowLayout());
    jCenterPanel.add(jNewComboPicker);
    jCenterPanel.add(jAddButton);
    jCenterPanel.add(jExisting);
    jCenterPanel.add(jRemoveButton);

    getContentPane().setLayout(new java.awt.BorderLayout());
    getContentPane().add(jCenterPanel, java.awt.BorderLayout.CENTER);
    //Buttons at the button
    getContentPane().add(new OKCancelButtonPanel(this, oHelpBroker, m_sHelpID),
                         java.awt.BorderLayout.PAGE_END);

    this.pack();
    this.setLocationRelativeTo(null);
    this.setVisible(true);
  }

  /**
     * Performs actions for clicked buttons.
     * @param e ActionEvent object.
     *
     * <br>Edit history:
     * <br>------------------
     * <br>April 28, 2004: Submitted in beta version (LEM)
     */
    public void actionPerformed(java.awt.event.ActionEvent e) {
      if (e.getActionCommand().equals("Cancel")) {
        this.setVisible(false);
        this.dispose();
      }
      else if (e.getActionCommand().equals("OK")) {
        try {
          //Sets the combos listed
          //Find the behavior to which this applies
          ComboDisplay oCombo;
          int i;

          m_oBehavior.ClearSpeciesTypeCombos();
          for (i = 0; i < m_jAssignedCombosListModel.size(); i++) {
            oCombo = (ComboDisplay) m_jAssignedCombosListModel.elementAt(i);
            m_oBehavior.AddSpeciesTypeCombo(new SpeciesTypeCombo(oCombo.
                GetSpecies(), oCombo.GetType(), m_oPop));
          }

          this.setVisible(false);
          this.dispose();
        }
        catch (ModelException oErr) {
          ErrorGUI oHandler = new ErrorGUI(this);
          oHandler.WriteErrorMessage(oErr);
        }
      }
      else if (e.getActionCommand().equals("Add")) {

        //Adding a new type/species combo
        Object[] p_oSelectedSpecies, p_oSelectedTypes;
        ComboDisplay oCombo;
        String sSpecies, sType;
        int iSpecies, iType, i, j, k;
        boolean bDuplicate;

        //Get the species
        p_oSelectedSpecies = m_jSpecies.getSelectedValues();
        if (p_oSelectedSpecies == null || p_oSelectedSpecies.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a tree species.");
          return; //no species picked
        }

        //Get the types
        p_oSelectedTypes = m_jTypes.getSelectedValues();
        if (p_oSelectedTypes == null || p_oSelectedTypes.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a tree type.");
          return; //no types picked
        }

        for (i = 0; i < p_oSelectedSpecies.length; i++) {
          sSpecies = (String) p_oSelectedSpecies[i];
          iSpecies = m_oPop.GetSpeciesCodeFromName(sSpecies.replace(' ', '_'));
          for (j = 0; j < p_oSelectedTypes.length; j++) {
            bDuplicate = false;

            //Make sure this species/type combo doesn't already exist
            sType = (String) p_oSelectedTypes[j];
            iType = TreePopulation.GetTypeCodeFromName(sType);

            for (k = 0; k < m_jAssignedCombosListModel.size(); k++) {
              oCombo = (ComboDisplay) m_jAssignedCombosListModel.elementAt(k);
              if (iSpecies == oCombo.GetSpecies() && iType == oCombo.GetType()) {
                bDuplicate = true;
              }
            }

            //Add the new combo
            if (bDuplicate == false) {
              oCombo = new ComboDisplay(m_oPop, iSpecies, iType);
              m_jAssignedCombosListModel.addElement(oCombo);
            }
          }
        }
      }
      else if (e.getActionCommand().equals("Remove")) {

        //Removing one or more type/species combos
        Object[] p_oSelected = m_jAssignedCombosList.getSelectedValues();
        if (p_oSelected == null || p_oSelected.length == 0) {
          JOptionPane.showMessageDialog(this, "Please select a tree type.");
          return; //Nothing selected
        }
        for (int i = 0; i < p_oSelected.length; i++) {
          m_jAssignedCombosListModel.removeElement(p_oSelected[i]);
        }
      }
    }
}
