package javawrapper;

import java.util.Vector;

/**
 * This class encapsulates the data for a single harvest event - being for
 * one timestep, for one cut type.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>November 19, 2005:  Added species removal capacity (LEM)
 * <br>February 4, 2008: Updated to conform to Java 6 compliance (LEM)
 */

public class HarvestData {

  /**Species codes to which to apply this harvest cut*/
  private Vector<Integer> mp_iSpecies = new Vector<Integer>(0);
  /**The list of cells to which to apply this cut (vector of Cell objects)*/
  private Vector<Cell> mp_oCells = new Vector<Cell>(0);
  /**The list of cut ranges and cut amounts (CutRange objects)*/
  private Vector<CutRange> mp_oCutRanges = new Vector<CutRange>(0);

  /**Timestep to which to apply this harvest cut*/
  private int m_iTimestep = -1;
      /**The cut type - possible values come from DisturbanceBehaviors and are
             PARTIAL_CUT, GAP_CUT, and CLEAR_CUT*/
  private int m_iCutType = -1;
      /**How the cut amount is specified - possible values come from DisturbanceBehaviors
       and are PERCENTAGE_BASAL_AREA, ABSOLUTE_BASAL_AREA, PERCENTAGE_DENSITY, and
             ABSOLUTE_DENSITY.*/
  private int m_iCutAmountType = -1;
  /**Maximum number of allowed cut ranges*/
  public static int NUMBER_ALLOWED_CUT_RANGES = 4;

  /**
   * Gets the cut amount type.  Possible values come from DisturbanceBehaviors and
   * are PERCENTAGE_BASAL_AREA, ABSOLUTE_BASAL_AREA, PERCENTAGE_DENSITY, and
   * ABSOLUTE_DENSITY, or -1 if the value has not yet been set.
   * @return The cut amount type.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetCutAmountType() {
    return m_iCutAmountType;
  }

  /**
   * Sets the cut amount type.
   * @param iCut the cut amount type.  Possible values come from
   * DisturbanceBehaviors and are PERCENTAGE_BASAL_AREA, ABSOLUTE_BASAL_AREA,
   * PERCENTAGE_DENSITY, and ABSOLUTE_DENSITY.
   * @throws ModelException if the cut amount type is unrecognized.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetCutAmountType(int iCut) throws ModelException {
    if (iCut != DisturbanceBehaviors.PERCENTAGE_BASAL_AREA &&
        iCut != DisturbanceBehaviors.ABSOLUTE_BASAL_AREA &&
        iCut != DisturbanceBehaviors.PERCENTAGE_DENSITY &&
        iCut != DisturbanceBehaviors.ABSOLUTE_DENSITY) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                "The harvest cut amount type \"" + iCut +
                                "\" is not recognized."));
    }

    m_iCutAmountType = iCut;
  }

  /**
   * Gets the cut type.
   * @return The cut type.  Possible values come from DisturbanceBehaviors and are
   * PARTIAL_CUT, GAP_CUT, and CLEAR_CUT, or -1 if the value has not been set.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetCutType() {
    return m_iCutType;
  }

  /**
   * Sets the cut type.
   * @param iCut The cut type.  Values come from DisturbanceBehaviors and are
   * PARTIAL_CUT, GAP_CUT, and CLEAR_CUT.
   * @throws ModelException if the cut type is unrecognized.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetCutType(int iCut) throws ModelException {
    if (iCut != DisturbanceBehaviors.PARTIAL_CUT &&
        iCut != DisturbanceBehaviors.GAP_CUT &&
        iCut != DisturbanceBehaviors.CLEAR_CUT) {
      throw(new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA", "The cut type \""
                               + iCut + "\" is not recognized."));
    }

    m_iCutType = iCut;
  }

  /**
   * Gets the timestep for this cut event.
   * @return The timestep.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetTimestep() {
    return m_iTimestep;
  }

  /**
   * Sets the timestep for this cut event.
   * @param iTimestep The timestep to set.
   * @throws ModelException if the timestep is less than 0.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void SetTimestep(int iTimestep) throws ModelException {
    if (iTimestep < 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                "The timestep for this harvest event cannot be"
                                + " less than 0."));
    }

    m_iTimestep = iTimestep;
  }

  /**
   * Gets the number of species to which this cut has been applied.
   * @return The number of species.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetNumberOfSpecies() {
    return mp_iSpecies.size();
  }

  /**
   * Gets a species at a certain index within the species list.
   * @param iIndex The species index.
   * @return The species value.
   * @throws ModelException If the index is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetSpecies(int iIndex) throws ModelException {
    if (iIndex < 0 || iIndex >= mp_iSpecies.size()) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                "Invalid species index for this cut type."));
    }

    Integer iValue = (Integer) mp_iSpecies.elementAt(iIndex);
    return iValue.intValue();
  }

  /**
   * Adds a new species to apply this harvest cut to.  If this species is
   * already on the list, it is not added again.
   * @param iSpecies Species index.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddSpecies(int iSpecies) {
    Integer iValue;
    int i, iNumSpecies = mp_iSpecies.size();
    boolean bFound = false;

    //Make sure this species is not already on the list
    for (i = 0; i < iNumSpecies; i++) {
      iValue = (Integer) mp_iSpecies.elementAt(i);
      if (iValue.intValue() == iSpecies) {
        bFound = true;
      }
    }

    if (!bFound) {
      mp_iSpecies.add(new Integer(iSpecies));
    }
  }

  /**
   * Removes a species from the list to which to apply this harvest.  If this
   * species is not on the list, the function quietly exits.
   * @param iSpecies Species number (number identifying the species, NOT the
   * index number in the vector).
   */
  public void RemoveSpecies(int iSpecies) {
    Integer iValue;
    int i, iNumSpecies = mp_iSpecies.size();

    for (i = 0; i < iNumSpecies; i++) {
      iValue = (Integer) mp_iSpecies.elementAt(i);
      if (iValue.intValue() == iSpecies) {
        mp_iSpecies.remove(i);
        return;
      }
    }
  }

  /**
   * Clears the list of species
   */
  public void ClearSpecies() {
    mp_iSpecies.clear();
  }

  /**
   * Gets the number of cut ranges.
   * @return Number of cut ranges.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetNumberOfCutRanges() {
    return mp_oCutRanges.size();
  }

  /**
   * Adds a new cut range.
   * @param fLow Lower bound of the cut range, as a dbh in cm.
   * @param fHigh Upper bound of the cut range, as a dbh in cm.
   * @param fAmountToCut Amount to cut.  If the amount cut type is as a
   * percentage of either basal area or density, this is a percentage value
   * between 0 and 100.  If it is amount of basal area, this is basal area in
   * square meters per hectare.  If this is amount of density, this is number
   * of trees per hectare.
   * @throws ModelException if any of the following are true:
   * <ul>
   * <li>This adds more than the maximum number of cut ranges allowed
   * <li>A dbh bound value is less than 0
   * <li>The upper bound value is less than the lower bound value
   * <li>The amount to cut is negative
   * <li>If the amount to cut is a percentage, the value is either less than 0
   * or greater than 100
   * </ul>
   */
  public void AddCutRange(float fLow, float fHigh, float fAmountToCut) throws
      ModelException {
    if (mp_oCutRanges.size() == NUMBER_ALLOWED_CUT_RANGES) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                               "Too many cut ranges for this harvest cut."));
    }

    mp_oCutRanges.add(new CutRange(fLow, fHigh, fAmountToCut, this));
  }

  /**
   * Gets the lower dbh bound value for a cut range.
   * @param iCutRangeIndex The cut range index.
   * @return The lower bound.
   * @throws ModelException If the cut range is invalid.
   */
  public float GetLowerBound(int iCutRangeIndex) throws ModelException {
    if (iCutRangeIndex < 0 || iCutRangeIndex >= mp_oCutRanges.size()) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                               "Invalid cut range index for this cut."));
    }

    CutRange oCutRange = (CutRange) mp_oCutRanges.elementAt(iCutRangeIndex);
    return oCutRange.GetLowBound();
  }

  /**
   * Gets the upper dbh bound value for a cut range.
   * @param iCutRangeIndex The cut range index.
   * @return The upper bound.
   * @throws ModelException If the cut range is invalid.
   */
  public float GetUpperBound(int iCutRangeIndex) throws ModelException {

    if (iCutRangeIndex < 0 || iCutRangeIndex >= mp_oCutRanges.size()) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                "Invalid cut range index for this cut."));
    }

    CutRange oCutRange = (CutRange) mp_oCutRanges.elementAt(iCutRangeIndex);
    return oCutRange.GetHighBound();
  }

  /**
   * Gets the cut amount for a cut range.
   * @param iCutRangeIndex The cut range index.
   * @return The cut amount.
   * @throws ModelException If the cut range is invalid.
   */
  public float GetCutAmount(int iCutRangeIndex) throws ModelException {

    if (iCutRangeIndex < 0 || iCutRangeIndex >= mp_oCutRanges.size()) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                "Invalid cut range index for this cut."));
    }

    CutRange oCutRange = (CutRange) mp_oCutRanges.elementAt(iCutRangeIndex);
    return oCutRange.GetAmountToCut();
  }

  /**
   * Gets the number of cells in the cut range's cut area.
   * @return Number of cells.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public int GetNumberOfCells() {
    return mp_oCells.size();
  }

  /**
   * Adds a cell to the harvest cut range's area.  Duplicates will not be
   * added.
   * @param iX X cell number.
   * @param iY Y cell number.
   * @param oPlot a Plot object.
   * @throws ModelException if the cell coordinates are invalid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddCell(int iX, int iY, Plot oPlot) throws ModelException {
    int i, iNumCells = GetNumberOfCells();
    Cell oCell;

    for (i = 0; i < iNumCells; i++) {
      oCell = (Cell) mp_oCells.elementAt(i);
      if (iX == oCell.GetX() && iY == oCell.GetY()) {
        return;
      }
    }

    mp_oCells.add(new Cell(iX, iY, oPlot));
  }

  /**
   * Adds a cell to the harvest cut range's area.  Duplicates will not be
   * added.
   * @param oNewCell Cell object to add.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void AddCell(Cell oNewCell) {
    Cell oExistingCell;
    int i, iNumCells = GetNumberOfCells();

    for (i = 0; i < iNumCells; i++) {
      oExistingCell = (Cell) mp_oCells.elementAt(i);
      if (oNewCell.GetX() == oExistingCell.GetX() &&
          oNewCell.GetY() == oExistingCell.GetY()) {
        return;
      }
    }

    mp_oCells.add(oNewCell);
  }

  /**
   * Removes a cell from the harvest cut range's area.
   * @param iIndex int Cell index. Invalid values are ignored.
   */
  public void RemoveCell(int iIndex) {
    if (iIndex < 0 || iIndex >= mp_oCells.size()) {
      return;
    }
    mp_oCells.remove(iIndex);
  }

  /**
   * Gets a cell in a harvest's cut area.
   * @param iIndex The index of the object.
   * @return The cell object.
   * @throws ModelException If the index is not valid.
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public Cell GetCell(int iIndex) throws ModelException {
    if (iIndex < 0 || iIndex >= mp_oCells.size()) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                "Invalid cell index for this cut."));
    }

    return (Cell) mp_oCells.elementAt(iIndex);
  }

  /**
   * Validates the current set of data in this cut.  This makes sure the
   * following are all true:
   * <ul>
   * <li>there is at least one species
   * <li>all species are valid
   * <li>a timestep has been specified
   * <li>a cut type has been specified
   * <li>a cut amount type have been specified
   * <li>there is at least one grid cell to which to apply the cut
   * <li>there is at least one cut range to which to apply the cut
   * </ul>
   * @param oPop TreePopulation object, to help verify data
   * @param oPlot Plot object for timestep verification
   * @throws ModelException if any of the above conditions is not true
   *
   * <br>Edit history:
   * <br>------------------
   * <br>April 28, 2004: Submitted in beta version (LEM)
   */
  public void ValidateCut(TreePopulation oPop, Plot oPlot) throws ModelException {

    int i, iNumSpecies = GetNumberOfSpecies();

    if (iNumSpecies == 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "This harvest cut must have at least one species."));
    }

    //Let the tree population object check out all the species
    for (i = 0; i < iNumSpecies; i++) {
      Integer iVal = (Integer) mp_iSpecies.elementAt(i);
      if (oPop.GetSpeciesNameFromCode(iVal.intValue()).length() == 0) {
        throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                 "Invalid species number - " +
                                 String.valueOf(iVal) + "."));
      }
    }

    if (m_iTimestep < 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "This harvest cut must have a timestep specified."));
    }

    if (m_iTimestep > oPlot.GetNumberOfTimesteps()) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "The harvest timestep is after the end of the run."));
    }

    if (m_iCutType < 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "This harvest cut must have a cut type specified."));
    }

    if (m_iCutAmountType < 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "This harvest cut must have a cut amount type specified."));
    }

    if (GetNumberOfCells() == 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "This harvest cut must be applied to at least one grid cell."));
    }

    if (GetNumberOfCutRanges() == 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "This harvest cut must be applied to at least one cut range."));
    }
  }
}

/**
 * This class defines a cut range for a harvest event.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 */
class CutRange {

  /**Low end of the dbh range*/
  private float m_fLowDbh,
      /**High end of the dbh range*/
      m_fHighDbh,
      /**Amount to cut, according to the cut type*/
      m_fAmountToCut;

  /**
   * Constructor.
   * @param fLow Lower bound of the cut range, as a dbh in cm.
   * @param fHigh Upper bound of the cut range, as a dbh in cm.
   * @param fAmountToCut Amount to cut.  If the amount cut type is as a
   * percentage of either basal area or density, this is a percentage value
   * between 0 and 100.  If it is amount of basal area, this is basal area in
   * square meters per hectare.  If this is amount of density, this is number
   * of trees per hectare.
   * @param oParent Parent HarvestData object.
   * @throws ModelException if any of the following are true:
   * <ul>
   * <li>A dbh bound value is less than 0
   * <li>The upper bound value is less than the lower bound value
   * <li>The amount to cut is negative
   * <li>If the amount to cut is a percentage, the value is either less than 0
   * or greater than 100
   * </ul>
   */
  public CutRange(float fLow, float fHigh, float fAmountToCut,
                  HarvestData oParent) throws ModelException {
    if (fLow < 0 || fHigh < 0 || fAmountToCut < 0) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
                                "The low bound, high bound, and amount to cut "
                                + "for a harvest must be greater than 0."));
    }
    if (oParent.GetCutAmountType() == DisturbanceBehaviors.PERCENTAGE_BASAL_AREA ||
        oParent.GetCutAmountType() == DisturbanceBehaviors.PERCENTAGE_DENSITY) {
      if (fAmountToCut > 100) {
        throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
        "The percentage to cut for a harvest must not be greater than 100."));
      }
    }

    if (fHigh < fLow) {
      throw( new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
      "The low bound for this harvest cut cannot be less than the high bound."));
    }

    m_fLowDbh = fLow;
    m_fHighDbh = fHigh;
    m_fAmountToCut = fAmountToCut;
  }

  /**
   * Gets the lower bound value of the dbh cut range.
   * @return The lower dbh.
   */
  public float GetLowBound() {
    return m_fLowDbh;
  }

  /**
   * Gets the upper bound value of the dbh cut range.
   * @return The upper dbh.
   */
  public float GetHighBound() {
    return m_fHighDbh;
  }

  /**
   * Gets the amount to cut.
   * @return The amount to cut.
   */
  public float GetAmountToCut() {
    return m_fAmountToCut;
  }
}
