package javawrapper;

import javax.swing.*;

import uihelpers.SortieFont;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

/**
 * Window for harvest editing.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */

public class HarvestEdit
    extends EpisodicEventsEdit
    implements ActionListener {
  
  

  /**List of checkboxes, one for each species, for which species to apply
   * the harvest to*/
  protected JCheckBox[] mp_jSpeciesChex;

  /**Edit box where timestep is entered*/
  protected JTextField m_jTimestepEdit = new JTextField("       ");

  /**Edit box where the cut range 1 minimum DBH is entered*/
  protected JTextField m_jCutRange1MinDBH = new JTextField();
  /**Edit box where the cut range 2 minimum DBH is entered*/
  protected JTextField m_jCutRange2MinDBH = new JTextField();
  /**Edit box where the cut range 3 minimum DBH is entered*/
  protected JTextField m_jCutRange3MinDBH = new JTextField();
  /**Edit box where the cut range 4 minimum DBH is entered*/
  protected JTextField m_jCutRange4MinDBH = new JTextField();

  /**Edit box where the cut range 1 maximum DBH is entered*/
  protected JTextField m_jCutRange1MaxDBH = new JTextField();
  /**Edit box where the cut range 2 maximum DBH is entered*/
  protected JTextField m_jCutRange2MaxDBH = new JTextField();
  /**Edit box where the cut range 3 maximum DBH is entered*/
  protected JTextField m_jCutRange3MaxDBH = new JTextField();
  /**Edit box where the cut range 4 maximum DBH is entered*/
  protected JTextField m_jCutRange4MaxDBH = new JTextField();

  /**Edit box where the cut range 1 cut amount is entered*/
  protected JTextField m_jCutRange1Amt = new JTextField();
  /**Edit box where the cut range 2 cut amount is entered*/
  protected JTextField m_jCutRange2Amt = new JTextField();
  /**Edit box where the cut range 3 cut amount is entered*/
  protected JTextField m_jCutRange3Amt = new JTextField();
  /**Edit box where the cut range 4 cut amount is entered*/
  protected JTextField m_jCutRange4Amt = new JTextField();

  /**Radio button for partial cut*/
  protected JRadioButton m_jPartialCutButton = new JRadioButton("Partial cut");
  /**Radio button for gap cut*/
  protected JRadioButton m_jGapCutButton = new JRadioButton("Gap cut");
  /**Radio button for clear cut*/
  protected JRadioButton m_jClearCutButton = new JRadioButton("Clear cut");

  /**Radio button for percent of density*/
  protected JRadioButton m_jPercentDensityButton = new JRadioButton(
      "% of density");
  /**Radio button for absolute density*/
  protected JRadioButton m_jAbsDensityButton = new JRadioButton(
      "Amt. density (#/ha)");
  /**Radio button for percent of basal area*/
  protected JRadioButton m_jPercentBAButton = new JRadioButton(
      "% of basal area");
  /**Radio button for absolute basal area*/
  protected JRadioButton m_jAbsBAButton = new JRadioButton(
      "Amt. basal area (m2/ha)");

  /**This is where our new harvest event goes.  In the case of an edited
   * event, this will make sure that it is inserted where it used to be.*/
  protected int m_iHarvestIndex = -1;

  /**Help ID string*/
  private String m_sHelpID = "windows.edit_harvest_window";

  /**
   * Constructor.  Creates the window.
   * @param oWindow Object to exchange harvest data with.
   * @throws ModelException If the window cannot be set up correctly.
   */
  public HarvestEdit(EpisodicEventsWindow oWindow) throws ModelException {
    super(oWindow, "Edit Harvest");

    //Get the cell numbers and lengths
    m_iNumXCells = m_oEpisodicEventsWindow.m_iNumHarvestXCells;
    m_iNumYCells = m_oEpisodicEventsWindow.m_iNumHarvestYCells;
    m_fLengthXCells = (float) m_oEpisodicEventsWindow.m_iPlotLengthX /
        (float) m_iNumXCells;
    m_fLengthYCells = (float) m_oEpisodicEventsWindow.m_iPlotLengthY /
        (float) m_iNumYCells;
    m_jColor = EpisodicEventsWindow.HARVEST_COLOR;

    MakeGUI();
    SetUpCharting();
    RefreshChart();

    //Help ID
    m_oEpisodicEventsWindow.GetHelpBroker().enableHelpKey(this.getRootPane(), m_sHelpID, null);
  }

  /**
   * Constructor that displays a harvest event to edit.
   * @param oWindow Object to exchange harvest data with.
   * @param oToDisplay Harvest event to display for editing.
   * @throws ModelException If the window cannot be set up correctly.
   */
  public HarvestEdit(EpisodicEventsWindow oWindow, HarvestData oToDisplay) throws
      ModelException {
    this(oWindow);

    //Get the index of this harvest
    m_iHarvestIndex = oWindow.mp_oHarvestData.indexOf(oToDisplay);

    int iTemp, i;

    //Set timestep
    m_jTimestepEdit.setText(String.valueOf(oToDisplay.GetTimestep())+ "     ");

    //Set the cut type
    iTemp = oToDisplay.GetCutType();
    if (DisturbanceBehaviors.CLEAR_CUT == iTemp) {
      m_jClearCutButton.setSelected(true);
    }
    else if (DisturbanceBehaviors.PARTIAL_CUT == iTemp) {
      m_jPartialCutButton.setSelected(true);
    }
    else if (DisturbanceBehaviors.GAP_CUT == iTemp) {
      m_jGapCutButton.setSelected(true);
    }

    //Set the cut amount type
    iTemp = oToDisplay.GetCutAmountType();
    if (DisturbanceBehaviors.ABSOLUTE_BASAL_AREA == iTemp) {
      m_jAbsBAButton.setSelected(true);
    }
    else if (DisturbanceBehaviors.ABSOLUTE_DENSITY == iTemp) {
      m_jAbsDensityButton.setSelected(true);
    }
    else if (DisturbanceBehaviors.PERCENTAGE_BASAL_AREA == iTemp) {
      m_jPercentBAButton.setSelected(true);
    }
    else if (DisturbanceBehaviors.PERCENTAGE_DENSITY == iTemp) {
      m_jPercentDensityButton.setSelected(true);
    }

    //Set the cut range data
    iTemp = oToDisplay.GetNumberOfCutRanges();
    if (iTemp >= 1) {
      m_jCutRange1MinDBH.setText(String.valueOf(oToDisplay.GetLowerBound(0)));
      m_jCutRange1MaxDBH.setText(String.valueOf(oToDisplay.GetUpperBound(0)));
      m_jCutRange1Amt.setText(String.valueOf(oToDisplay.GetCutAmount(0)));
    }
    if (iTemp >= 2) {
      m_jCutRange2MinDBH.setText(String.valueOf(oToDisplay.GetLowerBound(1)));
      m_jCutRange2MaxDBH.setText(String.valueOf(oToDisplay.GetUpperBound(1)));
      m_jCutRange2Amt.setText(String.valueOf(oToDisplay.GetCutAmount(1)));
    }
    if (iTemp >= 3) {
      m_jCutRange3MinDBH.setText(String.valueOf(oToDisplay.GetLowerBound(2)));
      m_jCutRange3MaxDBH.setText(String.valueOf(oToDisplay.GetUpperBound(2)));
      m_jCutRange3Amt.setText(String.valueOf(oToDisplay.GetCutAmount(2)));
    }
    if (iTemp == 4) {
      m_jCutRange4MinDBH.setText(String.valueOf(oToDisplay.GetLowerBound(3)));
      m_jCutRange4MaxDBH.setText(String.valueOf(oToDisplay.GetUpperBound(3)));
      m_jCutRange4Amt.setText(String.valueOf(oToDisplay.GetCutAmount(3)));
    }

    //Set the species
    iTemp = oToDisplay.GetNumberOfSpecies();
    for (i = 0; i < iTemp; i++) {
      mp_jSpeciesChex[oToDisplay.GetSpecies(i)].setSelected(true);
    }

    //Set the cells
    iTemp = oToDisplay.GetNumberOfCells();
    for (i = 0; i < iTemp; i++) {
      Cell oCell = oToDisplay.GetCell(i);
      m_oDataset.mp_bData[1][oCell.GetX()][oCell.GetY()] = true;
    }

    //Refresh the display
    RefreshChart();
  }

  /**
   * Makes the GUI.  This draws all the window components and places them
   * appropriately.
   */
  protected void MakeGUI() {
    int i;
    TreePopulation oPop = m_oEpisodicEventsWindow.GetTreePopulation();

    //Set fonts
    mp_jSpeciesChex = new JCheckBox[m_iNumSpecies];
    m_jTimestepEdit.setFont(new SortieFont());
    m_jClearCutButton.setFont(new SortieFont());
    m_jPartialCutButton.setFont(new SortieFont());
    m_jGapCutButton.setFont(new SortieFont());
    m_jPercentBAButton.setFont(new SortieFont());
    m_jPercentDensityButton.setFont(new SortieFont());
    m_jAbsBAButton.setFont(new SortieFont());
    m_jAbsDensityButton.setFont(new SortieFont());
    m_jCutRange1Amt.setFont(new SortieFont());
    m_jCutRange2Amt.setFont(new SortieFont());
    m_jCutRange3Amt.setFont(new SortieFont());
    m_jCutRange4Amt.setFont(new SortieFont());
    m_jCutRange1MinDBH.setFont(new SortieFont());
    m_jCutRange2MinDBH.setFont(new SortieFont());
    m_jCutRange3MinDBH.setFont(new SortieFont());
    m_jCutRange4MinDBH.setFont(new SortieFont());
    m_jCutRange1MaxDBH.setFont(new SortieFont());
    m_jCutRange2MaxDBH.setFont(new SortieFont());
    m_jCutRange3MaxDBH.setFont(new SortieFont());
    m_jCutRange4MaxDBH.setFont(new SortieFont());

    //*************************
     // Choices panel
     //*************************
    JPanel jDataPanel = new JPanel();
    jDataPanel.setLayout(new BoxLayout(jDataPanel, BoxLayout.PAGE_AXIS));
    jDataPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

    //Timestep
    JLabel jTimestepLabel = new JLabel("Timestep:");
    jTimestepLabel.setHorizontalAlignment(SwingConstants.LEFT);
    jTimestepLabel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTimestepLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    FlowLayout jFlowLayout = new FlowLayout();
    jFlowLayout.setAlignment(FlowLayout.LEFT);
    JPanel jTimestepPanel = new JPanel(jFlowLayout);
    jTimestepPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jTimestepPanel.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0, java.awt.Color.black));
    jTimestepPanel.add(jTimestepLabel);
    jTimestepPanel.add(m_jTimestepEdit);
    jDataPanel.add(jTimestepPanel);

    //Species
    JPanel jSpeciesPanel = new JPanel();
    jSpeciesPanel.setLayout(new GridLayout(0, 1));
    jSpeciesPanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);

    JLabel jSpeciesLabel = new JLabel("Species applied to:");
    jSpeciesLabel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jSpeciesLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    jSpeciesPanel.add(jSpeciesLabel);

    for (i = 0; i < m_iNumSpecies; i++) {
      mp_jSpeciesChex[i] = new JCheckBox(oPop.GetSpeciesNameFromCode(i).replace(
          '_', ' '));
      mp_jSpeciesChex[i].setFont(new SortieFont());
      mp_jSpeciesChex[i].setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
      jSpeciesPanel.add(mp_jSpeciesChex[i]);
    }
    if (jSpeciesPanel.getPreferredSize().height > 200) {
      JScrollPane jScroller = new JScrollPane(jSpeciesPanel);
      jScroller.setPreferredSize(new Dimension((int)jSpeciesPanel.getPreferredSize().getWidth() + 14,
          200));
      jDataPanel.add(jScroller);
    }
    else {
      jSpeciesPanel.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0, java.awt.Color.black));
      jDataPanel.add(jSpeciesPanel);
    }

    //Cut type
    JPanel jCutTypePanel = new JPanel();
    jCutTypePanel.setLayout(new GridLayout(0, 1));
    JLabel jCutLabel = new JLabel("Cut type:");
    jCutLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    jCutLabel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jCutTypePanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jCutTypePanel.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0, java.awt.Color.black));
    jCutTypePanel.add(jCutLabel);
    //Group our radio buttons and add action listeners for each
    m_jClearCutButton.setActionCommand("Chose clear cut");
    m_jClearCutButton.addActionListener(this);
    m_jClearCutButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);

    m_jGapCutButton.setActionCommand("Chose gap cut");
    m_jGapCutButton.addActionListener(this);
    m_jGapCutButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);

    m_jPartialCutButton.setActionCommand("Chose partial cut");
    m_jPartialCutButton.addActionListener(this);
    m_jGapCutButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);

    ButtonGroup jCutGroup = new ButtonGroup();
    jCutGroup.add(m_jClearCutButton);
    jCutGroup.add(m_jPartialCutButton);
    jCutGroup.add(m_jGapCutButton);
    jCutTypePanel.add(m_jClearCutButton);
    jCutTypePanel.add(m_jGapCutButton);
    jCutTypePanel.add(m_jPartialCutButton);
    jDataPanel.add(jCutTypePanel);

    //Cut amount type
    JPanel jCutAmountTypePanel = new JPanel();
    jCutAmountTypePanel.setLayout(new GridLayout(0, 1));
    jCutAmountTypePanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jCutAmountTypePanel.setBorder(BorderFactory.createMatteBorder(0, 0, 1, 0, java.awt.Color.black));
    JLabel jCutAmtLabel = new JLabel("Cut amount type:");
    jCutAmtLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    jCutAmtLabel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    jCutAmountTypePanel.add(jCutAmtLabel);
    m_jAbsBAButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    m_jAbsDensityButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    m_jPercentBAButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    m_jPercentDensityButton.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    //Group our radio buttons
    ButtonGroup jCutAmtGroup = new ButtonGroup();
    jCutAmtGroup.add(m_jAbsBAButton);
    jCutAmtGroup.add(m_jAbsDensityButton);
    jCutAmtGroup.add(m_jPercentBAButton);
    jCutAmtGroup.add(m_jPercentDensityButton);
    jCutAmountTypePanel.add(m_jPercentDensityButton);
    jCutAmountTypePanel.add(m_jAbsDensityButton);
    jCutAmountTypePanel.add(m_jPercentBAButton);
    jCutAmountTypePanel.add(m_jAbsBAButton);
    jDataPanel.add(jCutAmountTypePanel);

    //Cut ranges
    JPanel jCutRangePanel = new JPanel(new GridLayout(0, 5));
    jCutRangePanel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
    JLabel jTempLabel = new JLabel("Min");
    jTempLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    jCutRangePanel.add(jTempLabel);
    jCutRangePanel.add(m_jCutRange1MinDBH);
    jCutRangePanel.add(m_jCutRange2MinDBH);
    jCutRangePanel.add(m_jCutRange3MinDBH);
    jCutRangePanel.add(m_jCutRange4MinDBH);
    jTempLabel = new JLabel("Max");
    jTempLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    jCutRangePanel.add(jTempLabel);
    jCutRangePanel.add(m_jCutRange1MaxDBH);
    jCutRangePanel.add(m_jCutRange2MaxDBH);
    jCutRangePanel.add(m_jCutRange3MaxDBH);
    jCutRangePanel.add(m_jCutRange4MaxDBH);
    jTempLabel = new JLabel("Amt");
    jTempLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    jCutRangePanel.add(jTempLabel);
    jCutRangePanel.add(m_jCutRange1Amt);
    jCutRangePanel.add(m_jCutRange2Amt);
    jCutRangePanel.add(m_jCutRange3Amt);
    jCutRangePanel.add(m_jCutRange4Amt);
    jTempLabel = new JLabel("Diameter Range(s) to cut:");
    jTempLabel.setFont(new SortieFont(java.awt.Font.BOLD));
    JPanel jTempPanel2 = new JPanel();
    jTempPanel2.setLayout(new BoxLayout(jTempPanel2, BoxLayout.PAGE_AXIS));
    jTempPanel2.add(jTempLabel);
    jTempPanel2.add(jCutRangePanel);
    jTempPanel2.setBorder(BorderFactory.createEmptyBorder(0, 5, 5, 5));
    jDataPanel.add(jTempPanel2);

    //**********************************
     //Chart panel
     //**********************************
    JLabel jInstructionLabel = new JLabel("Click on a map cell to select it for this harvest.  Click on a selected cell to de-select it.");
    jInstructionLabel.setFont(new SortieFont());
    jInstructionLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
    m_jChartPanel.setLayout(new BorderLayout());
    m_jChartPanel.add(jInstructionLabel, BorderLayout.NORTH);
    m_jChartPanel.setBorder(BorderFactory.createMatteBorder(0, 1, 0, 0,
        java.awt.Color.black));

    //**********************************
     // Legend panel
     //**********************************
    JPanel jLegendPanel = MakeTreeLegendPanel(oPop);    

    //Put it all together
    JPanel jContentPanel = new JPanel(new BorderLayout());

    jContentPanel.add(jDataPanel, BorderLayout.WEST);
    jContentPanel.add(m_jChartPanel, BorderLayout.CENTER);
    int iExpectedChartHeight = Math.min(600, m_oEpisodicEventsWindow.m_iPlotLengthY * 3);
    if (jLegendPanel.getPreferredSize().getHeight() > iExpectedChartHeight) {
      JScrollPane jScroller = new JScrollPane(jLegendPanel);
      jScroller.setPreferredSize(new Dimension((int)jLegendPanel.getPreferredSize().getWidth() + 14,
          iExpectedChartHeight));
      jContentPanel.add(jScroller, BorderLayout.EAST);       
    } else {
      jContentPanel.add(jLegendPanel, BorderLayout.EAST);     
    }

    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(new JScrollPane(jContentPanel), BorderLayout.CENTER);
    getContentPane().add(new OKCancelButtonPanel(this,
                                                 m_oEpisodicEventsWindow.
                                                 GetHelpBroker(), m_sHelpID),
                         BorderLayout.SOUTH);

  }

  /**
   * Puts the new harvest into DisturbanceWindow.
   * @throws ModelException if any of the values cannot be translated into
   * numbers, or if the new HarvestData object throws an error during its
   * validation.
   */
  protected void AddNewHarvest() throws ModelException {

    //Create a new HarvestData object
    HarvestData oNewHarvest = new HarvestData();
    int iTemp, i, j;

    //Set the timestep, making sure it's formated correctly
    String sTemp = m_jTimestepEdit.getText().trim();
    if (sTemp.length() == 0) {
      throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA", "Invalid timestep."));
    }
    try {
      iTemp = new Integer(sTemp).intValue();
    }
    catch (java.lang.NumberFormatException oEx) {
      throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA", "Invalid timestep."));
    }
    oNewHarvest.SetTimestep(iTemp);

    //Set the species
    for (i = 0; i < mp_jSpeciesChex.length; i++) {
      if (mp_jSpeciesChex[i].isSelected()) {
        oNewHarvest.AddSpecies(i);
      }
    }

    //Set the cut type
    if (m_jClearCutButton.isSelected()) {
      oNewHarvest.SetCutType(DisturbanceBehaviors.CLEAR_CUT);
    }
    else if (m_jPartialCutButton.isSelected()) {
      oNewHarvest.SetCutType(DisturbanceBehaviors.PARTIAL_CUT);
    }
    else if (m_jGapCutButton.isSelected()) {
      oNewHarvest.SetCutType(DisturbanceBehaviors.GAP_CUT);
    }

    //Set the cut amount type
    if (m_jAbsBAButton.isSelected()) {
      oNewHarvest.SetCutAmountType(DisturbanceBehaviors.ABSOLUTE_BASAL_AREA);
    }
    else if (m_jAbsDensityButton.isSelected()) {
      oNewHarvest.SetCutAmountType(DisturbanceBehaviors.ABSOLUTE_DENSITY);
    }
    else if (m_jPercentBAButton.isSelected()) {
      oNewHarvest.SetCutAmountType(DisturbanceBehaviors.PERCENTAGE_BASAL_AREA);
    }
    else if (m_jPercentDensityButton.isSelected()) {
      oNewHarvest.SetCutAmountType(DisturbanceBehaviors.PERCENTAGE_DENSITY);
    }

    //Now each cut range
    float fMinDBH, fMaxDBH, fAmt;
    try {
      //Cut range 1
      sTemp = m_jCutRange1MinDBH.getText().trim();
      fMinDBH = new Float(m_jCutRange1MinDBH.getText()).floatValue();
      fMaxDBH = new Float(m_jCutRange1MaxDBH.getText()).floatValue();
      fAmt = new Float(m_jCutRange1Amt.getText()).floatValue();
      oNewHarvest.AddCutRange(fMinDBH, fMaxDBH, fAmt);

      //Cut range 2
      fMinDBH = -1;
      fMaxDBH = -1;
      fAmt = -1;
      sTemp = m_jCutRange2MinDBH.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fMinDBH = new Float(sTemp).floatValue();
        }
      }
      sTemp = m_jCutRange2MaxDBH.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fMaxDBH = new Float(sTemp).floatValue();
        }
      }
      sTemp = m_jCutRange2Amt.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fAmt = new Float(sTemp).floatValue();
        }
      }
      if (fMinDBH > 0 || fMaxDBH > 0 || fAmt > 0) {
        oNewHarvest.AddCutRange(fMinDBH, fMaxDBH, fAmt);
      }

      //Cut range 3
      fMinDBH = -1;
      fMaxDBH = -1;
      fAmt = -1;
      sTemp = m_jCutRange3MinDBH.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fMinDBH = new Float(sTemp).floatValue();
        }
      }
      sTemp = m_jCutRange3MaxDBH.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fMaxDBH = new Float(sTemp).floatValue();
        }
      }
      sTemp = m_jCutRange3Amt.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fAmt = new Float(sTemp).floatValue();
        }
      }
      if (fMinDBH > 0 || fMaxDBH > 0 || fAmt > 0) {
        oNewHarvest.AddCutRange(fMinDBH, fMaxDBH, fAmt);
      }

      //Cut range 4
      fMinDBH = -1;
      fMaxDBH = -1;
      fAmt = -1;
      sTemp = m_jCutRange4MinDBH.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fMinDBH = new Float(sTemp).floatValue();
        }
      }
      sTemp = m_jCutRange4MaxDBH.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fMaxDBH = new Float(sTemp).floatValue();
        }
      }
      sTemp = m_jCutRange4Amt.getText();
      if (sTemp != null) {
        sTemp = sTemp.trim();
        if (sTemp.length() > 0) {
          fAmt = new Float(sTemp).floatValue();
        }
      }
      if (fMinDBH > 0 || fMaxDBH > 0 || fAmt > 0) {
        oNewHarvest.AddCutRange(fMinDBH, fMaxDBH, fAmt);
      }
    }
    catch (java.lang.NumberFormatException oEx) {
      throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                               "Invalid cut range value."));
    }

    //Grid cells
    Plot oPlot = m_oEpisodicEventsWindow.m_oDisturbanceBehaviors.m_oManager.
        GetPlot();
    for (i = 0; i < m_oDataset.mp_bData[1].length; i++) {
      for (j = 0; j < m_oDataset.mp_bData[1][i].length; j++) {
        if (m_oDataset.mp_bData[1][i][j]) {
          oNewHarvest.AddCell(i, j, oPlot);
        }
      }
    }

    oNewHarvest.ValidateCut(m_oEpisodicEventsWindow.m_oDisturbanceBehaviors.
                            m_oManager.GetTreePopulation(), m_oEpisodicEventsWindow.m_oDisturbanceBehaviors.m_oManager.GetPlot());

    //Add this new harvest to the list in DisturbanceWindow
    if (m_iHarvestIndex == -1) {
      m_oEpisodicEventsWindow.mp_oHarvestData.add(oNewHarvest);
    }
    else {
      //This was the edit of an existing harvest
      m_oEpisodicEventsWindow.mp_oHarvestData.remove(m_iHarvestIndex);
      m_oEpisodicEventsWindow.mp_oHarvestData.add(m_iHarvestIndex, oNewHarvest);
      //Cause this harvest to be displayed
      m_oEpisodicEventsWindow.DisplayHarvest(oNewHarvest);
    }

    //Update the number of harvest events
    m_oEpisodicEventsWindow.m_jNumHarvestEvents.setText(String.valueOf(
        m_oEpisodicEventsWindow.mp_oHarvestData.size()));
  }

  /**
   * Responds to window events.
   * @param oEvent Event triggering this call.
   */
  public void actionPerformed(ActionEvent oEvent) {
    if (oEvent.getActionCommand().equals("OK")) {
      try {
        AddNewHarvest();
        this.setVisible(false);
        this.dispose();
      }
      catch (ModelException oErr) {
        ErrorGUI oHandler = new ErrorGUI(this);
        oHandler.WriteErrorMessage(oErr);
      }
    }
    else if (oEvent.getActionCommand().equals("Cancel")) {
      //Close this window
      this.setVisible(false);
      this.dispose();
    }
    else if (oEvent.getActionCommand().equals("Chose clear cut") ||
             oEvent.getActionCommand().equals("Chose gap cut")) {

      //Disable cut type amount buttons, and set to percent density
      m_jPercentDensityButton.setSelected(true);
      m_jPercentDensityButton.setEnabled(false);
      m_jPercentBAButton.setEnabled(false);
      m_jAbsBAButton.setEnabled(false);
      m_jAbsDensityButton.setEnabled(false);

      //Set the cut ranges to 0 for range 1 min and 100 for range 1 amt, and
      //the rest disabled
      m_jCutRange1MinDBH.setText("0");
      m_jCutRange1MaxDBH.setText("3000");
      m_jCutRange1MaxDBH.setEnabled(false);
      m_jCutRange1Amt.setText("100");

      m_jCutRange2MinDBH.setText("0");
      m_jCutRange2MaxDBH.setText("0");
      m_jCutRange2Amt.setText("0");
      m_jCutRange2MinDBH.setEnabled(false);
      m_jCutRange2MaxDBH.setEnabled(false);
      m_jCutRange2Amt.setEnabled(false);

      m_jCutRange3MinDBH.setText("0");
      m_jCutRange3MaxDBH.setText("0");
      m_jCutRange3Amt.setText("0");
      m_jCutRange3MinDBH.setEnabled(false);
      m_jCutRange3MaxDBH.setEnabled(false);
      m_jCutRange3Amt.setEnabled(false);

      m_jCutRange4MinDBH.setText("0");
      m_jCutRange4MaxDBH.setText("0");
      m_jCutRange4Amt.setText("0");
      m_jCutRange4MinDBH.setEnabled(false);
      m_jCutRange4MaxDBH.setEnabled(false);
      m_jCutRange4Amt.setEnabled(false);
    }
    else if (oEvent.getActionCommand().equals("Chose partial cut")) {

      //Enable cut type amount buttons
      m_jPercentDensityButton.setEnabled(true);
      m_jPercentBAButton.setEnabled(true);
      m_jAbsBAButton.setEnabled(true);
      m_jAbsDensityButton.setEnabled(true);

      //Enable all the cut ranges
      m_jCutRange1MaxDBH.setEnabled(true);

      m_jCutRange2MinDBH.setEnabled(true);
      m_jCutRange2MaxDBH.setEnabled(true);
      m_jCutRange2Amt.setEnabled(true);

      m_jCutRange3MinDBH.setEnabled(true);
      m_jCutRange3MaxDBH.setEnabled(true);
      m_jCutRange3Amt.setEnabled(true);

      m_jCutRange4MinDBH.setEnabled(true);
      m_jCutRange4MaxDBH.setEnabled(true);
      m_jCutRange4Amt.setEnabled(true);
    }
  }

}
