package javawrapper;

import java.awt.*;
import javax.swing.*;
import java.awt.event.ActionEvent;
import java.util.Vector;
import java.io.File;
import java.io.FileWriter;
import javax.help.CSH;
import uihelpers.SortieFont;
import uihelpers.TextFileFilter;

/**
 * Window used to trigger the parameter editing process.  This displays the
 * list of parameter groups so a user can choose what to display.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 *
 * <br>Edit history:
 * <br>------------------
 * <br>April 28, 2004: Submitted in beta version (LEM)
 * <br>February 4, 2008: Updated to conform to Java 6 compliance (LEM)
 */

public class ParameterEdit
    extends JDialog
    implements java.awt.event.ActionListener {
	
  static final long serialVersionUID = 1;
  
  /** The parent frame over which dialogs are displayed.*/
  protected JFrame m_jParentFrame;
  /** The set of checkboxes which the user uses to choose the parameters to
   * display. */
  protected JCheckBox[] mp_jCheckBoxes;

  /**GUIManager object from which to extract data.*/
  protected GUIManager m_oManager;

  /**System help id*/
  private String m_sHelpID = "windows.parameters_window";

  /**
   * Constructor.
   * @param jFrame Parent frame for these dialogs.
   * @param oManager GUIManager object.
   */
  public ParameterEdit(JFrame jFrame, GUIManager oManager) {
    super(jFrame, "Edit Parameters", true);
    m_oManager = oManager;
    m_jParentFrame = jFrame;

    //Help ID
    oManager.GetHelpBroker().enableHelpKey(this.getRootPane(), m_sHelpID, null);

    getContentPane().setLayout(new BorderLayout());

    JPanel jPanel = new JPanel();
    BoxLayout jLayout = new BoxLayout(jPanel, BoxLayout.PAGE_AXIS);
    jPanel.setLayout(jLayout);

    //A label at the top
    JLabel jLabel = new JLabel("Choose which parameters to display:");
    jLabel.setFont(new SortieFont());
    jPanel.add(jLabel);

    //A checkbox for each possible group
    WorkerBase[] p_oObjects = m_oManager.GetAllObjects();
    int i, iNumAutoDisplay = 0;

    //Count the number of objects to auto-display
    for (i = 0; i < p_oObjects.length; i++) {
      if (p_oObjects[i].GetAutoDisplayParameters()) {
        iNumAutoDisplay++;
      }
    }

    mp_jCheckBoxes = new JCheckBox[iNumAutoDisplay + 1];
    iNumAutoDisplay = 0;
    for (i = 0; i < p_oObjects.length; i++) {
      if (p_oObjects[i].GetAutoDisplayParameters()) {
        mp_jCheckBoxes[iNumAutoDisplay] = new JCheckBox(p_oObjects[i].GetName());
        mp_jCheckBoxes[iNumAutoDisplay].setFont(new java.awt.Font("Dialog", 0,
            14));
        jPanel.add(mp_jCheckBoxes[iNumAutoDisplay]);
        iNumAutoDisplay++;
      }
    }
    mp_jCheckBoxes[mp_jCheckBoxes.length - 1] = new JCheckBox("All");
    mp_jCheckBoxes[mp_jCheckBoxes.length -
        1].setFont(new SortieFont());

    jPanel.add(mp_jCheckBoxes[mp_jCheckBoxes.length - 1]);

    //Build the GUI
    getContentPane().add(jPanel, BorderLayout.CENTER);
    getContentPane().add(new OKCancelButtonPanel(this, m_oManager.GetHelpBroker(),
                                                 m_sHelpID),
                         BorderLayout.PAGE_END);
  }

  /**
   * Responds to button events.  If OK, then the parameter window is
   * constructed and this window is closed.  If Cancel, then this window is
   * closed.
   * @param e ActionEvent object.
   */
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Cancel")) {
      this.setVisible(false);
      this.dispose();
    }
    else if (e.getActionCommand().equals("OK")) {
      this.setVisible(false);
      DisplayParameters oEditor = new DisplayParameters(m_oManager,
          (MainWindow) m_jParentFrame, this);
      oEditor.pack();
      oEditor.setLocationRelativeTo(null);
      oEditor.setVisible(true);
      this.dispose();
    }
  }

}

/**
 * Displays a window with parameter data for the user to edit.  This window
 * displays the data in a set of EnhancedTables.  The user can edit data, copy
 * and paste it between a spreadsheet, or print it off.
 * Copyright: Copyright (c) Charles D. Canham 2003</p>
 * Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */
class DisplayParameters
    extends JDialog
    implements java.awt.event.ActionListener, EnhancedTableWindow {
	
  static final long serialVersionUID = 1;	

  /**GUIManager object containing the data to display.*/
  GUIManager m_oManager;

  /**Parameter chooser window*/
  ParameterEdit m_oChoicesWindow;

  /**Vector holding all of the EnhancedTable objects in this window*/
  Vector<EnhancedTable> mp_oAllTables;

  /**Main window - we will need to communicate directories with it*/
  MainWindow m_oMainWindow;

  /**Help ID*/
  private String m_sHelpID = "windows.parameters_window";

  /**
   * Constructor.  Builds the GUI.
   * @param oWindow Window that's the parent of this dialog.
   * @param oManager GUIManager object.
   * @param oChoicesWindow Which choices are picked
   */
  DisplayParameters(GUIManager oManager, MainWindow oWindow,
                    ParameterEdit oChoicesWindow) {
    super(oWindow, "Edit Parameters", true);
    m_oManager = oManager;
    m_oChoicesWindow = oChoicesWindow;
    m_oMainWindow = oWindow;
    mp_oAllTables = new Vector<EnhancedTable>();

    //Help ID
    oWindow.m_oHelpBroker.enableHelpKey(this.getRootPane(), m_sHelpID, null);

    TreePopulation oPop = m_oManager.GetTreePopulation();
    WorkerBase[] p_oObjects = m_oManager.GetAllObjects();
    int i, j;
    Vector<TableData> p_oData; //each object's vector of table vectors
    TableData oTableData; //one table's worth of data
    boolean bAll = false, bThisObject = false, bAutoDisplay;

    //Is "All" chosen?  If so, we'll load everything.  Also, check for any
    //parameters chosen
    for (i = 0; i < oChoicesWindow.mp_jCheckBoxes.length; i++) {
      if (oChoicesWindow.mp_jCheckBoxes[i].isSelected()) {
        bThisObject = true;
        if (oChoicesWindow.mp_jCheckBoxes[i].getText().equalsIgnoreCase("all")) {
          bAll = true;
        }
      }
    }

    //If there is nothing selected for display, make an error message box
    //instead
    if (!bThisObject) {
      getContentPane().setLayout(new BorderLayout());
      JLabel jLabel = new JLabel("Please choose parameters to display.");
      jLabel.setFont(new SortieFont());
      getContentPane().add(jLabel, BorderLayout.CENTER);
      //Now create OK and Cancel buttons and wrap in a panel
      JButton jOKButton = new JButton("OK");
      jOKButton.setFont(new SortieFont());
      JPanel jButtonPanel = new JPanel();
      FlowLayout jFlowLayout = new FlowLayout();
      jFlowLayout.setAlignment(FlowLayout.CENTER);
      jButtonPanel.setLayout(jFlowLayout);
      jButtonPanel.add(jOKButton);
      jButtonPanel.setBorder(BorderFactory.createMatteBorder(1, 0, 0, 0,
          Color.BLACK));
      jOKButton.setActionCommand("NoParamsOK");
      jOKButton.addActionListener(this);
      getContentPane().add(jButtonPanel, BorderLayout.PAGE_END);
      return;
    }

    //Create a main panel in which to place everything else
    JPanel jMainPanel = new QuickScrollingPanel();
    jMainPanel.setName("Main Panel");
    jMainPanel.setLayout(new BoxLayout(jMainPanel, BoxLayout.Y_AXIS));
    jMainPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

    //Add the data to the window
    for (i = 0; i < p_oObjects.length; i++) {

      //Is this one to be added?
      bThisObject = false;
      bAutoDisplay = p_oObjects[i].GetAutoDisplayParameters();
      if (!bAll) {
        for (j = 0; j < oChoicesWindow.mp_jCheckBoxes.length; j++) {
          if (oChoicesWindow.mp_jCheckBoxes[j].getText().equals(p_oObjects[i].
              GetName()) &&
              oChoicesWindow.mp_jCheckBoxes[j].isSelected()) {
            bThisObject = true;
            break;
          }
        }
      }

      if ( (bAll || bThisObject) && bAutoDisplay) {
        p_oData = p_oObjects[i].FormatDataForDisplay(oPop);

        //Create a panel for this object's data
        JPanel jObjectPanel = new JPanel();
        jObjectPanel.setName("Object Panel");
        jObjectPanel.setLayout(new BoxLayout(jObjectPanel, BoxLayout.Y_AXIS));
        jObjectPanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 0, 0));

        //Add a label with the object's name
        JLabel jLabel = new JLabel(p_oObjects[i].GetName());
        jLabel.setName("Object Name");
        jLabel.setFont(new SortieFont(java.awt.Font.BOLD, 2));
        jLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
        jObjectPanel.add(jLabel);

        //Add a more info button
        if (p_oObjects[i].GetHelpID().length() > 0) {
          JButton jMoreInfoButton = new JButton("More info...");
          jMoreInfoButton.setFont(new SortieFont());
          jMoreInfoButton.addActionListener(new CSH.DisplayHelpFromSource(
              oManager.GetHelpBroker()));
          CSH.setHelpIDString(jMoreInfoButton,
                              p_oObjects[i].GetHelpID() + "_parameters");
          jMoreInfoButton.setAlignmentX(Component.LEFT_ALIGNMENT);
          jObjectPanel.add(Box.createRigidArea(new Dimension(0, 5)));
          jObjectPanel.add(jMoreInfoButton);
        }

        if (null == p_oData) {

          jLabel = new JLabel("No parameters needed.");
          jLabel.setFont(new SortieFont());
          jLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
          jObjectPanel.add(jLabel);

        }
        else {
          //Each vector element represents a table to add.  Process them one at
          //a time
          for (j = 0; j < p_oData.size(); j++) {

            oTableData = p_oData.elementAt(j);

            //Create the table from the data
            EnhancedTable oTable = new EnhancedTable(oTableData.mp_oTableData, oTableData.mp_oHeaderRow, this, p_oObjects[i].GetName());
            mp_oAllTables.add(oTable);

            //Create a table panel with an empty panel in CENTER so that the
            //table won't automatically fill the space
            JPanel jTablePanel = new JPanel(new BorderLayout());
            jTablePanel.setName("Table Panel");
            jTablePanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 0, 0));
            jTablePanel.setAlignmentX(Component.LEFT_ALIGNMENT);
            jTablePanel.add(oTable, BorderLayout.LINE_START);
            jTablePanel.add(new JPanel(), BorderLayout.CENTER);

            //Put the table panel in the object's panel
            jObjectPanel.add(jTablePanel);
          }
        }
        //Add the object's panel to the main panel
        jMainPanel.add(jObjectPanel);
      }
    }

    //Lay out the dialog
    JScrollPane jScroller = new JScrollPane();
    jScroller.setName("Scroller");
    //Make sure window isn't larger than background window
    jScroller.getViewport().add(jMainPanel);
    jScroller.setPreferredSize(new Dimension( (int) jMainPanel.getPreferredSize().
                                             getWidth() + 20,
                                             (int) jScroller.getPreferredSize().
                                             getHeight()));
    int iHeight = (int) jScroller.getPreferredSize().getHeight();
    int iWidth = (int) jScroller.getPreferredSize().getWidth();
    iHeight = java.lang.Math.min(iHeight, oWindow.getHeight() - 100);
    iWidth = java.lang.Math.min(iWidth, oWindow.getWidth() - 100);
    jScroller.setPreferredSize(new Dimension(iWidth, iHeight));

    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(jScroller, BorderLayout.CENTER);
    getContentPane().add(new OKCancelButtonPanel(this,
                                                 m_oMainWindow.m_oHelpBroker,
                                                 m_sHelpID), BorderLayout.SOUTH);

    //Create our menu
    JMenuBar jMenuBar = new JMenuBar();
    jMenuBar.setBackground(SystemColor.control);

    JMenu jMenuFile = new JMenu();
    jMenuFile.setText("File");
    jMenuFile.setFont(new SortieFont());
    jMenuFile.setMnemonic(java.awt.event.KeyEvent.VK_F);
    jMenuFile.setBackground(SystemColor.control);
    JMenuItem jSave = new JMenuItem("Save window as file",
                                    java.awt.event.KeyEvent.VK_S);
    jSave.setAccelerator(KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S,
                                                ActionEvent.CTRL_MASK));
    jSave.setFont(new SortieFont());
    jSave.setActionCommand("Save");
    jSave.addActionListener(this);
    jSave.setBackground(SystemColor.control);
    jMenuFile.add(jSave);

    JMenu jMenuEdit = new JMenu();
    jMenuEdit.setText("Edit");
    jMenuEdit.setFont(new SortieFont());
    jMenuEdit.setMnemonic(java.awt.event.KeyEvent.VK_E);
    jMenuEdit.setBackground(SystemColor.control);
    JMenuItem jCopy = new JMenuItem("Copy", java.awt.event.KeyEvent.VK_C);
    jCopy.setAccelerator(KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_C,
                                                ActionEvent.CTRL_MASK));
    jCopy.setFont(new SortieFont());
    jCopy.setActionCommand("Copy");
    jCopy.addActionListener(this);
    jCopy.setBackground(SystemColor.control);
    JMenuItem jPaste = new JMenuItem("Paste", java.awt.event.KeyEvent.VK_P);
    jPaste.setAccelerator(KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_V,
                                                 ActionEvent.CTRL_MASK));
    jPaste.setFont(new SortieFont());
    jPaste.setActionCommand("Paste");
    jPaste.addActionListener(this);
    jPaste.setBackground(SystemColor.control);
    jMenuEdit.add(jCopy);
    jMenuEdit.add(jPaste);

    jMenuBar.add(jMenuFile);
    jMenuBar.add(jMenuEdit);
    setJMenuBar(jMenuBar);
  }

  /**
   * Responds to button events.  If OK, then the parameter window is
   * constructed and this window is closed.  If Cancel, then this window is
   * closed.
   * @param e ActionEvent object.
   */
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Cancel")) {
      this.setVisible(false);
      this.dispose();
    }
    else if (e.getActionCommand().equals("OK")) {
      try {
        //Make sure that we capture in-process edits
        int i;
        EnhancedTable oTable = null;
        for (i = 0; i < mp_oAllTables.size(); i++) {
          oTable = (EnhancedTable) mp_oAllTables.elementAt(i);
          if (oTable.isEditing()) {
            oTable.getCellEditor(oTable.getEditingRow(),
                                 oTable.getEditingColumn()).stopCellEditing();
          }
        }
        m_oManager.ExtractParameters(this);
        this.setVisible(false);
        this.dispose();
      }
      catch (ModelException oErr) {
        ErrorGUI oHandler = new ErrorGUI(this);
        oHandler.WriteErrorMessage(oErr);
      }
    }
    else if (e.getActionCommand().equals("NoParamsOK")) {
      this.setVisible(false);
      m_oChoicesWindow.setVisible(true);
      this.dispose();
    }
    else if (e.getActionCommand().equals("Copy")) {
      EnhancedTable oTable = GetLastTouchedTable();
      if (oTable != null) {
        oTable.Copy();
      }
    }
    else if (e.getActionCommand().equals("Paste")) {
      EnhancedTable oTable = GetLastTouchedTable();
      if (oTable != null) {
        oTable.Paste();
      }
    }
    else if (e.getActionCommand().equals("Save")) {

      //Allow the user to save a text file
      ModelFileChooser jChooser = new ModelFileChooser();
      jChooser.setFileFilter(new TextFileFilter());

      int iReturnVal = jChooser.showSaveDialog(this);
      if (iReturnVal == JFileChooser.APPROVE_OPTION) {
        //User chose a file - trigger the save
        File oFile = jChooser.getSelectedFile();
        if (oFile.exists()) {
          iReturnVal = JOptionPane.showConfirmDialog(this,
                                              "Do you wish to overwrite the existing file?",
                                              "Model",
                                              JOptionPane.YES_NO_OPTION);
        }
        if (!oFile.exists() || iReturnVal == JOptionPane.YES_OPTION) {
          String sFileName = oFile.getAbsolutePath();
          if (sFileName.endsWith(".txt") == false) {
            sFileName += ".txt";
          }
          SaveCurrentWindow(sFileName);
        }
      }
    }
  }

  /**
   * Saves the data in the current set of tables as a tab-delimited text file.
   * @param sFileName File name to save to.
   */
  public void SaveCurrentWindow(String sFileName) {
    try {
      FileWriter oOut = new FileWriter(sFileName);
      EnhancedTable oTable;
      String sVal, sTitle = "", sTemp = "";
      Object oValue;
      int i, iRow, iCol;

      for (i = 0; i < mp_oAllTables.size(); i++) {
        oTable = (EnhancedTable) mp_oAllTables.elementAt(i);
        sTemp = oTable.GetTitle();
        if (sTemp.equals(sTitle) == false) {
          sTitle = sTemp;
          oOut.write("\n" + sTitle + "\n");
        }

        //Write each row
        for (iRow = 0; iRow < oTable.getRowCount(); iRow++) {
          for (iCol = 0; iCol < oTable.getColumnCount(); iCol++) {
            oValue = oTable.getValueAt(iRow, iCol);
            sVal = String.valueOf(oValue);
            //If a combo box - strip out the text
            if (sVal.startsWith("&&")) {
              sVal = EnhancedTable.GetComboValue(sVal);
            }
            oOut.write(sVal + "\t");
          }
          oOut.write("\n");
        }
      }
      oOut.close();

    }
    catch (java.io.IOException oErr) {
      ModelException oExp = new ModelException(ErrorGUI.BAD_FILE, "JAVA",
         "There was a problem writing the file " + sFileName + ".");
      ErrorGUI oHandler = new ErrorGUI(this);
      oHandler.WriteErrorMessage(oExp);
    }
  }

  /**
   * Sets a table as last touched by the user.
   * @param oTable Table to be set as last touched.
   */
  public void SetLastTouched(EnhancedTable oTable) {
    EnhancedTable oATable;
    int i;
    for (i = 0; i < mp_oAllTables.size(); i++) {
      oATable = (EnhancedTable) mp_oAllTables.elementAt(i);
      oATable.SetLastTouched(false);
    }
    oTable.SetLastTouched(true);
  }

  /**
   * Discovers which EnhancedTable in the parameter editing window has
   * focus.
   * @return EnhancedTable object with focus, or null if none has
   * focus.
   */
  protected EnhancedTable GetLastTouchedTable() {
    EnhancedTable oTable;
    int i;
    for (i = 0; i < mp_oAllTables.size(); i++) {
      oTable = (EnhancedTable) mp_oAllTables.elementAt(i);
      if (oTable.GetLastTouched()) {
        return oTable;
      }
    }

    return null;
  }
}

/**
 * This class creates a JPanel that scrolls at a normal speed through a
 * JScrollPane.  The default scrolling increment for a JPanel is 1 pixel at a
 * time, which is wicked slow.  This boosts that number to make scrolling
 * faster.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */
class QuickScrollingPanel
    extends JPanel
    implements javax.swing.Scrollable {
  
  static final long serialVersionUID = 1;

  /**
   * Get preferred size - this is nothing special
   * @return Dimension Preferred size
   */
  public java.awt.Dimension getPreferredScrollableViewportSize() {
    return getPreferredSize();
  }

  /**
   * Scrolls 75% of the visible pane when paging up or down.
   * @param r Rectangle Ignored
   * @param orientation int Ignored
   * @param direction int Ignored
   * @return int 75% of the viewable height, in pixels
   */
  public int getScrollableBlockIncrement(java.awt.Rectangle r,
                                         int orientation, int direction) {
    return (int)(getVisibleRect().getHeight() * 0.75);
  }

  /**
   * Always false.
   * @return boolean Always false.
   */
  public boolean getScrollableTracksViewportHeight() {
    return false;
  }

  /**
   * Always false.
   * @return boolean Always false.
   */
  public boolean getScrollableTracksViewportWidth() {
    return false;
  }

  /**
   * Causes scrolling to go 20 pixels at a time when scrolling continuously.
   * @param r Rectangle Ignored
   * @param orientation int Ignored
   * @param direction int Ignored
   * @return int 20 pixels
   */
  public int getScrollableUnitIncrement(java.awt.Rectangle r,
                                        int orientation, int direction) {
    return 20;
  }
}
