package javawrapper;

import java.util.Vector;
import java.io.BufferedWriter;
import java.io.FileReader;
import org.xml.sax.Attributes;

/**
 * This holds all data relating to the tree population. It also manages the
 * windows necessary to display and retrieve information about the trees.
 * Copyright: Copyright (c) Charles D. Canham 2003
 * Company: Institute of Ecosystem Studies
 * 
 * @author Lora E. Murphy
 * @version 1.1
 * 
 * <br>
 * Edit history: <br>
 * ------------------ <br>
 * April 28, 2004: Submitted in beta version (LEM) <br>
 * March 8, 2005: Added seedling height classes for initial densities and made
 * version 1.1 (LEM) <br>
 * March 28, 2005: Added new tab-delimited tree map format (LEM) <br>
 * May 2, 2005: Added new seedling max height (LEM) <br>
 * February 4, 2008: Updated to conform to Java 6 compliance (LEM) <br>
 * November 11, 2009: Added height validation (LEM)
 */

public class TreePopulation extends WorkerBase {

  /** Seed tree life history stage - matches definition in C++ */
  public final static int SEED = 0;
  /** Seedling tree life history stage - matches definition in C++ */
  public final static int SEEDLING = 1;
  /** Sapling tree life history stage - matches definition in C++ */
  public final static int SAPLING = 2;
  /** Adult tree life history stage - matches definition in C++ */
  public final static int ADULT = 3;
  /** Stump tree life history stage - matches definition in C++ */
  public final static int STUMP = 4;
  /** Snag tree life history stage - matches definition in C++ */
  public final static int SNAG = 5;
  /** Woody debris tree life history stage - matches definition in C++ */
  public final static int WOODY_DEBRIS = 6;

  private ModelVector
  /**
   * List of species names - ignore XML tags since this data can't use the
   * automated XML read-write system
   */
  mp_sSpeciesNames = new ModelVector("Species names", "", "", 0,
      ModelVector.STRING);

  /**
   * Size class list - the values in this are floats which represent the upper
   * dbh limit of a class. The lower limit of a class is the upper limit of the
   * size class below it. A size class with an upper limit of 0 is for
   * seedlings. Ignore XML tags since this data can't use the automated XML
   * read-write system.
   */
  private ModelVector mp_fSizeClasses = new ModelVector("Size classes", "", "",
      0, ModelVector.FLOAT);

  /**
   * Species-specific - contains the initial density for each class. Each vector
   * bucket contains another vector with the values for each species for that
   * size class. Ignore XML tags since this data can't use the automated XML
   * read-write system.
   */
  private ModelVector mp_fInitialDensities = new ModelVector(
      "Initial Densities", "", "", 0, ModelVector.FLOAT);

  /** Species-specific - the minimum adult dbh value. */
  private ModelVector mp_fMinAdultDbh = new ModelVector("Minimum Adult DBH",
      "tr_minAdultDBH", "tr_madVal", 0, ModelVector.FLOAT, true);

  /** Species-specific - the maximum seedling height value. */
  private ModelVector mp_fMaxSeedlingHeight = new ModelVector(
      "Max Seedling Height (meters)", "tr_maxSeedlingHeight", "tr_mshVal", 0,
      ModelVector.FLOAT, true);

  /** Species-specific - initial densities for seedling height class 1. */
  protected ModelVector mp_fSeedlingClass1Density = new ModelVector(
      "Initial Density (#/ha) - Seedling Height Class 1",
      "tr_seedlingHeight1Density", "tr_sh1dVal", 0, ModelVector.FLOAT, true);

  /** Species-specific - initial densities for seedling height class 2. */
  protected ModelVector mp_fSeedlingClass2Density = new ModelVector(
      "Initial Density (#/ha) - Seedling Height Class 2",
      "tr_seedlingHeight2Density", "tr_sh2dVal", 0, ModelVector.FLOAT, true);

  /** Species-specific - initial densities for seedling height class 3. */
  protected ModelVector mp_fSeedlingClass3Density = new ModelVector(
      "Initial Density (#/ha) - Seedling Height Class 3",
      "tr_seedlingHeight3Density", "tr_sh3dVal", 0, ModelVector.FLOAT, true);

  /** Holds Tree objects to be written to a tree map */
  protected Vector<Tree> mp_oTrees = new Vector<Tree>(0);

  protected Vector<Vector<Vector<String>>>
  /**
   * Tree integer data members. First array index is species, second is type.
   * The vector is a set of strings with the data member names. The position of
   * the String in the vector corresponds to the data member code in the tree
   * map.
   */
  mp_sTreeIntDataMembers,
  /**
   * Tree float data members. First array index is species, second is type. The
   * vector is a set of strings with the data member names. The position of the
   * String in the vector corresponds to the data member code in the tree map.
   */
  mp_sTreeFloatDataMembers,
  /**
   * Tree char data members. First array index is species, second is type. The
   * vector is a set of strings with the data member names. The position of the
   * String in the vector corresponds to the data member code in the tree map.
   */
  mp_sTreeCharDataMembers,
  /**
   * Tree bool data members. First array index is species, second is type. The
   * vector is a set of strings with the data member names. The position of the
   * String in the vector corresponds to the data member code in the tree map.
   */
  mp_sTreeBoolDataMembers;

  /**
   * Index positions for tree integer data members. The current tree map being
   * read may have different data member indexes, and this will help translate
   * to mp_sTreeIntDataMembers. First index is species, second is type. Vector
   * position equals the data member position in the map being read, and vector
   * value is the position in mp_sTreeIntDataMembers of the data member (and
   * thus the tree index).
   */
  protected Vector<Vector<Vector<Integer>>> mp_iTreeIntTransforms,
  /**
   * Index positions for tree float data members. The current tree map being
   * read may have different data member indexes, and this will help translate
   * to mp_sTreeFloatDataMembers. First index is species, second is type. Vector
   * position equals the data member position in the map being read, and vector
   * value is the position in mp_sTreeFloatDataMembers of the data member (and
   * thus the tree index).
   */
  mp_iTreeFloatTransforms,
  /**
   * Index positions for tree char data members. The current tree map being read
   * may have different data member indexes, and this will help translate to
   * mp_sTreeCharDataMembers. First index is species, second is type. Vector
   * position equals the data member position in the map being read, and vector
   * value is the position in mp_sTreeCharDataMembers of the data member (and
   * thus the tree index).
   */
  mp_iTreeCharTransforms,
  /**
   * Index positions for tree bool data members. The current tree map being read
   * may have different data member indexes, and this will help translate to
   * mp_sTreeBoolDataMembers. First index is species, second is type. Vector
   * position equals the data member position in the map being read, and vector
   * value is the position in mp_sTreeBoolDataMembers of the data member (and
   * thus the tree index).
   */
  mp_iTreeBoolTransforms;

  protected ModelFloat m_fInitialSeedlingSize = new ModelFloat(
      "New Seedling Diameter at 10 cm", "tr_seedDiam10Cm");
  /** <Initial seedling diam10. */

  /** Upper limit of seedling 1 initial densities height class, in cm */
  protected ModelFloat m_fSeedlingHeightClass1 = new ModelFloat(0,
      "Seedling Height Class 1 Upper Bound, in cm", "tr_seedlingHeightClass1");

  /** Upper limit of seedling 2 initial densities height class, in cm */
  protected ModelFloat m_fSeedlingHeightClass2 = new ModelFloat(0,
      "Seedling Height Class 2 Upper Bound, in cm", "tr_seedlingHeightClass2");

  /** Filename for text tree map, if present */
  protected ModelString m_sTextTreeMap = new ModelString("",
      "Tree Map To Add As Text", "tr_treemapFile");

  /**
   * When parsing tree maps, this will convert in case the species list is in a
   * different order. The array index is the code to convert; the value in the
   * array at that point is the corresponding old species number.
   */
  protected int[] mp_iSpeciesTransforms;

  /** Counter for tree map species list when parsing tree map files. */
  protected int m_iSpeciesCounter;

  /** When parsing tree map settings, the current tree type. */
  protected int m_iCurrentTreeType,
  /** When parsing tree map settings, the current tree species. */
  m_iCurrentSpecies;

  /**
   * Constructor
   * 
   * @param oManager GUIManager object
   */
  public TreePopulation(GUIManager oManager) {
    super(oManager, "Tree Population", "trees");

    m_sHelpID = "data.trees";

    m_iCurrentSpecies = -1;
    m_iCurrentTreeType = -1;

    mp_sSpeciesNames.SetIsSpeciesSpecific(false);
    mp_fSizeClasses.SetIsSpeciesSpecific(false);

    mp_oAllData.add(mp_fMinAdultDbh);
    mp_oAllData.add(mp_fMaxSeedlingHeight);
    mp_oAllData.add(m_fInitialSeedlingSize);
    mp_oAllData.add(m_fSeedlingHeightClass1);
    mp_oAllData.add(m_fSeedlingHeightClass2);
    mp_oAllData.add(mp_fSeedlingClass1Density);
    mp_oAllData.add(mp_fSeedlingClass2Density);
    mp_oAllData.add(mp_fSeedlingClass3Density);
    mp_oAllData.add(m_sTextTreeMap);
  }

  /**
   * Returns the number of species
   * 
   * @return Number of species
   */
  public int GetNumberOfSpecies() {
    return mp_sSpeciesNames.GetValue().size();
  }

  /**
   * Gets the list of trees.
   * 
   * @return A vector of Tree objects.
   */
  public Vector<Tree> GetTrees() {
    return mp_oTrees;
  }

  /**
   * Get the total number of tree types (life history stages).
   * 
   * @return Total number of tree types
   */
  public static int GetNumberOfTypes() {
    return 7;
  }

  /**
   * When passed the name of a species, will give back the corresponding species
   * number code. If there are spaces in the name, they should be replaced with
   * underscores BEFORE being passed.
   * 
   * @param sName The name of the species.
   * @return The code, or -1 if the name is not recognized.
   */
  public int GetSpeciesCodeFromName(String sName) {
    if (sName == null) {
      return -1;
    }
    int iCode = -1, i;

    for (i = 0; i < mp_sSpeciesNames.GetValue().size(); i++) {
      if (sName.equalsIgnoreCase((String) mp_sSpeciesNames.GetValue()
          .elementAt(i))) {
        iCode = i;
        break;
      }
    }
    return iCode;
  }

  /**
   * Adds the trees from a tab-delimited text tree map file.
   * 
   * @param sFileName The file name of the tree map file.
   * @param oWindow Window to which to display messages
   * @throws ModelException
   *           if:
   *           <ul>
   *           <li>There is an unrecognized species</li>
   *           <li>There is an unrecognized life history stage</li>
   *           <li>There is a negative diameter</li>
   *           <li>There is a negative height</li>
   *           </ul>
   */
  public void AddTabDelimTreeMapFile(javax.swing.JFrame oWindow,
      String sFileName) throws ModelException {
    try {
      Plot oPlot = m_oManager.GetPlot();
      Allometry oAllom = m_oManager.GetAllometry();
      FileReader oIn = new FileReader(sFileName);
      Vector<String> oLine = new Vector<String>(0); // one line of data
      Vector<Tree> oNewTrees = new Vector<Tree>(0); // the trees contained in
                                                    // this file
      String sData = "";
      float fPlotLenX = oPlot.GetPlotXLength(), 
            fPlotLenY = oPlot.GetPlotYLength(), fX, fY, fDiam, 
            fMaxHeight, fHeight = 0;
      int iSpecies, iType, i, j, iNumSpecies = GetNumberOfSpecies(), iNumTypes = GetNumberOfTypes();
      int[][]
      // data member codes for X
      p_iXCodes = new int[iNumSpecies][iNumTypes],
      // data member codes for Y
      p_iYCodes = new int[iNumSpecies][iNumTypes],
      // data member codes for diameter (diam10 for saplings and DBH for all
      // others)
      p_iDiamCodes = new int[iNumSpecies][iNumTypes],
      // data member codes for height
      p_iHeightCodes = new int[iNumSpecies][iNumTypes];

      boolean bWarned = false, bReplaceTrees = false, bUseTree;
      
      // Get the maximum possible height
      fMaxHeight = -1;
      for (i = 0; i < iNumSpecies; i++) {
        fHeight = new Float(oAllom.mp_fMaxCanopyHeight.GetValue().elementAt(i).toString()).floatValue();
        if (fHeight > fMaxHeight)          
          fMaxHeight = fHeight;
      }

      // Ask the user whether they want to add the trees to the parameter file
      // or keep it separate
      Object[] oButtonOptions = { "Add To Parameter File",
          "Keep As Separate File", "Cancel" };
      int iChoice = javax.swing.JOptionPane
          .showOptionDialog(
              oWindow,
              "Would you like to add the trees to the parameter file or keep them\n"
                  + "as a separate file? It is recommended you only keep it separate\n"
                  + "if the file is very large.", "",
              javax.swing.JOptionPane.YES_NO_OPTION,
              javax.swing.JOptionPane.QUESTION_MESSAGE, null, // don't use a
                                                              // custom Icon
              oButtonOptions, // the titles of buttons
              oButtonOptions[0]); // default button title

      if (iChoice == 2) {
        return;
      }
      if (iChoice == 1) {
        m_sTextTreeMap.SetValue(sFileName);
        return;
      }

      // If there are already trees, ask the user if they want to overwrite
      if (mp_oTrees.size() > 0) {
        oButtonOptions = new Object[] { "Replace", "Add To", "Cancel" };
        iChoice = javax.swing.JOptionPane.showOptionDialog(oWindow,
            "What would you like to do with existing tree map data?", "",
            javax.swing.JOptionPane.YES_NO_OPTION,
            javax.swing.JOptionPane.QUESTION_MESSAGE, null, // don't use a
                                                            // custom Icon
            oButtonOptions, // the titles of buttons
            oButtonOptions[0]); // default button title

        if (iChoice == 2) {
          return;
        }
        if (iChoice == 0) {
          bReplaceTrees = true;
        } else {
          bReplaceTrees = false;
        }
      }

      // Get data member codes
      for (i = 0; i < iNumSpecies; i++) {
        for (j = SEEDLING; j <= SNAG; j++) {
          if (j != STUMP) {
            p_iXCodes[i][j] = GetCodeForDataMember(mp_sTreeFloatDataMembers
                .elementAt(i).elementAt(j), "X");
            p_iYCodes[i][j] = GetCodeForDataMember(mp_sTreeFloatDataMembers
                .elementAt(i).elementAt(j), "Y");
            p_iHeightCodes[i][j] = GetCodeForDataMember(
                mp_sTreeFloatDataMembers.elementAt(i).elementAt(j), "Height");
          }
        }
      }

      for (i = 0; i < iNumSpecies; i++) {
        p_iDiamCodes[i][SEEDLING] = GetCodeForDataMember(
            mp_sTreeFloatDataMembers.elementAt(i).elementAt(SEEDLING), "Diam10");
        p_iDiamCodes[i][SAPLING] = GetCodeForDataMember(
            mp_sTreeFloatDataMembers.elementAt(i).elementAt(SAPLING), "DBH");
        p_iDiamCodes[i][ADULT] = GetCodeForDataMember(mp_sTreeFloatDataMembers
            .elementAt(i).elementAt(ADULT), "DBH");
        p_iDiamCodes[i][SNAG] = GetCodeForDataMember(mp_sTreeFloatDataMembers
            .elementAt(i).elementAt(SNAG), "DBH");
      }

      // Skip the first line - just column headers
      ModelFileFunctions.SkipLine(oIn);

      oLine = ModelFileFunctions.ReadLine(oIn);
      while (oLine.size() > 0) {
        i = 0;
        bUseTree = true;

        // Verify that all data is present
        if (oLine.size() != 6) {
          throw (new ModelException(ErrorGUI.BAD_FILE, "JAVA", "The file \""
              + sFileName + "\" has an invalid format."));
        }

        fX = 0;
        fY = 0;
        try {
          fX = new Float((String) oLine.elementAt(i)).floatValue();
          i++;
          fY = new Float((String) oLine.elementAt(i)).floatValue();
          i++;
        } catch (java.lang.NumberFormatException oErr) {
          throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "The tree map"
              + " file appears to be in the wrong format."
              + "  X and Y are expected as the first two " + "columns."));
        }

        // If a tree is outside the plot, warn the user
        if (fX >= fPlotLenX || fX < 0 || fY >= fPlotLenY || fY < 0) {
          if (bWarned == false) {
            iChoice = javax.swing.JOptionPane.showConfirmDialog(oWindow,
                "One or more trees is outside the plot.\n"
                    + " These trees will be discarded.  Continue?",
                "SORTIE-ND", javax.swing.JOptionPane.YES_NO_OPTION);
            if (iChoice == javax.swing.JOptionPane.NO_OPTION) {
              javax.swing.JOptionPane.showMessageDialog(oWindow,
                  "No trees were added.");
              return;
            }
            bWarned = true;
          }
          bUseTree = false;
        }

        if (bUseTree) {

          // Get species
          sData = ((String) oLine.elementAt(i)).replace(' ', '_');
          i++;
          iSpecies = GetSpeciesCodeFromName(sData);
          if (iSpecies == -1) {
            throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "The file \""
                + sFileName + "\" has an invalid species: \"" + sData + "\"."));
          }

          // Get tree type
          sData = (String) oLine.elementAt(i);
          i++;
          iType = GetTypeCodeFromName(sData);
          if (iType < SEEDLING || iType > SNAG || iType == STUMP) {
            throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "The file \""
                + sFileName + "\" has an invalid tree type: \"" + sData + "\"."));
          }

          // Get diameter
          fDiam = new Float((String) oLine.elementAt(i)).floatValue();
          i++;
          if (fDiam <= 0) {
            throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "The file \""
                + sFileName + "\" has an invalid tree diameter: \"" + fDiam
                + "\"."));
          }

          // Get the height
          fHeight = new Float((String) oLine.elementAt(i)).floatValue();
          if (fHeight < 0) {
            throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "The file \""
                + sFileName + "\" has an invalid tree height: \"" + fHeight
                + "\"."));
          }
          if (fHeight > fMaxHeight) {
            throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", 
                "The following tree has a height above the maximum canopy " +
                "height: X = " + fX + ", Y = " + fY + ", Diam = " + fDiam +
                ", Species = " + GetSpeciesNameFromCode(iSpecies) + ", Type = "
                + iType + ", Height = " + fHeight + "."));
          }

          Tree oTree = new Tree(iType, iSpecies, mp_sTreeFloatDataMembers
              .elementAt(iSpecies).elementAt(iType).size(),
              mp_sTreeIntDataMembers.elementAt(iSpecies).elementAt(iType)
                  .size(), mp_sTreeCharDataMembers.elementAt(iSpecies)
                  .elementAt(iType).size(), mp_sTreeBoolDataMembers.elementAt(
                  iSpecies).elementAt(iType).size(), this);
          oTree.SetValue(p_iXCodes[iSpecies][iType], new Float(fX));
          oTree.SetValue(p_iYCodes[iSpecies][iType], new Float(fY));
          oTree.SetValue(p_iDiamCodes[iSpecies][iType], new Float(fDiam));
          if (java.lang.Math.abs(fHeight) > 0.01) {
            oTree.SetValue(p_iHeightCodes[iSpecies][iType], new Float(fHeight));
          }

          oNewTrees.add(oTree);
        }

        oLine = ModelFileFunctions.ReadLine(oIn);
      }
      oIn.close();

      // Add the new trees
      if (bReplaceTrees) {
        mp_oTrees = null;
        mp_oTrees = new Vector<Tree>(0);
      }

      for (i = 0; i < oNewTrees.size(); i++) {
        mp_oTrees.add(oNewTrees.elementAt(i));
      }
      oNewTrees = null;
    } catch (java.io.IOException e) {
      throw (new ModelException(ErrorGUI.BAD_FILE, "WriteTreeMap", sFileName));
    }
  }

  /**
   * Returns the number of size classes.
   * 
   * @return Number of size classes
   */
  public int GetNumberOfSizeClasses() {
    return mp_fSizeClasses.GetValue().size();
  }

  /**
   * Returns the size class at a specific index.
   * 
   * @param iIndex Index of size class to return.
   * @return Upper limit of size class.
   */
  public Float GetSizeClass(int iIndex) {
    return (Float) mp_fSizeClasses.GetValue().elementAt(iIndex);
  }

  /**
   * Sets the value of the diameter at 10 cm for new seedlings.
   * 
   * @param fVal Diameter at 10 cm, in cm.
   * @throws ModelException if value passed is less than or equal to 0.
   */
  public void SetInitialSeedlingSize(float fVal) throws ModelException {
    ModelFloat oTest = new ModelFloat(fVal, m_fInitialSeedlingSize
        .GetDescriptor(), "");
    MakeSureGreaterThan(oTest, 0);
    m_fInitialSeedlingSize.SetValue(fVal);
  }

  /**
   * Sets the size classes. The size class vector will be sized to match the
   * array of Floats passed. The values in the array represent the upper dbh
   * limits of each size class. A value of 0 is assumed to be for seedlings.
   * 
   * @param p_fVals A Float array of size class values.
   * @throws ModelException if any size class value is less than 0.
   */
  public void SetSizeClasses(Float[] p_fVals) throws ModelException {
    MakeSureAllNonNegative(p_fVals, mp_fSizeClasses.GetDescriptor());
    SetVectorValues(mp_fSizeClasses, p_fVals);

    // Reset all initial densities
    ModelVector oDensity;
    int i, j, iNumSpecies = GetNumberOfSpecies();

    for (i = 0; i < mp_oAllData.size(); i++) {
      ModelData oData = (ModelData) mp_oAllData.elementAt(i);
      // Find all that are initial density but not seedling height class
      // densities and remove them
      if (oData.GetDescriptor().indexOf("Initial Density") > -1
          && oData.GetDescriptor().indexOf("Height") == -1) {
        mp_oAllData.remove(i);
        i--;
      }
    }

    mp_fInitialDensities.GetValue().clear();
    for (i = 0; i < mp_fSizeClasses.GetValue().size(); i++) {
      if (i == 0) {
        Float fVal = (Float) mp_fSizeClasses.GetValue().elementAt(i);
        if (fVal.floatValue() == 0) {
          oDensity = new ModelVector("Initial Density Seedlings (#/ha)", "",
              "", iNumSpecies, ModelVector.FLOAT, true);
        } else {
          oDensity = new ModelVector("Initial Density (#/ha) - 0-"
              + String.valueOf(mp_fSizeClasses.GetValue().elementAt(i))
              + " cm DBH", "", "", iNumSpecies, ModelVector.FLOAT, true);
        }
      } else {
        oDensity = new ModelVector("Initial Density (#/ha) - "
            + String.valueOf(mp_fSizeClasses.GetValue().elementAt(i - 1)) + "-"
            + String.valueOf(mp_fSizeClasses.GetValue().elementAt(i)) + " cm",
            "", "", iNumSpecies, ModelVector.FLOAT, true);
      }

      // Initialize to all zeroes so it will look nice in the parameter window
      for (j = 0; j < iNumSpecies; j++) {
        oDensity.GetValue().add(new Float(0));
      }
      mp_fInitialDensities.GetValue().add(i, oDensity);
      mp_oAllData.add(oDensity);

    }
  }

  /**
   * Sets the minimum adult dbhs. There must be one for each species.
   * 
   * @param p_fVals A Float array of values.
   * @throws ModelException if any value is less than 0.
   */
  public void SetMinimumAdultDbh(Float[] p_fVals) throws ModelException {
    MakeSureAllPositive(p_fVals, mp_fMinAdultDbh.GetDescriptor());
    SetVectorValues(mp_fMinAdultDbh, p_fVals);
  }

  /**
   * Sets the species name list. The species vector will be sized to match the
   * array of Strings passed. All spaces in the species names are replaced with
   * underscores.
   * 
   * If there are species currently defined, this calls ChangeOfSpecies() for
   * all objects.
   * 
   * This calls the GUIManager::DoSetup() method so that behavior groups can
   * adjust for the new species list if species were already defined.
   * 
   * This function also declares certain arrays, since this is the first
   * opportunity to know how many species there are.
   * 
   * @param p_sVals A String array of species names.
   * @throws ModelException if there is a problem setting the species names.
   */
  public void SetSpeciesNames(String[] p_sVals) throws ModelException {
    int i;
    boolean bSpeciesExisted = false;

    // Replace the spaces in the names with underscores
    for (i = 0; i < p_sVals.length; i++) {
      p_sVals[i] = p_sVals[i].replace(' ', '_');
    }

    if (mp_sSpeciesNames.GetValue().size() != 0) {

      // This is not the first time this has been called
      bSpeciesExisted = true;

      // Null the arrays to reset them before they are redeclared
      mp_sTreeIntDataMembers = null;
      mp_sTreeFloatDataMembers = null;
      mp_sTreeCharDataMembers = null;
      mp_sTreeBoolDataMembers = null;

      mp_iTreeBoolTransforms = null;
      mp_iTreeIntTransforms = null;
      mp_iTreeCharTransforms = null;
      mp_iTreeFloatTransforms = null;

      // Prepare to call ChangeOfSpecies for all objects
      int[] p_iIndexer = new int[p_sVals.length];
      int iNumOldSpecies = mp_sSpeciesNames.GetValue().size(), j;

      // Set the values in the indexer
      for (i = 0; i < p_iIndexer.length; i++) {
        p_iIndexer[i] = -1;
        // Does this species already exist? If so, grab its index
        for (j = 0; j < mp_sSpeciesNames.GetValue().size(); j++) {
          if (p_sVals[i].equals((String) mp_sSpeciesNames.GetValue().elementAt(
              j))) {
            p_iIndexer[i] = j;
          }
        }
      }

      // Now get the index translator the other way
      int[] p_iOldToNewTranslator = new int[mp_sSpeciesNames.GetValue().size()];
      for (i = 0; i < p_iOldToNewTranslator.length; i++) {
        p_iOldToNewTranslator[i] = -1;

        // Does this species still exist in the new list? If so, grab its index
        for (j = 0; j < p_sVals.length; j++) {
          if (p_sVals[j].equals((String) mp_sSpeciesNames.GetValue().elementAt(
              i))) {
            p_iOldToNewTranslator[i] = j;
          }
        }
      }

      // Trigger the change of species
      m_oManager.ChangeOfSpecies(iNumOldSpecies, p_iIndexer, p_sVals);
    }

    int iNumSpecies, iNumTypes, j;

    mp_sSpeciesNames.GetValue().clear();

    for (i = 0; i < p_sVals.length; i++) {
      mp_sSpeciesNames.GetValue().add(i, new String(p_sVals[i]));
    }

    // Declare the tree data member arrays
    iNumSpecies = GetNumberOfSpecies();
    iNumTypes = GetNumberOfTypes();
    mp_sTreeIntDataMembers = new Vector<Vector<Vector<String>>>(iNumSpecies);
    mp_sTreeFloatDataMembers = new Vector<Vector<Vector<String>>>(iNumSpecies);
    mp_sTreeCharDataMembers = new Vector<Vector<Vector<String>>>(iNumSpecies);
    mp_sTreeBoolDataMembers = new Vector<Vector<Vector<String>>>(iNumSpecies);

    mp_iTreeBoolTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);
    mp_iTreeIntTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);
    mp_iTreeCharTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);
    mp_iTreeFloatTransforms = new Vector<Vector<Vector<Integer>>>(iNumSpecies);

    for (i = 0; i < iNumSpecies; i++) {
      mp_sTreeIntDataMembers.add(i, new Vector<Vector<String>>(iNumTypes));
      mp_sTreeFloatDataMembers.add(i, new Vector<Vector<String>>(iNumTypes));
      mp_sTreeCharDataMembers.add(i, new Vector<Vector<String>>(iNumTypes));
      mp_sTreeBoolDataMembers.add(i, new Vector<Vector<String>>(iNumTypes));

      mp_iTreeBoolTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
      mp_iTreeIntTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
      mp_iTreeCharTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
      mp_iTreeFloatTransforms.add(i, new Vector<Vector<Integer>>(iNumTypes));
      for (j = 0; j < iNumTypes; j++) {
        mp_sTreeIntDataMembers.elementAt(i).add(j, new Vector<String>(0));
        mp_sTreeFloatDataMembers.elementAt(i).add(j, new Vector<String>(0));
        mp_sTreeCharDataMembers.elementAt(i).add(j, new Vector<String>(0));
        mp_sTreeBoolDataMembers.elementAt(i).add(j, new Vector<String>(0));

        mp_iTreeBoolTransforms.elementAt(i).add(j, new Vector<Integer>(0));
        mp_iTreeIntTransforms.elementAt(i).add(j, new Vector<Integer>(0));
        mp_iTreeCharTransforms.elementAt(i).add(j, new Vector<Integer>(0));
        mp_iTreeFloatTransforms.elementAt(i).add(j, new Vector<Integer>(0));
      }
    }

    mp_iSpeciesTransforms = new int[iNumSpecies];
    for (i = 0; i < iNumSpecies; i++) {
      mp_iSpeciesTransforms[i] = i;
    }

    // Now trigger setup if species existed before
    if (bSpeciesExisted) {
      m_oManager.DoSetup();
    }
  }

  /**
   * Performs setup. Causes the max seedling height to default to 1.35 for all
   * species.
   * 
   * @param oPop TreePopulation This.
   */
  public void DoSetup(TreePopulation oPop) {

    int iNumSpecies = GetNumberOfSpecies(), i;
    if (mp_fMaxSeedlingHeight.GetValue().size() == 0) {
      for (i = 0; i < iNumSpecies; i++) {
        mp_fMaxSeedlingHeight.GetValue().add(new Float(1.35));
      }
    }
  }

  /**
   * Overridden from WorkerBase to take care of other arrays.
   * 
   * @param iOldNumSpecies says how many species there used to be.
   * @param p_iIndexer is an array, sized to the new number of species. For 
   * each bucket (representing the index number of a species on the new list), 
   * the value is either the index of that same species in the old species list,
   *  or -1 if the species is new.
   * @param p_sNewSpecies The new species list.
   * @throws ModelException if anything goes wrong.
   */
  public void ChangeOfSpecies(int iOldNumSpecies, int[] p_iIndexer,
      String[] p_sNewSpecies) throws ModelException {

    super.ChangeOfSpecies(iOldNumSpecies, p_iIndexer, p_sNewSpecies);

    // Delete all existing tree map trees
    mp_oTrees.clear();
  }

  /**
   * Sets the initial densities for a given species. Any species number greater
   * than zero is assumed to be valid.
   * 
   * @param iSizeClass size class number
   * @param p_fVals set of initial densities for each species
   * @throws ModelException if the size class number is less than 0 or greater 
   * than the number of size classes defined.
   */
  public void SetInitialDensities(int iSizeClass, Float[] p_fVals)
      throws ModelException {
    if (iSizeClass >= mp_fSizeClasses.GetValue().size() || iSizeClass < 0) {
      throw (new ModelException(ErrorGUI.ILLEGAL_OP, "JAVA",
          "Invalid size class number."));
    }

    // Set up the new vector of initial densities
    ModelVector p_fNewVector = (ModelVector) mp_fInitialDensities.GetValue()
        .elementAt(iSizeClass);
    SetVectorValues(p_fNewVector, p_fVals);
  }

  /**
   * For a given species code, returns its name. If this name is for display, a
   * replace should be done on underscores to spaces.
   * 
   * @param iSpecies The species code.
   * @return The species name.
   */
  public String GetSpeciesNameFromCode(int iSpecies) {
    if (iSpecies < 0 || iSpecies >= mp_sSpeciesNames.GetValue().size()) {
      return "";
    }
    String sName = (String) mp_sSpeciesNames.GetValue().elementAt(iSpecies);
    return sName;
  }

  /**
   * For a given type code, returns its name.
   * 
   * @param iType The type code.
   * @return The type name, or an empty string if the code isn't recognized.
   */
  public static String GetTypeNameFromCode(int iType) {
    if (iType < 0 || iType >= GetNumberOfTypes()) {
      return "";
    }

    if (SEED == iType) {
      return "Seed";
    } else if (SEEDLING == iType) {
      return "Seedling";
    } else if (SAPLING == iType) {
      return "Sapling";
    } else if (ADULT == iType) {
      return "Adult";
    } else if (STUMP == iType) {
      return "Stump";
    } else if (SNAG == iType) {
      return "Snag";
    } else {
      return "Woody debris";
    }
  }

  /**
   * For a given type name, returns its code.
   * 
   * @param sName The type name.
   * @return The type code, or -1 if the name is invalid.
   */
  public static int GetTypeCodeFromName(String sName) {
    if (sName == null) {
      return -1;
    }
    if (sName.equals("Seed")) {
      return SEED;
    } else if (sName.equalsIgnoreCase("seedling")) {
      return SEEDLING;
    } else if (sName.equalsIgnoreCase("Sapling")) {
      return SAPLING;
    } else if (sName.equalsIgnoreCase("Adult")) {
      return ADULT;
    } else if (sName.equalsIgnoreCase("Stump")) {
      return STUMP;
    } else if (sName.equalsIgnoreCase("Snag")) {
      return SNAG;
    } else if (sName.equalsIgnoreCase("Woody debris")) {
      return WOODY_DEBRIS;
    } else {
      return -1;
    }
  }

  /**
   * Gets the new seedling diameter at 10 cm value, in cm.
   * 
   * @return The new seedling diameter at 10 cm value, in cm.
   */
  public float GetNewSeedlingDiam10() {
    return m_fInitialSeedlingSize.GetValue();
  }

  /**
   * Writes the settings to XML for the parameter file for the tree population.
   * 
   * @param out Out stream for the file.
   * @param oPop TreePopulation object.
   * @throws ModelException If something goes wrong with the writing.
   */
  public void WriteXML(BufferedWriter out, TreePopulation oPop)
      throws ModelException {
    try {
      ModelVector vData;
      String sData;
      float[] p_fSpeciesData;
      Float fData;
      boolean bHasContents;
      int i, j;

      // Write the opening tag
      out.write("<trees>");

      // Create the species list
      out.write("<tr_speciesList>");
      for (i = 0; i < mp_sSpeciesNames.GetValue().size(); i++) {
        sData = (String) mp_sSpeciesNames.GetValue().elementAt(i);
        out.write("<tr_species speciesName=\"");
        out.write(sData.replace(' ', '_'));
        out.write("\"/>");
      }
      out.write("</tr_speciesList>");

      // Write size classes, if present
      if (mp_fSizeClasses.GetValue().size() > 0) {
        // Open tag
        out.write("<tr_sizeClasses>");

        for (i = 0; i < mp_fSizeClasses.GetValue().size(); i++) {

          fData = (Float) mp_fSizeClasses.GetValue().elementAt(i);
          out.write("<tr_sizeClass sizeKey=\"");

          if (fData.floatValue() == 0) {
            // If the size class is 0, treat it as seedling
            out.write("Seedling");
          } else {
            out.write("s");
            out.write(String.valueOf(fData.floatValue()));
          }

          out.write("\"/>");
        }

        // Closing size classes tag
        out.write("</tr_sizeClasses>");
      }

      // Write initial densities, if present
      if (mp_fInitialDensities.GetValue().size() > 0) {
        // Opening tag
        out.write("<tr_initialDensities>");

        p_fSpeciesData = new float[mp_fSizeClasses.GetValue().size()];

        for (i = 0; i < mp_sSpeciesNames.GetValue().size(); i++) {

          // The data is organized as size class first, then species; we want
          // to re-organize that into the data for one species
          for (j = 0; j < mp_fSizeClasses.GetValue().size(); j++) {
            p_fSpeciesData[j] = 0;
          }
          for (j = 0; j < mp_fSizeClasses.GetValue().size(); j++) {
            if (null != mp_fInitialDensities.GetValue().elementAt(j)) {
              vData = (ModelVector) mp_fInitialDensities.GetValue()
                  .elementAt(j);
              if (vData.GetValue().size() > 0) {
                fData = (Float) vData.GetValue().elementAt(i);
                if (fData != null) {
                  p_fSpeciesData[j] = fData.floatValue();
                }
              }
            }
          }

          // Now our array should hold the values for that species. Is there
          // anything in it?
          bHasContents = false;
          for (j = 0; j < mp_fSizeClasses.GetValue().size(); j++) {
            if (p_fSpeciesData[j] > 0) {
              bHasContents = true;

              // If there was anything, write it
            }
          }
          if (bHasContents) {
            // Write opening tag
            out.write("<tr_idVals whatSpecies=\"");
            sData = (String) mp_sSpeciesNames.GetValue().elementAt(i);
            out.write(sData.replace(' ', '_'));
            out.write("\">");

            for (j = 0; j < p_fSpeciesData.length; j++) {
              if (p_fSpeciesData[j] > 0) {
                fData = (Float) mp_fSizeClasses.GetValue().elementAt(j);
                if (fData.floatValue() == 0) {
                  out.write("<tr_initialDensity sizeClass=\"Seedling\">");
                } else {
                  out.write("<tr_initialDensity sizeClass=\"s");
                  out.write(String.valueOf(fData.floatValue()));
                  out.write("\">");
                }
                out.write(String.valueOf(p_fSpeciesData[j]));
                out.write("</tr_initialDensity>");

              }
            }

            // Write closing tag
            out.write("</tr_idVals>");

          }
        }
        // Closing tag
        out.write("</tr_initialDensities>");
      }

      // Write out the seedling height classes and densities, if used
      if (m_fSeedlingHeightClass1.GetValue() > 0
          || m_fSeedlingHeightClass2.GetValue() > 0) {
        WriteDataToFile(out, m_fSeedlingHeightClass1);
        WriteDataToFile(out, m_fSeedlingHeightClass2);
        WriteSpeciesSpecificValue(out, mp_fSeedlingClass1Density, this);
        WriteSpeciesSpecificValue(out, mp_fSeedlingClass2Density, this);
        WriteSpeciesSpecificValue(out, mp_fSeedlingClass3Density, this);
      }

      // Write seedling initial values
      WriteDataToFile(out, m_fInitialSeedlingSize);

      // Write minimum adult dbh
      WriteSpeciesSpecificValue(out, mp_fMinAdultDbh, this);

      // Write the max seedling height
      WriteSpeciesSpecificValue(out, mp_fMaxSeedlingHeight, this);

      // Write a tree map, if it exists
      WriteTreeMap(out);

      if (m_sTextTreeMap.GetValue().trim().length() > 0)
        WriteDataToFile(out, m_sTextTreeMap);

      // Write the closing tag
      out.write("</trees>");
    } catch (java.io.IOException e) {
      throw (new ModelException(ErrorGUI.BAD_FILE, "JAVA",
          "There was a problem writing the parameter file."));
    }
  }

  /**
   * This makes sure all data is valid and can be used to run the model.
   * 
   * @throws ModelException in any of the following cases:
   *           <ul>
   *           <li> Any of the species-specific vectors is not sized equal to
   *           the number of species.</li>
   *           <li> There are no species defined.</li>
   *           <li> The initial seedling size is not greater than zero.</li>
   *           <li> There is any value less than or equal to zero in any of the
   *           following vectors:
   *           <ul>
   *           <li>mp_fCanradToDbhRatio</li>
   *           <li>mp_fDiam10ToDbhConversion</li>
   *           <li>mp_fSlopeOfAsympCanDepth</li>
   *           <li>mp_fInitialDensities</li>
   *           <li>mp_fMaxCanopyHeight</li>
   *           <li>mp_fMinAdultDbh</li>
   *           <li>mp_fSizeClasses</li>
   *           <li>mp_fSlopeOfAsymptoticHeight</li>
   *           </ul>
   *           </li>
   *           <li>Either of the seedling initial densities height classes is
   *           less than 0 or greater than 135.</li>
   *           <li>The first seedling height class upper bound is greater than
   *           the second one.</li>
   *           <li>Any of the seedling initial densities is less than 0.</li>
   *           </ul>
   * @param oPop
   *          Not used.
   */
  public void ValidateData(TreePopulation oPop) throws ModelException {
    ModelVector vData;
    int iNumSpecies = mp_sSpeciesNames.GetValue().size(), iSize, i;
    boolean bUsed;

    if (iNumSpecies == 0) {
      throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
          "There must be some species defined."));
    }

    MakeSureGreaterThan(m_fInitialSeedlingSize, 0);

    // Validate the size of the species-specific vectors
    MakeSureRightSize(mp_fMinAdultDbh, iNumSpecies);

    // Make sure that the values of the vectors are positive
    for (i = 0; i < mp_fInitialDensities.GetValue().size(); i++) {
      vData = (ModelVector) mp_fInitialDensities.GetValue().elementAt(i);
      MakeSureAllNonNegative(vData);
    }
    MakeSureAllNonNegative(mp_fMinAdultDbh);
    MakeSureAllNonNegative(mp_fSizeClasses);

    // LEM 03/08/05 - validation of tree height class parameters
    MakeSureGreaterThanEqualTo(m_fSeedlingHeightClass1, 0);
    MakeSureGreaterThanEqualTo(m_fSeedlingHeightClass2, 0);
    if (m_fSeedlingHeightClass1.GetValue() >= 135) {
      throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "Values for \""
          + m_fSeedlingHeightClass1.GetDescriptor()
          + "\" must be less than 135."));
    }

    if (m_fSeedlingHeightClass2.GetValue() >= 135) {
      throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "Values for \""
          + m_fSeedlingHeightClass2.GetDescriptor()
          + "\" must be less than 135."));
    }
    // Make sure the seedling height classes do not overlap
    if (m_fSeedlingHeightClass1.GetValue() > m_fSeedlingHeightClass2.GetValue()) {
      throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA", "The value for \""
          + m_fSeedlingHeightClass2.GetDescriptor()
          + "\" must be greater than the value for \""
          + m_fSeedlingHeightClass1.GetDescriptor() + "\"."));
    }

    MakeSureAllNonNegative(mp_fSeedlingClass1Density);
    MakeSureAllNonNegative(mp_fSeedlingClass2Density);
    MakeSureAllNonNegative(mp_fSeedlingClass3Density);

    // LEM 02/14/06 - if there's something in the densities, make sure there's
    // something in the size class boundaries
    bUsed = false;
    iSize = mp_fSeedlingClass1Density.GetValue().size();
    for (i = 0; i < iSize; i++) {
      if (((Float) mp_fSeedlingClass1Density.GetValue().elementAt(i))
          .floatValue() > 0) {
        bUsed = true;
        break;
      }
    }
    if (false == bUsed) {
      iSize = mp_fSeedlingClass2Density.GetValue().size();
      for (i = 0; i < iSize; i++) {
        if (((Float) mp_fSeedlingClass2Density.GetValue().elementAt(i))
            .floatValue() > 0) {
          bUsed = true;
          break;
        }
      }
    }
    if (false == bUsed) {
      iSize = mp_fSeedlingClass3Density.GetValue().size();
      for (i = 0; i < iSize; i++) {
        if (((Float) mp_fSeedlingClass3Density.GetValue().elementAt(i))
            .floatValue() > 0) {
          bUsed = true;
          break;
        }
      }
    }
    if (bUsed && m_fSeedlingHeightClass1.GetValue() == 0
        && m_fSeedlingHeightClass2.GetValue() == 0) {
      throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
          "You must define the values in \""
              + m_fSeedlingHeightClass1.GetDescriptor() + "\" and \""
              + m_fSeedlingHeightClass2.GetDescriptor()
              + "\" if you want to specify " + "seedling densities."));
    }
    
    // Validate tree map trees and dump any without X, Y, and diameter or height
    Tree oTree;
    int iXCode, iYCode, iDiamCode;
    boolean bOK;

    for (i = 0; i < mp_oTrees.size(); i++) {
      oTree = (Tree) mp_oTrees.elementAt(i);
      iXCode = GetFloatDataCode("X", oTree.m_iSpecies, oTree.m_iType);
      iYCode = GetFloatDataCode("Y", oTree.m_iSpecies, oTree.m_iType);
      iDiamCode = -1;
      bOK = false;

      if (iXCode != -1 && oTree.mp_fFloats[iXCode] != null && iYCode != -1
          && oTree.mp_fFloats[iYCode] != null) {

        if (oTree.m_iType > TreePopulation.SAPLING) {
          iDiamCode = GetFloatDataCode("DBH", oTree.m_iSpecies, oTree.m_iType);
          if (iDiamCode != -1 && oTree.mp_fFloats[iDiamCode] != null)
            bOK = true;
        } else if (oTree.m_iType == TreePopulation.SAPLING) {
          // Either DBH or diam10 is acceptable
          iDiamCode = GetFloatDataCode("DBH", oTree.m_iSpecies, oTree.m_iType);
          if (iDiamCode != -1 && oTree.mp_fFloats[iDiamCode] != null)
            bOK = true;
          else {
            iDiamCode = GetFloatDataCode("Diam10", oTree.m_iSpecies,
                oTree.m_iType);
            if (iDiamCode != -1 && oTree.mp_fFloats[iDiamCode] != null)
              bOK = true;
          }
        } else if (oTree.m_iType == TreePopulation.SEEDLING) {
          iDiamCode = GetFloatDataCode("Diam10", oTree.m_iSpecies,
              oTree.m_iType);
          if (iDiamCode != -1 && oTree.mp_fFloats[iDiamCode] != null)
            bOK = true;
        }

        // If diameter was not present, check for height
        if (false == bOK) {
          iDiamCode = GetFloatDataCode("Height", oTree.m_iSpecies,
              oTree.m_iType);
          if (iDiamCode != -1 && oTree.mp_fFloats[iDiamCode] != null)
            bOK = true;
        }
      }

      if (false == bOK) {
        mp_oTrees.remove(i);
        i--;
      }
    }

  }

  /**
   * Writes an XML tree map if there are any trees to write.
   * 
   * @param jOut The XML file to write to.
   * @throws ModelException If anything goes wrong. This could wrap another exception.
   * @todo subplot support
   */
  private void WriteTreeMap(BufferedWriter jOut) throws ModelException {
    try {
      if (mp_oTrees.size() == 0) {
        return;
      }
      int i, j, k;

      // Write the tree map header
      jOut.write("<tr_treemap>");

      // Tree settings for each species
      for (i = 0; i < GetNumberOfSpecies(); i++) {
        for (j = 0; j < GetNumberOfTypes(); j++) {

          // Only include if there are any settings
          if (mp_sTreeBoolDataMembers.elementAt(i).elementAt(j).size() > 0
              || mp_sTreeFloatDataMembers.elementAt(i).elementAt(j).size() > 0
              || mp_sTreeIntDataMembers.elementAt(i).elementAt(j).size() > 0
              || mp_sTreeCharDataMembers.elementAt(i).elementAt(j).size() > 0) {

            jOut.write("<tm_treeSettings sp=\"" + GetSpeciesNameFromCode(i)
                + "\" tp=\"" + String.valueOf(j) + "\">");

            // Int data members
            if (mp_sTreeIntDataMembers.elementAt(i).elementAt(j).size() > 0) {
              jOut.write("<tm_intCodes>");

              for (k = 0; k < mp_sTreeIntDataMembers.elementAt(i).elementAt(j)
                  .size(); k++) {
                jOut.write("<tm_intCode label=\""
                    + (String) mp_sTreeIntDataMembers.elementAt(i).elementAt(j)
                        .elementAt(k) + "\">" + k + "</tm_intCode>");
              }
              jOut.write("</tm_intCodes>");
            }

            // Float data members
            if (mp_sTreeFloatDataMembers.elementAt(i).elementAt(j).size() > 0) {
              jOut.write("<tm_floatCodes>");

              for (k = 0; k < mp_sTreeFloatDataMembers.elementAt(i)
                  .elementAt(j).size(); k++) {
                jOut.write("<tm_floatCode label=\""
                    + (String) mp_sTreeFloatDataMembers.elementAt(i).elementAt(
                        j).elementAt(k) + "\">" + k + "</tm_floatCode>");
              }
              jOut.write("</tm_floatCodes>");
            }

            // Char data members
            if (mp_sTreeCharDataMembers.elementAt(i).elementAt(j).size() > 0) {
              jOut.write("<tm_charCodes>");

              for (k = 0; k < mp_sTreeCharDataMembers.elementAt(i).elementAt(j)
                  .size(); k++) {
                jOut.write("<tm_charCode label=\""
                    + (String) mp_sTreeCharDataMembers.elementAt(i)
                        .elementAt(j).elementAt(k) + "\">" + k
                    + "</tm_charCode>");
              }
              jOut.write("</tm_charCodes>");
            }

            // Bool data members
            if (mp_sTreeBoolDataMembers.elementAt(i).elementAt(j).size() > 0) {
              jOut.write("<tm_boolCodes>");

              for (k = 0; k < mp_sTreeBoolDataMembers.elementAt(i).elementAt(j)
                  .size(); k++) {
                jOut.write("<tm_boolCode label=\""
                    + (String) mp_sTreeBoolDataMembers.elementAt(i)
                        .elementAt(j).elementAt(k) + "\">" + k
                    + "</tm_boolCode>");
              }
              jOut.write("</tm_boolCodes>");
            }

            jOut.write("</tm_treeSettings>");
          }
        }
      }

      // Now write all the trees
      Tree oTree;
      for (i = 0; i < mp_oTrees.size(); i++) {
        oTree = (Tree) mp_oTrees.elementAt(i);
        oTree.WriteXML(jOut);
      }

      jOut.write("</tr_treemap>");
    } catch (java.io.IOException e) {
      ModelException oErr = new ModelException();
      oErr.iErrorCode = ErrorGUI.BAD_FILE;
      oErr.sCallingFunction = "WriteTreeMap";
      throw oErr;
    }
  }

  /**
   * Override this function in order to be able to handle initial densities and
   * species names.
   * 
   * @param sXMLTag Parent XML tag of data vector whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param p_oData Vector of data values appropriate to the data type
   * @param p_sChildXMLTags The XML tags of the child elements
   * @param p_bAppliesTo Array of booleans saying which of the vector values 
   * should be set. This is important in the case of species-specifics - the 
   * vector index is the species number but not all species are set.
   * @param oParentAttributes Attributes of parent tag. May be useful when 
   * overridding this for unusual tags.
   * @param p_oAttributes Attributes passed from parser. This may be needed when
   *  overriding this function. Basic species-specific values are already 
   *  handled by this function.
   * @return true if the value was set successfully; false if the value could 
   * not be found. (This would not be an error, because I need a way to cycle 
   * through the objects until one of the objects comes up with a match.) If a 
   * match to a data object is made via XML tag, but the found object is not a 
   * ModelVector, this returns false.
   * @throws ModelException if the value could not be assigned to the data object.
   */
  public boolean SetVectorValueByXMLTag(String sXMLTag, String sXMLParentTag,
      Vector<String> p_oData, String[] p_sChildXMLTags, boolean[] p_bAppliesTo,
      Attributes oParentAttributes, Attributes[] p_oAttributes)
      throws ModelException {

    if (sXMLTag.equals("tr_idVals")) {

      // Get the species
      int iSpecies = GetSpeciesCodeFromName(oParentAttributes
          .getValue("whatSpecies")), iClass, // index of the size class
      i, j;
      if (iSpecies == -1) {
        throw (new ModelException(
            ErrorGUI.BAD_DATA,
            "JAVA",
            "Setting initial densities - XML tag \"tr_idVals\""
                + " missing \"whatSpecies\" attribute, or species unrecognized."));
      }
      for (i = 0; i < p_oData.size(); i++) {

        // Get the size class
        String sSizeClass = p_oAttributes[i].getValue("sizeClass");
        Float fSizeClass;
        if (sSizeClass.equalsIgnoreCase("seedling")) {
          fSizeClass = new Float(0);
        } else {
          sSizeClass = sSizeClass.substring(1); // trim off inital letter
          fSizeClass = new Float(sSizeClass);
        }
        // Find the index of the size class
        iClass = -1;
        for (j = 0; j < mp_fSizeClasses.GetValue().size(); j++) {
          if (fSizeClass
              .equals((Float) mp_fSizeClasses.GetValue().elementAt(j))) {
            iClass = j;
            break;
          }
        }
        if (iClass == -1) {
          throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
              "Invalid size class in initial densities - " + sSizeClass));
        }
        ModelVector p_oDensityArray = (ModelVector) mp_fInitialDensities
            .GetValue().elementAt(iClass);
        if (p_oDensityArray.GetValue().size() <= iSpecies) {
          p_oDensityArray.GetValue().setSize(iSpecies + 1);
        }
        p_oDensityArray.GetValue().remove(iSpecies);
        p_oDensityArray.GetValue().add(iSpecies,
            new Float((String) p_oData.elementAt(i)));
      }
      return true;
    } else if (sXMLTag.equals("tr_speciesList")) {

      // Species list
      // Create an array of species name strings
      String[] p_sNames = new String[p_oAttributes.length];
      int i;
      for (i = 0; i < p_sNames.length; i++) {
        p_sNames[i] = p_oAttributes[i].getValue("speciesName");
      }
      SetSpeciesNames(p_sNames);
      return true;

    } else if (sXMLTag.equals("tr_sizeClasses")) {

      // Size classes
      Float[] p_fSizeClasses = new Float[p_oAttributes.length];
      int i;
      for (i = 0; i < p_fSizeClasses.length; i++) {
        String sSizeKey = p_oAttributes[i].getValue("sizeKey");
        if (sSizeKey.equalsIgnoreCase("seedling")) {
          p_fSizeClasses[i] = new Float(0);
        } else {
          p_fSizeClasses[i] = new Float(sSizeKey.substring(1));
        }
      }
      SetSizeClasses(p_fSizeClasses);
      return true;
    } else if (sXMLTag.equals("tr_initialDensities")) {
      return true; // ignore this tag
    } else {
      // These may be tree map codes rolled up into a vector - handle them
      // like individuals
      int i;
      boolean bReturn = false;
      for (i = 0; i < p_sChildXMLTags.length; i++) {
        if (SetSingleValueByXMLTag(p_sChildXMLTags[i], sXMLParentTag,
            p_oAttributes[i], p_oData.elementAt(i))) {
          bReturn = true;
        }
      }

      if (bReturn) {
        return bReturn;
      }
    }
    return super.SetVectorValueByXMLTag(sXMLTag, sXMLParentTag, p_oData,
        p_sChildXMLTags, p_bAppliesTo, oParentAttributes, p_oAttributes);
  }

  /**
   * Gets the data member code for a data member label. If this is not a
   * currently listed data member, it will be added.
   * 
   * @param p_oMemberList List of data members.
   * @param sDataMember Data member name.
   * @return Data member code.
   */
  protected int GetCodeForDataMember(Vector<String> p_oMemberList,
      String sDataMember) {

    int i;
    for (i = 0; i < p_oMemberList.size(); i++) {
      if (sDataMember.equals(p_oMemberList.elementAt(i))) {
        return i;
      }
    }

    p_oMemberList.add(sDataMember);
    return p_oMemberList.size() - 1;
  }

  /**
   * Gets the data code for accessing a float tree data member. You can use the
   * return code to get the data from a Tree class object.
   * 
   * @param sDataMember Name of data member.
   * @param iSpecies Tree species.
   * @param iType Tree type.
   * @return Data code, or -1 if the data member name is unrecognized.
   */
  public int GetFloatDataCode(String sDataMember, int iSpecies, int iType) {
    int i, iReturnCode = -1;
    for (i = 0; i < mp_sTreeFloatDataMembers.elementAt(iSpecies).elementAt(
        iType).size(); i++) {
      if (sDataMember.equalsIgnoreCase(mp_sTreeFloatDataMembers.elementAt(
          iSpecies).elementAt(iType).elementAt(i).toString())) {
        return i;
      }
    }
    return iReturnCode;
  }

  /**
   * Accepts an XML parent tag (empty, no data) from the parser. This method
   * watches for the following tags:
   * <ul>
   * <li>tm_treeSettings
   * <li>tree
   * <li>tm_speciesList
   * <li>tm_species
   * <li>grid
   * </ul>
   * 
   * @param sXMLTag The XML tag.
   * @param oAttributes The attributes of this object.
   * @throws ModelException if there is a problem reading this data.
   */
  public void ReadXMLParentTag(String sXMLTag, Attributes oAttributes)
      throws ModelException {

    if (sXMLTag.equals("tree")) {

      // Extract species and type
      m_iCurrentSpecies = new Integer(oAttributes.getValue("sp")).intValue();
      m_iCurrentTreeType = new Integer(oAttributes.getValue("tp")).intValue();

      if (m_iCurrentSpecies < 0 || m_iCurrentSpecies >= GetNumberOfSpecies()
          || m_iCurrentTreeType < 0 || m_iCurrentTreeType >= GetNumberOfTypes()) {
        throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
            "tm_tree has missing or invalid attributes \"" + "sp\" or \"tp\""));
      }

      // Put the current species through the transform list
      m_iCurrentSpecies = mp_iSpeciesTransforms[m_iCurrentSpecies];

      // Create the new tree
      Tree oTree = new Tree(m_iCurrentTreeType, m_iCurrentSpecies,
          mp_sTreeFloatDataMembers.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).size(), mp_sTreeIntDataMembers.elementAt(
              m_iCurrentSpecies).elementAt(m_iCurrentTreeType).size(),
          mp_sTreeCharDataMembers.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).size(), mp_sTreeBoolDataMembers.elementAt(
              m_iCurrentSpecies).elementAt(m_iCurrentTreeType).size(), this);
      mp_oTrees.add(oTree);
    } else if (sXMLTag.equals("tm_treeSettings")) {

      // Extract species and type
      m_iCurrentSpecies = GetSpeciesCodeFromName(oAttributes.getValue("sp"));
      m_iCurrentTreeType = new Integer(oAttributes.getValue("tp")).intValue();

      if (m_iCurrentSpecies == -1 || m_iCurrentTreeType < 0
          || m_iCurrentTreeType >= GetNumberOfTypes()) {
        throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
            "tm_treeSettings has missing or invalid attributes \""
                + "sp\" or \"tp\""));
      }
    } else if (sXMLTag.equals("grid")) {
      // Reset our species and type so we know we're not reading tree map
      m_iCurrentSpecies = -1;
      m_iCurrentTreeType = -1;
    } else if (sXMLTag.equals("tm_species")) {

      // This is a species designator in our tree map
      // Extract the species name and turn into a code
      int iSpeciesCode = GetSpeciesCodeFromName(oAttributes
          .getValue("speciesName"));

      // If unrecognized, throw an error
      if (iSpeciesCode == -1) {
        throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
            "Unrecognized or absent species value in tree map species list."));
      }

      // Put the number in the transform list at the species counter
      mp_iSpeciesTransforms[m_iSpeciesCounter] = iSpeciesCode;
      m_iSpeciesCounter++;

    } else if (sXMLTag.equals("tm_speciesList")) {
      // Reset the species counter
      m_iSpeciesCounter = 0;
    }
    super.ReadXMLParentTag(sXMLTag, oAttributes);
  }

  /**
   * Clears all trees
   */
  public void ClearTrees() {
    mp_oTrees = null;
    mp_oTrees = new Vector<Tree>(0);
  }

  /**
   * This method is looking for tm_floatCode, tm_intCode, tm_charCode,
   * tm_boolCode, fl, int, ch, and bl. For the last four, they are only set if
   * m_iCurrentSpecies and m_iCurrentTreeType are greater than -1. Thus grid
   * values with these tags will be ignored.
   * 
   * @param sXMLTag XML tag of data object whose value is to be set.
   * @param sXMLParentTag The immediate parent tag that sXMLTag is within.
   * @param oAttributes Attributes of the object. Ignored, but may be needed by 
   * overriding objects.
   * @param oData Data value appropriate to the data type
   * @return true if the value was set successfully; false if the value could 
   * not be found. (This would not be an error, because I need a way to cycle 
   * through the objects until one of the objects comes up with a match.)
   * @throws ModelException if the value could not be assigned to the data 
   * object.
   */
  public boolean SetSingleValueByXMLTag(String sXMLTag, String sXMLParentTag,
      Attributes oAttributes, Object oData) throws ModelException {
    try {
      if (sXMLTag != null) {
        if (sXMLTag.equals("fl")) {
          if (m_iCurrentSpecies > -1 && m_iCurrentTreeType > -1) {
            // Get the last added tree
            Tree oTree = (Tree) mp_oTrees.elementAt(mp_oTrees.size() - 1);
            // Get the code
            int iCode = new Integer(oAttributes.getValue("c")).intValue();
            Float fValue = new Float(String.valueOf(oData));
            Integer iIndex = mp_iTreeFloatTransforms.elementAt(
                m_iCurrentSpecies).elementAt(m_iCurrentTreeType).elementAt(
                iCode);
            if (iIndex == null) {
              throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                  "There was an error reading the tree map. One "
                      + "of the float data codes appears to be invalid"
                      + " for species "
                      + GetSpeciesNameFromCode(m_iCurrentSpecies) + ", type "
                      + GetTypeNameFromCode(m_iCurrentTreeType) + "."));
            }

            // Validate if X or Y
            int iIndVal = iIndex.intValue();
            float fVal = fValue.floatValue();
            String sName = (String) mp_sTreeFloatDataMembers.elementAt(
                oTree.m_iSpecies).elementAt(oTree.m_iType).elementAt(iIndVal);
            if ((sName.equals("X") && (fVal < 0 || fVal >= m_oManager.GetPlot()
                .GetPlotXLength()))
                || (sName.equals("Y") && (fVal < 0 || fVal >= m_oManager
                    .GetPlot().GetPlotYLength()))) {
              ModelException oErr = new ModelException(ErrorGUI.BAD_DATA,
                  "JAVA",
                  "The tree map contains a tree whose X or Y coordinate is outside the plot.");
              throw (oErr);
            }

            oTree.SetValue(iIndVal, fValue);
            return true;
          }
        } else if (sXMLTag.equals("int")) {
          if (m_iCurrentSpecies > -1 && m_iCurrentTreeType > -1) {
            // Get the last added tree
            Tree oTree = (Tree) mp_oTrees.elementAt(mp_oTrees.size() - 1);
            // Get the code
            int iCode = new Integer(oAttributes.getValue("c")).intValue();
            Integer iValue = new Integer(String.valueOf(oData)), iIndex = (Integer) mp_iTreeIntTransforms
                .elementAt(m_iCurrentSpecies).elementAt(m_iCurrentTreeType)
                .elementAt(iCode);
            if (iIndex == null) {
              throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                  "There was an error reading the tree map.  "
                      + "One of the integer data codes appears to be "
                      + "invalid for species "
                      + GetSpeciesNameFromCode(m_iCurrentSpecies) + ", type "
                      + GetTypeNameFromCode(m_iCurrentTreeType) + "."));
            }
            oTree.SetValue(iIndex.intValue(), iValue);
            return true;
          }
        } else if (sXMLTag.equals("ch")) {
          if (m_iCurrentSpecies > -1 && m_iCurrentTreeType > -1) {
            // Get the last added tree
            Tree oTree = (Tree) mp_oTrees.elementAt(mp_oTrees.size() - 1);
            // Get the code
            int iCode = new Integer(oAttributes.getValue("c")).intValue();
            Integer iIndex = (Integer) mp_iTreeCharTransforms.elementAt(
                m_iCurrentSpecies).elementAt(m_iCurrentTreeType).elementAt(
                iCode);
            if (iIndex == null) {
              throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                  "There was an error reading the tree map. One "
                      + "of the string data codes appears to be invalid"
                      + " for species "
                      + GetSpeciesNameFromCode(m_iCurrentSpecies) + ", type "
                      + GetTypeNameFromCode(m_iCurrentTreeType) + "."));
            }
            oTree.SetValue(iIndex.intValue(), String.valueOf(oData));
            return true;
          }
        } else if (sXMLTag.equals("bl")) {
          // 
          
          if (m_iCurrentSpecies > -1 && m_iCurrentTreeType > -1) {
            // Get the last added tree
            Tree oTree = (Tree) mp_oTrees.elementAt(mp_oTrees.size() - 1);
            // Get the code
            int iCode = new Integer(oAttributes.getValue("c")).intValue();
            Boolean bValue = new Boolean(String.valueOf(oData));
            Integer iIndex = (Integer) mp_iTreeBoolTransforms.elementAt(
                m_iCurrentSpecies).elementAt(m_iCurrentTreeType).elementAt(
                iCode);
            if (iIndex == null) {
              throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                  "There was an error reading the tree map. One "
                      + "of the bool data codes appears to be invalid"
                      + " for species "
                      + GetSpeciesNameFromCode(m_iCurrentSpecies) + ", type "
                      + GetTypeNameFromCode(m_iCurrentTreeType) + "."));

            }
            oTree.SetValue(iIndex.intValue(), bValue);
            return true;
          }
        } else if (sXMLTag.equals("tm_floatCode")) {
          String sLabel = oAttributes.getValue("label");
          int iIndex = new Integer((String) oData).intValue(), iOfficialIndex = GetCodeForDataMember(
              mp_sTreeFloatDataMembers.elementAt(m_iCurrentSpecies).elementAt(
                  m_iCurrentTreeType), sLabel);
          if (mp_iTreeFloatTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).size() <= iIndex) {
            mp_iTreeFloatTransforms.elementAt(m_iCurrentSpecies).elementAt(
                m_iCurrentTreeType).setSize(iIndex + 1);
          }
          mp_iTreeFloatTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).remove(iIndex);
          mp_iTreeFloatTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).add(iIndex, new Integer(iOfficialIndex));

          return true;
        } else if (sXMLTag.equals("tm_intCode")) {
          String sLabel = oAttributes.getValue("label");
          int iIndex = new Integer((String) oData).intValue(), iOfficialIndex = GetCodeForDataMember(
              mp_sTreeIntDataMembers.elementAt(m_iCurrentSpecies).elementAt(
                  m_iCurrentTreeType), sLabel);
          if (mp_iTreeIntTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).size() <= iIndex) {
            mp_iTreeIntTransforms.elementAt(m_iCurrentSpecies).elementAt(
                m_iCurrentTreeType).setSize(iIndex + 1);
          }
          mp_iTreeIntTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).remove(iIndex);
          mp_iTreeIntTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).add(iIndex, new Integer(iOfficialIndex));

          return true;
        } else if (sXMLTag.equals("tm_boolCode")) {
          String sLabel = oAttributes.getValue("label");
          
          //Trap for "dead" changing to int from bool - just ignore 
          if (sLabel.equals("dead")) {
            return true;
          }
          
          int iIndex = new Integer((String) oData).intValue(), iOfficialIndex = GetCodeForDataMember(
              mp_sTreeBoolDataMembers.elementAt(m_iCurrentSpecies).elementAt(
                  m_iCurrentTreeType), sLabel);
          if (mp_iTreeBoolTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).size() <= iIndex) {
            mp_iTreeBoolTransforms.elementAt(m_iCurrentSpecies).elementAt(
                m_iCurrentTreeType).setSize(iIndex + 1);
          }
          mp_iTreeBoolTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).remove(iIndex);
          mp_iTreeBoolTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).add(iIndex, new Integer(iOfficialIndex));

          return true;
        } else if (sXMLTag.equals("tm_charCode")) {
          String sLabel = oAttributes.getValue("label");
          int iIndex = new Integer((String) oData).intValue(), iOfficialIndex = GetCodeForDataMember(
              mp_sTreeCharDataMembers.elementAt(m_iCurrentSpecies).elementAt(
                  m_iCurrentTreeType), sLabel);
          if (mp_iTreeCharTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).size() <= iIndex) {
            mp_iTreeCharTransforms.elementAt(m_iCurrentSpecies).elementAt(
                m_iCurrentTreeType).setSize(iIndex + 1);
          }
          mp_iTreeCharTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).remove(iIndex);
          mp_iTreeCharTransforms.elementAt(m_iCurrentSpecies).elementAt(
              m_iCurrentTreeType).add(iIndex, new Integer(iOfficialIndex));

          return true;
        }
      }
    } catch (java.lang.NumberFormatException oE) {
      throw (new ModelException(ErrorGUI.BAD_DATA, "JAVA",
          "Could not turn the value \"" + oData + "\" into a number."));
    }
    return super.SetSingleValueByXMLTag(sXMLTag, sXMLParentTag, oAttributes,
        oData);

  }

  /**
   * Changes the name of a species. Everything about the species remains the
   * same except for the name. This is an easier process than actually changing
   * the species list.
   * 
   * @param sOldSpecies String Old name of the species
   * @param sNewSpecies String New name of the species
   * @throws ModelException if there is a problem.
   */
  public void ChangeSpeciesName(String sOldSpecies, String sNewSpecies)
      throws ModelException {
    int i;
    boolean bSpeciesExisted = false;

    // Replace the spaces in the names with underscores
    sOldSpecies = sOldSpecies.replace(' ', '_');
    sNewSpecies = sNewSpecies.replace(' ', '_');

    // Make sure the species exists before we try to change it
    for (i = 0; i < mp_sSpeciesNames.GetValue().size(); i++) {
      if (((String) mp_sSpeciesNames.GetValue().elementAt(i))
          .equals(sOldSpecies)) {
        bSpeciesExisted = true;
        mp_sSpeciesNames.GetValue().setElementAt(sNewSpecies, i);
        break;
      }
    }

    if (false == bSpeciesExisted)
      return;
    m_oManager.ChangeOfSpeciesName(sOldSpecies, sNewSpecies);
  }

  /**
   * Makes one species exactly like another. Any difference from the target
   * species is obliterated. Both species names must already exist (i.e. add a
   * new species before making it into a copy of another). This triggers the
   * change through the other GUI objects.
   * 
   * @param sCopyFrom String Name of species to copy.
   * @param sCopyTo String Name of species that is the copy.
   * @throws ModelException
   */
  public void DoCopySpecies(String sCopyFrom, String sCopyTo)
      throws ModelException {
    // Verify that the old and new species both exist
    int iCopyFrom = GetSpeciesCodeFromName(sCopyFrom), iCopyTo = GetSpeciesCodeFromName(sCopyTo);
    if (-1 == iCopyFrom || -1 == iCopyTo) {
      throw new ModelException(ErrorGUI.BAD_COMMAND, "JAVA",
          "I can't copy species \"" + sCopyFrom + "\" to species \"" + sCopyTo
              + "\" because they don't both exist.");
    }

    m_oManager.CopySpecies(iCopyFrom, iCopyTo);
  }
}

/**
 * Represents a single tree's data. Individuals can thus be stored and turned
 * into a tree map for the model.
 * <p>
 * Copyright: Copyright (c) Charles D. Canham 2003
 * </p>
 * <p>
 * Company: Institute of Ecosystem Studies
 * </p>
 * 
 * @author Lora E. Murphy
 * @version 1.0
 */
class Tree extends DataMemberData {
  protected int
  /** Tree's species */
  m_iSpecies,
  /** Tree's type */
  m_iType;

  /**
   * Constructor. Creates the arrays.
   * 
   * @param iType Tree's type
   * @param iSpecies Tree's species
   * @param iNumFloats Number of float data members
   * @param iNumInts Number of integer data members
   * @param iNumChars Number of char (String) data members
   * @param iNumBools Number of bool data members
   * @param oPop TreePopulation object
   * @throws ModelException if the species or type numbers are invalid
   */
  public Tree(int iType, int iSpecies, int iNumFloats, int iNumInts,
      int iNumChars, int iNumBools, TreePopulation oPop) throws ModelException {

    super(iNumFloats, iNumInts, iNumChars, iNumBools);

    if (iSpecies < 0 || iSpecies >= oPop.GetNumberOfSpecies() || iType < 0
        || iType >= TreePopulation.GetNumberOfTypes()) {
      throw (new ModelException(ErrorGUI.BAD_ARGUMENT, "JAVA",
          "Tree contains bad species or type number. Species = \""
              + String.valueOf(iSpecies) + "\", type = \""
              + String.valueOf(iType) + "\"."));
    }
    m_iSpecies = iSpecies;
    m_iType = iType;
  }

  /**
   * Writes this tree's data to XML.
   * 
   * @param jOut File to write to.
   * @throws java.io.IOException if there is a problem writing to file.
   */
  public void WriteXML(BufferedWriter jOut) throws java.io.IOException {
    int i;

    // Opening tree tag
    jOut.write("<tree sp=\"" + String.valueOf(m_iSpecies) + "\" tp=\""
        + String.valueOf(m_iType) + "\">");

    // Write the ints
    for (i = 0; i < mp_iInts.length; i++) {
      if (mp_iInts[i] != null) {
        jOut.write("<int c=\"" + String.valueOf(i) + "\">"
            + mp_iInts[i].toString() + "</int>");
      }
    }

    // Write the floats
    for (i = 0; i < mp_fFloats.length; i++) {
      if (mp_fFloats[i] != null) {
        jOut.write("<fl c=\"" + String.valueOf(i) + "\">"
            + mp_fFloats[i].toString() + "</fl>");
      }
    }

    // Write the chars
    for (i = 0; i < mp_sChars.length; i++) {
      if (mp_sChars[i] != null) {
        jOut.write("<ch c=\"" + String.valueOf(i) + "\">" + mp_sChars[i]
            + "</ch>");
      }
    }

    // Write the bools
    for (i = 0; i < mp_bBools.length; i++) {
      if (mp_bBools[i] != null) {
        jOut.write("<bl c=\"" + String.valueOf(i) + "\">"
            + mp_bBools[i].toString() + "</bl>");
      }
    }

    jOut.write("</tree>");
  }
}
