package javawrapper;

import junit.framework.*;
import java.io.FileWriter;
import java.io.File;


/**
 * Tests the OutputBehaviors class.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 * <br>April 28, 2004: Submitted in beta version (LEM)

 */
public class TestOutputBehaviors
    extends TestCase {
  /** Object for testing. */
  private OutputBehaviors mp_oOutputBehaviors = null;

  /**
   * Does set up by creating testing objects.
   * @throws Exception if there is a problem doing setup.
   */
  protected void setUp() throws Exception {
    super.setUp();

    try {
      TreePopulation oPop = new TreePopulation(null);
      oPop.SetSpeciesNames(new String[] {"ACSA", "ACRU", "QURU"});
      mp_oOutputBehaviors = new OutputBehaviors(null);
      mp_oOutputBehaviors.DoSetup(oPop);
    }
    catch (ModelException oErr) {
      throw new Exception(oErr.getMessage());
    }
    assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
    assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
    assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedGridSettings());
  }

  /**
   * Tests the AddDetailedGridSettings method.
   */
  public void testAddDetailedGridSettings() {
    DetailedGridSettings oSettings = null, oTest;
    mp_oOutputBehaviors.AddDetailedGridSettings(oSettings);

    //Normal processing

    //First added
    oSettings = new DetailedGridSettings("Grid 1");
    oSettings.AddChar("Char1", "Char1");
    oSettings.AddInt("Integer 1", "Integer 1");
    oSettings.SetSaveFrequency(2);
    mp_oOutputBehaviors.AddDetailedGridSettings(oSettings);
    assertEquals(1, mp_oOutputBehaviors.GetNumberOfDetailedGridSettings());
    oTest = mp_oOutputBehaviors.GetDetailedGridSetting(0);
    assertTrue(testEquals(oSettings, oTest));

    //Second added
    oSettings = null;
    oSettings = new DetailedGridSettings("Grid 2");
    oSettings.AddChar("Char1", "Char1");
    oSettings.AddInt("Integer 1", "Integer 1");
    oSettings.AddBool("OneBool", "OneBool");
    oSettings.AddFloat("OneFloat", "OneFloat");
    oSettings.AddFloat("OtherFloat", "OtherFloat");
    oSettings.AddFloat("TwoFloat", "TwoFloat");
    oSettings.AddChar("OneChar", "OneChar");
    oSettings.SetSaveFrequency(2);
    mp_oOutputBehaviors.AddDetailedGridSettings(oSettings);
    assertEquals(2, mp_oOutputBehaviors.GetNumberOfDetailedGridSettings());
    oTest = mp_oOutputBehaviors.GetDetailedGridSetting(1);
    assertTrue(testEquals(oSettings, oTest));
    assertFalse(testEquals(oSettings,
                           mp_oOutputBehaviors.GetDetailedGridSetting(0)));

    //Same type and species as first - should replace
    oSettings = null;
    oSettings = new DetailedGridSettings("Grid 1");
    oSettings.AddFloat("OneFloat", "OneFloat");
    oSettings.AddFloat("OtherFloat", "OtherFloat");
    oSettings.SetSaveFrequency(0);
    mp_oOutputBehaviors.AddDetailedGridSettings(oSettings);
    assertEquals(2, mp_oOutputBehaviors.GetNumberOfDetailedGridSettings());
    oTest = mp_oOutputBehaviors.GetDetailedGridSetting(1);
    assertTrue(testEquals(oSettings, oTest));

    System.out.println(
        "Normal processing succeeded for AddDetailedGridSettings.");
  }

  /**
   * Tests the AddDetailedTreeSettings method.
   */
  public void testAddDetailedTreeSettings() {
    DetailedTreeSettings oSettings = null,
        oTest;

    try {
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);

      //Normal processing

      //First added
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(1, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      oTest = mp_oOutputBehaviors.GetDetailedLiveTreeSetting(0);
      assertTrue(testEquals(oSettings, oTest));

      //No settings - should not be added
      /*      oSettings = null;
            oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1);
            oSettings.SetSaveFrequency(1);
            mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
            mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
       assertEquals(1, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
            oSettings = mp_oOutputBehaviors.GetDetailedLiveTreeSetting(0);
            assertTrue(testEquals(oSettings, oTest));*/

      //Second added - specify NOTDEAD
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1, OutputBehaviors.NOTDEAD);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("TwoFloat", "TwoFloat");
      oSettings.AddChar("OneChar", "OneChar");
      oSettings.SetSaveFrequency(0);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(2, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      oTest = mp_oOutputBehaviors.GetDetailedLiveTreeSetting(1);
      assertTrue(testEquals(oSettings, oTest));
      
      //Same type and species as second but dead - should not replace
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1, OutputBehaviors.NATURAL);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("TwoFloat", "TwoFloat");
      oSettings.AddChar("OneChar", "OneChar");
      oSettings.SetSaveFrequency(0);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(2, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(1, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      oTest = mp_oOutputBehaviors.GetDetailedDeadTreeSetting(0);
      assertTrue(testEquals(oSettings, oTest));

      //Same type and species as second live - should replace
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("DifferentFloat", "DifferentFloat");
      oSettings.AddChar("OneChar", "OneChar");
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(2, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(1, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      oSettings = mp_oOutputBehaviors.GetDetailedLiveTreeSetting(1);
      assertFalse(testEquals(oSettings, oTest)); //oTest is still from before

      //Same type but not same species - should add
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 0);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("DifferentFloat", "DifferentFloat");
      oSettings.AddChar("OneChar", "OneChar");
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(3, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(1, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      
      //Same species but not same type - should add
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 1);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("DifferentFloat", "DifferentFloat");
      oSettings.AddChar("OneChar", "OneChar");
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(4, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(1, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      
      //Dead: Same type and species but different dead code - should add 
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1, OutputBehaviors.HARVEST);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("TwoFloat", "TwoFloat");
      oSettings.AddChar("OneChar", "OneChar");
      oSettings.SetSaveFrequency(0);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(4, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(2, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      oTest = mp_oOutputBehaviors.GetDetailedDeadTreeSetting(1);
      assertTrue(testEquals(oSettings, oTest));
      
      //Dead: Same type and species and dead code as first - should replace 
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 1, OutputBehaviors.NATURAL);
      oSettings.AddBool("DeadBool", "OneBool");
      oSettings.AddFloat("OneFloat", "DeadFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("TwoFloat", "TwoFloat");
      oSettings.AddChar("OneChar", "OneChar");
      oSettings.SetSaveFrequency(0);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      assertEquals(4, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(2, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      oTest = mp_oOutputBehaviors.GetDetailedDeadTreeSetting(1);
      assertTrue(testEquals(oSettings, oTest));
    }
    catch (ModelException oErr) {
      fail("Normal processing failed for AddDetailedTreeSettings with message " +
           oErr.getMessage());
    }

    System.out.println(
        "Normal processing succeeded for AddDetailedTreeSettings.");

    //Exception processing

    //Invalid species
    try {
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 5);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("DifferentFloat", "DifferentFloat");
      oSettings.AddChar("OneChar", "OneChar");
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      fail("AddDetailedTreeSettings failed to catch invalid species");
    }
    catch (ModelException oErr) {
      ;
    }

    //Invalid type
    try {
      oSettings = null;
      oSettings = new DetailedTreeSettings(17, 0);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("DifferentFloat", "DifferentFloat");
      oSettings.AddChar("OneChar", "OneChar");
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      fail("AddDetailedTreeSettings failed to catch invalid type");
    }
    catch (ModelException oErr) {
      ;
    }
    
    //Invalid dead code
    try {
      oSettings = null;
      oSettings = new DetailedTreeSettings(TreePopulation.SEEDLING, 0, 10);
      oSettings.AddBool("OneBool", "OneBool");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("OneFloat", "OneFloat");
      oSettings.AddFloat("DifferentFloat", "DifferentFloat");
      oSettings.AddChar("OneChar", "OneChar");
      mp_oOutputBehaviors.AddDetailedTreeSettings(oSettings);
      fail("AddDetailedTreeSettings failed to catch invalid dead code");
    }
    catch (ModelException oErr) {
      ;
    }
    System.out.println(
        "Exception processing succeeded for AddDetailedTreeSettings.");
  }
  
  
  /**
   * Tests the ClearDetailedGridSettings method.
   */
  public void testClearDetailedGridSettings() {
    try {
      mp_oOutputBehaviors.ClearDetailedGridSettings();
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedGridSettings());
      mp_oOutputBehaviors.ClearDetailedGridSettings();

      //Make sure only grids are cleared, not trees
      DetailedTreeSettings oTreeSettings = new DetailedTreeSettings(
          TreePopulation.ADULT, 0);
      oTreeSettings.AddBool("OneBool", "OneBool");
      oTreeSettings.AddFloat("OneFloat", "OneFloat");
      oTreeSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oTreeSettings);

      DetailedGridSettings oGridSettings = new DetailedGridSettings("Grid 1");
      oGridSettings.AddBool("OneBool", "OneBool");
      oGridSettings.AddFloat("OneFloat", "OneFloat");
      oGridSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedGridSettings(oGridSettings);

      mp_oOutputBehaviors.ClearDetailedGridSettings();
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedGridSettings());
      assertTrue(mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings() > 0);
    }
    catch (ModelException oErr) {
      fail("ClearDetailedGridSettings failed with message " + oErr.getMessage());
    }
  }

  /**
   * Tests the ClearDetailedLiveTreeSettings method.
   */
  public void testClearDetailedLiveTreeSettings() {
    try {
      mp_oOutputBehaviors.ClearDetailedLiveTreeSettings();
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      mp_oOutputBehaviors.ClearDetailedLiveTreeSettings();

      //Make sure only live trees are cleared, not dead trees or grids
      DetailedTreeSettings oTreeSettings = new DetailedTreeSettings(
          TreePopulation.ADULT, 0);
      oTreeSettings.AddBool("OneBool", "OneBool");
      oTreeSettings.AddFloat("OneFloat", "OneFloat");
      oTreeSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oTreeSettings);
      
      oTreeSettings = new DetailedTreeSettings(
          TreePopulation.ADULT, 0, OutputBehaviors.DISEASE);
      oTreeSettings.AddBool("OneBool", "OneBool");
      oTreeSettings.AddFloat("OneFloat", "OneFloat");
      oTreeSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oTreeSettings);
      
      DetailedGridSettings oGridSettings = new DetailedGridSettings("Grid 1");
      oGridSettings.AddBool("OneBool", "OneBool");
      oGridSettings.AddFloat("OneFloat", "OneFloat");
      oGridSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedGridSettings(oGridSettings);

      mp_oOutputBehaviors.ClearDetailedLiveTreeSettings();
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings());
      assertTrue(mp_oOutputBehaviors.GetNumberOfDetailedGridSettings() > 0);
      assertTrue(mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings() > 0);
    }
    catch (ModelException oErr) {
      fail("ClearDetailedLiveTreeSettings failed with message " + oErr.getMessage());
    }
  }
  
  /**
   * Tests the ClearDetailedDeadTreeSettings method.
   */
  public void testClearDetailedDeadTreeSettings() {
    try {
      mp_oOutputBehaviors.ClearDetailedDeadTreeSettings();
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());

      //Make sure only dead trees are cleared, not live trees or grids
      DetailedTreeSettings oTreeSettings = new DetailedTreeSettings(
          TreePopulation.ADULT, 0);
      oTreeSettings.AddBool("OneBool", "OneBool");
      oTreeSettings.AddFloat("OneFloat", "OneFloat");
      oTreeSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oTreeSettings);
      
      oTreeSettings = new DetailedTreeSettings(
          TreePopulation.ADULT, 0, OutputBehaviors.DISEASE);
      oTreeSettings.AddBool("OneBool", "OneBool");
      oTreeSettings.AddFloat("OneFloat", "OneFloat");
      oTreeSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedTreeSettings(oTreeSettings);
      
      DetailedGridSettings oGridSettings = new DetailedGridSettings("Grid 1");
      oGridSettings.AddBool("OneBool", "OneBool");
      oGridSettings.AddFloat("OneFloat", "OneFloat");
      oGridSettings.SetSaveFrequency(2);
      mp_oOutputBehaviors.AddDetailedGridSettings(oGridSettings);

      mp_oOutputBehaviors.ClearDetailedDeadTreeSettings();
      assertEquals(0, mp_oOutputBehaviors.GetNumberOfDetailedDeadTreeSettings());
      assertTrue(mp_oOutputBehaviors.GetNumberOfDetailedGridSettings() > 0);
      assertTrue(mp_oOutputBehaviors.GetNumberOfDetailedLiveTreeSettings() > 0);
    }
    catch (ModelException oErr) {
      fail("ClearDetailedDeadTreeSettings failed with message " + oErr.getMessage());
    }

  }

  /**
   * Tests reading of parameter file XML rundata settings. Unfortunately,
   * we can't do the grids this way anymore now that I validate data members
   * against existing grids. But we're doing at least some testing of grid
   * reading with testChangeOfSpecies.
   */
  public void testReadXMLDetailedOutputSettings() {
    String sFileName = null;
    try {
      GUIManager oManager = new GUIManager(null);
      int i;
      //Force a new behavior group
      /*WorkerBase[] p_oExisting = oManager.mp_oManagedObjects;
      oManager.mp_oManagedObjects = new WorkerBase[p_oExisting.length+1];
      for (i = 0; i < p_oExisting.length; i++) {
        oManager.mp_oManagedObjects[i] = p_oExisting[i];
      }
      oManager.mp_oManagedObjects[p_oExisting.length] = new TestBehaviorGroup(oManager); */
      sFileName = WriteXMLFile1();
      oManager.InputXMLParameterFile(sFileName);
      DetailedTreeSettings oSettings;
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();
      assertEquals(10, oOutput.GetNumberOfDetailedGridSettings());
      assertEquals(6, oOutput.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(7, oOutput.GetNumberOfDetailedDeadTreeSettings());

      for (i = 0; i < oOutput.GetNumberOfDetailedLiveTreeSettings(); i++) {
        oSettings = oOutput.GetDetailedLiveTreeSetting(i);
        if (0 == oSettings.GetSpecies()) {
          assertEquals(1, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals("species 1 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("species 1 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("species 1 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("species 1 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (1 == oSettings.GetSpecies()) {

          assertEquals(2, oSettings.GetSaveFrequency());
          assertEquals(2, oSettings.GetNumberOfBools());
          assertEquals(2, oSettings.GetNumberOfChars());
          assertEquals(2, oSettings.GetNumberOfFloats());
          assertEquals(2, oSettings.GetNumberOfInts());
          assertEquals("species 2 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("species 2 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("species 2 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("species 2 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("species 2 test float 2",
                       oSettings.GetFloat(1).GetCodeName());
          assertEquals("species 2 test char 2",
                       oSettings.GetChar(1).GetCodeName());
          assertEquals("species 2 test int 2",
                       oSettings.GetInt(1).GetCodeName());
          assertEquals("species 2 test bool 2",
                       oSettings.GetBool(1).GetCodeName());

        }
        else if (2 == oSettings.GetSpecies()) {

          assertEquals(3, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals("species 3 test char 1",
                       oSettings.GetChar(0).GetCodeName());

        }
        else if (3 == oSettings.GetSpecies()) {

          assertEquals(4, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals("species 4 test float 1",
                       oSettings.GetFloat(0).GetCodeName());

        }
        else if (4 == oSettings.GetSpecies()) {

          assertEquals(5, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals("species 5 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (5 == oSettings.GetSpecies()) {

          assertEquals(6, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals("species 6 test int 1",
                       oSettings.GetInt(0).GetCodeName());

        }
        else {
          throw(new ModelException(ErrorGUI.BAD_DATA, "JAVA",
                                   "Unrecognized species in tree."));
        }
      }
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(0);
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(OutputBehaviors.HARVEST, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetSaveFrequency());
      assertEquals(1, oSettings.GetNumberOfBools());
      assertEquals(1, oSettings.GetNumberOfChars());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals(1, oSettings.GetNumberOfInts());
      assertEquals("dead species 1 test float 1 c " + OutputBehaviors.HARVEST, oSettings.GetFloat(0).GetCodeName());
      assertEquals("dead species 1 test char 1 c " + OutputBehaviors.HARVEST, oSettings.GetChar(0).GetCodeName());
      assertEquals("dead species 1 test int 1 c " + OutputBehaviors.HARVEST, oSettings.GetInt(0).GetCodeName());
      assertEquals("dead species 1 test bool 1 c " + OutputBehaviors.HARVEST, oSettings.GetBool(0).GetCodeName());
      
      //Natural
      oSettings = oOutput.GetDetailedDeadTreeSetting(1);
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(OutputBehaviors.NATURAL, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetSaveFrequency());
      assertEquals(1, oSettings.GetNumberOfBools());
      assertEquals(1, oSettings.GetNumberOfChars());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals(1, oSettings.GetNumberOfInts());
      assertEquals("dead species 1 test float 1 c " + OutputBehaviors.NATURAL, oSettings.GetFloat(0).GetCodeName());
      assertEquals("dead species 1 test char 1 c " + OutputBehaviors.NATURAL, oSettings.GetChar(0).GetCodeName());
      assertEquals("dead species 1 test int 1 c " + OutputBehaviors.NATURAL, oSettings.GetInt(0).GetCodeName());
      assertEquals("dead species 1 test bool 1 c " + OutputBehaviors.NATURAL, oSettings.GetBool(0).GetCodeName());
      
      //Insects
      oSettings = oOutput.GetDetailedDeadTreeSetting(2);
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(OutputBehaviors.INSECTS, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetSaveFrequency());
      assertEquals(1, oSettings.GetNumberOfBools());
      assertEquals(1, oSettings.GetNumberOfChars());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals(1, oSettings.GetNumberOfInts());
      assertEquals("dead species 1 test float 1 c " + OutputBehaviors.INSECTS, oSettings.GetFloat(0).GetCodeName());
      assertEquals("dead species 1 test char 1 c " + OutputBehaviors.INSECTS, oSettings.GetChar(0).GetCodeName());
      assertEquals("dead species 1 test int 1 c " + OutputBehaviors.INSECTS, oSettings.GetInt(0).GetCodeName());
      assertEquals("dead species 1 test bool 1 c " + OutputBehaviors.INSECTS, oSettings.GetBool(0).GetCodeName());
      
      //Disease
      oSettings = oOutput.GetDetailedDeadTreeSetting(3);
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(OutputBehaviors.DISEASE, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetSaveFrequency());
      assertEquals(1, oSettings.GetNumberOfBools());
      assertEquals(1, oSettings.GetNumberOfChars());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals(1, oSettings.GetNumberOfInts());
      assertEquals("dead species 1 test float 1 c " + OutputBehaviors.DISEASE, oSettings.GetFloat(0).GetCodeName());
      assertEquals("dead species 1 test char 1 c " + OutputBehaviors.DISEASE, oSettings.GetChar(0).GetCodeName());
      assertEquals("dead species 1 test int 1 c " + OutputBehaviors.DISEASE, oSettings.GetInt(0).GetCodeName());
      assertEquals("dead species 1 test bool 1 c " + OutputBehaviors.DISEASE, oSettings.GetBool(0).GetCodeName());
      
      //Fire
      oSettings = oOutput.GetDetailedDeadTreeSetting(4);
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(OutputBehaviors.FIRE, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetSaveFrequency());
      assertEquals(1, oSettings.GetNumberOfBools());
      assertEquals(1, oSettings.GetNumberOfChars());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals(1, oSettings.GetNumberOfInts());
      assertEquals("dead species 1 test float 1 c " + OutputBehaviors.FIRE, oSettings.GetFloat(0).GetCodeName());
      assertEquals("dead species 1 test char 1 c " + OutputBehaviors.FIRE, oSettings.GetChar(0).GetCodeName());
      assertEquals("dead species 1 test int 1 c " + OutputBehaviors.FIRE, oSettings.GetInt(0).GetCodeName());
      assertEquals("dead species 1 test bool 1 c " + OutputBehaviors.FIRE, oSettings.GetBool(0).GetCodeName());
      
      //Storm
      oSettings = oOutput.GetDetailedDeadTreeSetting(5);
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(OutputBehaviors.STORM, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetSaveFrequency());
      assertEquals(1, oSettings.GetNumberOfBools());
      assertEquals(1, oSettings.GetNumberOfChars());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals(1, oSettings.GetNumberOfInts());
      assertEquals("dead species 1 test float 1 c " + OutputBehaviors.STORM, oSettings.GetFloat(0).GetCodeName());
      assertEquals("dead species 1 test char 1 c " + OutputBehaviors.STORM, oSettings.GetChar(0).GetCodeName());
      assertEquals("dead species 1 test int 1 c " + OutputBehaviors.STORM, oSettings.GetInt(0).GetCodeName());
      assertEquals("dead species 1 test bool 1 c " + OutputBehaviors.STORM, oSettings.GetBool(0).GetCodeName());
      
      //Multiples of each - not grouped together by type
      oSettings = oOutput.GetDetailedDeadTreeSetting(6);
      assertEquals(1, oSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(OutputBehaviors.STORM, oSettings.GetDeadCode());
      assertEquals(2, oSettings.GetSaveFrequency());
      assertEquals(2, oSettings.GetNumberOfBools());
      assertEquals(2, oSettings.GetNumberOfChars());
      assertEquals(2, oSettings.GetNumberOfFloats());
      assertEquals(2, oSettings.GetNumberOfInts());

      /*for (i = 0; i < oOutput.GetNumberOfDetailedGridSettings(); i++) {
        DetailedGridSettings oSettings = (DetailedGridSettings) oOutput.
            GetDetailedGridSetting(i);
        if (oSettings.GetName().equals("grid 1")) {
          assertEquals(1, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 1 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 1 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 1 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 1 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 2")) {

          assertEquals(2, oSettings.GetSaveFrequency());
          assertEquals(2, oSettings.GetNumberOfBools());
          assertEquals(2, oSettings.GetNumberOfChars());
          assertEquals(2, oSettings.GetNumberOfFloats());
          assertEquals(2, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 2 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 2 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 2 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 2 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("grid 2 test float 2",
                       oSettings.GetFloat(1).GetCodeName());
          assertEquals("grid 2 test char 2",
                       oSettings.GetChar(1).GetCodeName());
          assertEquals("grid 2 test int 2",
                       oSettings.GetInt(1).GetCodeName());
          assertEquals("grid 2 test bool 2",
                       oSettings.GetBool(1).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 3")) {

          assertEquals(3, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 3 test char 1",
                       oSettings.GetChar(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 4")) {

          assertEquals(4, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 4 test float 1",
                       oSettings.GetFloat(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 5")) {

          assertEquals(5, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 5 test bool 1",
                       oSettings.GetBool(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 6")) {

          assertEquals(6, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals(0, oSettings.GetNumberOfPackageBools());
          assertEquals(0, oSettings.GetNumberOfPackageChars());
          assertEquals(0, oSettings.GetNumberOfPackageFloats());
          assertEquals(0, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 6 test int 1",
                       oSettings.GetInt(0).GetCodeName());

        }
        else if (oSettings.GetName().equals("grid 7")) {

          assertEquals(7, oSettings.GetSaveFrequency());
          assertEquals(1, oSettings.GetNumberOfBools());
          assertEquals(1, oSettings.GetNumberOfChars());
          assertEquals(1, oSettings.GetNumberOfFloats());
          assertEquals(1, oSettings.GetNumberOfInts());
          assertEquals(1, oSettings.GetNumberOfPackageBools());
          assertEquals(1, oSettings.GetNumberOfPackageChars());
          assertEquals(1, oSettings.GetNumberOfPackageFloats());
          assertEquals(1, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 7 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 7 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 7 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 7 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("grid 7 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 7 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 7 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 7 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
        }
        else if (oSettings.GetName().equals("grid 8")) {

          assertEquals(8, oSettings.GetSaveFrequency());
          assertEquals(2, oSettings.GetNumberOfBools());
          assertEquals(2, oSettings.GetNumberOfChars());
          assertEquals(2, oSettings.GetNumberOfFloats());
          assertEquals(2, oSettings.GetNumberOfInts());
          assertEquals(5, oSettings.GetNumberOfPackageBools());
          assertEquals(4, oSettings.GetNumberOfPackageChars());
          assertEquals(3, oSettings.GetNumberOfPackageFloats());
          assertEquals(2, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 8 test float 1",
                       oSettings.GetFloat(0).GetCodeName());
          assertEquals("grid 8 test char 1",
                       oSettings.GetChar(0).GetCodeName());
          assertEquals("grid 8 test int 1",
                       oSettings.GetInt(0).GetCodeName());
          assertEquals("grid 8 test bool 1",
                       oSettings.GetBool(0).GetCodeName());
          assertEquals("grid 8 test float 2",
                       oSettings.GetFloat(1).GetCodeName());
          assertEquals("grid 8 test char 2",
                       oSettings.GetChar(1).GetCodeName());
          assertEquals("grid 8 test int 2",
                       oSettings.GetInt(1).GetCodeName());
          assertEquals("grid 8 test bool 2",
                       oSettings.GetBool(1).GetCodeName());
          assertEquals("grid 8 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 8 test package int 2",
                       oSettings.GetPackageInt(1).GetCodeName());
          assertEquals("grid 8 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 8 test package float 2",
                       oSettings.GetPackageFloat(1).GetCodeName());
          assertEquals("grid 8 test package float 3",
                       oSettings.GetPackageFloat(2).GetCodeName());
          assertEquals("grid 8 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 8 test package char 2",
                       oSettings.GetPackageChar(1).GetCodeName());
          assertEquals("grid 8 test package char 3",
                       oSettings.GetPackageChar(2).GetCodeName());
          assertEquals("grid 8 test package char 4",
                       oSettings.GetPackageChar(3).GetCodeName());
          assertEquals("grid 8 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
          assertEquals("grid 8 test package bool 2",
                       oSettings.GetPackageBool(1).GetCodeName());
          assertEquals("grid 8 test package bool 3",
                       oSettings.GetPackageBool(2).GetCodeName());
          assertEquals("grid 8 test package bool 4",
                       oSettings.GetPackageBool(3).GetCodeName());
          assertEquals("grid 8 test package bool 5",
                       oSettings.GetPackageBool(4).GetCodeName());
        }
        else if (oSettings.GetName().equals("grid 9")) {

          assertEquals(9, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(1, oSettings.GetNumberOfPackageBools());
          assertEquals(1, oSettings.GetNumberOfPackageChars());
          assertEquals(1, oSettings.GetNumberOfPackageFloats());
          assertEquals(1, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 9 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 9 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 9 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 9 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
        }
        else if (oSettings.GetName().equals("grid 10")) {

          assertEquals(10, oSettings.GetSaveFrequency());
          assertEquals(0, oSettings.GetNumberOfBools());
          assertEquals(0, oSettings.GetNumberOfChars());
          assertEquals(0, oSettings.GetNumberOfFloats());
          assertEquals(0, oSettings.GetNumberOfInts());
          assertEquals(5, oSettings.GetNumberOfPackageBools());
          assertEquals(4, oSettings.GetNumberOfPackageChars());
          assertEquals(3, oSettings.GetNumberOfPackageFloats());
          assertEquals(2, oSettings.GetNumberOfPackageInts());
          assertEquals("grid 10 test package int 1",
                       oSettings.GetPackageInt(0).GetCodeName());
          assertEquals("grid 10 test package int 2",
                       oSettings.GetPackageInt(1).GetCodeName());
          assertEquals("grid 10 test package float 1",
                       oSettings.GetPackageFloat(0).GetCodeName());
          assertEquals("grid 10 test package float 2",
                       oSettings.GetPackageFloat(1).GetCodeName());
          assertEquals("grid 10 test package float 3",
                       oSettings.GetPackageFloat(2).GetCodeName());
          assertEquals("grid 10 test package char 1",
                       oSettings.GetPackageChar(0).GetCodeName());
          assertEquals("grid 10 test package char 2",
                       oSettings.GetPackageChar(1).GetCodeName());
          assertEquals("grid 10 test package char 3",
                       oSettings.GetPackageChar(2).GetCodeName());
          assertEquals("grid 10 test package char 4",
                       oSettings.GetPackageChar(3).GetCodeName());
          assertEquals("grid 10 test package bool 1",
                       oSettings.GetPackageBool(0).GetCodeName());
          assertEquals("grid 10 test package bool 2",
                       oSettings.GetPackageBool(1).GetCodeName());
          assertEquals("grid 10 test package bool 3",
                       oSettings.GetPackageBool(2).GetCodeName());
          assertEquals("grid 10 test package bool 4",
                       oSettings.GetPackageBool(3).GetCodeName());
          assertEquals("grid 10 test package bool 5",
                       oSettings.GetPackageBool(4).GetCodeName());
        }
        else {
          ModelException oErr = new ModelException();
          oErr.sMessage = "Unrecognized grid settings.";
          throw (oErr);
        }
      } */

      //Subplots
      assertEquals(2, oOutput.mp_oDetailedOutputSubplots.size());

      Subplot oSubplot = (Subplot) oOutput.mp_oDetailedOutputSubplots.elementAt(0);
      assertEquals("test 1", oSubplot.GetSubplotName());
      assertEquals(1, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX());
      assertEquals(12, oSubplot.GetCell(0).GetY());

      oSubplot = (Subplot) oOutput.mp_oDetailedOutputSubplots.elementAt(1);
      assertEquals("test 2", oSubplot.GetSubplotName());
      assertEquals(5, oSubplot.GetNumberOfCells());
      assertEquals(6, oSubplot.GetCell(0).GetX());
      assertEquals(14, oSubplot.GetCell(0).GetY());
      assertEquals(7, oSubplot.GetCell(1).GetX());
      assertEquals(15, oSubplot.GetCell(1).GetY());
      assertEquals(8, oSubplot.GetCell(2).GetX());
      assertEquals(16, oSubplot.GetCell(2).GetY());
      assertEquals(9, oSubplot.GetCell(3).GetX());
      assertEquals(17, oSubplot.GetCell(3).GetY());
      assertEquals(10, oSubplot.GetCell(4).GetX());
      assertEquals(18, oSubplot.GetCell(4).GetY());

      assertEquals(0, oOutput.mp_oShortOutputSubplots.size());
      System.out.println("XML rundata reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML rundata reading test failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
    }

  }

  /**
   * Tests shrinking of subplots when the plot resolution is changed.
   */
  public void testChangeOfPlotResolution() {
    String sFileName = null;
    try {
      GUIManager oManager = new GUIManager(null);

      sFileName = WriteXMLFile6();
      oManager.InputXMLParameterFile(sFileName);
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();

      //*****************************
      //Verify the initial data
      //*****************************
      assertEquals(3, oOutput.mp_oShortOutputSubplots.size());
      assertEquals(3, oOutput.mp_oDetailedOutputSubplots.size());

      //Detailed output subplots
      Subplot oSubplot = (Subplot) oOutput.mp_oDetailedOutputSubplots.elementAt(0);
      assertEquals("plot 1", oSubplot.GetSubplotName());
      assertEquals(25, oSubplot.GetNumberOfCells());
      assertEquals(20, oSubplot.GetCell(0).GetX()); assertEquals(20, oSubplot.GetCell(0).GetY());
      assertEquals(20, oSubplot.GetCell(1).GetX()); assertEquals(21, oSubplot.GetCell(1).GetY());
      assertEquals(20, oSubplot.GetCell(2).GetX()); assertEquals(22, oSubplot.GetCell(2).GetY());
      assertEquals(20, oSubplot.GetCell(3).GetX()); assertEquals(23, oSubplot.GetCell(3).GetY());
      assertEquals(20, oSubplot.GetCell(4).GetX()); assertEquals(24, oSubplot.GetCell(4).GetY());
      assertEquals(21, oSubplot.GetCell(5).GetX()); assertEquals(20, oSubplot.GetCell(5).GetY());
      assertEquals(21, oSubplot.GetCell(6).GetX()); assertEquals(21, oSubplot.GetCell(6).GetY());
      assertEquals(21, oSubplot.GetCell(7).GetX()); assertEquals(22, oSubplot.GetCell(7).GetY());
      assertEquals(21, oSubplot.GetCell(8).GetX()); assertEquals(23, oSubplot.GetCell(8).GetY());
      assertEquals(21, oSubplot.GetCell(9).GetX()); assertEquals(24, oSubplot.GetCell(9).GetY());
      assertEquals(22, oSubplot.GetCell(10).GetX()); assertEquals(20, oSubplot.GetCell(10).GetY());
      assertEquals(22, oSubplot.GetCell(11).GetX()); assertEquals(21, oSubplot.GetCell(11).GetY());
      assertEquals(22, oSubplot.GetCell(12).GetX()); assertEquals(22, oSubplot.GetCell(12).GetY());
      assertEquals(22, oSubplot.GetCell(13).GetX()); assertEquals(23, oSubplot.GetCell(13).GetY());
      assertEquals(22, oSubplot.GetCell(14).GetX()); assertEquals(24, oSubplot.GetCell(14).GetY());
      assertEquals(23, oSubplot.GetCell(15).GetX()); assertEquals(20, oSubplot.GetCell(15).GetY());
      assertEquals(23, oSubplot.GetCell(16).GetX()); assertEquals(21, oSubplot.GetCell(16).GetY());
      assertEquals(23, oSubplot.GetCell(17).GetX()); assertEquals(22, oSubplot.GetCell(17).GetY());
      assertEquals(23, oSubplot.GetCell(18).GetX()); assertEquals(23, oSubplot.GetCell(18).GetY());
      assertEquals(23, oSubplot.GetCell(19).GetX()); assertEquals(24, oSubplot.GetCell(19).GetY());
      assertEquals(24, oSubplot.GetCell(20).GetX()); assertEquals(20, oSubplot.GetCell(20).GetY());
      assertEquals(24, oSubplot.GetCell(21).GetX()); assertEquals(21, oSubplot.GetCell(21).GetY());
      assertEquals(24, oSubplot.GetCell(22).GetX()); assertEquals(22, oSubplot.GetCell(22).GetY());
      assertEquals(24, oSubplot.GetCell(23).GetX()); assertEquals(23, oSubplot.GetCell(23).GetY());

      oSubplot = (Subplot) oOutput.mp_oDetailedOutputSubplots.elementAt(1);
      assertEquals("plot 2", oSubplot.GetSubplotName());
      assertEquals(48, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(0, oSubplot.GetCell(6).GetX()); assertEquals(6, oSubplot.GetCell(6).GetY());
      assertEquals(0, oSubplot.GetCell(7).GetX()); assertEquals(7, oSubplot.GetCell(7).GetY());
      assertEquals(0, oSubplot.GetCell(8).GetX()); assertEquals(8, oSubplot.GetCell(8).GetY());
      assertEquals(0, oSubplot.GetCell(9).GetX()); assertEquals(9, oSubplot.GetCell(9).GetY());
      assertEquals(0, oSubplot.GetCell(10).GetX()); assertEquals(10, oSubplot.GetCell(10).GetY());
      assertEquals(0, oSubplot.GetCell(11).GetX()); assertEquals(11, oSubplot.GetCell(11).GetY());
      assertEquals(0, oSubplot.GetCell(12).GetX()); assertEquals(12, oSubplot.GetCell(12).GetY());
      assertEquals(0, oSubplot.GetCell(13).GetX()); assertEquals(13, oSubplot.GetCell(13).GetY());
      assertEquals(0, oSubplot.GetCell(14).GetX()); assertEquals(14, oSubplot.GetCell(14).GetY());
      assertEquals(0, oSubplot.GetCell(15).GetX()); assertEquals(15, oSubplot.GetCell(15).GetY());
      assertEquals(0, oSubplot.GetCell(16).GetX()); assertEquals(16, oSubplot.GetCell(16).GetY());
      assertEquals(0, oSubplot.GetCell(17).GetX()); assertEquals(17, oSubplot.GetCell(17).GetY());
      assertEquals(0, oSubplot.GetCell(18).GetX()); assertEquals(24, oSubplot.GetCell(18).GetY());
      assertEquals(1, oSubplot.GetCell(19).GetX()); assertEquals(24, oSubplot.GetCell(19).GetY());
      assertEquals(2, oSubplot.GetCell(20).GetX()); assertEquals(0, oSubplot.GetCell(20).GetY());
      assertEquals(2, oSubplot.GetCell(21).GetX()); assertEquals(1, oSubplot.GetCell(21).GetY());
      assertEquals(2, oSubplot.GetCell(22).GetX()); assertEquals(2, oSubplot.GetCell(22).GetY());
      assertEquals(2, oSubplot.GetCell(23).GetX()); assertEquals(3, oSubplot.GetCell(23).GetY());
      assertEquals(2, oSubplot.GetCell(24).GetX()); assertEquals(4, oSubplot.GetCell(24).GetY());
      assertEquals(2, oSubplot.GetCell(25).GetX()); assertEquals(5, oSubplot.GetCell(25).GetY());
      assertEquals(2, oSubplot.GetCell(26).GetX()); assertEquals(6, oSubplot.GetCell(26).GetY());
      assertEquals(2, oSubplot.GetCell(27).GetX()); assertEquals(7, oSubplot.GetCell(27).GetY());
      assertEquals(2, oSubplot.GetCell(28).GetX()); assertEquals(8, oSubplot.GetCell(28).GetY());
      assertEquals(2, oSubplot.GetCell(29).GetX()); assertEquals(9, oSubplot.GetCell(29).GetY());
      assertEquals(2, oSubplot.GetCell(30).GetX()); assertEquals(10, oSubplot.GetCell(30).GetY());
      assertEquals(2, oSubplot.GetCell(31).GetX()); assertEquals(11, oSubplot.GetCell(31).GetY());
      assertEquals(2, oSubplot.GetCell(32).GetX()); assertEquals(12, oSubplot.GetCell(32).GetY());
      assertEquals(2, oSubplot.GetCell(33).GetX()); assertEquals(13, oSubplot.GetCell(33).GetY());
      assertEquals(2, oSubplot.GetCell(34).GetX()); assertEquals(14, oSubplot.GetCell(34).GetY());
      assertEquals(2, oSubplot.GetCell(35).GetX()); assertEquals(15, oSubplot.GetCell(35).GetY());
      assertEquals(2, oSubplot.GetCell(36).GetX()); assertEquals(16, oSubplot.GetCell(36).GetY());
      assertEquals(2, oSubplot.GetCell(37).GetX()); assertEquals(17, oSubplot.GetCell(37).GetY());
      assertEquals(2, oSubplot.GetCell(38).GetX()); assertEquals(24, oSubplot.GetCell(38).GetY());
      assertEquals(3, oSubplot.GetCell(39).GetX()); assertEquals(24, oSubplot.GetCell(39).GetY());
      assertEquals(4, oSubplot.GetCell(40).GetX()); assertEquals(24, oSubplot.GetCell(40).GetY());
      assertEquals(5, oSubplot.GetCell(41).GetX()); assertEquals(24, oSubplot.GetCell(41).GetY());
      assertEquals(24, oSubplot.GetCell(42).GetX()); assertEquals(0, oSubplot.GetCell(42).GetY());
      assertEquals(24, oSubplot.GetCell(43).GetX()); assertEquals(1, oSubplot.GetCell(43).GetY());
      assertEquals(24, oSubplot.GetCell(44).GetX()); assertEquals(2, oSubplot.GetCell(44).GetY());
      assertEquals(24, oSubplot.GetCell(45).GetX()); assertEquals(3, oSubplot.GetCell(45).GetY());
      assertEquals(24, oSubplot.GetCell(46).GetX()); assertEquals(4, oSubplot.GetCell(46).GetY());
      assertEquals(24, oSubplot.GetCell(47).GetX()); assertEquals(5, oSubplot.GetCell(47).GetY());

      oSubplot = (Subplot) oOutput.mp_oDetailedOutputSubplots.elementAt(2);
      assertEquals("plot 3", oSubplot.GetSubplotName());
      assertEquals(72, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(1, oSubplot.GetCell(6).GetX()); assertEquals(0, oSubplot.GetCell(6).GetY());
      assertEquals(1, oSubplot.GetCell(7).GetX()); assertEquals(1, oSubplot.GetCell(7).GetY());
      assertEquals(1, oSubplot.GetCell(8).GetX()); assertEquals(2, oSubplot.GetCell(8).GetY());
      assertEquals(1, oSubplot.GetCell(9).GetX()); assertEquals(3, oSubplot.GetCell(9).GetY());
      assertEquals(1, oSubplot.GetCell(10).GetX()); assertEquals(4, oSubplot.GetCell(10).GetY());
      assertEquals(1, oSubplot.GetCell(11).GetX()); assertEquals(5, oSubplot.GetCell(11).GetY());
      assertEquals(2, oSubplot.GetCell(12).GetX()); assertEquals(0, oSubplot.GetCell(12).GetY());
      assertEquals(2, oSubplot.GetCell(13).GetX()); assertEquals(1, oSubplot.GetCell(13).GetY());
      assertEquals(2, oSubplot.GetCell(14).GetX()); assertEquals(2, oSubplot.GetCell(14).GetY());
      assertEquals(2, oSubplot.GetCell(15).GetX()); assertEquals(3, oSubplot.GetCell(15).GetY());
      assertEquals(2, oSubplot.GetCell(16).GetX()); assertEquals(4, oSubplot.GetCell(16).GetY());
      assertEquals(2, oSubplot.GetCell(17).GetX()); assertEquals(5, oSubplot.GetCell(17).GetY());
      assertEquals(3, oSubplot.GetCell(18).GetX()); assertEquals(0, oSubplot.GetCell(18).GetY());
      assertEquals(3, oSubplot.GetCell(19).GetX()); assertEquals(1, oSubplot.GetCell(19).GetY());
      assertEquals(3, oSubplot.GetCell(20).GetX()); assertEquals(2, oSubplot.GetCell(20).GetY());
      assertEquals(3, oSubplot.GetCell(21).GetX()); assertEquals(3, oSubplot.GetCell(21).GetY());
      assertEquals(3, oSubplot.GetCell(22).GetX()); assertEquals(4, oSubplot.GetCell(22).GetY());
      assertEquals(3, oSubplot.GetCell(23).GetX()); assertEquals(5, oSubplot.GetCell(23).GetY());
      assertEquals(4, oSubplot.GetCell(24).GetX()); assertEquals(0, oSubplot.GetCell(24).GetY());
      assertEquals(4, oSubplot.GetCell(25).GetX()); assertEquals(1, oSubplot.GetCell(25).GetY());
      assertEquals(4, oSubplot.GetCell(26).GetX()); assertEquals(2, oSubplot.GetCell(26).GetY());
      assertEquals(4, oSubplot.GetCell(27).GetX()); assertEquals(3, oSubplot.GetCell(27).GetY());
      assertEquals(4, oSubplot.GetCell(28).GetX()); assertEquals(4, oSubplot.GetCell(28).GetY());
      assertEquals(4, oSubplot.GetCell(29).GetX()); assertEquals(5, oSubplot.GetCell(29).GetY());
      assertEquals(5, oSubplot.GetCell(30).GetX()); assertEquals(0, oSubplot.GetCell(30).GetY());
      assertEquals(5, oSubplot.GetCell(31).GetX()); assertEquals(1, oSubplot.GetCell(31).GetY());
      assertEquals(5, oSubplot.GetCell(32).GetX()); assertEquals(2, oSubplot.GetCell(32).GetY());
      assertEquals(5, oSubplot.GetCell(33).GetX()); assertEquals(3, oSubplot.GetCell(33).GetY());
      assertEquals(5, oSubplot.GetCell(34).GetX()); assertEquals(4, oSubplot.GetCell(34).GetY());
      assertEquals(5, oSubplot.GetCell(35).GetX()); assertEquals(5, oSubplot.GetCell(35).GetY());
      assertEquals(6, oSubplot.GetCell(36).GetX()); assertEquals(0, oSubplot.GetCell(36).GetY());
      assertEquals(6, oSubplot.GetCell(37).GetX()); assertEquals(1, oSubplot.GetCell(37).GetY());
      assertEquals(6, oSubplot.GetCell(38).GetX()); assertEquals(2, oSubplot.GetCell(38).GetY());
      assertEquals(6, oSubplot.GetCell(39).GetX()); assertEquals(3, oSubplot.GetCell(39).GetY());
      assertEquals(6, oSubplot.GetCell(40).GetX()); assertEquals(4, oSubplot.GetCell(40).GetY());
      assertEquals(6, oSubplot.GetCell(41).GetX()); assertEquals(5, oSubplot.GetCell(41).GetY());
      assertEquals(7, oSubplot.GetCell(42).GetX()); assertEquals(0, oSubplot.GetCell(42).GetY());
      assertEquals(7, oSubplot.GetCell(43).GetX()); assertEquals(1, oSubplot.GetCell(43).GetY());
      assertEquals(7, oSubplot.GetCell(44).GetX()); assertEquals(2, oSubplot.GetCell(44).GetY());
      assertEquals(7, oSubplot.GetCell(45).GetX()); assertEquals(3, oSubplot.GetCell(45).GetY());
      assertEquals(7, oSubplot.GetCell(46).GetX()); assertEquals(4, oSubplot.GetCell(46).GetY());
      assertEquals(7, oSubplot.GetCell(47).GetX()); assertEquals(5, oSubplot.GetCell(47).GetY());
      assertEquals(8, oSubplot.GetCell(48).GetX()); assertEquals(0, oSubplot.GetCell(48).GetY());
      assertEquals(8, oSubplot.GetCell(49).GetX()); assertEquals(1, oSubplot.GetCell(49).GetY());
      assertEquals(8, oSubplot.GetCell(50).GetX()); assertEquals(2, oSubplot.GetCell(50).GetY());
      assertEquals(8, oSubplot.GetCell(51).GetX()); assertEquals(3, oSubplot.GetCell(51).GetY());
      assertEquals(8, oSubplot.GetCell(52).GetX()); assertEquals(4, oSubplot.GetCell(52).GetY());
      assertEquals(8, oSubplot.GetCell(53).GetX()); assertEquals(5, oSubplot.GetCell(53).GetY());
      assertEquals(9, oSubplot.GetCell(54).GetX()); assertEquals(0, oSubplot.GetCell(54).GetY());
      assertEquals(9, oSubplot.GetCell(55).GetX()); assertEquals(1, oSubplot.GetCell(55).GetY());
      assertEquals(9, oSubplot.GetCell(56).GetX()); assertEquals(2, oSubplot.GetCell(56).GetY());
      assertEquals(9, oSubplot.GetCell(57).GetX()); assertEquals(3, oSubplot.GetCell(57).GetY());
      assertEquals(9, oSubplot.GetCell(58).GetX()); assertEquals(4, oSubplot.GetCell(58).GetY());
      assertEquals(9, oSubplot.GetCell(59).GetX()); assertEquals(5, oSubplot.GetCell(59).GetY());
      assertEquals(10, oSubplot.GetCell(60).GetX()); assertEquals(0, oSubplot.GetCell(60).GetY());
      assertEquals(10, oSubplot.GetCell(61).GetX()); assertEquals(1, oSubplot.GetCell(61).GetY());
      assertEquals(10, oSubplot.GetCell(62).GetX()); assertEquals(2, oSubplot.GetCell(62).GetY());
      assertEquals(10, oSubplot.GetCell(63).GetX()); assertEquals(3, oSubplot.GetCell(63).GetY());
      assertEquals(10, oSubplot.GetCell(64).GetX()); assertEquals(4, oSubplot.GetCell(64).GetY());
      assertEquals(10, oSubplot.GetCell(65).GetX()); assertEquals(5, oSubplot.GetCell(65).GetY());
      assertEquals(11, oSubplot.GetCell(66).GetX()); assertEquals(0, oSubplot.GetCell(66).GetY());
      assertEquals(11, oSubplot.GetCell(67).GetX()); assertEquals(1, oSubplot.GetCell(67).GetY());
      assertEquals(11, oSubplot.GetCell(68).GetX()); assertEquals(2, oSubplot.GetCell(68).GetY());
      assertEquals(11, oSubplot.GetCell(69).GetX()); assertEquals(3, oSubplot.GetCell(69).GetY());
      assertEquals(11, oSubplot.GetCell(70).GetX()); assertEquals(4, oSubplot.GetCell(70).GetY());
      assertEquals(11, oSubplot.GetCell(71).GetX()); assertEquals(5, oSubplot.GetCell(71).GetY());

      //Summary output subplots
      oSubplot = (Subplot) oOutput.mp_oShortOutputSubplots.elementAt(0);
      assertEquals("plot 1", oSubplot.GetSubplotName());
      assertEquals(25, oSubplot.GetNumberOfCells());
      assertEquals(20, oSubplot.GetCell(0).GetX()); assertEquals(20, oSubplot.GetCell(0).GetY());
      assertEquals(20, oSubplot.GetCell(1).GetX()); assertEquals(21, oSubplot.GetCell(1).GetY());
      assertEquals(20, oSubplot.GetCell(2).GetX()); assertEquals(22, oSubplot.GetCell(2).GetY());
      assertEquals(20, oSubplot.GetCell(3).GetX()); assertEquals(23, oSubplot.GetCell(3).GetY());
      assertEquals(20, oSubplot.GetCell(4).GetX()); assertEquals(24, oSubplot.GetCell(4).GetY());
      assertEquals(21, oSubplot.GetCell(5).GetX()); assertEquals(20, oSubplot.GetCell(5).GetY());
      assertEquals(21, oSubplot.GetCell(6).GetX()); assertEquals(21, oSubplot.GetCell(6).GetY());
      assertEquals(21, oSubplot.GetCell(7).GetX()); assertEquals(22, oSubplot.GetCell(7).GetY());
      assertEquals(21, oSubplot.GetCell(8).GetX()); assertEquals(23, oSubplot.GetCell(8).GetY());
      assertEquals(21, oSubplot.GetCell(9).GetX()); assertEquals(24, oSubplot.GetCell(9).GetY());
      assertEquals(22, oSubplot.GetCell(10).GetX()); assertEquals(20, oSubplot.GetCell(10).GetY());
      assertEquals(22, oSubplot.GetCell(11).GetX()); assertEquals(21, oSubplot.GetCell(11).GetY());
      assertEquals(22, oSubplot.GetCell(12).GetX()); assertEquals(22, oSubplot.GetCell(12).GetY());
      assertEquals(22, oSubplot.GetCell(13).GetX()); assertEquals(23, oSubplot.GetCell(13).GetY());
      assertEquals(22, oSubplot.GetCell(14).GetX()); assertEquals(24, oSubplot.GetCell(14).GetY());
      assertEquals(23, oSubplot.GetCell(15).GetX()); assertEquals(20, oSubplot.GetCell(15).GetY());
      assertEquals(23, oSubplot.GetCell(16).GetX()); assertEquals(21, oSubplot.GetCell(16).GetY());
      assertEquals(23, oSubplot.GetCell(17).GetX()); assertEquals(22, oSubplot.GetCell(17).GetY());
      assertEquals(23, oSubplot.GetCell(18).GetX()); assertEquals(23, oSubplot.GetCell(18).GetY());
      assertEquals(23, oSubplot.GetCell(19).GetX()); assertEquals(24, oSubplot.GetCell(19).GetY());
      assertEquals(24, oSubplot.GetCell(20).GetX()); assertEquals(20, oSubplot.GetCell(20).GetY());
      assertEquals(24, oSubplot.GetCell(21).GetX()); assertEquals(21, oSubplot.GetCell(21).GetY());
      assertEquals(24, oSubplot.GetCell(22).GetX()); assertEquals(22, oSubplot.GetCell(22).GetY());
      assertEquals(24, oSubplot.GetCell(23).GetX()); assertEquals(23, oSubplot.GetCell(23).GetY());

      oSubplot = (Subplot) oOutput.mp_oShortOutputSubplots.elementAt(1);
      assertEquals("plot 2", oSubplot.GetSubplotName());
      assertEquals(48, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(0, oSubplot.GetCell(6).GetX()); assertEquals(6, oSubplot.GetCell(6).GetY());
      assertEquals(0, oSubplot.GetCell(7).GetX()); assertEquals(7, oSubplot.GetCell(7).GetY());
      assertEquals(0, oSubplot.GetCell(8).GetX()); assertEquals(8, oSubplot.GetCell(8).GetY());
      assertEquals(0, oSubplot.GetCell(9).GetX()); assertEquals(9, oSubplot.GetCell(9).GetY());
      assertEquals(0, oSubplot.GetCell(10).GetX()); assertEquals(10, oSubplot.GetCell(10).GetY());
      assertEquals(0, oSubplot.GetCell(11).GetX()); assertEquals(11, oSubplot.GetCell(11).GetY());
      assertEquals(0, oSubplot.GetCell(12).GetX()); assertEquals(12, oSubplot.GetCell(12).GetY());
      assertEquals(0, oSubplot.GetCell(13).GetX()); assertEquals(13, oSubplot.GetCell(13).GetY());
      assertEquals(0, oSubplot.GetCell(14).GetX()); assertEquals(14, oSubplot.GetCell(14).GetY());
      assertEquals(0, oSubplot.GetCell(15).GetX()); assertEquals(15, oSubplot.GetCell(15).GetY());
      assertEquals(0, oSubplot.GetCell(16).GetX()); assertEquals(16, oSubplot.GetCell(16).GetY());
      assertEquals(0, oSubplot.GetCell(17).GetX()); assertEquals(17, oSubplot.GetCell(17).GetY());
      assertEquals(0, oSubplot.GetCell(18).GetX()); assertEquals(24, oSubplot.GetCell(18).GetY());
      assertEquals(1, oSubplot.GetCell(19).GetX()); assertEquals(24, oSubplot.GetCell(19).GetY());
      assertEquals(2, oSubplot.GetCell(20).GetX()); assertEquals(0, oSubplot.GetCell(20).GetY());
      assertEquals(2, oSubplot.GetCell(21).GetX()); assertEquals(1, oSubplot.GetCell(21).GetY());
      assertEquals(2, oSubplot.GetCell(22).GetX()); assertEquals(2, oSubplot.GetCell(22).GetY());
      assertEquals(2, oSubplot.GetCell(23).GetX()); assertEquals(3, oSubplot.GetCell(23).GetY());
      assertEquals(2, oSubplot.GetCell(24).GetX()); assertEquals(4, oSubplot.GetCell(24).GetY());
      assertEquals(2, oSubplot.GetCell(25).GetX()); assertEquals(5, oSubplot.GetCell(25).GetY());
      assertEquals(2, oSubplot.GetCell(26).GetX()); assertEquals(6, oSubplot.GetCell(26).GetY());
      assertEquals(2, oSubplot.GetCell(27).GetX()); assertEquals(7, oSubplot.GetCell(27).GetY());
      assertEquals(2, oSubplot.GetCell(28).GetX()); assertEquals(8, oSubplot.GetCell(28).GetY());
      assertEquals(2, oSubplot.GetCell(29).GetX()); assertEquals(9, oSubplot.GetCell(29).GetY());
      assertEquals(2, oSubplot.GetCell(30).GetX()); assertEquals(10, oSubplot.GetCell(30).GetY());
      assertEquals(2, oSubplot.GetCell(31).GetX()); assertEquals(11, oSubplot.GetCell(31).GetY());
      assertEquals(2, oSubplot.GetCell(32).GetX()); assertEquals(12, oSubplot.GetCell(32).GetY());
      assertEquals(2, oSubplot.GetCell(33).GetX()); assertEquals(13, oSubplot.GetCell(33).GetY());
      assertEquals(2, oSubplot.GetCell(34).GetX()); assertEquals(14, oSubplot.GetCell(34).GetY());
      assertEquals(2, oSubplot.GetCell(35).GetX()); assertEquals(15, oSubplot.GetCell(35).GetY());
      assertEquals(2, oSubplot.GetCell(36).GetX()); assertEquals(16, oSubplot.GetCell(36).GetY());
      assertEquals(2, oSubplot.GetCell(37).GetX()); assertEquals(17, oSubplot.GetCell(37).GetY());
      assertEquals(2, oSubplot.GetCell(38).GetX()); assertEquals(24, oSubplot.GetCell(38).GetY());
      assertEquals(3, oSubplot.GetCell(39).GetX()); assertEquals(24, oSubplot.GetCell(39).GetY());
      assertEquals(4, oSubplot.GetCell(40).GetX()); assertEquals(24, oSubplot.GetCell(40).GetY());
      assertEquals(5, oSubplot.GetCell(41).GetX()); assertEquals(24, oSubplot.GetCell(41).GetY());
      assertEquals(24, oSubplot.GetCell(42).GetX()); assertEquals(0, oSubplot.GetCell(42).GetY());
      assertEquals(24, oSubplot.GetCell(43).GetX()); assertEquals(1, oSubplot.GetCell(43).GetY());
      assertEquals(24, oSubplot.GetCell(44).GetX()); assertEquals(2, oSubplot.GetCell(44).GetY());
      assertEquals(24, oSubplot.GetCell(45).GetX()); assertEquals(3, oSubplot.GetCell(45).GetY());
      assertEquals(24, oSubplot.GetCell(46).GetX()); assertEquals(4, oSubplot.GetCell(46).GetY());
      assertEquals(24, oSubplot.GetCell(47).GetX()); assertEquals(5, oSubplot.GetCell(47).GetY());

      oSubplot = (Subplot) oOutput.mp_oShortOutputSubplots.elementAt(2);
      assertEquals("plot 3", oSubplot.GetSubplotName());
      assertEquals(72, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(1, oSubplot.GetCell(6).GetX()); assertEquals(0, oSubplot.GetCell(6).GetY());
      assertEquals(1, oSubplot.GetCell(7).GetX()); assertEquals(1, oSubplot.GetCell(7).GetY());
      assertEquals(1, oSubplot.GetCell(8).GetX()); assertEquals(2, oSubplot.GetCell(8).GetY());
      assertEquals(1, oSubplot.GetCell(9).GetX()); assertEquals(3, oSubplot.GetCell(9).GetY());
      assertEquals(1, oSubplot.GetCell(10).GetX()); assertEquals(4, oSubplot.GetCell(10).GetY());
      assertEquals(1, oSubplot.GetCell(11).GetX()); assertEquals(5, oSubplot.GetCell(11).GetY());
      assertEquals(2, oSubplot.GetCell(12).GetX()); assertEquals(0, oSubplot.GetCell(12).GetY());
      assertEquals(2, oSubplot.GetCell(13).GetX()); assertEquals(1, oSubplot.GetCell(13).GetY());
      assertEquals(2, oSubplot.GetCell(14).GetX()); assertEquals(2, oSubplot.GetCell(14).GetY());
      assertEquals(2, oSubplot.GetCell(15).GetX()); assertEquals(3, oSubplot.GetCell(15).GetY());
      assertEquals(2, oSubplot.GetCell(16).GetX()); assertEquals(4, oSubplot.GetCell(16).GetY());
      assertEquals(2, oSubplot.GetCell(17).GetX()); assertEquals(5, oSubplot.GetCell(17).GetY());
      assertEquals(3, oSubplot.GetCell(18).GetX()); assertEquals(0, oSubplot.GetCell(18).GetY());
      assertEquals(3, oSubplot.GetCell(19).GetX()); assertEquals(1, oSubplot.GetCell(19).GetY());
      assertEquals(3, oSubplot.GetCell(20).GetX()); assertEquals(2, oSubplot.GetCell(20).GetY());
      assertEquals(3, oSubplot.GetCell(21).GetX()); assertEquals(3, oSubplot.GetCell(21).GetY());
      assertEquals(3, oSubplot.GetCell(22).GetX()); assertEquals(4, oSubplot.GetCell(22).GetY());
      assertEquals(3, oSubplot.GetCell(23).GetX()); assertEquals(5, oSubplot.GetCell(23).GetY());
      assertEquals(4, oSubplot.GetCell(24).GetX()); assertEquals(0, oSubplot.GetCell(24).GetY());
      assertEquals(4, oSubplot.GetCell(25).GetX()); assertEquals(1, oSubplot.GetCell(25).GetY());
      assertEquals(4, oSubplot.GetCell(26).GetX()); assertEquals(2, oSubplot.GetCell(26).GetY());
      assertEquals(4, oSubplot.GetCell(27).GetX()); assertEquals(3, oSubplot.GetCell(27).GetY());
      assertEquals(4, oSubplot.GetCell(28).GetX()); assertEquals(4, oSubplot.GetCell(28).GetY());
      assertEquals(4, oSubplot.GetCell(29).GetX()); assertEquals(5, oSubplot.GetCell(29).GetY());
      assertEquals(5, oSubplot.GetCell(30).GetX()); assertEquals(0, oSubplot.GetCell(30).GetY());
      assertEquals(5, oSubplot.GetCell(31).GetX()); assertEquals(1, oSubplot.GetCell(31).GetY());
      assertEquals(5, oSubplot.GetCell(32).GetX()); assertEquals(2, oSubplot.GetCell(32).GetY());
      assertEquals(5, oSubplot.GetCell(33).GetX()); assertEquals(3, oSubplot.GetCell(33).GetY());
      assertEquals(5, oSubplot.GetCell(34).GetX()); assertEquals(4, oSubplot.GetCell(34).GetY());
      assertEquals(5, oSubplot.GetCell(35).GetX()); assertEquals(5, oSubplot.GetCell(35).GetY());
      assertEquals(6, oSubplot.GetCell(36).GetX()); assertEquals(0, oSubplot.GetCell(36).GetY());
      assertEquals(6, oSubplot.GetCell(37).GetX()); assertEquals(1, oSubplot.GetCell(37).GetY());
      assertEquals(6, oSubplot.GetCell(38).GetX()); assertEquals(2, oSubplot.GetCell(38).GetY());
      assertEquals(6, oSubplot.GetCell(39).GetX()); assertEquals(3, oSubplot.GetCell(39).GetY());
      assertEquals(6, oSubplot.GetCell(40).GetX()); assertEquals(4, oSubplot.GetCell(40).GetY());
      assertEquals(6, oSubplot.GetCell(41).GetX()); assertEquals(5, oSubplot.GetCell(41).GetY());
      assertEquals(7, oSubplot.GetCell(42).GetX()); assertEquals(0, oSubplot.GetCell(42).GetY());
      assertEquals(7, oSubplot.GetCell(43).GetX()); assertEquals(1, oSubplot.GetCell(43).GetY());
      assertEquals(7, oSubplot.GetCell(44).GetX()); assertEquals(2, oSubplot.GetCell(44).GetY());
      assertEquals(7, oSubplot.GetCell(45).GetX()); assertEquals(3, oSubplot.GetCell(45).GetY());
      assertEquals(7, oSubplot.GetCell(46).GetX()); assertEquals(4, oSubplot.GetCell(46).GetY());
      assertEquals(7, oSubplot.GetCell(47).GetX()); assertEquals(5, oSubplot.GetCell(47).GetY());
      assertEquals(8, oSubplot.GetCell(48).GetX()); assertEquals(0, oSubplot.GetCell(48).GetY());
      assertEquals(8, oSubplot.GetCell(49).GetX()); assertEquals(1, oSubplot.GetCell(49).GetY());
      assertEquals(8, oSubplot.GetCell(50).GetX()); assertEquals(2, oSubplot.GetCell(50).GetY());
      assertEquals(8, oSubplot.GetCell(51).GetX()); assertEquals(3, oSubplot.GetCell(51).GetY());
      assertEquals(8, oSubplot.GetCell(52).GetX()); assertEquals(4, oSubplot.GetCell(52).GetY());
      assertEquals(8, oSubplot.GetCell(53).GetX()); assertEquals(5, oSubplot.GetCell(53).GetY());
      assertEquals(9, oSubplot.GetCell(54).GetX()); assertEquals(0, oSubplot.GetCell(54).GetY());
      assertEquals(9, oSubplot.GetCell(55).GetX()); assertEquals(1, oSubplot.GetCell(55).GetY());
      assertEquals(9, oSubplot.GetCell(56).GetX()); assertEquals(2, oSubplot.GetCell(56).GetY());
      assertEquals(9, oSubplot.GetCell(57).GetX()); assertEquals(3, oSubplot.GetCell(57).GetY());
      assertEquals(9, oSubplot.GetCell(58).GetX()); assertEquals(4, oSubplot.GetCell(58).GetY());
      assertEquals(9, oSubplot.GetCell(59).GetX()); assertEquals(5, oSubplot.GetCell(59).GetY());
      assertEquals(10, oSubplot.GetCell(60).GetX()); assertEquals(0, oSubplot.GetCell(60).GetY());
      assertEquals(10, oSubplot.GetCell(61).GetX()); assertEquals(1, oSubplot.GetCell(61).GetY());
      assertEquals(10, oSubplot.GetCell(62).GetX()); assertEquals(2, oSubplot.GetCell(62).GetY());
      assertEquals(10, oSubplot.GetCell(63).GetX()); assertEquals(3, oSubplot.GetCell(63).GetY());
      assertEquals(10, oSubplot.GetCell(64).GetX()); assertEquals(4, oSubplot.GetCell(64).GetY());
      assertEquals(10, oSubplot.GetCell(65).GetX()); assertEquals(5, oSubplot.GetCell(65).GetY());
      assertEquals(11, oSubplot.GetCell(66).GetX()); assertEquals(0, oSubplot.GetCell(66).GetY());
      assertEquals(11, oSubplot.GetCell(67).GetX()); assertEquals(1, oSubplot.GetCell(67).GetY());
      assertEquals(11, oSubplot.GetCell(68).GetX()); assertEquals(2, oSubplot.GetCell(68).GetY());
      assertEquals(11, oSubplot.GetCell(69).GetX()); assertEquals(3, oSubplot.GetCell(69).GetY());
      assertEquals(11, oSubplot.GetCell(70).GetX()); assertEquals(4, oSubplot.GetCell(70).GetY());
      assertEquals(11, oSubplot.GetCell(71).GetX()); assertEquals(5, oSubplot.GetCell(71).GetY());

      //*****************************
      //Change the plot resolution
      //*****************************
      oManager.ChangeOfPlotResolution(200, 200, 100, 120);

      assertEquals(2, oOutput.mp_oShortOutputSubplots.size());
      assertEquals(2, oOutput.mp_oDetailedOutputSubplots.size());

      //Detailed output subplots
      oSubplot = (Subplot) oOutput.mp_oDetailedOutputSubplots.elementAt(0);
      assertEquals("plot 2", oSubplot.GetSubplotName());
      assertEquals(30, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(0, oSubplot.GetCell(6).GetX()); assertEquals(6, oSubplot.GetCell(6).GetY());
      assertEquals(0, oSubplot.GetCell(7).GetX()); assertEquals(7, oSubplot.GetCell(7).GetY());
      assertEquals(0, oSubplot.GetCell(8).GetX()); assertEquals(8, oSubplot.GetCell(8).GetY());
      assertEquals(0, oSubplot.GetCell(9).GetX()); assertEquals(9, oSubplot.GetCell(9).GetY());
      assertEquals(0, oSubplot.GetCell(10).GetX()); assertEquals(10, oSubplot.GetCell(10).GetY());
      assertEquals(0, oSubplot.GetCell(11).GetX()); assertEquals(11, oSubplot.GetCell(11).GetY());
      assertEquals(0, oSubplot.GetCell(12).GetX()); assertEquals(12, oSubplot.GetCell(12).GetY());
      assertEquals(0, oSubplot.GetCell(13).GetX()); assertEquals(13, oSubplot.GetCell(13).GetY());
      assertEquals(0, oSubplot.GetCell(14).GetX()); assertEquals(14, oSubplot.GetCell(14).GetY());
      assertEquals(2, oSubplot.GetCell(15).GetX()); assertEquals(0, oSubplot.GetCell(15).GetY());
      assertEquals(2, oSubplot.GetCell(16).GetX()); assertEquals(1, oSubplot.GetCell(16).GetY());
      assertEquals(2, oSubplot.GetCell(17).GetX()); assertEquals(2, oSubplot.GetCell(17).GetY());
      assertEquals(2, oSubplot.GetCell(18).GetX()); assertEquals(3, oSubplot.GetCell(18).GetY());
      assertEquals(2, oSubplot.GetCell(19).GetX()); assertEquals(4, oSubplot.GetCell(19).GetY());
      assertEquals(2, oSubplot.GetCell(20).GetX()); assertEquals(5, oSubplot.GetCell(20).GetY());
      assertEquals(2, oSubplot.GetCell(21).GetX()); assertEquals(6, oSubplot.GetCell(21).GetY());
      assertEquals(2, oSubplot.GetCell(22).GetX()); assertEquals(7, oSubplot.GetCell(22).GetY());
      assertEquals(2, oSubplot.GetCell(23).GetX()); assertEquals(8, oSubplot.GetCell(23).GetY());
      assertEquals(2, oSubplot.GetCell(24).GetX()); assertEquals(9, oSubplot.GetCell(24).GetY());
      assertEquals(2, oSubplot.GetCell(25).GetX()); assertEquals(10, oSubplot.GetCell(25).GetY());
      assertEquals(2, oSubplot.GetCell(26).GetX()); assertEquals(11, oSubplot.GetCell(26).GetY());
      assertEquals(2, oSubplot.GetCell(27).GetX()); assertEquals(12, oSubplot.GetCell(27).GetY());
      assertEquals(2, oSubplot.GetCell(28).GetX()); assertEquals(13, oSubplot.GetCell(28).GetY());
      assertEquals(2, oSubplot.GetCell(29).GetX()); assertEquals(14, oSubplot.GetCell(29).GetY());

      oSubplot = (Subplot) oOutput.mp_oDetailedOutputSubplots.elementAt(1);
      assertEquals("plot 3", oSubplot.GetSubplotName());
      assertEquals(72, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(1, oSubplot.GetCell(6).GetX()); assertEquals(0, oSubplot.GetCell(6).GetY());
      assertEquals(1, oSubplot.GetCell(7).GetX()); assertEquals(1, oSubplot.GetCell(7).GetY());
      assertEquals(1, oSubplot.GetCell(8).GetX()); assertEquals(2, oSubplot.GetCell(8).GetY());
      assertEquals(1, oSubplot.GetCell(9).GetX()); assertEquals(3, oSubplot.GetCell(9).GetY());
      assertEquals(1, oSubplot.GetCell(10).GetX()); assertEquals(4, oSubplot.GetCell(10).GetY());
      assertEquals(1, oSubplot.GetCell(11).GetX()); assertEquals(5, oSubplot.GetCell(11).GetY());
      assertEquals(2, oSubplot.GetCell(12).GetX()); assertEquals(0, oSubplot.GetCell(12).GetY());
      assertEquals(2, oSubplot.GetCell(13).GetX()); assertEquals(1, oSubplot.GetCell(13).GetY());
      assertEquals(2, oSubplot.GetCell(14).GetX()); assertEquals(2, oSubplot.GetCell(14).GetY());
      assertEquals(2, oSubplot.GetCell(15).GetX()); assertEquals(3, oSubplot.GetCell(15).GetY());
      assertEquals(2, oSubplot.GetCell(16).GetX()); assertEquals(4, oSubplot.GetCell(16).GetY());
      assertEquals(2, oSubplot.GetCell(17).GetX()); assertEquals(5, oSubplot.GetCell(17).GetY());
      assertEquals(3, oSubplot.GetCell(18).GetX()); assertEquals(0, oSubplot.GetCell(18).GetY());
      assertEquals(3, oSubplot.GetCell(19).GetX()); assertEquals(1, oSubplot.GetCell(19).GetY());
      assertEquals(3, oSubplot.GetCell(20).GetX()); assertEquals(2, oSubplot.GetCell(20).GetY());
      assertEquals(3, oSubplot.GetCell(21).GetX()); assertEquals(3, oSubplot.GetCell(21).GetY());
      assertEquals(3, oSubplot.GetCell(22).GetX()); assertEquals(4, oSubplot.GetCell(22).GetY());
      assertEquals(3, oSubplot.GetCell(23).GetX()); assertEquals(5, oSubplot.GetCell(23).GetY());
      assertEquals(4, oSubplot.GetCell(24).GetX()); assertEquals(0, oSubplot.GetCell(24).GetY());
      assertEquals(4, oSubplot.GetCell(25).GetX()); assertEquals(1, oSubplot.GetCell(25).GetY());
      assertEquals(4, oSubplot.GetCell(26).GetX()); assertEquals(2, oSubplot.GetCell(26).GetY());
      assertEquals(4, oSubplot.GetCell(27).GetX()); assertEquals(3, oSubplot.GetCell(27).GetY());
      assertEquals(4, oSubplot.GetCell(28).GetX()); assertEquals(4, oSubplot.GetCell(28).GetY());
      assertEquals(4, oSubplot.GetCell(29).GetX()); assertEquals(5, oSubplot.GetCell(29).GetY());
      assertEquals(5, oSubplot.GetCell(30).GetX()); assertEquals(0, oSubplot.GetCell(30).GetY());
      assertEquals(5, oSubplot.GetCell(31).GetX()); assertEquals(1, oSubplot.GetCell(31).GetY());
      assertEquals(5, oSubplot.GetCell(32).GetX()); assertEquals(2, oSubplot.GetCell(32).GetY());
      assertEquals(5, oSubplot.GetCell(33).GetX()); assertEquals(3, oSubplot.GetCell(33).GetY());
      assertEquals(5, oSubplot.GetCell(34).GetX()); assertEquals(4, oSubplot.GetCell(34).GetY());
      assertEquals(5, oSubplot.GetCell(35).GetX()); assertEquals(5, oSubplot.GetCell(35).GetY());
      assertEquals(6, oSubplot.GetCell(36).GetX()); assertEquals(0, oSubplot.GetCell(36).GetY());
      assertEquals(6, oSubplot.GetCell(37).GetX()); assertEquals(1, oSubplot.GetCell(37).GetY());
      assertEquals(6, oSubplot.GetCell(38).GetX()); assertEquals(2, oSubplot.GetCell(38).GetY());
      assertEquals(6, oSubplot.GetCell(39).GetX()); assertEquals(3, oSubplot.GetCell(39).GetY());
      assertEquals(6, oSubplot.GetCell(40).GetX()); assertEquals(4, oSubplot.GetCell(40).GetY());
      assertEquals(6, oSubplot.GetCell(41).GetX()); assertEquals(5, oSubplot.GetCell(41).GetY());
      assertEquals(7, oSubplot.GetCell(42).GetX()); assertEquals(0, oSubplot.GetCell(42).GetY());
      assertEquals(7, oSubplot.GetCell(43).GetX()); assertEquals(1, oSubplot.GetCell(43).GetY());
      assertEquals(7, oSubplot.GetCell(44).GetX()); assertEquals(2, oSubplot.GetCell(44).GetY());
      assertEquals(7, oSubplot.GetCell(45).GetX()); assertEquals(3, oSubplot.GetCell(45).GetY());
      assertEquals(7, oSubplot.GetCell(46).GetX()); assertEquals(4, oSubplot.GetCell(46).GetY());
      assertEquals(7, oSubplot.GetCell(47).GetX()); assertEquals(5, oSubplot.GetCell(47).GetY());
      assertEquals(8, oSubplot.GetCell(48).GetX()); assertEquals(0, oSubplot.GetCell(48).GetY());
      assertEquals(8, oSubplot.GetCell(49).GetX()); assertEquals(1, oSubplot.GetCell(49).GetY());
      assertEquals(8, oSubplot.GetCell(50).GetX()); assertEquals(2, oSubplot.GetCell(50).GetY());
      assertEquals(8, oSubplot.GetCell(51).GetX()); assertEquals(3, oSubplot.GetCell(51).GetY());
      assertEquals(8, oSubplot.GetCell(52).GetX()); assertEquals(4, oSubplot.GetCell(52).GetY());
      assertEquals(8, oSubplot.GetCell(53).GetX()); assertEquals(5, oSubplot.GetCell(53).GetY());
      assertEquals(9, oSubplot.GetCell(54).GetX()); assertEquals(0, oSubplot.GetCell(54).GetY());
      assertEquals(9, oSubplot.GetCell(55).GetX()); assertEquals(1, oSubplot.GetCell(55).GetY());
      assertEquals(9, oSubplot.GetCell(56).GetX()); assertEquals(2, oSubplot.GetCell(56).GetY());
      assertEquals(9, oSubplot.GetCell(57).GetX()); assertEquals(3, oSubplot.GetCell(57).GetY());
      assertEquals(9, oSubplot.GetCell(58).GetX()); assertEquals(4, oSubplot.GetCell(58).GetY());
      assertEquals(9, oSubplot.GetCell(59).GetX()); assertEquals(5, oSubplot.GetCell(59).GetY());
      assertEquals(10, oSubplot.GetCell(60).GetX()); assertEquals(0, oSubplot.GetCell(60).GetY());
      assertEquals(10, oSubplot.GetCell(61).GetX()); assertEquals(1, oSubplot.GetCell(61).GetY());
      assertEquals(10, oSubplot.GetCell(62).GetX()); assertEquals(2, oSubplot.GetCell(62).GetY());
      assertEquals(10, oSubplot.GetCell(63).GetX()); assertEquals(3, oSubplot.GetCell(63).GetY());
      assertEquals(10, oSubplot.GetCell(64).GetX()); assertEquals(4, oSubplot.GetCell(64).GetY());
      assertEquals(10, oSubplot.GetCell(65).GetX()); assertEquals(5, oSubplot.GetCell(65).GetY());
      assertEquals(11, oSubplot.GetCell(66).GetX()); assertEquals(0, oSubplot.GetCell(66).GetY());
      assertEquals(11, oSubplot.GetCell(67).GetX()); assertEquals(1, oSubplot.GetCell(67).GetY());
      assertEquals(11, oSubplot.GetCell(68).GetX()); assertEquals(2, oSubplot.GetCell(68).GetY());
      assertEquals(11, oSubplot.GetCell(69).GetX()); assertEquals(3, oSubplot.GetCell(69).GetY());
      assertEquals(11, oSubplot.GetCell(70).GetX()); assertEquals(4, oSubplot.GetCell(70).GetY());
      assertEquals(11, oSubplot.GetCell(71).GetX()); assertEquals(5, oSubplot.GetCell(71).GetY());

      //Summary output subplots
      oSubplot = (Subplot) oOutput.mp_oShortOutputSubplots.elementAt(0);
      assertEquals("plot 2", oSubplot.GetSubplotName());
      assertEquals(30, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(0, oSubplot.GetCell(6).GetX()); assertEquals(6, oSubplot.GetCell(6).GetY());
      assertEquals(0, oSubplot.GetCell(7).GetX()); assertEquals(7, oSubplot.GetCell(7).GetY());
      assertEquals(0, oSubplot.GetCell(8).GetX()); assertEquals(8, oSubplot.GetCell(8).GetY());
      assertEquals(0, oSubplot.GetCell(9).GetX()); assertEquals(9, oSubplot.GetCell(9).GetY());
      assertEquals(0, oSubplot.GetCell(10).GetX()); assertEquals(10, oSubplot.GetCell(10).GetY());
      assertEquals(0, oSubplot.GetCell(11).GetX()); assertEquals(11, oSubplot.GetCell(11).GetY());
      assertEquals(0, oSubplot.GetCell(12).GetX()); assertEquals(12, oSubplot.GetCell(12).GetY());
      assertEquals(0, oSubplot.GetCell(13).GetX()); assertEquals(13, oSubplot.GetCell(13).GetY());
      assertEquals(0, oSubplot.GetCell(14).GetX()); assertEquals(14, oSubplot.GetCell(14).GetY());
      assertEquals(2, oSubplot.GetCell(15).GetX()); assertEquals(0, oSubplot.GetCell(15).GetY());
      assertEquals(2, oSubplot.GetCell(16).GetX()); assertEquals(1, oSubplot.GetCell(16).GetY());
      assertEquals(2, oSubplot.GetCell(17).GetX()); assertEquals(2, oSubplot.GetCell(17).GetY());
      assertEquals(2, oSubplot.GetCell(18).GetX()); assertEquals(3, oSubplot.GetCell(18).GetY());
      assertEquals(2, oSubplot.GetCell(19).GetX()); assertEquals(4, oSubplot.GetCell(19).GetY());
      assertEquals(2, oSubplot.GetCell(20).GetX()); assertEquals(5, oSubplot.GetCell(20).GetY());
      assertEquals(2, oSubplot.GetCell(21).GetX()); assertEquals(6, oSubplot.GetCell(21).GetY());
      assertEquals(2, oSubplot.GetCell(22).GetX()); assertEquals(7, oSubplot.GetCell(22).GetY());
      assertEquals(2, oSubplot.GetCell(23).GetX()); assertEquals(8, oSubplot.GetCell(23).GetY());
      assertEquals(2, oSubplot.GetCell(24).GetX()); assertEquals(9, oSubplot.GetCell(24).GetY());
      assertEquals(2, oSubplot.GetCell(25).GetX()); assertEquals(10, oSubplot.GetCell(25).GetY());
      assertEquals(2, oSubplot.GetCell(26).GetX()); assertEquals(11, oSubplot.GetCell(26).GetY());
      assertEquals(2, oSubplot.GetCell(27).GetX()); assertEquals(12, oSubplot.GetCell(27).GetY());
      assertEquals(2, oSubplot.GetCell(28).GetX()); assertEquals(13, oSubplot.GetCell(28).GetY());
      assertEquals(2, oSubplot.GetCell(29).GetX()); assertEquals(14, oSubplot.GetCell(29).GetY());

      oSubplot = (Subplot) oOutput.mp_oShortOutputSubplots.elementAt(1);
      assertEquals("plot 3", oSubplot.GetSubplotName());
      assertEquals(72, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX()); assertEquals(0, oSubplot.GetCell(0).GetY());
      assertEquals(0, oSubplot.GetCell(1).GetX()); assertEquals(1, oSubplot.GetCell(1).GetY());
      assertEquals(0, oSubplot.GetCell(2).GetX()); assertEquals(2, oSubplot.GetCell(2).GetY());
      assertEquals(0, oSubplot.GetCell(3).GetX()); assertEquals(3, oSubplot.GetCell(3).GetY());
      assertEquals(0, oSubplot.GetCell(4).GetX()); assertEquals(4, oSubplot.GetCell(4).GetY());
      assertEquals(0, oSubplot.GetCell(5).GetX()); assertEquals(5, oSubplot.GetCell(5).GetY());
      assertEquals(1, oSubplot.GetCell(6).GetX()); assertEquals(0, oSubplot.GetCell(6).GetY());
      assertEquals(1, oSubplot.GetCell(7).GetX()); assertEquals(1, oSubplot.GetCell(7).GetY());
      assertEquals(1, oSubplot.GetCell(8).GetX()); assertEquals(2, oSubplot.GetCell(8).GetY());
      assertEquals(1, oSubplot.GetCell(9).GetX()); assertEquals(3, oSubplot.GetCell(9).GetY());
      assertEquals(1, oSubplot.GetCell(10).GetX()); assertEquals(4, oSubplot.GetCell(10).GetY());
      assertEquals(1, oSubplot.GetCell(11).GetX()); assertEquals(5, oSubplot.GetCell(11).GetY());
      assertEquals(2, oSubplot.GetCell(12).GetX()); assertEquals(0, oSubplot.GetCell(12).GetY());
      assertEquals(2, oSubplot.GetCell(13).GetX()); assertEquals(1, oSubplot.GetCell(13).GetY());
      assertEquals(2, oSubplot.GetCell(14).GetX()); assertEquals(2, oSubplot.GetCell(14).GetY());
      assertEquals(2, oSubplot.GetCell(15).GetX()); assertEquals(3, oSubplot.GetCell(15).GetY());
      assertEquals(2, oSubplot.GetCell(16).GetX()); assertEquals(4, oSubplot.GetCell(16).GetY());
      assertEquals(2, oSubplot.GetCell(17).GetX()); assertEquals(5, oSubplot.GetCell(17).GetY());
      assertEquals(3, oSubplot.GetCell(18).GetX()); assertEquals(0, oSubplot.GetCell(18).GetY());
      assertEquals(3, oSubplot.GetCell(19).GetX()); assertEquals(1, oSubplot.GetCell(19).GetY());
      assertEquals(3, oSubplot.GetCell(20).GetX()); assertEquals(2, oSubplot.GetCell(20).GetY());
      assertEquals(3, oSubplot.GetCell(21).GetX()); assertEquals(3, oSubplot.GetCell(21).GetY());
      assertEquals(3, oSubplot.GetCell(22).GetX()); assertEquals(4, oSubplot.GetCell(22).GetY());
      assertEquals(3, oSubplot.GetCell(23).GetX()); assertEquals(5, oSubplot.GetCell(23).GetY());
      assertEquals(4, oSubplot.GetCell(24).GetX()); assertEquals(0, oSubplot.GetCell(24).GetY());
      assertEquals(4, oSubplot.GetCell(25).GetX()); assertEquals(1, oSubplot.GetCell(25).GetY());
      assertEquals(4, oSubplot.GetCell(26).GetX()); assertEquals(2, oSubplot.GetCell(26).GetY());
      assertEquals(4, oSubplot.GetCell(27).GetX()); assertEquals(3, oSubplot.GetCell(27).GetY());
      assertEquals(4, oSubplot.GetCell(28).GetX()); assertEquals(4, oSubplot.GetCell(28).GetY());
      assertEquals(4, oSubplot.GetCell(29).GetX()); assertEquals(5, oSubplot.GetCell(29).GetY());
      assertEquals(5, oSubplot.GetCell(30).GetX()); assertEquals(0, oSubplot.GetCell(30).GetY());
      assertEquals(5, oSubplot.GetCell(31).GetX()); assertEquals(1, oSubplot.GetCell(31).GetY());
      assertEquals(5, oSubplot.GetCell(32).GetX()); assertEquals(2, oSubplot.GetCell(32).GetY());
      assertEquals(5, oSubplot.GetCell(33).GetX()); assertEquals(3, oSubplot.GetCell(33).GetY());
      assertEquals(5, oSubplot.GetCell(34).GetX()); assertEquals(4, oSubplot.GetCell(34).GetY());
      assertEquals(5, oSubplot.GetCell(35).GetX()); assertEquals(5, oSubplot.GetCell(35).GetY());
      assertEquals(6, oSubplot.GetCell(36).GetX()); assertEquals(0, oSubplot.GetCell(36).GetY());
      assertEquals(6, oSubplot.GetCell(37).GetX()); assertEquals(1, oSubplot.GetCell(37).GetY());
      assertEquals(6, oSubplot.GetCell(38).GetX()); assertEquals(2, oSubplot.GetCell(38).GetY());
      assertEquals(6, oSubplot.GetCell(39).GetX()); assertEquals(3, oSubplot.GetCell(39).GetY());
      assertEquals(6, oSubplot.GetCell(40).GetX()); assertEquals(4, oSubplot.GetCell(40).GetY());
      assertEquals(6, oSubplot.GetCell(41).GetX()); assertEquals(5, oSubplot.GetCell(41).GetY());
      assertEquals(7, oSubplot.GetCell(42).GetX()); assertEquals(0, oSubplot.GetCell(42).GetY());
      assertEquals(7, oSubplot.GetCell(43).GetX()); assertEquals(1, oSubplot.GetCell(43).GetY());
      assertEquals(7, oSubplot.GetCell(44).GetX()); assertEquals(2, oSubplot.GetCell(44).GetY());
      assertEquals(7, oSubplot.GetCell(45).GetX()); assertEquals(3, oSubplot.GetCell(45).GetY());
      assertEquals(7, oSubplot.GetCell(46).GetX()); assertEquals(4, oSubplot.GetCell(46).GetY());
      assertEquals(7, oSubplot.GetCell(47).GetX()); assertEquals(5, oSubplot.GetCell(47).GetY());
      assertEquals(8, oSubplot.GetCell(48).GetX()); assertEquals(0, oSubplot.GetCell(48).GetY());
      assertEquals(8, oSubplot.GetCell(49).GetX()); assertEquals(1, oSubplot.GetCell(49).GetY());
      assertEquals(8, oSubplot.GetCell(50).GetX()); assertEquals(2, oSubplot.GetCell(50).GetY());
      assertEquals(8, oSubplot.GetCell(51).GetX()); assertEquals(3, oSubplot.GetCell(51).GetY());
      assertEquals(8, oSubplot.GetCell(52).GetX()); assertEquals(4, oSubplot.GetCell(52).GetY());
      assertEquals(8, oSubplot.GetCell(53).GetX()); assertEquals(5, oSubplot.GetCell(53).GetY());
      assertEquals(9, oSubplot.GetCell(54).GetX()); assertEquals(0, oSubplot.GetCell(54).GetY());
      assertEquals(9, oSubplot.GetCell(55).GetX()); assertEquals(1, oSubplot.GetCell(55).GetY());
      assertEquals(9, oSubplot.GetCell(56).GetX()); assertEquals(2, oSubplot.GetCell(56).GetY());
      assertEquals(9, oSubplot.GetCell(57).GetX()); assertEquals(3, oSubplot.GetCell(57).GetY());
      assertEquals(9, oSubplot.GetCell(58).GetX()); assertEquals(4, oSubplot.GetCell(58).GetY());
      assertEquals(9, oSubplot.GetCell(59).GetX()); assertEquals(5, oSubplot.GetCell(59).GetY());
      assertEquals(10, oSubplot.GetCell(60).GetX()); assertEquals(0, oSubplot.GetCell(60).GetY());
      assertEquals(10, oSubplot.GetCell(61).GetX()); assertEquals(1, oSubplot.GetCell(61).GetY());
      assertEquals(10, oSubplot.GetCell(62).GetX()); assertEquals(2, oSubplot.GetCell(62).GetY());
      assertEquals(10, oSubplot.GetCell(63).GetX()); assertEquals(3, oSubplot.GetCell(63).GetY());
      assertEquals(10, oSubplot.GetCell(64).GetX()); assertEquals(4, oSubplot.GetCell(64).GetY());
      assertEquals(10, oSubplot.GetCell(65).GetX()); assertEquals(5, oSubplot.GetCell(65).GetY());
      assertEquals(11, oSubplot.GetCell(66).GetX()); assertEquals(0, oSubplot.GetCell(66).GetY());
      assertEquals(11, oSubplot.GetCell(67).GetX()); assertEquals(1, oSubplot.GetCell(67).GetY());
      assertEquals(11, oSubplot.GetCell(68).GetX()); assertEquals(2, oSubplot.GetCell(68).GetY());
      assertEquals(11, oSubplot.GetCell(69).GetX()); assertEquals(3, oSubplot.GetCell(69).GetY());
      assertEquals(11, oSubplot.GetCell(70).GetX()); assertEquals(4, oSubplot.GetCell(70).GetY());
      assertEquals(11, oSubplot.GetCell(71).GetX()); assertEquals(5, oSubplot.GetCell(71).GetY());

      System.out.println("ChangeOfPlotResolution test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML rundata reading test failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
    }
  }

  /**
   * Tests reading of parameter file XML settings.
   */
  public void testReadXMLShortSettings1() {
    String sFileName = null;
    try {
      GUIManager oManager = new GUIManager(null);
      sFileName = WriteXMLFile2();
      oManager.InputXMLParameterFile(sFileName);
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SEEDLING, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.SEEDLING));

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.SAPLING));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeBasalArea(TreePopulation.SAPLING));

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.ADULT));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeBasalArea(TreePopulation.ADULT));

      //Subplots
      assertEquals(2, oOutput.mp_oShortOutputSubplots.size());

      Subplot oSubplot = (Subplot) oOutput.mp_oShortOutputSubplots.elementAt(0);
      assertEquals("test 1", oSubplot.GetSubplotName());
      assertEquals(1, oSubplot.GetNumberOfCells());
      assertEquals(0, oSubplot.GetCell(0).GetX());
      assertEquals(12, oSubplot.GetCell(0).GetY());

      oSubplot = (Subplot) oOutput.mp_oShortOutputSubplots.elementAt(1);
      assertEquals("test 2", oSubplot.GetSubplotName());
      assertEquals(5, oSubplot.GetNumberOfCells());
      assertEquals(6, oSubplot.GetCell(0).GetX());
      assertEquals(14, oSubplot.GetCell(0).GetY());
      assertEquals(7, oSubplot.GetCell(1).GetX());
      assertEquals(15, oSubplot.GetCell(1).GetY());
      assertEquals(8, oSubplot.GetCell(2).GetX());
      assertEquals(16, oSubplot.GetCell(2).GetY());
      assertEquals(9, oSubplot.GetCell(3).GetX());
      assertEquals(17, oSubplot.GetCell(3).GetY());
      assertEquals(10, oSubplot.GetCell(4).GetX());
      assertEquals(18, oSubplot.GetCell(4).GetY());

      assertEquals(0, oOutput.mp_oDetailedOutputSubplots.size());
      System.out.println("XML short output reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML short output reading test failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
    }
  }

  /**
   * Tests reading of parameter file XML rundata settings.
   */
  public void testReadXMLShortSettings2() {
    String sFileName = null;
    try {
      GUIManager oManager = new GUIManager(null);
      sFileName = WriteXMLFile3();
      oManager.InputXMLParameterFile(sFileName);
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();

      assertFalse(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeDensity(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SEEDLING, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.SEEDLING));

      assertFalse(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.SAPLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeBasalArea(TreePopulation.SAPLING));

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeDensity(TreePopulation.ADULT));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.ADULT));
      System.out.println("XML short output reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML short output reading test failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
    }
  }
  
  public void testUpdateIsEnabled() {
    String sFileName = null;
    try {
      GUIManager oManager = new GUIManager(null);
      sFileName = WriteXMLFile2();
      oManager.InputXMLParameterFile(sFileName);
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();      
      
      setShortOutputSettingsToFalse(oOutput);      
      Behavior oBeh = oOutput.GetBehaviorByKey("short output");
      assertFalse(oBeh.m_bIsEnabled); 
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.NOTDEAD, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.NOTDEAD, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.NOTDEAD, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.NOTDEAD, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.NOTDEAD, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveRelativeBasalArea(TreePopulation.SAPLING, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveRelativeBasalArea(TreePopulation.ADULT, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveRelativeBasalArea(TreePopulation.SNAG, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveRelativeDensity(TreePopulation.SEEDLING, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveRelativeDensity(TreePopulation.SAPLING, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveRelativeDensity(TreePopulation.ADULT, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveRelativeDensity(TreePopulation.SNAG, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.DISEASE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.DISEASE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.DISEASE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.DISEASE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.DISEASE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.DISEASE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.DISEASE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.FIRE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.FIRE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.FIRE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.FIRE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.FIRE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.FIRE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.FIRE, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.HARVEST, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.HARVEST, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.HARVEST, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.HARVEST, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.HARVEST, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.HARVEST, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.HARVEST, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.INSECTS, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.INSECTS, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.INSECTS, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.INSECTS, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.INSECTS, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.INSECTS, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.INSECTS, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.NATURAL, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.NATURAL, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.NATURAL, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.NATURAL, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.NATURAL, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.NATURAL, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.NATURAL, true);
      assertTrue(oBeh.m_bIsEnabled); 
      setShortOutputSettingsToFalse(oOutput);
      
      System.out.println("Short output enabling test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML short output reading test failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
    }
  }
  
  /**
   * This tests writing and reading short output settings, including dead. 
   */
  public void testWriteXMLShort() {
    String sFileName = null, sFileName2 = "tester.xml"; 
    try {
      GUIManager oManager = new GUIManager(null);
      sFileName = WriteXMLFile3();
      oManager.InputXMLParameterFile(sFileName);
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.DISEASE, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.DISEASE, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.DISEASE, true);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.DISEASE, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.DISEASE, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.DISEASE, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.DISEASE, true);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.FIRE, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.FIRE, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.FIRE, true);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.FIRE, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.FIRE, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.FIRE, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.FIRE, true);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.HARVEST, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.HARVEST, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.HARVEST, true);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.HARVEST, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.HARVEST, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.HARVEST, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.HARVEST, true);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.INSECTS, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.INSECTS, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.INSECTS, true);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.INSECTS, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.INSECTS, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.INSECTS, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.INSECTS, true);
      
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.NATURAL, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.NATURAL, true);
      oOutput.SetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.NATURAL, true);
      
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.NATURAL, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.NATURAL, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.NATURAL, true);
      oOutput.SetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.NATURAL, true);      
      
      oManager.WriteParameterFile(sFileName2); 
      oManager.InputXMLParameterFile(sFileName2);
      oOutput = oManager.GetOutputBehaviors();
      
      assertFalse(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeDensity(TreePopulation.SEEDLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SEEDLING, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.SEEDLING));

      assertFalse(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeDensity(TreePopulation.SAPLING));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.NOTDEAD));
      assertTrue(oOutput.GetSaveRelativeBasalArea(TreePopulation.SAPLING));

      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeDensity(TreePopulation.ADULT));
      assertFalse(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.NOTDEAD));
      assertFalse(oOutput.GetSaveRelativeBasalArea(TreePopulation.ADULT));
      
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.DISEASE));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.DISEASE));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.DISEASE));
      
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.DISEASE));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.DISEASE));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.DISEASE));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.DISEASE));
      
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.FIRE));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.FIRE));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.FIRE));
      
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.FIRE));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.FIRE));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.FIRE));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.FIRE));
      
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.HARVEST));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.HARVEST));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.HARVEST));
      
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.HARVEST));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.HARVEST));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.HARVEST));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.HARVEST));
      
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.INSECTS));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.INSECTS));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.INSECTS));
      
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.INSECTS));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.INSECTS));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.INSECTS));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.INSECTS));
      
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SAPLING, OutputBehaviors.NATURAL));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.ADULT, OutputBehaviors.NATURAL));
      assertTrue(oOutput.GetSaveAbsoluteBasalArea(TreePopulation.SNAG, OutputBehaviors.NATURAL));
      
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SEEDLING, OutputBehaviors.NATURAL));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SAPLING, OutputBehaviors.NATURAL));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.ADULT, OutputBehaviors.NATURAL));
      assertTrue(oOutput.GetSaveAbsoluteDensity(TreePopulation.SNAG, OutputBehaviors.NATURAL)); 
      System.out.println("XML short output reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML short output reading test failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
      new File(sFileName2).delete();
    }
  }
  
  /**
   * This tests writing and reading output settings, including dead. 
   */
  public void testWriteXMLDetailed() {
    String sFileName = null, sFileName2 = "tester.xml"; 
    DetailedTreeSettings oSettings;
    try {
      GUIManager oManager = new GUIManager(null);
      sFileName = WriteXMLFile3();
      oManager.InputXMLParameterFile(sFileName);
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();
      
      oOutput.SetDetailedOutputFileName("figs");
      
      //Try each dead reason code
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0, OutputBehaviors.NOTDEAD);
      oSettings.AddFloat("X", "X");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0, OutputBehaviors.DISEASE);
      oSettings.AddFloat("X", "X");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0, OutputBehaviors.NATURAL);
      oSettings.AddFloat("X", "X");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0, OutputBehaviors.FIRE);
      oSettings.AddFloat("X", "X");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0, OutputBehaviors.STORM);
      oSettings.AddFloat("X", "X");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0, OutputBehaviors.HARVEST);
      oSettings.AddFloat("X", "X");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.ADULT, 0, OutputBehaviors.INSECTS);
      oSettings.AddFloat("X", "X");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 0, OutputBehaviors.NOTDEAD);
      oSettings.AddFloat("Y", "Y");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 0, OutputBehaviors.DISEASE);
      oSettings.AddFloat("Y", "Y");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 0, OutputBehaviors.NATURAL);
      oSettings.AddFloat("Y", "Y");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 0, OutputBehaviors.FIRE);
      oSettings.AddFloat("Y", "Y");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 0, OutputBehaviors.STORM);
      oSettings.AddFloat("Y", "Y");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 0, OutputBehaviors.HARVEST);
      oSettings.AddFloat("Y", "Y");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
      oSettings = new DetailedTreeSettings(TreePopulation.SAPLING, 0, OutputBehaviors.INSECTS);
      oSettings.AddFloat("Y", "Y");
      oSettings.SetSaveFrequency(2);
      oOutput.AddDetailedTreeSettings(oSettings);
      
           
      oManager.WriteParameterFile(sFileName2); 
      oManager.InputXMLParameterFile(sFileName2);
      oOutput = oManager.GetOutputBehaviors();
      
      assertEquals(2, oOutput.GetNumberOfDetailedLiveTreeSettings());
      assertEquals(12, oOutput.GetNumberOfDetailedDeadTreeSettings());
      
      oSettings = oOutput.GetDetailedLiveTreeSetting(0);
      assertEquals(TreePopulation.ADULT, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.NOTDEAD, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("X", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
            
      oSettings = oOutput.GetDetailedDeadTreeSetting(0);
      assertEquals(TreePopulation.ADULT, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.DISEASE, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("X", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(1);
      assertEquals(TreePopulation.ADULT, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.NATURAL, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("X", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(2);
      assertEquals(TreePopulation.ADULT, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.FIRE, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("X", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(3);
      assertEquals(TreePopulation.ADULT, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.STORM, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("X", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(4);
      assertEquals(TreePopulation.ADULT, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.HARVEST, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("X", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(5);
      assertEquals(TreePopulation.ADULT, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.INSECTS, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("X", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedLiveTreeSetting(1);
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.NOTDEAD, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("Y", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(6);
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.DISEASE, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("Y", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(7);
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.NATURAL, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("Y", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(8);
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.FIRE, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("Y", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(9);
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.STORM, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("Y", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(10);
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.HARVEST, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("Y", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      oSettings = oOutput.GetDetailedDeadTreeSetting(11);
      assertEquals(TreePopulation.SAPLING, oSettings.GetType());
      assertEquals(0, oSettings.GetSpecies());
      assertEquals(OutputBehaviors.INSECTS, oSettings.GetDeadCode());
      assertEquals(1, oSettings.GetNumberOfFloats());
      assertEquals("Y", oSettings.GetFloat(0).GetDisplayName());
      assertEquals(2, oSettings.GetSaveFrequency());
      
      
      System.out.println("XML detailed output reading test succeeded.");
    }
    catch (ModelException oErr) {
      fail("XML detailed output reading test failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
      new File(sFileName2).delete();
    }
  }

  /**
   * This writes an XML file to test value setting.  This file contains
   * detailed output only.
   * @return Name of file.
   * @throws ModelException if there is an IO problem.
   */
  protected String WriteXMLFile1() throws ModelException {
    try {
      String sFileName = "\\loratestxml1.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      oOut.write("<paramFile fileCode=\"06010101\">");

      //Plot - so the subplot cells can be set up
      oOut.write("<plot>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>55.37</plot_latitude>");
      oOut.write("</plot>");

      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("<tr_species speciesName=\"Amabilis_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Subalpine_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Hybrid_Spruce\"/>");
      oOut.write("<tr_species speciesName=\"Lodgepole_Pine\"/>");
      oOut.write("<tr_species speciesName=\"Trembling_Aspen\"/>");
      oOut.write("<tr_species speciesName=\"Black_Cottonwood\"/>");
      oOut.write("<tr_species speciesName=\"Paper_Birch\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("</trees>");

      //Include an empty behavior list - this triggers setup
      oOut.write("<behaviorList>");
      oOut.write("</behaviorList>");

      oOut.write("<output>");

      //Straight-up - one per data type
      oOut.write("<ou_filename>OutputSizeTest</ou_filename>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>species 1 test float 1</ou_float>");
      oOut.write("<ou_char>species 1 test char 1</ou_char>");
      oOut.write("<ou_int>species 1 test int 1</ou_int>");
      oOut.write("<ou_bool>species 1 test bool 1</ou_bool>");
      oOut.write("</ou_treeInfo>");

      //Multiples of each - not grouped together by type
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>2</ou_saveFreq>");
      oOut.write("<ou_float>species 2 test float 1</ou_float>");
      oOut.write("<ou_char>species 2 test char 1</ou_char>");
      oOut.write("<ou_int>species 2 test int 1</ou_int>");
      oOut.write("<ou_bool>species 2 test bool 1</ou_bool>");
      oOut.write("<ou_float>species 2 test float 2</ou_float>");
      oOut.write("<ou_char>species 2 test char 2</ou_char>");
      oOut.write("<ou_int>species 2 test int 2</ou_int>");
      oOut.write("<ou_bool>species 2 test bool 2</ou_bool>");
      oOut.write("</ou_treeInfo>");

      //One char only
      oOut.write("<ou_treeInfo species=\"Amabilis_Fir\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>3</ou_saveFreq>");
      oOut.write("<ou_char>species 3 test char 1</ou_char>");
      oOut.write("</ou_treeInfo>");

      //One float only
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>4</ou_saveFreq>");
      oOut.write("<ou_float>species 4 test float 1</ou_float>");
      oOut.write("</ou_treeInfo>");

      //One bool only
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>5</ou_saveFreq>");
      oOut.write("<ou_bool>species 5 test bool 1</ou_bool>");
      oOut.write("</ou_treeInfo>");

      //One int only
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"sapling\">");
      oOut.write("<ou_saveFreq>6</ou_saveFreq>");
      oOut.write("<ou_int>species 6 test int 1</ou_int>");
      oOut.write("</ou_treeInfo>");
      
      //Dead trees
      //One per - do for each dead reason code
      //Harvest
      oOut.write("<ou_deadTreeInfo species=\"Western_Hemlock\" type=\"sapling\" code=\"harvest\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>dead species 1 test float 1 c " + OutputBehaviors.HARVEST + "</ou_float>");
      oOut.write("<ou_char>dead species 1 test char 1 c " + OutputBehaviors.HARVEST + "</ou_char>");
      oOut.write("<ou_int>dead species 1 test int 1 c " + OutputBehaviors.HARVEST + "</ou_int>");
      oOut.write("<ou_bool>dead species 1 test bool 1 c " + OutputBehaviors.HARVEST + "</ou_bool>");
      oOut.write("</ou_deadTreeInfo>");
      
      //Natural
      oOut.write("<ou_deadTreeInfo species=\"Western_Hemlock\" type=\"sapling\" code=\"natural\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>dead species 1 test float 1 c " + OutputBehaviors.NATURAL + "</ou_float>");
      oOut.write("<ou_char>dead species 1 test char 1 c " + OutputBehaviors.NATURAL + "</ou_char>");
      oOut.write("<ou_int>dead species 1 test int 1 c " + OutputBehaviors.NATURAL + "</ou_int>");
      oOut.write("<ou_bool>dead species 1 test bool 1 c " + OutputBehaviors.NATURAL + "</ou_bool>");
      oOut.write("</ou_deadTreeInfo>");
      
      //Insects
      oOut.write("<ou_deadTreeInfo species=\"Western_Hemlock\" type=\"sapling\" code=\"insects\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>dead species 1 test float 1 c " + OutputBehaviors.INSECTS + "</ou_float>");
      oOut.write("<ou_char>dead species 1 test char 1 c " + OutputBehaviors.INSECTS + "</ou_char>");
      oOut.write("<ou_int>dead species 1 test int 1 c " + OutputBehaviors.INSECTS + "</ou_int>");
      oOut.write("<ou_bool>dead species 1 test bool 1 c " + OutputBehaviors.INSECTS + "</ou_bool>");
      oOut.write("</ou_deadTreeInfo>");
      
      //Disease
      oOut.write("<ou_deadTreeInfo species=\"Western_Hemlock\" type=\"sapling\" code=\"disease\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>dead species 1 test float 1 c " + OutputBehaviors.DISEASE + "</ou_float>");
      oOut.write("<ou_char>dead species 1 test char 1 c " + OutputBehaviors.DISEASE + "</ou_char>");
      oOut.write("<ou_int>dead species 1 test int 1 c " + OutputBehaviors.DISEASE + "</ou_int>");
      oOut.write("<ou_bool>dead species 1 test bool 1 c " + OutputBehaviors.DISEASE + "</ou_bool>");
      oOut.write("</ou_deadTreeInfo>");
      
      //Fire
      oOut.write("<ou_deadTreeInfo species=\"Western_Hemlock\" type=\"sapling\" code=\"fire\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>dead species 1 test float 1 c " + OutputBehaviors.FIRE + "</ou_float>");
      oOut.write("<ou_char>dead species 1 test char 1 c " + OutputBehaviors.FIRE + "</ou_char>");
      oOut.write("<ou_int>dead species 1 test int 1 c " + OutputBehaviors.FIRE + "</ou_int>");
      oOut.write("<ou_bool>dead species 1 test bool 1 c " + OutputBehaviors.FIRE + "</ou_bool>");
      oOut.write("</ou_deadTreeInfo>");
      
      //Storm
      oOut.write("<ou_deadTreeInfo species=\"Western_Hemlock\" type=\"sapling\" code=\"storm\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>dead species 1 test float 1 c " + OutputBehaviors.STORM + "</ou_float>");
      oOut.write("<ou_char>dead species 1 test char 1 c " + OutputBehaviors.STORM + "</ou_char>");
      oOut.write("<ou_int>dead species 1 test int 1 c " + OutputBehaviors.STORM + "</ou_int>");
      oOut.write("<ou_bool>dead species 1 test bool 1 c " + OutputBehaviors.STORM + "</ou_bool>");
      oOut.write("</ou_deadTreeInfo>");
      
      //Multiples of each - not grouped together by type
      oOut.write("<ou_deadTreeInfo species=\"Western_Redcedar\" type=\"sapling\" code=\"storm\">");
      oOut.write("<ou_saveFreq>2</ou_saveFreq>");
      oOut.write("<ou_float>dead species 2 test float 1</ou_float>");
      oOut.write("<ou_char>dead species 2 test char 1</ou_char>");
      oOut.write("<ou_int>dead species 2 test int 1</ou_int>");
      oOut.write("<ou_bool>dead species 2 test bool 1</ou_bool>");
      oOut.write("<ou_float>dead species 2 test float 2</ou_float>");
      oOut.write("<ou_char>dead species 2 test char 2</ou_char>");
      oOut.write("<ou_int>dead species 2 test int 2</ou_int>");
      oOut.write("<ou_bool>dead species 2 test bool 2</ou_bool>");
      oOut.write("</ou_deadTreeInfo>");

      //Now grids - same for all
      //Straight-up - one per data type - no package
      oOut.write("<ou_gridInfo gridName=\"grid 1\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>grid 1 test float 1</ou_float>");
      oOut.write("<ou_char>grid 1 test char 1</ou_char>");
      oOut.write("<ou_int>grid 1 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 1 test bool 1</ou_bool>");
      oOut.write("</ou_gridInfo>");

      //Multiples of each - not grouped together by type
      oOut.write("<ou_gridInfo gridName=\"grid 2\">");
      oOut.write("<ou_saveFreq>2</ou_saveFreq>");
      oOut.write("<ou_float>grid 2 test float 1</ou_float>");
      oOut.write("<ou_char>grid 2 test char 1</ou_char>");
      oOut.write("<ou_int>grid 2 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 2 test bool 1</ou_bool>");
      oOut.write("<ou_float>grid 2 test float 2</ou_float>");
      oOut.write("<ou_char>grid 2 test char 2</ou_char>");
      oOut.write("<ou_int>grid 2 test int 2</ou_int>");
      oOut.write("<ou_bool>grid 2 test bool 2</ou_bool>");
      oOut.write("</ou_gridInfo>");

      //One char only
      oOut.write("<ou_gridInfo gridName=\"grid 3\">");
      oOut.write("<ou_saveFreq>3</ou_saveFreq>");
      oOut.write("<ou_char>grid 3 test char 1</ou_char>");
      oOut.write("</ou_gridInfo>");

      //One float only
      oOut.write("<ou_gridInfo gridName=\"grid 4\">");
      oOut.write("<ou_saveFreq>4</ou_saveFreq>");
      oOut.write("<ou_float>grid 4 test float 1</ou_float>");
      oOut.write("</ou_gridInfo>");

      //One bool only
      oOut.write("<ou_gridInfo gridName=\"grid 5\">");
      oOut.write("<ou_saveFreq>5</ou_saveFreq>");
      oOut.write("<ou_bool>grid 5 test bool 1</ou_bool>");
      oOut.write("</ou_gridInfo>");

      //One int only
      oOut.write("<ou_gridInfo gridName=\"grid 6\">");
      oOut.write("<ou_saveFreq>6</ou_saveFreq>");
      oOut.write("<ou_int>grid 6 test int 1</ou_int>");
      oOut.write("</ou_gridInfo>");

      //One per data type, both package and regular
      oOut.write("<ou_gridInfo gridName=\"grid 7\">");
      oOut.write("<ou_saveFreq>7</ou_saveFreq>");
      oOut.write("<ou_float>grid 7 test float 1</ou_float>");
      oOut.write("<ou_char>grid 7 test char 1</ou_char>");
      oOut.write("<ou_int>grid 7 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 7 test bool 1</ou_bool>");
      oOut.write("<ou_packageInt>grid 7 test package int 1</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 7 test package float 1</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 7 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 7 test package bool 1</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      //Multiples of each, both package and regular
      oOut.write("<ou_gridInfo gridName=\"grid 8\">");
      oOut.write("<ou_saveFreq>8</ou_saveFreq>");
      oOut.write("<ou_float>grid 8 test float 1</ou_float>");
      oOut.write("<ou_char>grid 8 test char 1</ou_char>");
      oOut.write("<ou_int>grid 8 test int 1</ou_int>");
      oOut.write("<ou_bool>grid 8 test bool 1</ou_bool>");
      oOut.write("<ou_float>grid 8 test float 2</ou_float>");
      oOut.write("<ou_char>grid 8 test char 2</ou_char>");
      oOut.write("<ou_int>grid 8 test int 2</ou_int>");
      oOut.write("<ou_bool>grid 8 test bool 2</ou_bool>");
      oOut.write("<ou_packageInt>grid 8 test package int 1</ou_packageInt>");
      oOut.write("<ou_packageInt>grid 8 test package int 2</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 8 test package float 1</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 8 test package float 2</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 8 test package float 3</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 8 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 8 test package char 2</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 8 test package char 3</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 8 test package char 4</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 8 test package bool 1</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 2</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 3</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 4</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 8 test package bool 5</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      //One per data type, package only
      oOut.write("<ou_gridInfo gridName=\"grid 9\">");
      oOut.write("<ou_saveFreq>9</ou_saveFreq>");
      oOut.write("<ou_packageInt>grid 9 test package int 1</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 9 test package float 1</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 9 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 9 test package bool 1</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      //Multiples of each, package only
      oOut.write("<ou_gridInfo gridName=\"grid 10\">");
      oOut.write("<ou_saveFreq>10</ou_saveFreq>");
      oOut.write("<ou_packageInt>grid 10 test package int 1</ou_packageInt>");
      oOut.write("<ou_packageInt>grid 10 test package int 2</ou_packageInt>");
      oOut.write(
          "<ou_packageFloat>grid 10 test package float 1</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 10 test package float 2</ou_packageFloat>");
      oOut.write(
          "<ou_packageFloat>grid 10 test package float 3</ou_packageFloat>");
      oOut.write("<ou_packageChar>grid 10 test package char 1</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 10 test package char 2</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 10 test package char 3</ou_packageChar>");
      oOut.write("<ou_packageChar>grid 10 test package char 4</ou_packageChar>");
      oOut.write("<ou_packageBool>grid 10 test package bool 1</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 2</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 3</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 4</ou_packageBool>");
      oOut.write("<ou_packageBool>grid 10 test package bool 5</ou_packageBool>");
      oOut.write("</ou_gridInfo>");

      //Now subplots
      oOut.write("<ou_subplot>");
      oOut.write("<ou_subplotName>test 1</ou_subplotName>");
      //One point
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"0\" y=\"12\"/>");
      oOut.write("</pointSet>");
      oOut.write("</ou_subplot>");

      //5 points
      oOut.write("<ou_subplot>");
      oOut.write("<ou_subplotName>test 2</ou_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"6\" y=\"14\"/>");
      oOut.write("<po_point x=\"7\" y=\"15\"/>");
      oOut.write("<po_point x=\"8\" y=\"16\"/>");
      oOut.write("<po_point x=\"9\" y=\"17\"/>");
      oOut.write("<po_point x=\"10\" y=\"18\"/>");
      oOut.write("</pointSet>");
      oOut.write("</ou_subplot>");

      oOut.write("</output>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      throw(new ModelException(ErrorGUI.BAD_FILE, "", oE.getMessage()));
    }
  }

  /**
   * This writes an XML file to test value setting.  This file contains short
   * output settings only, for all (normal) types with everything set to true.
   * It also contains subplots.
   * @return Name of file.
   * @throws ModelException if there is an IO problem.
   */
  protected String WriteXMLFile2() throws ModelException {
    try {
      String sFileName = "\\loratestxml2.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      oOut.write("<paramFile fileCode=\"06010101\">");

      //Plot - so the subplot cells can be set up
      oOut.write("<plot>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>55.37</plot_latitude>");
      oOut.write("</plot>");

      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("<tr_species speciesName=\"Amabilis_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Subalpine_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Hybrid_Spruce\"/>");
      oOut.write("<tr_species speciesName=\"Lodgepole_Pine\"/>");
      oOut.write("<tr_species speciesName=\"Trembling_Aspen\"/>");
      oOut.write("<tr_species speciesName=\"Black_Cottonwood\"/>");
      oOut.write("<tr_species speciesName=\"Paper_Birch\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("</trees>");

      //Include an empty behavior list - this triggers setup
      oOut.write("<behaviorList>");
      oOut.write("</behaviorList>");

      oOut.write("<shortoutput>");

      //Set all save options to true for all recognized types
      oOut.write("<so_filename>testfile</so_filename>");

      oOut.write("<so_treeTypeInfo type=\"seedling\">");
      oOut.write("<so_saveRBA save=\"true\"/>"); //this shouldn't happen
      oOut.write("<so_saveABA save=\"true\"/>"); //this shouldn't happen
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("<so_treeTypeInfo type=\"sapling\">");
      oOut.write("<so_saveRBA save=\"true\"/>");
      oOut.write("<so_saveABA save=\"true\"/>");
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("<so_treeTypeInfo type=\"adult\">");
      oOut.write("<so_saveRBA save=\"true\"/>");
      oOut.write("<so_saveABA save=\"true\"/>");
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      //Now subplots
      oOut.write("<so_subplot>");
      oOut.write("<so_subplotName>test 1</so_subplotName>");
      //One point
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"0\" y=\"12\"/>");
      oOut.write("</pointSet>");
      oOut.write("</so_subplot>");

      //5 points
      oOut.write("<so_subplot>");
      oOut.write("<so_subplotName>test 2</so_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"6\" y=\"14\"/>");
      oOut.write("<po_point x=\"7\" y=\"15\"/>");
      oOut.write("<po_point x=\"8\" y=\"16\"/>");
      oOut.write("<po_point x=\"9\" y=\"17\"/>");
      oOut.write("<po_point x=\"10\" y=\"18\"/>");
      oOut.write("</pointSet>");
      oOut.write("</so_subplot>");

      oOut.write("</shortoutput>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      throw(new ModelException(ErrorGUI.BAD_FILE, "", oE.getMessage()));
    }
  }

  /**
   * This writes an XML file to test value setting.  This file contains short
   * output settings only, for all (normal) types with a mix of true, false,
   * and absent.
   * @return Name of file.
   * @throws ModelException if there is an IO problem.
   */
  protected String WriteXMLFile3() throws ModelException {
    try {
      String sFileName = "\\loratestxml3.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write("<paramFile fileCode=\"06100101\">");
      oOut.write("<plot>");
      oOut.write("<timesteps>500</timesteps>");
      oOut.write("<rt_timestep>0</rt_timestep>");
      oOut.write("<randomSeed>0</randomSeed>");
      oOut.write("<yearsPerTimestep>1.0</yearsPerTimestep>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>42.5</plot_latitude>");
      oOut.write("<plot_precip_mm_yr>1000.0</plot_precip_mm_yr>");
      oOut.write("<plot_temp_C>8.0</plot_temp_C>");
      oOut.write("<plot_title>Default Location</plot_title>");
      oOut.write("</plot>");
      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"ABBA\" />");
      oOut.write("</tr_speciesList>");
      oOut.write("<tr_seedDiam10Cm>0.2</tr_seedDiam10Cm>");
      oOut.write("<tr_minAdultDBH>");
      oOut.write("<tr_madVal species=\"ABBA\">12.7</tr_madVal>");
      oOut.write("</tr_minAdultDBH>");
      oOut.write("<tr_maxSeedlingHeight>");
      oOut.write("<tr_mshVal species=\"ABBA\">1.35</tr_mshVal>");
      oOut.write("</tr_maxSeedlingHeight>");
      oOut.write("</trees>");
      oOut.write("<behaviorList>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>removedead</behaviorName>");
      oOut.write("<version>2.0</version>");
      oOut.write("<applyTo species=\"ABBA\" type=\"Seedling\" />");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>short output</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("</behavior>");
      oOut.write("</behaviorList>");
      oOut.write("<allometry>");
      oOut.write("<tr_whatAdultHeightDiam>");
      oOut.write("<tr_wahdVal species=\"ABBA\">0</tr_wahdVal>");
      oOut.write("</tr_whatAdultHeightDiam>");
      oOut.write("<tr_whatSaplingHeightDiam>");
      oOut.write("<tr_wsahdVal species=\"ABBA\">0</tr_wsahdVal>");
      oOut.write("</tr_whatSaplingHeightDiam>");
      oOut.write("<tr_whatSeedlingHeightDiam>");
      oOut.write("<tr_wsehdVal species=\"ABBA\">0</tr_wsehdVal>");
      oOut.write("</tr_whatSeedlingHeightDiam>");
      oOut.write("<tr_whatAdultCrownRadDiam>");
      oOut.write("<tr_wacrdVal species=\"ABBA\">0</tr_wacrdVal>");
      oOut.write("</tr_whatAdultCrownRadDiam>");
      oOut.write("<tr_whatSaplingCrownRadDiam>");
      oOut.write("<tr_wscrdVal species=\"ABBA\">0</tr_wscrdVal>");
      oOut.write("</tr_whatSaplingCrownRadDiam>");
      oOut.write("<tr_whatAdultCrownHeightHeight>");
      oOut.write("<tr_wachhVal species=\"ABBA\">0</tr_wachhVal>");
      oOut.write("</tr_whatAdultCrownHeightHeight>");
      oOut.write("<tr_whatSaplingCrownHeightHeight>");
      oOut.write("<tr_wschhVal species=\"ABBA\">0</tr_wschhVal>");
      oOut.write("</tr_whatSaplingCrownHeightHeight>");
      oOut.write("<tr_canopyHeight>");
      oOut.write("<tr_chVal species=\"ABBA\">30.0</tr_chVal>");
      oOut.write("</tr_canopyHeight>");
      oOut.write("<tr_stdAsympCrownRad>");
      oOut.write("<tr_sacrVal species=\"ABBA\">0.1</tr_sacrVal>");
      oOut.write("</tr_stdAsympCrownRad>");
      oOut.write("<tr_stdCrownRadExp>");
      oOut.write("<tr_screVal species=\"ABBA\">1.0</tr_screVal>");
      oOut.write("</tr_stdCrownRadExp>");
      oOut.write("<tr_stdAsympCrownHt>");
      oOut.write("<tr_sachVal species=\"ABBA\">0.4</tr_sachVal>");
      oOut.write("</tr_stdAsympCrownHt>");
      oOut.write("<tr_stdCrownHtExp>");
      oOut.write("<tr_scheVal species=\"ABBA\">1.0</tr_scheVal>");
      oOut.write("</tr_stdCrownHtExp>");
      oOut.write("<tr_conversionDiam10ToDBH>");
      oOut.write("<tr_cdtdVal species=\"ABBA\">1.0</tr_cdtdVal>");
      oOut.write("</tr_conversionDiam10ToDBH>");
      oOut.write("<tr_interceptDiam10ToDBH>");
      oOut.write("<tr_idtdVal species=\"ABBA\">0.0</tr_idtdVal>");
      oOut.write("</tr_interceptDiam10ToDBH>");
      oOut.write("<tr_slopeOfAsymHeight>");
      oOut.write("<tr_soahVal species=\"ABBA\">0.063</tr_soahVal>");
      oOut.write("</tr_slopeOfAsymHeight>");
      oOut.write("<tr_slopeOfHeight-Diam10>");
      oOut.write("<tr_sohdVal species=\"ABBA\">0.03</tr_sohdVal>");
      oOut.write("</tr_slopeOfHeight-Diam10>");
      oOut.write("</allometry>");

      oOut.write("<shortoutput>");

      //Set all save options to true for all recognized types
      oOut.write("<so_filename>testfile</so_filename>");

      oOut.write("<so_treeTypeInfo type=\"sapling\">");
      oOut.write("<so_saveRBA save=\"true\"/>");
      oOut.write("<so_saveABA save=\"false\"/>");
      oOut.write("<so_saveRDN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("<so_treeTypeInfo type=\"adult\">");
      oOut.write("<so_saveRBA save=\"false\"/>");
      oOut.write("<so_saveADN save=\"true\"/>");
      oOut.write("</so_treeTypeInfo>");

      oOut.write("</shortoutput>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      throw(new ModelException(ErrorGUI.BAD_FILE, "", oE.getMessage()));
    }
  }

  /**
   * Writes a real-type file, so we can work with actual grids.
   * @throws ModelException if there is a problem writing the file.
   * @return String Filename just written.
   */
  protected String WriteXMLFile4() throws ModelException {
    try {
      String sFileName = "\\loratestxml3.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write("<paramFile fileCode=\"06070101\">");
      oOut.write("<plot>");
      oOut.write("<timesteps>100</timesteps>");
      oOut.write("<randomSeed>0</randomSeed>");
      oOut.write("<yearsPerTimestep>5.0</yearsPerTimestep>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>55.37</plot_latitude>");
      oOut.write("<plot_title>Default parameter file-use for testing only</plot_title>");
      oOut.write("</plot>");
      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("<tr_species speciesName=\"Amabilis_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Subalpine_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Hybrid_Spruce\"/>");
      oOut.write("<tr_species speciesName=\"Lodgepole_Pine\"/>");
      oOut.write("<tr_species speciesName=\"Trembling_Aspen\"/>");
      oOut.write("<tr_species speciesName=\"Black_Cottonwood\"/>");
      oOut.write("<tr_species speciesName=\"Paper_Birch\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("<tr_seedDiam10Cm>0.1</tr_seedDiam10Cm>");
      oOut.write("<tr_minAdultDBH>");
      oOut.write("<tr_madVal species=\"Western_Hemlock\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Western_Redcedar\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Amabilis_Fir\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Subalpine_Fir\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Hybrid_Spruce\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Lodgepole_Pine\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Trembling_Aspen\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Black_Cottonwood\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Paper_Birch\">10.0</tr_madVal>");
      oOut.write("</tr_minAdultDBH>");
      oOut.write("<tr_maxSeedlingHeight>");
      oOut.write("<tr_mshVal species=\"Western_Hemlock\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Western_Redcedar\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Amabilis_Fir\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Subalpine_Fir\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Hybrid_Spruce\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Lodgepole_Pine\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Trembling_Aspen\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Black_Cottonwood\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Paper_Birch\">1.35</tr_mshVal>");
      oOut.write("</tr_maxSeedlingHeight>");
      oOut.write("</trees>");
      oOut.write("<behaviorList>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>Windstorm</behaviorName>");
      oOut.write("<version>2.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>quadratlight</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>constradialgrowth</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>adultstochasticmort</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>substrate</behaviorName>");
      oOut.write("<version>2.1</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>removedead</behaviorName>");
      oOut.write("<version>2.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>gap disperse</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>Gap Substrate Seed Survival</behaviorName>");
      oOut.write("<version>1.1</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seed\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>Establishment</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seed\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>output</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("</behavior>");
      oOut.write("</behaviorList>");
      oOut.write("<allometry>");
      oOut.write("<tr_whatAdultHeightDiam>");
      oOut.write("<tr_wahdVal species=\"Western_Hemlock\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Western_Redcedar\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Amabilis_Fir\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Subalpine_Fir\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Hybrid_Spruce\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Lodgepole_Pine\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Trembling_Aspen\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Black_Cottonwood\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Paper_Birch\">0</tr_wahdVal>");
      oOut.write("</tr_whatAdultHeightDiam>");
      oOut.write("<tr_whatSaplingHeightDiam>");
      oOut.write("<tr_wsahdVal species=\"Western_Hemlock\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Western_Redcedar\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Amabilis_Fir\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Subalpine_Fir\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Hybrid_Spruce\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Lodgepole_Pine\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Trembling_Aspen\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Black_Cottonwood\">0</tr_wsahdVal>");
      oOut.write("<tr_wsahdVal species=\"Paper_Birch\">0</tr_wsahdVal>");
      oOut.write("</tr_whatSaplingHeightDiam>");
      oOut.write("<tr_whatSeedlingHeightDiam>");
      oOut.write("<tr_wsehdVal species=\"Western_Hemlock\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Western_Redcedar\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Amabilis_Fir\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Subalpine_Fir\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Hybrid_Spruce\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Lodgepole_Pine\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Trembling_Aspen\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Black_Cottonwood\">0</tr_wsehdVal>");
      oOut.write("<tr_wsehdVal species=\"Paper_Birch\">0</tr_wsehdVal>");
      oOut.write("</tr_whatSeedlingHeightDiam>");
      oOut.write("<tr_whatAdultCrownRadDiam>");
      oOut.write("<tr_wacrdVal species=\"Western_Hemlock\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Western_Redcedar\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Amabilis_Fir\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Subalpine_Fir\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Hybrid_Spruce\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Lodgepole_Pine\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Trembling_Aspen\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Black_Cottonwood\">0</tr_wacrdVal>");
      oOut.write("<tr_wacrdVal species=\"Paper_Birch\">0</tr_wacrdVal>");
      oOut.write("</tr_whatAdultCrownRadDiam>");
      oOut.write("<tr_whatSaplingCrownRadDiam>");
      oOut.write("<tr_wscrdVal species=\"Western_Hemlock\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Western_Redcedar\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Amabilis_Fir\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Subalpine_Fir\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Hybrid_Spruce\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Lodgepole_Pine\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Trembling_Aspen\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Black_Cottonwood\">0</tr_wscrdVal>");
      oOut.write("<tr_wscrdVal species=\"Paper_Birch\">0</tr_wscrdVal>");
      oOut.write("</tr_whatSaplingCrownRadDiam>");
      oOut.write("<tr_whatAdultCrownHeightHeight>");
      oOut.write("<tr_wachhVal species=\"Western_Hemlock\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Western_Redcedar\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Amabilis_Fir\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Subalpine_Fir\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Hybrid_Spruce\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Lodgepole_Pine\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Trembling_Aspen\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Black_Cottonwood\">0</tr_wachhVal>");
      oOut.write("<tr_wachhVal species=\"Paper_Birch\">0</tr_wachhVal>");
      oOut.write("</tr_whatAdultCrownHeightHeight>");
      oOut.write("<tr_whatSaplingCrownHeightHeight>");
      oOut.write("<tr_wschhVal species=\"Western_Hemlock\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Western_Redcedar\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Amabilis_Fir\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Subalpine_Fir\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Hybrid_Spruce\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Lodgepole_Pine\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Trembling_Aspen\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Black_Cottonwood\">0</tr_wschhVal>");
      oOut.write("<tr_wschhVal species=\"Paper_Birch\">0</tr_wschhVal>");
      oOut.write("</tr_whatSaplingCrownHeightHeight>");
      oOut.write("<tr_canopyHeight>");
      oOut.write("<tr_chVal species=\"Western_Hemlock\">39.48</tr_chVal>");
      oOut.write("<tr_chVal species=\"Western_Redcedar\">39.54</tr_chVal>");
      oOut.write("<tr_chVal species=\"Amabilis_Fir\">40.0</tr_chVal>");
      oOut.write("<tr_chVal species=\"Subalpine_Fir\">40.0</tr_chVal>");
      oOut.write("<tr_chVal species=\"Hybrid_Spruce\">45.0</tr_chVal>");
      oOut.write("<tr_chVal species=\"Lodgepole_Pine\">40.0</tr_chVal>");
      oOut.write("<tr_chVal species=\"Trembling_Aspen\">39.14</tr_chVal>");
      oOut.write("<tr_chVal species=\"Black_Cottonwood\">39.47</tr_chVal>");
      oOut.write("<tr_chVal species=\"Paper_Birch\">33.18</tr_chVal>");
      oOut.write("</tr_canopyHeight>");
      oOut.write("<tr_stdAsympCrownRad>");
      oOut.write("<tr_sacrVal species=\"Western_Hemlock\">0.0549</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Western_Redcedar\">0.0614</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Amabilis_Fir\">0.0242</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Subalpine_Fir\">0.0251</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Hybrid_Spruce\">0.0239</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Lodgepole_Pine\">0.0303</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Trembling_Aspen\">0.0328</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Black_Cottonwood\">0.0247</tr_sacrVal>");
      oOut.write("<tr_sacrVal species=\"Paper_Birch\">0.0484</tr_sacrVal>");
      oOut.write("</tr_stdAsympCrownRad>");
      oOut.write("<tr_stdCrownRadExp>");
      oOut.write("<tr_screVal species=\"Western_Hemlock\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Western_Redcedar\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Amabilis_Fir\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Subalpine_Fir\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Hybrid_Spruce\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Lodgepole_Pine\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Trembling_Aspen\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Black_Cottonwood\">1.0</tr_screVal>");
      oOut.write("<tr_screVal species=\"Paper_Birch\">1.0</tr_screVal>");
      oOut.write("</tr_stdCrownRadExp>");
      oOut.write("<tr_stdAsympCrownHt>");
      oOut.write("<tr_sachVal species=\"Western_Hemlock\">0.389</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Western_Redcedar\">0.368</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Amabilis_Fir\">0.464</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Subalpine_Fir\">0.454</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Hybrid_Spruce\">0.405</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Lodgepole_Pine\">0.201</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Trembling_Aspen\">0.301</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Black_Cottonwood\">0.42</tr_sachVal>");
      oOut.write("<tr_sachVal species=\"Paper_Birch\">0.315</tr_sachVal>");
      oOut.write("</tr_stdAsympCrownHt>");
      oOut.write("<tr_stdCrownHtExp>");
      oOut.write("<tr_scheVal species=\"Western_Hemlock\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Western_Redcedar\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Amabilis_Fir\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Subalpine_Fir\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Hybrid_Spruce\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Lodgepole_Pine\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Trembling_Aspen\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Black_Cottonwood\">1.0</tr_scheVal>");
      oOut.write("<tr_scheVal species=\"Paper_Birch\">1.0</tr_scheVal>");
      oOut.write("</tr_stdCrownHtExp>");
      oOut.write("<tr_conversionDiam10ToDBH>");
      oOut.write("<tr_cdtdVal species=\"Western_Hemlock\">0.8008</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Western_Redcedar\">0.5944</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Amabilis_Fir\">0.7059</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Subalpine_Fir\">0.7776</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Hybrid_Spruce\">0.6933</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Lodgepole_Pine\">0.8014</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Trembling_Aspen\">0.7992</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Black_Cottonwood\">0.7926</tr_cdtdVal>");
      oOut.write("<tr_cdtdVal species=\"Paper_Birch\">0.7803</tr_cdtdVal>");
      oOut.write("</tr_conversionDiam10ToDBH>");
      oOut.write("<tr_interceptDiam10ToDBH>");
      oOut.write("<tr_idtdVal species=\"Western_Hemlock\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Western_Redcedar\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Amabilis_Fir\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Subalpine_Fir\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Hybrid_Spruce\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Lodgepole_Pine\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Trembling_Aspen\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Black_Cottonwood\">0.0</tr_idtdVal>");
      oOut.write("<tr_idtdVal species=\"Paper_Birch\">0.0</tr_idtdVal>");
      oOut.write("</tr_interceptDiam10ToDBH>");
      oOut.write("<tr_slopeOfAsymHeight>");
      oOut.write("<tr_soahVal species=\"Western_Hemlock\">0.0299</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Western_Redcedar\">0.0241</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Amabilis_Fir\">0.0263</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Subalpine_Fir\">0.0264</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Hybrid_Spruce\">0.0264</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Lodgepole_Pine\">0.0333</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Trembling_Aspen\">0.0352</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Black_Cottonwood\">0.0347</tr_soahVal>");
      oOut.write("<tr_soahVal species=\"Paper_Birch\">0.0454</tr_soahVal>");
      oOut.write("</tr_slopeOfAsymHeight>");
      oOut.write("<tr_slopeOfHeight-Diam10>");
      oOut.write("<tr_sohdVal species=\"Western_Hemlock\">0.03418</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Western_Redcedar\">0.0269</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Amabilis_Fir\">0.02871</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Subalpine_Fir\">0.02815</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Hybrid_Spruce\">0.02871</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Lodgepole_Pine\">0.03224</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Trembling_Aspen\">0.04796</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Black_Cottonwood\">0.04681</tr_sohdVal>");
      oOut.write("<tr_sohdVal species=\"Paper_Birch\">0.04101</tr_sohdVal>");
      oOut.write("</tr_slopeOfHeight-Diam10>");
      oOut.write("</allometry>");
      oOut.write("<windstorm>");
      oOut.write("<ws_severityReturnInterval1>0.0</ws_severityReturnInterval1>");
      oOut.write("<ws_severityReturnInterval5>0.0</ws_severityReturnInterval5>");
      oOut.write("<ws_severityReturnInterval10>0.0</ws_severityReturnInterval10>");
      oOut.write("<ws_severityReturnInterval20>0.0</ws_severityReturnInterval20>");
      oOut.write("<ws_severityReturnInterval40>0.0</ws_severityReturnInterval40>");
      oOut.write("<ws_severityReturnInterval80>0.0</ws_severityReturnInterval80>");
      oOut.write("<ws_severityReturnInterval160>0.0</ws_severityReturnInterval160>");
      oOut.write("<ws_severityReturnInterval320>0.0</ws_severityReturnInterval320>");
      oOut.write("<ws_severityReturnInterval640>0.0</ws_severityReturnInterval640>");
      oOut.write("<ws_severityReturnInterval1280>0.0</ws_severityReturnInterval1280>");
      oOut.write("<ws_severityReturnInterval2560>0.0</ws_severityReturnInterval2560>");
      oOut.write("<ws_stmSSTPeriod>1.0</ws_stmSSTPeriod>");
      oOut.write("<ws_stmSineD>0.0</ws_stmSineD>");
      oOut.write("<ws_stmSineF>1.0</ws_stmSineF>");
      oOut.write("<ws_stmSineG>1.0</ws_stmSineG>");
      oOut.write("<ws_stmTrendSlopeM>0.0</ws_stmTrendSlopeM>");
      oOut.write("<ws_stmTrendInterceptI>1.0</ws_stmTrendInterceptI>");
      oOut.write("<ws_stmTSToStartStorms>0</ws_stmTSToStartStorms>");
      oOut.write("</windstorm>");
      oOut.write("<quadratGliLight>");
      oOut.write("<li_heightOfFishEyePhoto>0</li_heightOfFishEyePhoto>");
      oOut.write("<li_beamFractGlobalRad>0.5</li_beamFractGlobalRad>");
      oOut.write("<li_clearSkyTransCoeff>0.65</li_clearSkyTransCoeff>");
      oOut.write("<li_julianDayGrowthStarts>105</li_julianDayGrowthStarts>");
      oOut.write("<li_julianDayGrowthEnds>258</li_julianDayGrowthEnds>");
      oOut.write("<li_minSunAngle>0.785</li_minSunAngle>");
      oOut.write("<li_numAltGrids>12</li_numAltGrids>");
      oOut.write("<li_numAziGrids>18</li_numAziGrids>");
      oOut.write("<li_quadratLightHeight>0.675</li_quadratLightHeight>");
      oOut.write("<li_lightExtinctionCoefficient>");
      oOut.write("<li_lecVal species=\"Western_Hemlock\">0.08</li_lecVal>");
      oOut.write("<li_lecVal species=\"Western_Redcedar\">0.091</li_lecVal>");
      oOut.write("<li_lecVal species=\"Amabilis_Fir\">0.059</li_lecVal>");
      oOut.write("<li_lecVal species=\"Subalpine_Fir\">0.092</li_lecVal>");
      oOut.write("<li_lecVal species=\"Hybrid_Spruce\">0.114</li_lecVal>");
      oOut.write("<li_lecVal species=\"Lodgepole_Pine\">0.135</li_lecVal>");
      oOut.write("<li_lecVal species=\"Trembling_Aspen\">0.206</li_lecVal>");
      oOut.write("<li_lecVal species=\"Black_Cottonwood\">0.143</li_lecVal>");
      oOut.write("<li_lecVal species=\"Paper_Birch\">0.058</li_lecVal>");
      oOut.write("</li_lightExtinctionCoefficient>");
      oOut.write("<li_snagAgeClass1>0</li_snagAgeClass1>");
      oOut.write("<li_snagAgeClass2>0</li_snagAgeClass2>");
      oOut.write("</quadratGliLight>");
      oOut.write("<growth>");
      oOut.write("<gr_adultConstRadialInc>");
      oOut.write("<gr_acriVal species=\"Trembling_Aspen\">1.14</gr_acriVal>");
      oOut.write("<gr_acriVal species=\"Black_Cottonwood\">2.47</gr_acriVal>");
      oOut.write("<gr_acriVal species=\"Paper_Birch\">0.72</gr_acriVal>");
      oOut.write("</gr_adultConstRadialInc>");
      oOut.write("</growth>");
      oOut.write("<mortality>");
      oOut.write("<mo_nonSizeDepMort>");
      oOut.write("<mo_nsdmVal species=\"Western_Hemlock\">0.0050</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Western_Redcedar\">0.0050</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Amabilis_Fir\">0.0050</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Subalpine_Fir\">0.01</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Hybrid_Spruce\">0.01</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Lodgepole_Pine\">0.01</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Trembling_Aspen\">0.015</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Black_Cottonwood\">0.01</mo_nsdmVal>");
      oOut.write("<mo_nsdmVal species=\"Paper_Birch\">0.015</mo_nsdmVal>");
      oOut.write("</mo_nonSizeDepMort>");
      oOut.write("</mortality>");
      oOut.write("<substrate>");
      oOut.write("<su_scarSoilDecayAlpha>-5.1E-4</su_scarSoilDecayAlpha>");
      oOut.write("<su_scarSoilDecayBeta>4.4</su_scarSoilDecayBeta>");
      oOut.write("<su_tipupDecayAlpha>-7.0E-4</su_tipupDecayAlpha>");
      oOut.write("<su_tipupDecayBeta>4.3</su_tipupDecayBeta>");
      oOut.write("<su_freshLogDecayAlpha>-0.05</su_freshLogDecayAlpha>");
      oOut.write("<su_freshLogDecayBeta>6.5</su_freshLogDecayBeta>");
      oOut.write("<su_decayedLogDecayAlpha>-0.79850775</su_decayedLogDecayAlpha>");
      oOut.write("<su_decayedLogDecayBeta>1.0</su_decayedLogDecayBeta>");
      oOut.write("<su_maxNumberDecayYears>50</su_maxNumberDecayYears>");
      oOut.write("<su_initialScarSoil>0.0</su_initialScarSoil>");
      oOut.write("<su_initialTipup>0.0</su_initialTipup>");
      oOut.write("<su_initialFreshLog>0.01</su_initialFreshLog>");
      oOut.write("<su_initialDecayedLog>0.11</su_initialDecayedLog>");
      oOut.write("<su_partialCutScarSoil>0.17</su_partialCutScarSoil>");
      oOut.write("<su_partialCutTipup>0.15</su_partialCutTipup>");
      oOut.write("<su_partialCutFreshLog>0.06</su_partialCutFreshLog>");
      oOut.write("<su_partialCutDecayedLog>0.04</su_partialCutDecayedLog>");
      oOut.write("<su_gapCutScarSoil>0.45</su_gapCutScarSoil>");
      oOut.write("<su_gapCutTipup>0.34</su_gapCutTipup>");
      oOut.write("<su_gapCutFreshLog>0.09</su_gapCutFreshLog>");
      oOut.write("<su_gapCutDecayedLog>0.0</su_gapCutDecayedLog>");
      oOut.write("<su_clearCutScarSoil>0.36</su_clearCutScarSoil>");
      oOut.write("<su_clearCutTipup>0.34</su_clearCutTipup>");
      oOut.write("<su_clearCutFreshLog>0.17</su_clearCutFreshLog>");
      oOut.write("<su_clearCutDecayedLog>0.01</su_clearCutDecayedLog>");
      oOut.write("<su_propOfDeadFall>");
      oOut.write("<su_podfVal species=\"Western_Hemlock\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Western_Redcedar\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Amabilis_Fir\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Subalpine_Fir\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Hybrid_Spruce\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Lodgepole_Pine\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Trembling_Aspen\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Black_Cottonwood\">0.7</su_podfVal>");
      oOut.write("<su_podfVal species=\"Paper_Birch\">0.7</su_podfVal>");
      oOut.write("</su_propOfDeadFall>");
      oOut.write("<su_propOfFallUproot>");
      oOut.write("<su_pofuVal species=\"Western_Hemlock\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Western_Redcedar\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Amabilis_Fir\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Subalpine_Fir\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Hybrid_Spruce\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Lodgepole_Pine\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Trembling_Aspen\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Black_Cottonwood\">0.5</su_pofuVal>");
      oOut.write("<su_pofuVal species=\"Paper_Birch\">0.5</su_pofuVal>");
      oOut.write("</su_propOfFallUproot>");
      oOut.write("<su_propOfSnagsUproot>");
      oOut.write("<su_posuVal species=\"Western_Hemlock\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Western_Redcedar\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Amabilis_Fir\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Subalpine_Fir\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Hybrid_Spruce\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Lodgepole_Pine\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Trembling_Aspen\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Black_Cottonwood\">0.0</su_posuVal>");
      oOut.write("<su_posuVal species=\"Paper_Birch\">0.0</su_posuVal>");
      oOut.write("</su_propOfSnagsUproot>");
      oOut.write("<su_rootTipupFactor>3.1</su_rootTipupFactor>");
      oOut.write("<su_mossProportion>0.0</su_mossProportion>");
      oOut.write("<su_directionalTreeFall>0</su_directionalTreeFall>");
      oOut.write("</substrate>");
      oOut.write("<disperse>");
      oOut.write("<di_minDbhForReproduction>");
      oOut.write("<di_mdfrVal species=\"Western_Hemlock\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Western_Redcedar\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Amabilis_Fir\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Subalpine_Fir\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Hybrid_Spruce\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Lodgepole_Pine\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Trembling_Aspen\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Black_Cottonwood\">15.0</di_mdfrVal>");
      oOut.write("<di_mdfrVal species=\"Paper_Birch\">15.0</di_mdfrVal>");
      oOut.write("</di_minDbhForReproduction>");
      oOut.write("<di_weibullCanopySTR>");
      oOut.write("<di_wcsVal species=\"Western_Hemlock\">3.8650002</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Western_Redcedar\">2.4514</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Amabilis_Fir\">1.3255999</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Subalpine_Fir\">0.48839998</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Hybrid_Spruce\">3.4411998</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Lodgepole_Pine\">0.6114</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Trembling_Aspen\">5.9544</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Black_Cottonwood\">0.3178</di_wcsVal>");
      oOut.write("<di_wcsVal species=\"Paper_Birch\">3.2012</di_wcsVal>");
      oOut.write("</di_weibullCanopySTR>");
      oOut.write("<di_weibullCanopyBeta>");
      oOut.write("<di_wcbVal species=\"Western_Hemlock\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Western_Redcedar\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Amabilis_Fir\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Subalpine_Fir\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Hybrid_Spruce\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Lodgepole_Pine\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Trembling_Aspen\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Black_Cottonwood\">2.0</di_wcbVal>");
      oOut.write("<di_wcbVal species=\"Paper_Birch\">2.0</di_wcbVal>");
      oOut.write("</di_weibullCanopyBeta>");
      oOut.write("<di_weibullCanopyDispersal>");
      oOut.write("<di_wcdVal species=\"Western_Hemlock\">1.76E-4</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Western_Redcedar\">1.82E-4</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Amabilis_Fir\">9.61E-5</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Subalpine_Fir\">1.32E-4</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Hybrid_Spruce\">3.0E-5</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Lodgepole_Pine\">2.8E-4</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Trembling_Aspen\">3.8E-5</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Black_Cottonwood\">1.52E-4</di_wcdVal>");
      oOut.write("<di_wcdVal species=\"Paper_Birch\">2.33E-4</di_wcdVal>");
      oOut.write("</di_weibullCanopyDispersal>");
      oOut.write("<di_weibullCanopyTheta>");
      oOut.write("<di_wctVal species=\"Western_Hemlock\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Western_Redcedar\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Amabilis_Fir\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Subalpine_Fir\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Hybrid_Spruce\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Lodgepole_Pine\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Trembling_Aspen\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Black_Cottonwood\">3.0</di_wctVal>");
      oOut.write("<di_wctVal species=\"Paper_Birch\">3.0</di_wctVal>");
      oOut.write("</di_weibullCanopyTheta>");
      oOut.write("<di_maxGapDensity>0</di_maxGapDensity>");
      oOut.write("<di_canopyFunction>");
      oOut.write("<di_cfVal species=\"Western_Hemlock\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Western_Redcedar\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Amabilis_Fir\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Subalpine_Fir\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Hybrid_Spruce\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Lodgepole_Pine\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Trembling_Aspen\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Black_Cottonwood\">0</di_cfVal>");
      oOut.write("<di_cfVal species=\"Paper_Birch\">0</di_cfVal>");
      oOut.write("</di_canopyFunction>");
      oOut.write("<di_seedDistributionMethod>0</di_seedDistributionMethod>");
      oOut.write("</disperse>");
      oOut.write("<establishment>");
      oOut.write("<es_scarifiedSoilCanopyFav>");
      oOut.write("<es_sscfVal species=\"Western_Hemlock\">0.129</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Western_Redcedar\">0.48</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Amabilis_Fir\">0.064</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Subalpine_Fir\">0.017</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Hybrid_Spruce\">0.391</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Lodgepole_Pine\">0.462</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Trembling_Aspen\">0.781</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Black_Cottonwood\">0.95</es_sscfVal>");
      oOut.write("<es_sscfVal species=\"Paper_Birch\">0.687</es_sscfVal>");
      oOut.write("</es_scarifiedSoilCanopyFav>");
      oOut.write("<es_tipUpCanopyFav>");
      oOut.write("<es_tucfVal species=\"Western_Hemlock\">0.983</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Western_Redcedar\">0.391</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Amabilis_Fir\">0.274</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Subalpine_Fir\">0.278</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Hybrid_Spruce\">0.778</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Lodgepole_Pine\">0.155</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Trembling_Aspen\">0.013</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Black_Cottonwood\">0.31</es_tucfVal>");
      oOut.write("<es_tucfVal species=\"Paper_Birch\">0.0010</es_tucfVal>");
      oOut.write("</es_tipUpCanopyFav>");
      oOut.write("<es_freshLogCanopyFav>");
      oOut.write("<es_flcfVal species=\"Western_Hemlock\">0.353</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Western_Redcedar\">0.424</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Amabilis_Fir\">0.0010</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Subalpine_Fir\">0.028</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Hybrid_Spruce\">0.07</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Lodgepole_Pine\">0.566</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Trembling_Aspen\">0.017</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Black_Cottonwood\">0.413</es_flcfVal>");
      oOut.write("<es_flcfVal species=\"Paper_Birch\">0.381</es_flcfVal>");
      oOut.write("</es_freshLogCanopyFav>");
      oOut.write("<es_decayedLogCanopyFav>");
      oOut.write("<es_dlcfVal species=\"Western_Hemlock\">0.058</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Western_Redcedar\">0.024</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Amabilis_Fir\">0.305</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Subalpine_Fir\">0.55</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Hybrid_Spruce\">0.054</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Lodgepole_Pine\">0.0080</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Trembling_Aspen\">0.18</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Black_Cottonwood\">0.771</es_dlcfVal>");
      oOut.write("<es_dlcfVal species=\"Paper_Birch\">0.8</es_dlcfVal>");
      oOut.write("</es_decayedLogCanopyFav>");
      oOut.write("<es_forestFloorLitterCanopyFav>");
      oOut.write("<es_fflcfVal species=\"Western_Hemlock\">0.0090</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Western_Redcedar\">0.01</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Amabilis_Fir\">0.815</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Subalpine_Fir\">0.019</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Hybrid_Spruce\">0.0090</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Lodgepole_Pine\">0.0040</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Trembling_Aspen\">0.01</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Black_Cottonwood\">0.315</es_fflcfVal>");
      oOut.write("<es_fflcfVal species=\"Paper_Birch\">0.12</es_fflcfVal>");
      oOut.write("</es_forestFloorLitterCanopyFav>");
      oOut.write("<es_forestFloorMossCanopyFav>");
      oOut.write("<es_ffmcfVal species=\"Western_Hemlock\">0.0090</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Western_Redcedar\">0.01</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Amabilis_Fir\">0.815</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Subalpine_Fir\">0.019</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Hybrid_Spruce\">0.0090</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Lodgepole_Pine\">0.0040</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Trembling_Aspen\">0.01</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Black_Cottonwood\">0.315</es_ffmcfVal>");
      oOut.write("<es_ffmcfVal species=\"Paper_Birch\">0.12</es_ffmcfVal>");
      oOut.write("</es_forestFloorMossCanopyFav>");
      oOut.write("</establishment>");
      oOut.write("<output>");
      oOut.write("<ou_filename>zxc.gz.tar</ou_filename>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>2</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>3</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>4</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>5</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>6</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>7</ou_saveFreq>");
      oOut.write("<ou_float>Y</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>8</ou_saveFreq>");
      oOut.write("<ou_float>Diam10</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>9</ou_saveFreq>");
      oOut.write("<ou_float>DBH</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>10</ou_saveFreq>");
      oOut.write("<ou_float>Height</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>11</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("<ou_float>Light</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>12</ou_saveFreq>");
      oOut.write("<ou_float>Growth</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>13</ou_saveFreq>");
      oOut.write("<ou_bool>dead</ou_bool>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>14</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>15</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Trembling_Aspen\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>16</ou_saveFreq>");
      oOut.write("<ou_float>Diam10</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Black_Cottonwood\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>17</ou_saveFreq>");
      oOut.write("<ou_float>Light</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Black_Cottonwood\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>18</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Paper_Birch\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>19</ou_saveFreq>");
      oOut.write("<ou_float>Y</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Paper_Birch\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>20</ou_saveFreq>");
      oOut.write("<ou_bool>dead</ou_bool>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_gridInfo gridName=\"Windstorm Results\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_packageFloat>ba_0</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_1</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_2</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_3</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_4</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_5</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_6</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_7</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_8</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_0</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_1</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_2</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_3</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_4</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_5</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_6</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_7</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_8</ou_packageFloat>");
      oOut.write("<ou_packageFloat>severity</ou_packageFloat>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Quadrat GLI\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>GLI</ou_float>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Substrate\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>scarsoil</ou_float>");
      oOut.write("<ou_float>tipup</ou_float>");
      oOut.write("<ou_float>freshlog</ou_float>");
      oOut.write("<ou_float>declog</ou_float>");
      oOut.write("<ou_float>ffmoss</ou_float>");
      oOut.write("<ou_float>fflitter</ou_float>");
      oOut.write("<ou_packageInt>age</ou_packageInt>");
      oOut.write("<ou_packageFloat>scarsoil</ou_packageFloat>");
      oOut.write("<ou_packageFloat>tipup</ou_packageFloat>");
      oOut.write("<ou_packageFloat>freshlog</ou_packageFloat>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Dispersed Seeds\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_int>count</ou_int>");
      oOut.write("<ou_float>seeds_0</ou_float>");
      oOut.write("<ou_float>seeds_1</ou_float>");
      oOut.write("<ou_float>seeds_2</ou_float>");
      oOut.write("<ou_float>seeds_3</ou_float>");
      oOut.write("<ou_float>seeds_4</ou_float>");
      oOut.write("<ou_float>seeds_5</ou_float>");
      oOut.write("<ou_float>seeds_6</ou_float>");
      oOut.write("<ou_float>seeds_7</ou_float>");
      oOut.write("<ou_float>seeds_8</ou_float>");
      oOut.write("<ou_bool>Is Gap</ou_bool>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Substrate Favorability\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>Favorability Index0</ou_float>");
      oOut.write("<ou_float>Favorability Index1</ou_float>");
      oOut.write("<ou_float>Favorability Index2</ou_float>");
      oOut.write("<ou_float>Favorability Index3</ou_float>");
      oOut.write("<ou_float>Favorability Index4</ou_float>");
      oOut.write("<ou_float>Favorability Index5</ou_float>");
      oOut.write("<ou_float>Favorability Index6</ou_float>");
      oOut.write("<ou_float>Favorability Index7</ou_float>");
      oOut.write("<ou_float>Favorability Index8</ou_float>");
      oOut.write("</ou_gridInfo>");
      oOut.write("</output>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      throw(new ModelException(ErrorGUI.BAD_FILE, "", oE.getMessage()));
    }
  }

  /**
   * Writes a real-type file, so we can work with actual grids.
   * @throws ModelException if there is a problem writing the file.
   * @return String Filename just written.
   */
  protected String WriteXMLFile5() throws ModelException {
    try {
      String sFileName = "\\loratestxml3.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write("<paramFile fileCode=\"06070101\">");
      oOut.write("<plot>");
      oOut.write("<timesteps>100</timesteps>");
      oOut.write("<randomSeed>0</randomSeed>");
      oOut.write("<yearsPerTimestep>5.0</yearsPerTimestep>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>55.37</plot_latitude>");
      oOut.write("<plot_title>Default parameter file-use for testing only</plot_title>");
      oOut.write("</plot>");
      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("<tr_species speciesName=\"Amabilis_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Subalpine_Fir\"/>");
      oOut.write("<tr_species speciesName=\"Hybrid_Spruce\"/>");
      oOut.write("<tr_species speciesName=\"Lodgepole_Pine\"/>");
      oOut.write("<tr_species speciesName=\"Trembling_Aspen\"/>");
      oOut.write("<tr_species speciesName=\"Black_Cottonwood\"/>");
      oOut.write("<tr_species speciesName=\"Paper_Birch\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("<tr_seedDiam10Cm>0.1</tr_seedDiam10Cm>");
      oOut.write("<tr_minAdultDBH>");
      oOut.write("<tr_madVal species=\"Western_Hemlock\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Western_Redcedar\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Amabilis_Fir\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Subalpine_Fir\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Hybrid_Spruce\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Lodgepole_Pine\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Trembling_Aspen\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Black_Cottonwood\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Paper_Birch\">10.0</tr_madVal>");
      oOut.write("</tr_minAdultDBH>");
      oOut.write("<tr_maxSeedlingHeight>");
      oOut.write("<tr_mshVal species=\"Western_Hemlock\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Western_Redcedar\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Amabilis_Fir\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Subalpine_Fir\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Hybrid_Spruce\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Lodgepole_Pine\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Trembling_Aspen\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Black_Cottonwood\">1.35</tr_mshVal>");
      oOut.write("<tr_mshVal species=\"Paper_Birch\">1.35</tr_mshVal>");
      oOut.write("</tr_maxSeedlingHeight>");
      oOut.write("</trees>");
      oOut.write("<behaviorList>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>Windstorm</behaviorName>");
      oOut.write("<version>2.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>quadratlight</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>constradialgrowth</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>adultstochasticmort</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>substrate</behaviorName>");
      oOut.write("<version>2.1</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>removedead</behaviorName>");
      oOut.write("<version>2.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seedling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Sapling\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seedling\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>gap disperse</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>Gap Substrate Seed Survival</behaviorName>");
      oOut.write("<version>1.1</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seed\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>Establishment</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Amabilis_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Subalpine_Fir\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Hybrid_Spruce\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Lodgepole_Pine\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Trembling_Aspen\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Black_Cottonwood\" type=\"Seed\"/>");
      oOut.write("<applyTo species=\"Paper_Birch\" type=\"Seed\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>output</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("</behavior>");
      oOut.write("</behaviorList>");
      oOut.write("<output>");
      oOut.write("<ou_filename>zxc.gz.tar</ou_filename>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>2</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>3</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>4</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>5</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Western_Redcedar\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>6</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>7</ou_saveFreq>");
      oOut.write("<ou_float>Y</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>8</ou_saveFreq>");
      oOut.write("<ou_float>Diam10</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Subalpine_Fir\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>9</ou_saveFreq>");
      oOut.write("<ou_float>DBH</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>10</ou_saveFreq>");
      oOut.write("<ou_float>Height</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>11</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("<ou_float>Light</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Hybrid_Spruce\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>12</ou_saveFreq>");
      oOut.write("<ou_float>Growth</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>13</ou_saveFreq>");
      oOut.write("<ou_bool>dead</ou_bool>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>14</ou_saveFreq>");
      oOut.write("<ou_int>yls</ou_int>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Lodgepole_Pine\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>15</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Trembling_Aspen\" type=\"Sapling\">");
      oOut.write("<ou_saveFreq>16</ou_saveFreq>");
      oOut.write("<ou_float>Diam10</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Black_Cottonwood\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>17</ou_saveFreq>");
      oOut.write("<ou_float>Light</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Black_Cottonwood\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>18</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Paper_Birch\" type=\"Seedling\">");
      oOut.write("<ou_saveFreq>19</ou_saveFreq>");
      oOut.write("<ou_float>Y</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_treeInfo species=\"Paper_Birch\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>20</ou_saveFreq>");
      oOut.write("<ou_bool>dead</ou_bool>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_gridInfo gridName=\"Windstorm Results\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_packageFloat>ba_0</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_1</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_2</ou_packageFloat>");
      oOut.write("<ou_packageFloat>ba_3</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_3</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_4</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_5</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_6</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_7</ou_packageFloat>");
      oOut.write("<ou_packageFloat>density_8</ou_packageFloat>");
      oOut.write("<ou_packageFloat>severity</ou_packageFloat>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Quadrat GLI\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>GLI</ou_float>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Substrate\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>scarsoil</ou_float>");
      oOut.write("<ou_float>tipup</ou_float>");
      oOut.write("<ou_float>freshlog</ou_float>");
      oOut.write("<ou_float>declog</ou_float>");
      oOut.write("<ou_float>ffmoss</ou_float>");
      oOut.write("<ou_float>fflitter</ou_float>");
      oOut.write("<ou_packageInt>age</ou_packageInt>");
      oOut.write("<ou_packageFloat>scarsoil</ou_packageFloat>");
      oOut.write("<ou_packageFloat>tipup</ou_packageFloat>");
      oOut.write("<ou_packageFloat>freshlog</ou_packageFloat>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Dispersed Seeds\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_int>count</ou_int>");
      oOut.write("<ou_float>seeds_0</ou_float>");
      oOut.write("<ou_float>seeds_2</ou_float>");
      oOut.write("<ou_float>seeds_4</ou_float>");
      oOut.write("<ou_float>seeds_5</ou_float>");
      oOut.write("<ou_float>seeds_6</ou_float>");
      oOut.write("<ou_float>seeds_8</ou_float>");
      oOut.write("<ou_bool>Is Gap</ou_bool>");
      oOut.write("</ou_gridInfo>");
      oOut.write("<ou_gridInfo gridName=\"Substrate Favorability\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>Favorability Index2</ou_float>");
      oOut.write("</ou_gridInfo>");
      oOut.write("</output>");
      oOut.write("</paramFile>");

      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      throw(new ModelException(ErrorGUI.BAD_FILE, "", oE.getMessage()));
    }
  }

  /**
   * This writes an XML file to test value setting.  This file contains
   * detailed output only.
   * @return Name of file.
   * @throws ModelException if there is an IO problem.
   */
  protected String WriteXMLFile6() throws ModelException {
    try {
      String sFileName = "\\loratestxml1.xml";
      FileWriter oOut = new FileWriter(sFileName);

      oOut.write(
          "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\" ?>");
      oOut.write("<paramFile fileCode=\"06010101\">");
      oOut.write("<plot>");
      oOut.write("<timesteps>100</timesteps>");
      oOut.write("<randomSeed>0</randomSeed>");
      oOut.write("<yearsPerTimestep>5.0</yearsPerTimestep>");
      oOut.write("<plot_lenX>200.0</plot_lenX>");
      oOut.write("<plot_lenY>200.0</plot_lenY>");
      oOut.write("<plot_latitude>55.37</plot_latitude>");
      oOut.write("<plot_title>Default parameter file</plot_title>");
      oOut.write("</plot>");
      oOut.write("<trees>");
      oOut.write("<tr_speciesList>");
      oOut.write("<tr_species speciesName=\"Western_Hemlock\"/>");
      oOut.write("<tr_species speciesName=\"Western_Redcedar\"/>");
      oOut.write("</tr_speciesList>");
      oOut.write("<tr_seedDiam10Cm>0.1</tr_seedDiam10Cm>");
      oOut.write("<tr_minAdultDBH>");
      oOut.write("<tr_madVal species=\"Western_Hemlock\">10.0</tr_madVal>");
      oOut.write("<tr_madVal species=\"Western_Redcedar\">10.0</tr_madVal>");
      oOut.write("</tr_minAdultDBH>");
      oOut.write("</trees>");
      oOut.write("<behaviorList>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>Constant GLI</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("<applyTo species=\"Western_Hemlock\" type=\"Adult\"/>");
      oOut.write("<applyTo species=\"Western_Redcedar\" type=\"Adult\"/>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>short output</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("</behavior>");
      oOut.write("<behavior>");
      oOut.write("<behaviorName>output</behaviorName>");
      oOut.write("<version>1.0</version>");
      oOut.write("</behavior>");
      oOut.write("</behaviorList>");
      oOut.write("<allometry>");
      oOut.write("<tr_whatAdultHeightDiam>");
      oOut.write("<tr_wahdVal species=\"Western_Hemlock\">0</tr_wahdVal>");
      oOut.write("<tr_wahdVal species=\"Western_Redcedar\">0</tr_wahdVal>");
      oOut.write("</tr_whatAdultHeightDiam>");
      oOut.write("</allometry>");
      oOut.write("<lightOther>");
      oOut.write("<li_constGLI>100.0</li_constGLI>");
      oOut.write("</lightOther>");
      oOut.write("<shortoutput>");
      oOut.write("<so_filename>C:\\lkjh.out</so_filename>");
      oOut.write("<so_treeTypeInfo type=\"Sapling\">");
      oOut.write("<so_saveRBA save=\"true\"/>");
      oOut.write("<so_saveABA save=\"false\"/>");
      oOut.write("<so_saveRDN save=\"false\"/>");
      oOut.write("<so_saveADN save=\"false\"/>");
      oOut.write("</so_treeTypeInfo>");
      oOut.write("<so_subplot>");
      oOut.write("<so_subplotName>plot 1</so_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"20\" y=\"20\"/>");
      oOut.write("<po_point x=\"20\" y=\"21\"/>");
      oOut.write("<po_point x=\"20\" y=\"22\"/>");
      oOut.write("<po_point x=\"20\" y=\"23\"/>");
      oOut.write("<po_point x=\"20\" y=\"24\"/>");
      oOut.write("<po_point x=\"21\" y=\"20\"/>");
      oOut.write("<po_point x=\"21\" y=\"21\"/>");
      oOut.write("<po_point x=\"21\" y=\"22\"/>");
      oOut.write("<po_point x=\"21\" y=\"23\"/>");
      oOut.write("<po_point x=\"21\" y=\"24\"/>");
      oOut.write("<po_point x=\"22\" y=\"20\"/>");
      oOut.write("<po_point x=\"22\" y=\"21\"/>");
      oOut.write("<po_point x=\"22\" y=\"22\"/>");
      oOut.write("<po_point x=\"22\" y=\"23\"/>");
      oOut.write("<po_point x=\"22\" y=\"24\"/>");
      oOut.write("<po_point x=\"23\" y=\"20\"/>");
      oOut.write("<po_point x=\"23\" y=\"21\"/>");
      oOut.write("<po_point x=\"23\" y=\"22\"/>");
      oOut.write("<po_point x=\"23\" y=\"23\"/>");
      oOut.write("<po_point x=\"23\" y=\"24\"/>");
      oOut.write("<po_point x=\"24\" y=\"20\"/>");
      oOut.write("<po_point x=\"24\" y=\"21\"/>");
      oOut.write("<po_point x=\"24\" y=\"22\"/>");
      oOut.write("<po_point x=\"24\" y=\"23\"/>");
      oOut.write("<po_point x=\"24\" y=\"24\"/>");
      oOut.write("</pointSet>");
      oOut.write("</so_subplot>");
      oOut.write("<so_subplot>");
      oOut.write("<so_subplotName>plot 2</so_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"0\" y=\"0\"/>");
      oOut.write("<po_point x=\"0\" y=\"1\"/>");
      oOut.write("<po_point x=\"0\" y=\"2\"/>");
      oOut.write("<po_point x=\"0\" y=\"3\"/>");
      oOut.write("<po_point x=\"0\" y=\"4\"/>");
      oOut.write("<po_point x=\"0\" y=\"5\"/>");
      oOut.write("<po_point x=\"0\" y=\"6\"/>");
      oOut.write("<po_point x=\"0\" y=\"7\"/>");
      oOut.write("<po_point x=\"0\" y=\"8\"/>");
      oOut.write("<po_point x=\"0\" y=\"9\"/>");
      oOut.write("<po_point x=\"0\" y=\"10\"/>");
      oOut.write("<po_point x=\"0\" y=\"11\"/>");
      oOut.write("<po_point x=\"0\" y=\"12\"/>");
      oOut.write("<po_point x=\"0\" y=\"13\"/>");
      oOut.write("<po_point x=\"0\" y=\"14\"/>");
      oOut.write("<po_point x=\"0\" y=\"15\"/>");
      oOut.write("<po_point x=\"0\" y=\"16\"/>");
      oOut.write("<po_point x=\"0\" y=\"17\"/>");
      oOut.write("<po_point x=\"0\" y=\"24\"/>");
      oOut.write("<po_point x=\"1\" y=\"24\"/>");
      oOut.write("<po_point x=\"2\" y=\"0\"/>");
      oOut.write("<po_point x=\"2\" y=\"1\"/>");
      oOut.write("<po_point x=\"2\" y=\"2\"/>");
      oOut.write("<po_point x=\"2\" y=\"3\"/>");
      oOut.write("<po_point x=\"2\" y=\"4\"/>");
      oOut.write("<po_point x=\"2\" y=\"5\"/>");
      oOut.write("<po_point x=\"2\" y=\"6\"/>");
      oOut.write("<po_point x=\"2\" y=\"7\"/>");
      oOut.write("<po_point x=\"2\" y=\"8\"/>");
      oOut.write("<po_point x=\"2\" y=\"9\"/>");
      oOut.write("<po_point x=\"2\" y=\"10\"/>");
      oOut.write("<po_point x=\"2\" y=\"11\"/>");
      oOut.write("<po_point x=\"2\" y=\"12\"/>");
      oOut.write("<po_point x=\"2\" y=\"13\"/>");
      oOut.write("<po_point x=\"2\" y=\"14\"/>");
      oOut.write("<po_point x=\"2\" y=\"15\"/>");
      oOut.write("<po_point x=\"2\" y=\"16\"/>");
      oOut.write("<po_point x=\"2\" y=\"17\"/>");
      oOut.write("<po_point x=\"2\" y=\"24\"/>");
      oOut.write("<po_point x=\"3\" y=\"24\"/>");
      oOut.write("<po_point x=\"4\" y=\"24\"/>");
      oOut.write("<po_point x=\"5\" y=\"24\"/>");
      oOut.write("<po_point x=\"24\" y=\"0\"/>");
      oOut.write("<po_point x=\"24\" y=\"1\"/>");
      oOut.write("<po_point x=\"24\" y=\"2\"/>");
      oOut.write("<po_point x=\"24\" y=\"3\"/>");
      oOut.write("<po_point x=\"24\" y=\"4\"/>");
      oOut.write("<po_point x=\"24\" y=\"5\"/>");
      oOut.write("</pointSet>");
      oOut.write("</so_subplot>");
      oOut.write("<so_subplot>");
      oOut.write("<so_subplotName>plot 3</so_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"0\" y=\"0\"/>");
      oOut.write("<po_point x=\"0\" y=\"1\"/>");
      oOut.write("<po_point x=\"0\" y=\"2\"/>");
      oOut.write("<po_point x=\"0\" y=\"3\"/>");
      oOut.write("<po_point x=\"0\" y=\"4\"/>");
      oOut.write("<po_point x=\"0\" y=\"5\"/>");
      oOut.write("<po_point x=\"1\" y=\"0\"/>");
      oOut.write("<po_point x=\"1\" y=\"1\"/>");
      oOut.write("<po_point x=\"1\" y=\"2\"/>");
      oOut.write("<po_point x=\"1\" y=\"3\"/>");
      oOut.write("<po_point x=\"1\" y=\"4\"/>");
      oOut.write("<po_point x=\"1\" y=\"5\"/>");
      oOut.write("<po_point x=\"2\" y=\"0\"/>");
      oOut.write("<po_point x=\"2\" y=\"1\"/>");
      oOut.write("<po_point x=\"2\" y=\"2\"/>");
      oOut.write("<po_point x=\"2\" y=\"3\"/>");
      oOut.write("<po_point x=\"2\" y=\"4\"/>");
      oOut.write("<po_point x=\"2\" y=\"5\"/>");
      oOut.write("<po_point x=\"3\" y=\"0\"/>");
      oOut.write("<po_point x=\"3\" y=\"1\"/>");
      oOut.write("<po_point x=\"3\" y=\"2\"/>");
      oOut.write("<po_point x=\"3\" y=\"3\"/>");
      oOut.write("<po_point x=\"3\" y=\"4\"/>");
      oOut.write("<po_point x=\"3\" y=\"5\"/>");
      oOut.write("<po_point x=\"4\" y=\"0\"/>");
      oOut.write("<po_point x=\"4\" y=\"1\"/>");
      oOut.write("<po_point x=\"4\" y=\"2\"/>");
      oOut.write("<po_point x=\"4\" y=\"3\"/>");
      oOut.write("<po_point x=\"4\" y=\"4\"/>");
      oOut.write("<po_point x=\"4\" y=\"5\"/>");
      oOut.write("<po_point x=\"5\" y=\"0\"/>");
      oOut.write("<po_point x=\"5\" y=\"1\"/>");
      oOut.write("<po_point x=\"5\" y=\"2\"/>");
      oOut.write("<po_point x=\"5\" y=\"3\"/>");
      oOut.write("<po_point x=\"5\" y=\"4\"/>");
      oOut.write("<po_point x=\"5\" y=\"5\"/>");
      oOut.write("<po_point x=\"6\" y=\"0\"/>");
      oOut.write("<po_point x=\"6\" y=\"1\"/>");
      oOut.write("<po_point x=\"6\" y=\"2\"/>");
      oOut.write("<po_point x=\"6\" y=\"3\"/>");
      oOut.write("<po_point x=\"6\" y=\"4\"/>");
      oOut.write("<po_point x=\"6\" y=\"5\"/>");
      oOut.write("<po_point x=\"7\" y=\"0\"/>");
      oOut.write("<po_point x=\"7\" y=\"1\"/>");
      oOut.write("<po_point x=\"7\" y=\"2\"/>");
      oOut.write("<po_point x=\"7\" y=\"3\"/>");
      oOut.write("<po_point x=\"7\" y=\"4\"/>");
      oOut.write("<po_point x=\"7\" y=\"5\"/>");
      oOut.write("<po_point x=\"8\" y=\"0\"/>");
      oOut.write("<po_point x=\"8\" y=\"1\"/>");
      oOut.write("<po_point x=\"8\" y=\"2\"/>");
      oOut.write("<po_point x=\"8\" y=\"3\"/>");
      oOut.write("<po_point x=\"8\" y=\"4\"/>");
      oOut.write("<po_point x=\"8\" y=\"5\"/>");
      oOut.write("<po_point x=\"9\" y=\"0\"/>");
      oOut.write("<po_point x=\"9\" y=\"1\"/>");
      oOut.write("<po_point x=\"9\" y=\"2\"/>");
      oOut.write("<po_point x=\"9\" y=\"3\"/>");
      oOut.write("<po_point x=\"9\" y=\"4\"/>");
      oOut.write("<po_point x=\"9\" y=\"5\"/>");
      oOut.write("<po_point x=\"10\" y=\"0\"/>");
      oOut.write("<po_point x=\"10\" y=\"1\"/>");
      oOut.write("<po_point x=\"10\" y=\"2\"/>");
      oOut.write("<po_point x=\"10\" y=\"3\"/>");
      oOut.write("<po_point x=\"10\" y=\"4\"/>");
      oOut.write("<po_point x=\"10\" y=\"5\"/>");
      oOut.write("<po_point x=\"11\" y=\"0\"/>");
      oOut.write("<po_point x=\"11\" y=\"1\"/>");
      oOut.write("<po_point x=\"11\" y=\"2\"/>");
      oOut.write("<po_point x=\"11\" y=\"3\"/>");
      oOut.write("<po_point x=\"11\" y=\"4\"/>");
      oOut.write("<po_point x=\"11\" y=\"5\"/>");
      oOut.write("</pointSet>");
      oOut.write("</so_subplot>");
      oOut.write("</shortoutput>");
      oOut.write("<output>");
      oOut.write("<ou_filename>C:\\lij.gz.tar</ou_filename>");
      oOut.write("<ou_treeInfo species=\"Western_Hemlock\" type=\"Adult\">");
      oOut.write("<ou_saveFreq>1</ou_saveFreq>");
      oOut.write("<ou_float>X</ou_float>");
      oOut.write("</ou_treeInfo>");
      oOut.write("<ou_subplot>");
      oOut.write("<ou_subplotName>plot 1</ou_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"20\" y=\"20\"/>");
      oOut.write("<po_point x=\"20\" y=\"21\"/>");
      oOut.write("<po_point x=\"20\" y=\"22\"/>");
      oOut.write("<po_point x=\"20\" y=\"23\"/>");
      oOut.write("<po_point x=\"20\" y=\"24\"/>");
      oOut.write("<po_point x=\"21\" y=\"20\"/>");
      oOut.write("<po_point x=\"21\" y=\"21\"/>");
      oOut.write("<po_point x=\"21\" y=\"22\"/>");
      oOut.write("<po_point x=\"21\" y=\"23\"/>");
      oOut.write("<po_point x=\"21\" y=\"24\"/>");
      oOut.write("<po_point x=\"22\" y=\"20\"/>");
      oOut.write("<po_point x=\"22\" y=\"21\"/>");
      oOut.write("<po_point x=\"22\" y=\"22\"/>");
      oOut.write("<po_point x=\"22\" y=\"23\"/>");
      oOut.write("<po_point x=\"22\" y=\"24\"/>");
      oOut.write("<po_point x=\"23\" y=\"20\"/>");
      oOut.write("<po_point x=\"23\" y=\"21\"/>");
      oOut.write("<po_point x=\"23\" y=\"22\"/>");
      oOut.write("<po_point x=\"23\" y=\"23\"/>");
      oOut.write("<po_point x=\"23\" y=\"24\"/>");
      oOut.write("<po_point x=\"24\" y=\"20\"/>");
      oOut.write("<po_point x=\"24\" y=\"21\"/>");
      oOut.write("<po_point x=\"24\" y=\"22\"/>");
      oOut.write("<po_point x=\"24\" y=\"23\"/>");
      oOut.write("<po_point x=\"24\" y=\"24\"/>");
      oOut.write("</pointSet>");
      oOut.write("</ou_subplot>");
      oOut.write("<ou_subplot>");
      oOut.write("<ou_subplotName>plot 2</ou_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"0\" y=\"0\"/>");
      oOut.write("<po_point x=\"0\" y=\"1\"/>");
      oOut.write("<po_point x=\"0\" y=\"2\"/>");
      oOut.write("<po_point x=\"0\" y=\"3\"/>");
      oOut.write("<po_point x=\"0\" y=\"4\"/>");
      oOut.write("<po_point x=\"0\" y=\"5\"/>");
      oOut.write("<po_point x=\"0\" y=\"6\"/>");
      oOut.write("<po_point x=\"0\" y=\"7\"/>");
      oOut.write("<po_point x=\"0\" y=\"8\"/>");
      oOut.write("<po_point x=\"0\" y=\"9\"/>");
      oOut.write("<po_point x=\"0\" y=\"10\"/>");
      oOut.write("<po_point x=\"0\" y=\"11\"/>");
      oOut.write("<po_point x=\"0\" y=\"12\"/>");
      oOut.write("<po_point x=\"0\" y=\"13\"/>");
      oOut.write("<po_point x=\"0\" y=\"14\"/>");
      oOut.write("<po_point x=\"0\" y=\"15\"/>");
      oOut.write("<po_point x=\"0\" y=\"16\"/>");
      oOut.write("<po_point x=\"0\" y=\"17\"/>");
      oOut.write("<po_point x=\"0\" y=\"24\"/>");
      oOut.write("<po_point x=\"1\" y=\"24\"/>");
      oOut.write("<po_point x=\"2\" y=\"0\"/>");
      oOut.write("<po_point x=\"2\" y=\"1\"/>");
      oOut.write("<po_point x=\"2\" y=\"2\"/>");
      oOut.write("<po_point x=\"2\" y=\"3\"/>");
      oOut.write("<po_point x=\"2\" y=\"4\"/>");
      oOut.write("<po_point x=\"2\" y=\"5\"/>");
      oOut.write("<po_point x=\"2\" y=\"6\"/>");
      oOut.write("<po_point x=\"2\" y=\"7\"/>");
      oOut.write("<po_point x=\"2\" y=\"8\"/>");
      oOut.write("<po_point x=\"2\" y=\"9\"/>");
      oOut.write("<po_point x=\"2\" y=\"10\"/>");
      oOut.write("<po_point x=\"2\" y=\"11\"/>");
      oOut.write("<po_point x=\"2\" y=\"12\"/>");
      oOut.write("<po_point x=\"2\" y=\"13\"/>");
      oOut.write("<po_point x=\"2\" y=\"14\"/>");
      oOut.write("<po_point x=\"2\" y=\"15\"/>");
      oOut.write("<po_point x=\"2\" y=\"16\"/>");
      oOut.write("<po_point x=\"2\" y=\"17\"/>");
      oOut.write("<po_point x=\"2\" y=\"24\"/>");
      oOut.write("<po_point x=\"3\" y=\"24\"/>");
      oOut.write("<po_point x=\"4\" y=\"24\"/>");
      oOut.write("<po_point x=\"5\" y=\"24\"/>");
      oOut.write("<po_point x=\"24\" y=\"0\"/>");
      oOut.write("<po_point x=\"24\" y=\"1\"/>");
      oOut.write("<po_point x=\"24\" y=\"2\"/>");
      oOut.write("<po_point x=\"24\" y=\"3\"/>");
      oOut.write("<po_point x=\"24\" y=\"4\"/>");
      oOut.write("<po_point x=\"24\" y=\"5\"/>");
      oOut.write("</pointSet>");
      oOut.write("</ou_subplot>");
      oOut.write("<ou_subplot>");
      oOut.write("<ou_subplotName>plot 3</ou_subplotName>");
      oOut.write("<pointSet>");
      oOut.write("<po_point x=\"0\" y=\"0\"/>");
      oOut.write("<po_point x=\"0\" y=\"1\"/>");
      oOut.write("<po_point x=\"0\" y=\"2\"/>");
      oOut.write("<po_point x=\"0\" y=\"3\"/>");
      oOut.write("<po_point x=\"0\" y=\"4\"/>");
      oOut.write("<po_point x=\"0\" y=\"5\"/>");
      oOut.write("<po_point x=\"1\" y=\"0\"/>");
      oOut.write("<po_point x=\"1\" y=\"1\"/>");
      oOut.write("<po_point x=\"1\" y=\"2\"/>");
      oOut.write("<po_point x=\"1\" y=\"3\"/>");
      oOut.write("<po_point x=\"1\" y=\"4\"/>");
      oOut.write("<po_point x=\"1\" y=\"5\"/>");
      oOut.write("<po_point x=\"2\" y=\"0\"/>");
      oOut.write("<po_point x=\"2\" y=\"1\"/>");
      oOut.write("<po_point x=\"2\" y=\"2\"/>");
      oOut.write("<po_point x=\"2\" y=\"3\"/>");
      oOut.write("<po_point x=\"2\" y=\"4\"/>");
      oOut.write("<po_point x=\"2\" y=\"5\"/>");
      oOut.write("<po_point x=\"3\" y=\"0\"/>");
      oOut.write("<po_point x=\"3\" y=\"1\"/>");
      oOut.write("<po_point x=\"3\" y=\"2\"/>");
      oOut.write("<po_point x=\"3\" y=\"3\"/>");
      oOut.write("<po_point x=\"3\" y=\"4\"/>");
      oOut.write("<po_point x=\"3\" y=\"5\"/>");
      oOut.write("<po_point x=\"4\" y=\"0\"/>");
      oOut.write("<po_point x=\"4\" y=\"1\"/>");
      oOut.write("<po_point x=\"4\" y=\"2\"/>");
      oOut.write("<po_point x=\"4\" y=\"3\"/>");
      oOut.write("<po_point x=\"4\" y=\"4\"/>");
      oOut.write("<po_point x=\"4\" y=\"5\"/>");
      oOut.write("<po_point x=\"5\" y=\"0\"/>");
      oOut.write("<po_point x=\"5\" y=\"1\"/>");
      oOut.write("<po_point x=\"5\" y=\"2\"/>");
      oOut.write("<po_point x=\"5\" y=\"3\"/>");
      oOut.write("<po_point x=\"5\" y=\"4\"/>");
      oOut.write("<po_point x=\"5\" y=\"5\"/>");
      oOut.write("<po_point x=\"6\" y=\"0\"/>");
      oOut.write("<po_point x=\"6\" y=\"1\"/>");
      oOut.write("<po_point x=\"6\" y=\"2\"/>");
      oOut.write("<po_point x=\"6\" y=\"3\"/>");
      oOut.write("<po_point x=\"6\" y=\"4\"/>");
      oOut.write("<po_point x=\"6\" y=\"5\"/>");
      oOut.write("<po_point x=\"7\" y=\"0\"/>");
      oOut.write("<po_point x=\"7\" y=\"1\"/>");
      oOut.write("<po_point x=\"7\" y=\"2\"/>");
      oOut.write("<po_point x=\"7\" y=\"3\"/>");
      oOut.write("<po_point x=\"7\" y=\"4\"/>");
      oOut.write("<po_point x=\"7\" y=\"5\"/>");
      oOut.write("<po_point x=\"8\" y=\"0\"/>");
      oOut.write("<po_point x=\"8\" y=\"1\"/>");
      oOut.write("<po_point x=\"8\" y=\"2\"/>");
      oOut.write("<po_point x=\"8\" y=\"3\"/>");
      oOut.write("<po_point x=\"8\" y=\"4\"/>");
      oOut.write("<po_point x=\"8\" y=\"5\"/>");
      oOut.write("<po_point x=\"9\" y=\"0\"/>");
      oOut.write("<po_point x=\"9\" y=\"1\"/>");
      oOut.write("<po_point x=\"9\" y=\"2\"/>");
      oOut.write("<po_point x=\"9\" y=\"3\"/>");
      oOut.write("<po_point x=\"9\" y=\"4\"/>");
      oOut.write("<po_point x=\"9\" y=\"5\"/>");
      oOut.write("<po_point x=\"10\" y=\"0\"/>");
      oOut.write("<po_point x=\"10\" y=\"1\"/>");
      oOut.write("<po_point x=\"10\" y=\"2\"/>");
      oOut.write("<po_point x=\"10\" y=\"3\"/>");
      oOut.write("<po_point x=\"10\" y=\"4\"/>");
      oOut.write("<po_point x=\"10\" y=\"5\"/>");
      oOut.write("<po_point x=\"11\" y=\"0\"/>");
      oOut.write("<po_point x=\"11\" y=\"1\"/>");
      oOut.write("<po_point x=\"11\" y=\"2\"/>");
      oOut.write("<po_point x=\"11\" y=\"3\"/>");
      oOut.write("<po_point x=\"11\" y=\"4\"/>");
      oOut.write("<po_point x=\"11\" y=\"5\"/>");
      oOut.write("</pointSet>");
      oOut.write("</ou_subplot>");
      oOut.write("</output>");
      oOut.write("</paramFile>");


      oOut.close();
      return sFileName;
    }
    catch (java.io.IOException oE) {
      throw(new ModelException(ErrorGUI.BAD_FILE, "", oE.getMessage()));
    }
  }

  /**
   * Test two rundata settings for equality.
   * @param o1 First settings to test.
   * @param o2 Second settings to test.
   * @return True if all settings are the same, false if not.
   */
  private boolean testEquals(DetailedTreeSettings o1, DetailedTreeSettings o2) {
    if (o1.GetNumberOfBools() != o2.GetNumberOfBools() ||
        o1.GetNumberOfFloats() != o2.GetNumberOfFloats() ||
        o1.GetNumberOfChars() != o2.GetNumberOfChars() ||
        o1.GetNumberOfInts() != o2.GetNumberOfInts() ||
        o1.GetSaveFrequency() != o2.GetSaveFrequency() ||
        o1.GetSpecies() != o2.GetSpecies() ||
        o1.GetType() != o2.GetType()) {
      return false;
    }

    int i;
    for (i = 0; i < o1.GetNumberOfBools(); i++) {
      if (o1.GetBool(i) != o2.GetBool(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfChars(); i++) {
      if (o1.GetChar(i) != o2.GetChar(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfFloats(); i++) {
      if (o1.GetFloat(i) != o2.GetFloat(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfInts(); i++) {
      if (o1.GetInt(i) != o2.GetInt(i)) {
        return false;
      }
    }

    return true;
  }

  /**
   * Test two rundata settings for equality.
   * @param o1 First settings to test.
   * @param o2 Second settings to test.
   * @return True if all settings are the same, false if not.
   */
  private boolean testEquals(DetailedGridSettings o1, DetailedGridSettings o2) {
    if (o1.GetNumberOfBools() != o2.GetNumberOfBools() ||
        o1.GetNumberOfFloats() != o2.GetNumberOfFloats() ||
        o1.GetNumberOfChars() != o2.GetNumberOfChars() ||
        o1.GetNumberOfInts() != o2.GetNumberOfInts() ||
        o1.GetNumberOfPackageBools() != o2.GetNumberOfPackageBools() ||
        o1.GetNumberOfPackageFloats() != o2.GetNumberOfPackageFloats() ||
        o1.GetNumberOfPackageChars() != o2.GetNumberOfPackageChars() ||
        o1.GetNumberOfPackageInts() != o2.GetNumberOfPackageInts() ||
        o1.GetSaveFrequency() != o2.GetSaveFrequency() ||
        o1.GetName() != o2.GetName()) {
      return false;
    }

    int i;
    for (i = 0; i < o1.GetNumberOfBools(); i++) {
      if (o1.GetBool(i) != o2.GetBool(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfChars(); i++) {
      if (o1.GetChar(i) != o2.GetChar(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfFloats(); i++) {
      if (o1.GetFloat(i) != o2.GetFloat(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfInts(); i++) {
      if (o1.GetInt(i) != o2.GetInt(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfPackageBools(); i++) {
      if (o1.GetPackageBool(i) != o2.GetPackageBool(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfPackageChars(); i++) {
      if (o1.GetPackageChar(i) != o2.GetPackageChar(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfPackageFloats(); i++) {
      if (o1.GetPackageFloat(i) != o2.GetPackageFloat(i)) {
        return false;
      }
    }

    for (i = 0; i < o1.GetNumberOfPackageInts(); i++) {
      if (o1.GetPackageInt(i) != o2.GetPackageInt(i)) {
        return false;
      }
    }

    return true;
  }

  /**
   * Tests grid settings when species are changed.
   */
  public void testChangeOfSpecies() {
    GUIManager oManager = null;
    String sFileName = null;
    try {
      oManager = new GUIManager(null);
      sFileName = WriteXMLFile4();
      oManager.InputXMLParameterFile(sFileName);

      //Verify the current grid settings
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();
      assertEquals(5, oOutput.GetNumberOfDetailedGridSettings());
      DetailedGridSettings oGridSettings = oOutput.GetDetailedGridSetting(0);
      assertEquals("Windstorm Results", oGridSettings.GetName());
      assertEquals(19, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("ba_0", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("ba_1", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("ba_2", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("ba_3", oGridSettings.GetPackageFloat(3).GetCodeName());
      assertEquals("ba_4", oGridSettings.GetPackageFloat(4).GetCodeName());
      assertEquals("ba_5", oGridSettings.GetPackageFloat(5).GetCodeName());
      assertEquals("ba_6", oGridSettings.GetPackageFloat(6).GetCodeName());
      assertEquals("ba_7", oGridSettings.GetPackageFloat(7).GetCodeName());
      assertEquals("ba_8", oGridSettings.GetPackageFloat(8).GetCodeName());
      assertEquals("density_0", oGridSettings.GetPackageFloat(9).GetCodeName());
      assertEquals("density_1", oGridSettings.GetPackageFloat(10).GetCodeName());
      assertEquals("density_2", oGridSettings.GetPackageFloat(11).GetCodeName());
      assertEquals("density_3", oGridSettings.GetPackageFloat(12).GetCodeName());
      assertEquals("density_4", oGridSettings.GetPackageFloat(13).GetCodeName());
      assertEquals("density_5", oGridSettings.GetPackageFloat(14).GetCodeName());
      assertEquals("density_6", oGridSettings.GetPackageFloat(15).GetCodeName());
      assertEquals("density_7", oGridSettings.GetPackageFloat(16).GetCodeName());
      assertEquals("density_8", oGridSettings.GetPackageFloat(17).GetCodeName());
      assertEquals("severity", oGridSettings.GetPackageFloat(18).GetCodeName());
      assertEquals("Basal Area Dead For Western Hemlock", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Basal Area Dead For Western Redcedar", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Basal Area Dead For Amabilis Fir", oGridSettings.GetPackageFloat(2).GetDisplayName());
      assertEquals("Basal Area Dead For Subalpine Fir", oGridSettings.GetPackageFloat(3).GetDisplayName());
      assertEquals("Basal Area Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(4).GetDisplayName());
      assertEquals("Basal Area Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(5).GetDisplayName());
      assertEquals("Basal Area Dead For Trembling Aspen", oGridSettings.GetPackageFloat(6).GetDisplayName());
      assertEquals("Basal Area Dead For Black Cottonwood", oGridSettings.GetPackageFloat(7).GetDisplayName());
      assertEquals("Basal Area Dead For Paper Birch", oGridSettings.GetPackageFloat(8).GetDisplayName());
      assertEquals("Density Dead For Western Hemlock", oGridSettings.GetPackageFloat(9).GetDisplayName());
      assertEquals("Density Dead For Western Redcedar", oGridSettings.GetPackageFloat(10).GetDisplayName());
      assertEquals("Density Dead For Amabilis Fir", oGridSettings.GetPackageFloat(11).GetDisplayName());
      assertEquals("Density Dead For Subalpine Fir", oGridSettings.GetPackageFloat(12).GetDisplayName());
      assertEquals("Density Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(13).GetDisplayName());
      assertEquals("Density Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(14).GetDisplayName());
      assertEquals("Density Dead For Trembling Aspen", oGridSettings.GetPackageFloat(15).GetDisplayName());
      assertEquals("Density Dead For Black Cottonwood", oGridSettings.GetPackageFloat(16).GetDisplayName());
      assertEquals("Density Dead For Paper Birch", oGridSettings.GetPackageFloat(17).GetDisplayName());
      assertEquals("Storm Severity", oGridSettings.GetPackageFloat(18).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(1);
      assertEquals("Quadrat GLI", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(2);
      assertEquals("Substrate", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("scarsoil", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("declog", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("ffmoss", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("fflitter", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Proportion of scarified soil", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Proportion of tip-up mounds", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Proportion of fresh logs", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Proportion of decayed logs", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Proportion of forest floor moss", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Proportion of forest floor litter", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfPackageInts());
      assertEquals("age", oGridSettings.GetPackageInt(0).GetCodeName());
      assertEquals("Substrate cohort age", oGridSettings.GetPackageInt(0).GetDisplayName());
      assertEquals(3, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("scarsoil", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("Substrate cohort new scarified soil substrate", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Substrate cohort new tip-up mounds substrate", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Substrate cohort new fresh logs", oGridSettings.GetPackageFloat(2).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(3);
      assertEquals("Dispersed Seeds", oGridSettings.GetName());
      assertEquals(9, oGridSettings.GetNumberOfFloats());
      assertEquals("seeds_0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("seeds_1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("seeds_2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("seeds_3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("seeds_4", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("seeds_5", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("seeds_6", oGridSettings.GetFloat(6).GetCodeName());
      assertEquals("seeds_7", oGridSettings.GetFloat(7).GetCodeName());
      assertEquals("seeds_8", oGridSettings.GetFloat(8).GetCodeName());
      assertEquals("Number of seeds for Western Hemlock", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Number of seeds for Western Redcedar", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Number of seeds for Amabilis Fir", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Number of seeds for Subalpine Fir", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Number of seeds for Hybrid Spruce", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Number of seeds for Lodgepole Pine", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals("Number of seeds for Trembling Aspen", oGridSettings.GetFloat(6).GetDisplayName());
      assertEquals("Number of seeds for Black Cottonwood", oGridSettings.GetFloat(7).GetDisplayName());
      assertEquals("Number of seeds for Paper Birch", oGridSettings.GetFloat(8).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfInts());
      assertEquals("count", oGridSettings.GetInt(0).GetCodeName());
      assertEquals("Adult tree count", oGridSettings.GetInt(0).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfBools());
      assertEquals("Is Gap", oGridSettings.GetBool(0).GetCodeName());
      assertEquals("Gap status", oGridSettings.GetBool(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(4);
      assertEquals("Substrate Favorability", oGridSettings.GetName());
      assertEquals(9, oGridSettings.GetNumberOfFloats());
      assertEquals("Favorability Index0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("Favorability Index1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("Favorability Index2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("Favorability Index3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("Favorability Index4", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("Favorability Index5", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Favorability Index6", oGridSettings.GetFloat(6).GetCodeName());
      assertEquals("Favorability Index7", oGridSettings.GetFloat(7).GetCodeName());
      assertEquals("Favorability Index8", oGridSettings.GetFloat(8).GetCodeName());
      assertEquals("Favorability Index - Western Hemlock", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Favorability Index - Western Redcedar", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Favorability Index - Amabilis Fir", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Favorability Index - Subalpine Fir", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Favorability Index - Hybrid Spruce", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Favorability Index - Lodgepole Pine", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals("Favorability Index - Trembling Aspen", oGridSettings.GetFloat(6).GetDisplayName());
      assertEquals("Favorability Index - Black Cottonwood", oGridSettings.GetFloat(7).GetDisplayName());
      assertEquals("Favorability Index - Paper Birch", oGridSettings.GetFloat(8).GetDisplayName());

      //Verify the current tree settings
      assertEquals(20, oOutput.GetNumberOfDetailedLiveTreeSettings());
      DetailedTreeSettings oTreeSettings = oOutput.GetDetailedLiveTreeSetting(0);
      assertEquals(1, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(1);
      assertEquals(2, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(2);
      assertEquals(3, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(3);
      assertEquals(4, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(4);
      assertEquals(5, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(5);
      assertEquals(6, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(6);
      assertEquals(7, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(7);
      assertEquals(8, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(8);
      assertEquals(9, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("DBH", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(9);
      assertEquals(10, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Height", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(10);
      assertEquals(11, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(11);
      assertEquals(12, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("Growth", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(12);
      assertEquals(13, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(13);
      assertEquals(14, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(14);
      assertEquals(15, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(15);
      assertEquals(16, oTreeSettings.GetSaveFrequency());
      assertEquals(6, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(16);
      assertEquals(17, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(17);
      assertEquals(18, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(18);
      assertEquals(19, oTreeSettings.GetSaveFrequency());
      assertEquals(8, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(19);
      assertEquals(20, oTreeSettings.GetSaveFrequency());
      assertEquals(8, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      /////////////////////////////
      //Remove the first species
      //////////////////////////////
      String[] p_sSpeciesList = new String[] {
          "Western Redcedar",
          "Amabilis Fir",
          "Subalpine Fir",
          "Hybrid Spruce",
          "Lodgepole Pine",
          "Trembling Aspen",
          "Black Cottonwood",
          "Paper Birch"
      };
      TreePopulation oPop = oManager.GetTreePopulation();
      oPop.SetSpeciesNames(p_sSpeciesList);
      assertEquals(5, oOutput.GetNumberOfDetailedGridSettings());
      oGridSettings = oOutput.GetDetailedGridSetting(0);
      assertEquals("Windstorm Results", oGridSettings.GetName());
      assertEquals(17, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("ba_0", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("ba_1", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("ba_2", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("ba_3", oGridSettings.GetPackageFloat(3).GetCodeName());
      assertEquals("ba_4", oGridSettings.GetPackageFloat(4).GetCodeName());
      assertEquals("ba_5", oGridSettings.GetPackageFloat(5).GetCodeName());
      assertEquals("ba_6", oGridSettings.GetPackageFloat(6).GetCodeName());
      assertEquals("ba_7", oGridSettings.GetPackageFloat(7).GetCodeName());
      assertEquals("density_0", oGridSettings.GetPackageFloat(8).GetCodeName());
      assertEquals("density_1", oGridSettings.GetPackageFloat(9).GetCodeName());
      assertEquals("density_2", oGridSettings.GetPackageFloat(10).GetCodeName());
      assertEquals("density_3", oGridSettings.GetPackageFloat(11).GetCodeName());
      assertEquals("density_4", oGridSettings.GetPackageFloat(12).GetCodeName());
      assertEquals("density_5", oGridSettings.GetPackageFloat(13).GetCodeName());
      assertEquals("density_6", oGridSettings.GetPackageFloat(14).GetCodeName());
      assertEquals("density_7", oGridSettings.GetPackageFloat(15).GetCodeName());
      assertEquals("severity", oGridSettings.GetPackageFloat(16).GetCodeName());
      assertEquals("Basal Area Dead For Western Redcedar", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Basal Area Dead For Amabilis Fir", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Basal Area Dead For Subalpine Fir", oGridSettings.GetPackageFloat(2).GetDisplayName());
      assertEquals("Basal Area Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(3).GetDisplayName());
      assertEquals("Basal Area Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(4).GetDisplayName());
      assertEquals("Basal Area Dead For Trembling Aspen", oGridSettings.GetPackageFloat(5).GetDisplayName());
      assertEquals("Basal Area Dead For Black Cottonwood", oGridSettings.GetPackageFloat(6).GetDisplayName());
      assertEquals("Basal Area Dead For Paper Birch", oGridSettings.GetPackageFloat(7).GetDisplayName());
      assertEquals("Density Dead For Western Redcedar", oGridSettings.GetPackageFloat(8).GetDisplayName());
      assertEquals("Density Dead For Amabilis Fir", oGridSettings.GetPackageFloat(9).GetDisplayName());
      assertEquals("Density Dead For Subalpine Fir", oGridSettings.GetPackageFloat(10).GetDisplayName());
      assertEquals("Density Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(11).GetDisplayName());
      assertEquals("Density Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(12).GetDisplayName());
      assertEquals("Density Dead For Trembling Aspen", oGridSettings.GetPackageFloat(13).GetDisplayName());
      assertEquals("Density Dead For Black Cottonwood", oGridSettings.GetPackageFloat(14).GetDisplayName());
      assertEquals("Density Dead For Paper Birch", oGridSettings.GetPackageFloat(15).GetDisplayName());
      assertEquals("Storm Severity", oGridSettings.GetPackageFloat(16).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(1);
      assertEquals("Quadrat GLI", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(2);
      assertEquals("Substrate", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("scarsoil", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("declog", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("ffmoss", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("fflitter", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Proportion of scarified soil", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Proportion of tip-up mounds", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Proportion of fresh logs", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Proportion of decayed logs", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Proportion of forest floor moss", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Proportion of forest floor litter", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfPackageInts());
      assertEquals("age", oGridSettings.GetPackageInt(0).GetCodeName());
      assertEquals("Substrate cohort age", oGridSettings.GetPackageInt(0).GetDisplayName());
      assertEquals(3, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("scarsoil", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("Substrate cohort new scarified soil substrate", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Substrate cohort new tip-up mounds substrate", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Substrate cohort new fresh logs", oGridSettings.GetPackageFloat(2).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(3);
      assertEquals("Dispersed Seeds", oGridSettings.GetName());
      assertEquals(8, oGridSettings.GetNumberOfFloats());
      assertEquals("seeds_0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("seeds_1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("seeds_2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("seeds_3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("seeds_4", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("seeds_5", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("seeds_6", oGridSettings.GetFloat(6).GetCodeName());
      assertEquals("seeds_7", oGridSettings.GetFloat(7).GetCodeName());
      assertEquals("Number of seeds for Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Number of seeds for Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Number of seeds for Subalpine Fir", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Number of seeds for Hybrid Spruce", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Number of seeds for Lodgepole Pine", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Number of seeds for Trembling Aspen", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals("Number of seeds for Black Cottonwood", oGridSettings.GetFloat(6).GetDisplayName());
      assertEquals("Number of seeds for Paper Birch", oGridSettings.GetFloat(7).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfInts());
      assertEquals("count", oGridSettings.GetInt(0).GetCodeName());
      assertEquals("Adult tree count", oGridSettings.GetInt(0).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfBools());
      assertEquals("Is Gap", oGridSettings.GetBool(0).GetCodeName());
      assertEquals("Gap status", oGridSettings.GetBool(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(4);
      assertEquals("Substrate Favorability", oGridSettings.GetName());
      assertEquals(8, oGridSettings.GetNumberOfFloats());
      assertEquals("Favorability Index0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("Favorability Index1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("Favorability Index2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("Favorability Index3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("Favorability Index4", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("Favorability Index5", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Favorability Index6", oGridSettings.GetFloat(6).GetCodeName());
      assertEquals("Favorability Index7", oGridSettings.GetFloat(7).GetCodeName());
      assertEquals("Favorability Index - Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Favorability Index - Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Favorability Index - Subalpine Fir", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Favorability Index - Hybrid Spruce", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Favorability Index - Lodgepole Pine", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Favorability Index - Trembling Aspen", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals("Favorability Index - Black Cottonwood", oGridSettings.GetFloat(6).GetDisplayName());
      assertEquals("Favorability Index - Paper Birch", oGridSettings.GetFloat(7).GetDisplayName());

      //Verify the tree settings
      assertEquals(17, oOutput.GetNumberOfDetailedLiveTreeSettings());
      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(0);
      assertEquals(4, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(1);
      assertEquals(5, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(2);
      assertEquals(6, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(3);
      assertEquals(7, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(4);
      assertEquals(8, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(5);
      assertEquals(9, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("DBH", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(6);
      assertEquals(10, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Height", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(7);
      assertEquals(11, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(8);
      assertEquals(12, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("Growth", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(9);
      assertEquals(13, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(10);
      assertEquals(14, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(11);
      assertEquals(15, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(12);
      assertEquals(16, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(13);
      assertEquals(17, oTreeSettings.GetSaveFrequency());
      assertEquals(6, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(14);
      assertEquals(18, oTreeSettings.GetSaveFrequency());
      assertEquals(6, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(15);
      assertEquals(19, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(16);
      assertEquals(20, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      /////////////////////////////
      // Remove the last species
      /////////////////////////////
      p_sSpeciesList = new String[] {
          "Western Redcedar",
          "Amabilis Fir",
          "Subalpine Fir",
          "Hybrid Spruce",
          "Lodgepole Pine",
          "Trembling Aspen",
          "Black Cottonwood"
      };
      oPop.SetSpeciesNames(p_sSpeciesList);
      assertEquals(5, oOutput.GetNumberOfDetailedGridSettings());
      oGridSettings = oOutput.GetDetailedGridSetting(0);
      assertEquals("Windstorm Results", oGridSettings.GetName());
      assertEquals(15, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("ba_0", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("ba_1", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("ba_2", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("ba_3", oGridSettings.GetPackageFloat(3).GetCodeName());
      assertEquals("ba_4", oGridSettings.GetPackageFloat(4).GetCodeName());
      assertEquals("ba_5", oGridSettings.GetPackageFloat(5).GetCodeName());
      assertEquals("ba_6", oGridSettings.GetPackageFloat(6).GetCodeName());
      assertEquals("density_0", oGridSettings.GetPackageFloat(7).GetCodeName());
      assertEquals("density_1", oGridSettings.GetPackageFloat(8).GetCodeName());
      assertEquals("density_2", oGridSettings.GetPackageFloat(9).GetCodeName());
      assertEquals("density_3", oGridSettings.GetPackageFloat(10).GetCodeName());
      assertEquals("density_4", oGridSettings.GetPackageFloat(11).GetCodeName());
      assertEquals("density_5", oGridSettings.GetPackageFloat(12).GetCodeName());
      assertEquals("density_6", oGridSettings.GetPackageFloat(13).GetCodeName());
      assertEquals("severity", oGridSettings.GetPackageFloat(14).GetCodeName());
      assertEquals("Basal Area Dead For Western Redcedar", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Basal Area Dead For Amabilis Fir", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Basal Area Dead For Subalpine Fir", oGridSettings.GetPackageFloat(2).GetDisplayName());
      assertEquals("Basal Area Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(3).GetDisplayName());
      assertEquals("Basal Area Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(4).GetDisplayName());
      assertEquals("Basal Area Dead For Trembling Aspen", oGridSettings.GetPackageFloat(5).GetDisplayName());
      assertEquals("Basal Area Dead For Black Cottonwood", oGridSettings.GetPackageFloat(6).GetDisplayName());
      assertEquals("Density Dead For Western Redcedar", oGridSettings.GetPackageFloat(7).GetDisplayName());
      assertEquals("Density Dead For Amabilis Fir", oGridSettings.GetPackageFloat(8).GetDisplayName());
      assertEquals("Density Dead For Subalpine Fir", oGridSettings.GetPackageFloat(9).GetDisplayName());
      assertEquals("Density Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(10).GetDisplayName());
      assertEquals("Density Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(11).GetDisplayName());
      assertEquals("Density Dead For Trembling Aspen", oGridSettings.GetPackageFloat(12).GetDisplayName());
      assertEquals("Density Dead For Black Cottonwood", oGridSettings.GetPackageFloat(13).GetDisplayName());
      assertEquals("Storm Severity", oGridSettings.GetPackageFloat(14).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(1);
      assertEquals("Quadrat GLI", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(2);
      assertEquals("Substrate", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("scarsoil", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("declog", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("ffmoss", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("fflitter", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Proportion of scarified soil", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Proportion of tip-up mounds", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Proportion of fresh logs", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Proportion of decayed logs", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Proportion of forest floor moss", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Proportion of forest floor litter", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfPackageInts());
      assertEquals("age", oGridSettings.GetPackageInt(0).GetCodeName());
      assertEquals("Substrate cohort age", oGridSettings.GetPackageInt(0).GetDisplayName());
      assertEquals(3, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("scarsoil", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("Substrate cohort new scarified soil substrate", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Substrate cohort new tip-up mounds substrate", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Substrate cohort new fresh logs", oGridSettings.GetPackageFloat(2).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(3);
      assertEquals("Dispersed Seeds", oGridSettings.GetName());
      assertEquals(7, oGridSettings.GetNumberOfFloats());
      assertEquals("seeds_0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("seeds_1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("seeds_2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("seeds_3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("seeds_4", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("seeds_5", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("seeds_6", oGridSettings.GetFloat(6).GetCodeName());
      assertEquals("Number of seeds for Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Number of seeds for Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Number of seeds for Subalpine Fir", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Number of seeds for Hybrid Spruce", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Number of seeds for Lodgepole Pine", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Number of seeds for Trembling Aspen", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals("Number of seeds for Black Cottonwood", oGridSettings.GetFloat(6).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfInts());
      assertEquals("count", oGridSettings.GetInt(0).GetCodeName());
      assertEquals("Adult tree count", oGridSettings.GetInt(0).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfBools());
      assertEquals("Is Gap", oGridSettings.GetBool(0).GetCodeName());
      assertEquals("Gap status", oGridSettings.GetBool(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(4);
      assertEquals("Substrate Favorability", oGridSettings.GetName());
      assertEquals(7, oGridSettings.GetNumberOfFloats());
      assertEquals("Favorability Index0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("Favorability Index1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("Favorability Index2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("Favorability Index3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("Favorability Index4", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("Favorability Index5", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Favorability Index6", oGridSettings.GetFloat(6).GetCodeName());
      assertEquals("Favorability Index - Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Favorability Index - Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Favorability Index - Subalpine Fir", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Favorability Index - Hybrid Spruce", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Favorability Index - Lodgepole Pine", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Favorability Index - Trembling Aspen", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals("Favorability Index - Black Cottonwood", oGridSettings.GetFloat(6).GetDisplayName());

      //Verify the tree settings
      assertEquals(15, oOutput.GetNumberOfDetailedLiveTreeSettings());
      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(0);
      assertEquals(4, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(1);
      assertEquals(5, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(2);
      assertEquals(6, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(3);
      assertEquals(7, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(4);
      assertEquals(8, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(5);
      assertEquals(9, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("DBH", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(6);
      assertEquals(10, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Height", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(7);
      assertEquals(11, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(8);
      assertEquals(12, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("Growth", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(9);
      assertEquals(13, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(10);
      assertEquals(14, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(11);
      assertEquals(15, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(12);
      assertEquals(16, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(13);
      assertEquals(17, oTreeSettings.GetSaveFrequency());
      assertEquals(6, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(14);
      assertEquals(18, oTreeSettings.GetSaveFrequency());
      assertEquals(6, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      /////////////////////////////
      // Remove only output settings - don't change species
      /////////////////////////////
      oGridSettings = oOutput.GetDetailedGridSetting(0);
      assertEquals("Windstorm Results", oGridSettings.GetName());
      int i;
      for (i = 0; i < oGridSettings.GetNumberOfPackageFloats(); i++) {
        if (oGridSettings.GetPackageFloat(i).GetCodeName().equals("ba_0")) {
          oGridSettings.RemovePackageFloat(i);
          break;
        }
      }
      for (i = 0; i < oGridSettings.GetNumberOfPackageFloats(); i++) {
        if (oGridSettings.GetPackageFloat(i).GetCodeName().equals("density_1")) {
          oGridSettings.RemovePackageFloat(i);
          break;
        }
      }
      assertEquals(13, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("ba_1", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("ba_2", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("ba_3", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("ba_4", oGridSettings.GetPackageFloat(3).GetCodeName());
      assertEquals("ba_5", oGridSettings.GetPackageFloat(4).GetCodeName());
      assertEquals("ba_6", oGridSettings.GetPackageFloat(5).GetCodeName());
      assertEquals("density_0", oGridSettings.GetPackageFloat(6).GetCodeName());
      assertEquals("density_2", oGridSettings.GetPackageFloat(7).GetCodeName());
      assertEquals("density_3", oGridSettings.GetPackageFloat(8).GetCodeName());
      assertEquals("density_4", oGridSettings.GetPackageFloat(9).GetCodeName());
      assertEquals("density_5", oGridSettings.GetPackageFloat(10).GetCodeName());
      assertEquals("density_6", oGridSettings.GetPackageFloat(11).GetCodeName());
      assertEquals("severity", oGridSettings.GetPackageFloat(12).GetCodeName());
      assertEquals("Basal Area Dead For Amabilis Fir", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Basal Area Dead For Subalpine Fir", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Basal Area Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(2).GetDisplayName());
      assertEquals("Basal Area Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(3).GetDisplayName());
      assertEquals("Basal Area Dead For Trembling Aspen", oGridSettings.GetPackageFloat(4).GetDisplayName());
      assertEquals("Basal Area Dead For Black Cottonwood", oGridSettings.GetPackageFloat(5).GetDisplayName());
      assertEquals("Density Dead For Western Redcedar", oGridSettings.GetPackageFloat(6).GetDisplayName());
      assertEquals("Density Dead For Subalpine Fir", oGridSettings.GetPackageFloat(7).GetDisplayName());
      assertEquals("Density Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(8).GetDisplayName());
      assertEquals("Density Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(9).GetDisplayName());
      assertEquals("Density Dead For Trembling Aspen", oGridSettings.GetPackageFloat(10).GetDisplayName());
      assertEquals("Density Dead For Black Cottonwood", oGridSettings.GetPackageFloat(11).GetDisplayName());
      assertEquals("Storm Severity", oGridSettings.GetPackageFloat(12).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(1);
      assertEquals("Quadrat GLI", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(2);
      assertEquals("Substrate", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("scarsoil", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("declog", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("ffmoss", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("fflitter", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Proportion of scarified soil", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Proportion of tip-up mounds", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Proportion of fresh logs", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Proportion of decayed logs", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Proportion of forest floor moss", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Proportion of forest floor litter", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfPackageInts());
      assertEquals("age", oGridSettings.GetPackageInt(0).GetCodeName());
      assertEquals("Substrate cohort age", oGridSettings.GetPackageInt(0).GetDisplayName());
      assertEquals(3, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("scarsoil", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("Substrate cohort new scarified soil substrate", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Substrate cohort new tip-up mounds substrate", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Substrate cohort new fresh logs", oGridSettings.GetPackageFloat(2).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(3);
      assertEquals("Dispersed Seeds", oGridSettings.GetName());
      oGridSettings.RemoveFloat(2);
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("seeds_0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("seeds_1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("seeds_3", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("seeds_4", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("seeds_5", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("seeds_6", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Number of seeds for Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Number of seeds for Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Number of seeds for Hybrid Spruce", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Number of seeds for Lodgepole Pine", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Number of seeds for Trembling Aspen", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Number of seeds for Black Cottonwood", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfInts());
      assertEquals("count", oGridSettings.GetInt(0).GetCodeName());
      assertEquals("Adult tree count", oGridSettings.GetInt(0).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfBools());
      assertEquals("Is Gap", oGridSettings.GetBool(0).GetCodeName());
      assertEquals("Gap status", oGridSettings.GetBool(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(4);
      assertEquals("Substrate Favorability", oGridSettings.GetName());
      oGridSettings.RemoveFloat(3);
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("Favorability Index0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("Favorability Index1", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("Favorability Index2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("Favorability Index4", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("Favorability Index5", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("Favorability Index6", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Favorability Index - Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Favorability Index - Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Favorability Index - Subalpine Fir", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Favorability Index - Lodgepole Pine", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Favorability Index - Trembling Aspen", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Favorability Index - Black Cottonwood", oGridSettings.GetFloat(5).GetDisplayName());

      /////////////////////////////
      // Remove two middle species and alphabetize the remainder
      /////////////////////////////
      p_sSpeciesList = new String[] {
          "Amabilis Fir",
          "Black Cottonwood",
          "Lodgepole Pine",
          "Trembling Aspen",
          "Western Redcedar"
      };
      oPop.SetSpeciesNames(p_sSpeciesList);
      assertEquals(5, oOutput.GetNumberOfDetailedGridSettings());
      oGridSettings = oOutput.GetDetailedGridSetting(0);
      assertEquals("Windstorm Results", oGridSettings.GetName());
      assertEquals(9, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("ba_0", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("ba_2", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("ba_3", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("ba_1", oGridSettings.GetPackageFloat(3).GetCodeName());
      assertEquals("density_4", oGridSettings.GetPackageFloat(4).GetCodeName());
      assertEquals("density_2", oGridSettings.GetPackageFloat(5).GetCodeName());
      assertEquals("density_3", oGridSettings.GetPackageFloat(6).GetCodeName());
      assertEquals("density_1", oGridSettings.GetPackageFloat(7).GetCodeName());
      assertEquals("severity", oGridSettings.GetPackageFloat(8).GetCodeName());
      assertEquals("Basal Area Dead For Amabilis Fir", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Basal Area Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Basal Area Dead For Trembling Aspen", oGridSettings.GetPackageFloat(2).GetDisplayName());
      assertEquals("Basal Area Dead For Black Cottonwood", oGridSettings.GetPackageFloat(3).GetDisplayName());
      assertEquals("Density Dead For Western Redcedar", oGridSettings.GetPackageFloat(4).GetDisplayName());
      assertEquals("Density Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(5).GetDisplayName());
      assertEquals("Density Dead For Trembling Aspen", oGridSettings.GetPackageFloat(6).GetDisplayName());
      assertEquals("Density Dead For Black Cottonwood", oGridSettings.GetPackageFloat(7).GetDisplayName());
      assertEquals("Storm Severity", oGridSettings.GetPackageFloat(8).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(1);
      assertEquals("Quadrat GLI", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(2);
      assertEquals("Substrate", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("scarsoil", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("declog", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("ffmoss", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("fflitter", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Proportion of scarified soil", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Proportion of tip-up mounds", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Proportion of fresh logs", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Proportion of decayed logs", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Proportion of forest floor moss", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Proportion of forest floor litter", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfPackageInts());
      assertEquals("age", oGridSettings.GetPackageInt(0).GetCodeName());
      assertEquals("Substrate cohort age", oGridSettings.GetPackageInt(0).GetDisplayName());
      assertEquals(3, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("scarsoil", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("Substrate cohort new scarified soil substrate", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Substrate cohort new tip-up mounds substrate", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Substrate cohort new fresh logs", oGridSettings.GetPackageFloat(2).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(3);
      assertEquals("Dispersed Seeds", oGridSettings.GetName());
      assertEquals(5, oGridSettings.GetNumberOfFloats());
      assertEquals("seeds_4", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("seeds_0", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("seeds_2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("seeds_3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("seeds_1", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("Number of seeds for Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Number of seeds for Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Number of seeds for Lodgepole Pine", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Number of seeds for Trembling Aspen", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Number of seeds for Black Cottonwood", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfInts());
      assertEquals("count", oGridSettings.GetInt(0).GetCodeName());
      assertEquals("Adult tree count", oGridSettings.GetInt(0).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfBools());
      assertEquals("Is Gap", oGridSettings.GetBool(0).GetCodeName());
      assertEquals("Gap status", oGridSettings.GetBool(0).GetDisplayName());
      oGridSettings = oOutput.GetDetailedGridSetting(4);
      assertEquals("Substrate Favorability", oGridSettings.GetName());
      assertEquals(5, oGridSettings.GetNumberOfFloats());
      assertEquals("Favorability Index4", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("Favorability Index0", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("Favorability Index2", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("Favorability Index3", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("Favorability Index1", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("Favorability Index - Western Redcedar", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Favorability Index - Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Favorability Index - Lodgepole Pine", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Favorability Index - Trembling Aspen", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Favorability Index - Black Cottonwood", oGridSettings.GetFloat(4).GetDisplayName());

      //Verify the tree settings
      assertEquals(9, oOutput.GetNumberOfDetailedLiveTreeSettings());
      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(0);
      assertEquals(4, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(1);
      assertEquals(5, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(2);
      assertEquals(6, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(3);
      assertEquals(13, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(4);
      assertEquals(14, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(5);
      assertEquals(15, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(6);
      assertEquals(16, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(7);
      assertEquals(17, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(8);
      assertEquals(18, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());


      System.out.println(
        "Normal processing succeeded for testChangeOfSpecies.");
    } catch (ModelException oErr) {
      fail("testChangeOfSpecies failed with message " + oErr.getMessage());
    } finally {
      new File(sFileName).delete();
    }
  }

  /**
   * Tests output settings when species are copied.
   */
  public void testCopySpecies() {
    GUIManager oManager = null;
    String sFileName = null;
    try {
      oManager = new GUIManager(null);
      sFileName = WriteXMLFile5();
      oManager.InputXMLParameterFile(sFileName);

      //Verify the current grid settings
      OutputBehaviors oOutput = oManager.GetOutputBehaviors();
      assertEquals(5, oOutput.GetNumberOfDetailedGridSettings());
      DetailedGridSettings oGridSettings = oOutput.GetDetailedGridSetting(0);
      assertEquals("Windstorm Results", oGridSettings.GetName());
      assertEquals(11, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("ba_0", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("ba_1", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("ba_2", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("ba_3", oGridSettings.GetPackageFloat(3).GetCodeName());
      assertEquals("density_3", oGridSettings.GetPackageFloat(4).GetCodeName());
      assertEquals("density_4", oGridSettings.GetPackageFloat(5).GetCodeName());
      assertEquals("density_5", oGridSettings.GetPackageFloat(6).GetCodeName());
      assertEquals("density_6", oGridSettings.GetPackageFloat(7).GetCodeName());
      assertEquals("density_7", oGridSettings.GetPackageFloat(8).GetCodeName());
      assertEquals("density_8", oGridSettings.GetPackageFloat(9).GetCodeName());
      assertEquals("severity", oGridSettings.GetPackageFloat(10).GetCodeName());
      assertEquals("Basal Area Dead For Western Hemlock", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Basal Area Dead For Western Redcedar", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Basal Area Dead For Amabilis Fir", oGridSettings.GetPackageFloat(2).GetDisplayName());
      assertEquals("Basal Area Dead For Subalpine Fir", oGridSettings.GetPackageFloat(3).GetDisplayName());
      assertEquals("Density Dead For Subalpine Fir", oGridSettings.GetPackageFloat(4).GetDisplayName());
      assertEquals("Density Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(5).GetDisplayName());
      assertEquals("Density Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(6).GetDisplayName());
      assertEquals("Density Dead For Trembling Aspen", oGridSettings.GetPackageFloat(7).GetDisplayName());
      assertEquals("Density Dead For Black Cottonwood", oGridSettings.GetPackageFloat(8).GetDisplayName());
      assertEquals("Density Dead For Paper Birch", oGridSettings.GetPackageFloat(9).GetDisplayName());
      assertEquals("Storm Severity", oGridSettings.GetPackageFloat(10).GetDisplayName());

      oGridSettings = oOutput.GetDetailedGridSetting(1);
      assertEquals("Quadrat GLI", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetDisplayName());

      oGridSettings = oOutput.GetDetailedGridSetting(2);
      assertEquals("Substrate", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("scarsoil", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("declog", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("ffmoss", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("fflitter", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Proportion of scarified soil", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Proportion of tip-up mounds", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Proportion of fresh logs", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Proportion of decayed logs", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Proportion of forest floor moss", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Proportion of forest floor litter", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfPackageInts());
      assertEquals("age", oGridSettings.GetPackageInt(0).GetCodeName());
      assertEquals("Substrate cohort age", oGridSettings.GetPackageInt(0).GetDisplayName());
      assertEquals(3, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("scarsoil", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("Substrate cohort new scarified soil substrate", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Substrate cohort new tip-up mounds substrate", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Substrate cohort new fresh logs", oGridSettings.GetPackageFloat(2).GetDisplayName());

      oGridSettings = oOutput.GetDetailedGridSetting(3);
      assertEquals("Dispersed Seeds", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("seeds_0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("seeds_2", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("seeds_4", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("seeds_5", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("seeds_6", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("seeds_8", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Number of seeds for Western Hemlock", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Number of seeds for Amabilis Fir", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Number of seeds for Hybrid Spruce", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Number of seeds for Lodgepole Pine", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Number of seeds for Trembling Aspen", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Number of seeds for Paper Birch", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfInts());
      assertEquals("count", oGridSettings.GetInt(0).GetCodeName());
      assertEquals("Adult tree count", oGridSettings.GetInt(0).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfBools());
      assertEquals("Is Gap", oGridSettings.GetBool(0).GetCodeName());
      assertEquals("Gap status", oGridSettings.GetBool(0).GetDisplayName());

      oGridSettings = oOutput.GetDetailedGridSetting(4);
      assertEquals("Substrate Favorability", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("Favorability Index2", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("Favorability Index - Amabilis Fir", oGridSettings.GetFloat(0).GetDisplayName());

      //Verify the current tree settings
      assertEquals(20, oOutput.GetNumberOfDetailedLiveTreeSettings());
      DetailedTreeSettings oTreeSettings = oOutput.GetDetailedLiveTreeSetting(0);
      assertEquals(1, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(1);
      assertEquals(2, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(2);
      assertEquals(3, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(3);
      assertEquals(4, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(4);
      assertEquals(5, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(5);
      assertEquals(6, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(6);
      assertEquals(7, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(7);
      assertEquals(8, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(8);
      assertEquals(9, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("DBH", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(9);
      assertEquals(10, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Height", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(10);
      assertEquals(11, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(11);
      assertEquals(12, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("Growth", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(12);
      assertEquals(13, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(13);
      assertEquals(14, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(14);
      assertEquals(15, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(15);
      assertEquals(16, oTreeSettings.GetSaveFrequency());
      assertEquals(6, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(16);
      assertEquals(17, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(17);
      assertEquals(18, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(18);
      assertEquals(19, oTreeSettings.GetSaveFrequency());
      assertEquals(8, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(19);
      assertEquals(20, oTreeSettings.GetSaveFrequency());
      assertEquals(8, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      //********************
      // Make Amabilis Fir a copy of Lodgepole Pine
      //********************
      oManager.GetTreePopulation().DoCopySpecies("Lodgepole_Pine", "Amabilis_Fir");

      assertEquals(4, oOutput.GetNumberOfDetailedGridSettings());
      oGridSettings = oOutput.GetDetailedGridSetting(0);
      assertEquals("Windstorm Results", oGridSettings.GetName());
      assertEquals(11, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("ba_0", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("ba_1", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("ba_3", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("density_3", oGridSettings.GetPackageFloat(3).GetCodeName());
      assertEquals("density_4", oGridSettings.GetPackageFloat(4).GetCodeName());
      assertEquals("density_5", oGridSettings.GetPackageFloat(5).GetCodeName());
      assertEquals("density_6", oGridSettings.GetPackageFloat(6).GetCodeName());
      assertEquals("density_7", oGridSettings.GetPackageFloat(7).GetCodeName());
      assertEquals("density_8", oGridSettings.GetPackageFloat(8).GetCodeName());
      assertEquals("severity", oGridSettings.GetPackageFloat(9).GetCodeName());
      assertEquals("density_2", oGridSettings.GetPackageFloat(10).GetCodeName());
      assertEquals("Basal Area Dead For Western Hemlock", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Basal Area Dead For Western Redcedar", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Basal Area Dead For Subalpine Fir", oGridSettings.GetPackageFloat(2).GetDisplayName());
      assertEquals("Density Dead For Subalpine Fir", oGridSettings.GetPackageFloat(3).GetDisplayName());
      assertEquals("Density Dead For Hybrid Spruce", oGridSettings.GetPackageFloat(4).GetDisplayName());
      assertEquals("Density Dead For Lodgepole Pine", oGridSettings.GetPackageFloat(5).GetDisplayName());
      assertEquals("Density Dead For Trembling Aspen", oGridSettings.GetPackageFloat(6).GetDisplayName());
      assertEquals("Density Dead For Black Cottonwood", oGridSettings.GetPackageFloat(7).GetDisplayName());
      assertEquals("Density Dead For Paper Birch", oGridSettings.GetPackageFloat(8).GetDisplayName());
      assertEquals("Storm Severity", oGridSettings.GetPackageFloat(9).GetDisplayName());
      assertEquals("Density Dead For Amabilis Fir", oGridSettings.GetPackageFloat(10).GetDisplayName());

      oGridSettings = oOutput.GetDetailedGridSetting(1);
      assertEquals("Quadrat GLI", oGridSettings.GetName());
      assertEquals(1, oGridSettings.GetNumberOfFloats());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("GLI", oGridSettings.GetFloat(0).GetDisplayName());

      oGridSettings = oOutput.GetDetailedGridSetting(2);
      assertEquals("Substrate", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("scarsoil", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("declog", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("ffmoss", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("fflitter", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Proportion of scarified soil", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Proportion of tip-up mounds", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Proportion of fresh logs", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Proportion of decayed logs", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Proportion of forest floor moss", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Proportion of forest floor litter", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfPackageInts());
      assertEquals("age", oGridSettings.GetPackageInt(0).GetCodeName());
      assertEquals("Substrate cohort age", oGridSettings.GetPackageInt(0).GetDisplayName());
      assertEquals(3, oGridSettings.GetNumberOfPackageFloats());
      assertEquals("scarsoil", oGridSettings.GetPackageFloat(0).GetCodeName());
      assertEquals("tipup", oGridSettings.GetPackageFloat(1).GetCodeName());
      assertEquals("freshlog", oGridSettings.GetPackageFloat(2).GetCodeName());
      assertEquals("Substrate cohort new scarified soil substrate", oGridSettings.GetPackageFloat(0).GetDisplayName());
      assertEquals("Substrate cohort new tip-up mounds substrate", oGridSettings.GetPackageFloat(1).GetDisplayName());
      assertEquals("Substrate cohort new fresh logs", oGridSettings.GetPackageFloat(2).GetDisplayName());

      oGridSettings = oOutput.GetDetailedGridSetting(3);
      assertEquals("Dispersed Seeds", oGridSettings.GetName());
      assertEquals(6, oGridSettings.GetNumberOfFloats());
      assertEquals("seeds_0", oGridSettings.GetFloat(0).GetCodeName());
      assertEquals("seeds_4", oGridSettings.GetFloat(1).GetCodeName());
      assertEquals("seeds_5", oGridSettings.GetFloat(2).GetCodeName());
      assertEquals("seeds_6", oGridSettings.GetFloat(3).GetCodeName());
      assertEquals("seeds_8", oGridSettings.GetFloat(4).GetCodeName());
      assertEquals("seeds_2", oGridSettings.GetFloat(5).GetCodeName());
      assertEquals("Number of seeds for Western Hemlock", oGridSettings.GetFloat(0).GetDisplayName());
      assertEquals("Number of seeds for Hybrid Spruce", oGridSettings.GetFloat(1).GetDisplayName());
      assertEquals("Number of seeds for Lodgepole Pine", oGridSettings.GetFloat(2).GetDisplayName());
      assertEquals("Number of seeds for Trembling Aspen", oGridSettings.GetFloat(3).GetDisplayName());
      assertEquals("Number of seeds for Paper Birch", oGridSettings.GetFloat(4).GetDisplayName());
      assertEquals("Number of seeds for Amabilis Fir", oGridSettings.GetFloat(5).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfInts());
      assertEquals("count", oGridSettings.GetInt(0).GetCodeName());
      assertEquals("Adult tree count", oGridSettings.GetInt(0).GetDisplayName());
      assertEquals(1, oGridSettings.GetNumberOfBools());
      assertEquals("Is Gap", oGridSettings.GetBool(0).GetCodeName());
      assertEquals("Gap status", oGridSettings.GetBool(0).GetDisplayName());

      //Verify the current tree settings
      assertEquals(23, oOutput.GetNumberOfDetailedLiveTreeSettings());
      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(0);
      assertEquals(1, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(1);
      assertEquals(2, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(2);
      assertEquals(3, oTreeSettings.GetSaveFrequency());
      assertEquals(0, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(3);
      assertEquals(4, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(4);
      assertEquals(5, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(5);
      assertEquals(6, oTreeSettings.GetSaveFrequency());
      assertEquals(1, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(6);
      assertEquals(7, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(7);
      assertEquals(8, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(8);
      assertEquals(9, oTreeSettings.GetSaveFrequency());
      assertEquals(3, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("DBH", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(9);
      assertEquals(10, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Height", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(10);
      assertEquals(11, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(11);
      assertEquals(12, oTreeSettings.GetSaveFrequency());
      assertEquals(4, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("Growth", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(12);
      assertEquals(13, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(13);
      assertEquals(14, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(14);
      assertEquals(15, oTreeSettings.GetSaveFrequency());
      assertEquals(5, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(15);
      assertEquals(16, oTreeSettings.GetSaveFrequency());
      assertEquals(6, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("Diam10", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(16);
      assertEquals(17, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Light", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(17);
      assertEquals(18, oTreeSettings.GetSaveFrequency());
      assertEquals(7, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(18);
      assertEquals(19, oTreeSettings.GetSaveFrequency());
      assertEquals(8, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("Y", oTreeSettings.GetFloat(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(19);
      assertEquals(20, oTreeSettings.GetSaveFrequency());
      assertEquals(8, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(20);
      assertEquals(13, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SEEDLING, oTreeSettings.GetType());
      assertEquals("dead", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(21);
      assertEquals(14, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.SAPLING, oTreeSettings.GetType());
      assertEquals("yls", oTreeSettings.GetInt(0).GetCodeName());

      oTreeSettings = oOutput.GetDetailedLiveTreeSetting(22);
      assertEquals(15, oTreeSettings.GetSaveFrequency());
      assertEquals(2, oTreeSettings.GetSpecies());
      assertEquals(TreePopulation.ADULT, oTreeSettings.GetType());
      assertEquals("X", oTreeSettings.GetFloat(0).GetCodeName());

      System.out.println(
          "Normal processing succeeded for testCopySpecies.");
    }
    catch (ModelException oErr) {
      fail("testChangeOfSpecies failed with message " + oErr.getMessage());
    }
    finally {
      new File(sFileName).delete();
    }
  }
  
  private void setShortOutputSettingsToFalse(OutputBehaviors oOutput) {
    TreeOutputSaveInfo oData;
    int i, j; 
    for (i = 0; i < TreePopulation.GetNumberOfTypes(); i++) {
      oData = oOutput.mp_oShortTreeSaveSettings.elementAt(i);
      oData.bSaveABA = false;
      oData.bSaveADN = false;
      oData.bSaveRBA = false;
      oData.bSaveRDN = false;
      
      for (j = 0; j > OutputBehaviors.NUMCODES; j++) {
        oOutput.mp_oShortDeadTreeSaveSettings[i][j].bSaveABA = false;
        oOutput.mp_oShortDeadTreeSaveSettings[i][j].bSaveADN = false;
        oOutput.mp_oShortDeadTreeSaveSettings[i][j].bSaveRBA = false;
        oOutput.mp_oShortDeadTreeSaveSettings[i][j].bSaveRDN = false;
      }
    }
    oOutput.UpdateIsEnabled();
  }
}
/**
 * Produces fake grids, properly registered, for our testing delight. Except
 * it's impossible to do this in conjunction with a parameter file, which is
 * what we're trying to test, so this is useless.
 * <p>Copyright: Copyright (c) Charles D. Canham 2003</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */
/*class TestBehaviorGroup extends WorkerBase {

  public TestBehaviorGroup(GUIManager oManager) throws ModelException {
    super(oManager, "", "");
    //Create grids that match the settings being entered
      Grid oNewGrid;
      String sGridName = "grid 1";
      DataMember[] p_oDataMembers = new DataMember[4];
      p_oDataMembers[0] = new DataMember("grid 1 test float 1 DISPLAY",
                                         "grid 1 test float 1",
                                         DataMember.FLOAT);
      p_oDataMembers[1] = new DataMember("grid 1 test int 1 DISPLAY",
                                         "grid 1 test int 1",
                                         DataMember.INTEGER);
      p_oDataMembers[2] = new DataMember("grid 1 test char 1 DISPLAY",
                                         "grid 1 test char 1",
                                         DataMember.CHAR);
      p_oDataMembers[3] = new DataMember("grid 1 test bool 1 DISPLAY",
                                         "grid 1 test bool 1",
                                         DataMember.BOOLEAN);
      oNewGrid = new Grid(sGridName, p_oDataMembers, null, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 2";
      p_oDataMembers = new DataMember[8];
      p_oDataMembers[0] = new DataMember("grid 2 test float 1 DISPLAY",
                                         "grid 2 test float 1",
                                         DataMember.FLOAT);
      p_oDataMembers[1] = new DataMember("grid 2 test int 1 DISPLAY",
                                         "grid 2 test int 1",
                                         DataMember.INTEGER);
      p_oDataMembers[2] = new DataMember("grid 2 test char 1 DISPLAY",
                                         "grid 2 test char 1",
                                         DataMember.CHAR);
      p_oDataMembers[3] = new DataMember("grid 2 test bool 1 DISPLAY",
                                         "grid 2 test bool 1",
                                         DataMember.BOOLEAN);
      p_oDataMembers[4] = new DataMember("grid 2 test bool 2 DISPLAY",
                                         "grid 2 test bool 2",
                                         DataMember.BOOLEAN);
      p_oDataMembers[5] = new DataMember("grid 2 test int 2 DISPLAY",
                                         "grid 2 test int 2",
                                         DataMember.BOOLEAN);
      p_oDataMembers[6] = new DataMember("grid 2 test char 2 DISPLAY",
                                         "grid 2 test char 2",
                                         DataMember.CHAR);
      p_oDataMembers[7] = new DataMember("grid 2 test float 2 DISPLAY",
                                         "grid 2 test float 2",
                                         DataMember.FLOAT);
      oNewGrid = new Grid(sGridName, p_oDataMembers, null, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 3";
      p_oDataMembers = new DataMember[1];
      p_oDataMembers[0] = new DataMember("grid 3 test char 1 DISPLAY",
                                         "grid 3 test char 1",
                                         DataMember.CHAR);
      oNewGrid = new Grid(sGridName, p_oDataMembers, null, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 4";
      p_oDataMembers = new DataMember[1];
      p_oDataMembers[0] = new DataMember("grid 4 test float 1 DISPLAY",
                                         "grid 4 test float 1",
                                         DataMember.FLOAT);
      oNewGrid = new Grid(sGridName, p_oDataMembers, null, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 5";
      p_oDataMembers = new DataMember[1];
      p_oDataMembers[0] = new DataMember("grid 5 test bool 1 DISPLAY",
                                         "grid 5 test bool 1",
                                         DataMember.BOOLEAN);
      oNewGrid = new Grid(sGridName, p_oDataMembers, null, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 6";
      p_oDataMembers = new DataMember[1];
      p_oDataMembers[0] = new DataMember("grid 6 test int 1 DISPLAY",
                                         "grid 6 test int 1",
                                         DataMember.INTEGER);
      oNewGrid = new Grid(sGridName, p_oDataMembers, null, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 7";
      p_oDataMembers = new DataMember[4];
      p_oDataMembers[0] = new DataMember("grid 7 test float 1 DISPLAY",
                                         "grid 7 test float 1",
                                         DataMember.FLOAT);
      p_oDataMembers[1] = new DataMember("grid 7 test int 1 DISPLAY",
                                         "grid 7 test int 1",
                                         DataMember.INTEGER);
      p_oDataMembers[2] = new DataMember("grid 7 test char 1 DISPLAY",
                                         "grid 7 test char 1",
                                         DataMember.CHAR);
      p_oDataMembers[3] = new DataMember("grid 7 test bool 1 DISPLAY",
                                         "grid 7 test bool 1",
                                         DataMember.BOOLEAN);
      DataMember[] p_oPackageDataMembers = new DataMember[4];
      p_oPackageDataMembers[0] = new DataMember("grid 7 test package float 1 DISPLAY",
                                         "grid 7 test package float 1",
                                         DataMember.FLOAT);
      p_oPackageDataMembers[1] = new DataMember("grid 7 test package int 1 DISPLAY",
                                         "grid 7 test package int 1",
                                         DataMember.INTEGER);
      p_oPackageDataMembers[2] = new DataMember("grid 7 test package char 1 DISPLAY",
                                         "grid 7 test char 1",
                                         DataMember.CHAR);
      p_oPackageDataMembers[3] = new DataMember("grid 7 test package bool 1 DISPLAY",
                                         "grid 7 test bool 1",
                                         DataMember.BOOLEAN);
      oNewGrid = new Grid(sGridName, p_oDataMembers, p_oPackageDataMembers, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 8";
      p_oDataMembers = new DataMember[8];
      p_oDataMembers[0] = new DataMember("grid 8 test float 1 DISPLAY",
                                         "grid 8 test float 1",
                                         DataMember.FLOAT);
      p_oDataMembers[1] = new DataMember("grid 8 test int 1 DISPLAY",
                                         "grid 8 test int 1",
                                         DataMember.INTEGER);
      p_oDataMembers[2] = new DataMember("grid 8 test char 1 DISPLAY",
                                         "grid 8 test char 1",
                                         DataMember.CHAR);
      p_oDataMembers[3] = new DataMember("grid 8 test bool 1 DISPLAY",
                                         "grid 8 test bool 1",
                                         DataMember.BOOLEAN);
      p_oDataMembers[4] = new DataMember("grid 8 test bool 2 DISPLAY",
                                         "grid 8 test bool 2",
                                         DataMember.BOOLEAN);
      p_oDataMembers[5] = new DataMember("grid 8 test int 2 DISPLAY",
                                         "grid 8 test int 2",
                                         DataMember.BOOLEAN);
      p_oDataMembers[6] = new DataMember("grid 8 test char 2 DISPLAY",
                                         "grid 8 test char 2",
                                         DataMember.CHAR);
      p_oDataMembers[7] = new DataMember("grid 8 test float 2 DISPLAY",
                                         "grid 8 test float 2",
                                         DataMember.FLOAT);
      p_oPackageDataMembers = new DataMember[14];
      p_oPackageDataMembers[0] = new DataMember("grid 8 test package float 1 DISPLAY",
                                         "grid 8 test package float 1",
                                         DataMember.FLOAT);
      p_oPackageDataMembers[1] = new DataMember("grid 8 test package int 1 DISPLAY",
                                         "grid 8 test package int 1",
                                         DataMember.INTEGER);
      p_oPackageDataMembers[2] = new DataMember("grid 8 test package char 1 DISPLAY",
                                         "grid 8 test package char 1",
                                         DataMember.CHAR);
      p_oPackageDataMembers[3] = new DataMember("grid 8 test package bool 1 DISPLAY",
                                         "grid 8 test package bool 1",
                                         DataMember.BOOLEAN);
      p_oPackageDataMembers[4] = new DataMember("grid 8 test package bool 2 DISPLAY",
                                         "grid 8 test package bool 2",
                                         DataMember.BOOLEAN);
      p_oPackageDataMembers[5] = new DataMember("grid 8 test package int 2 DISPLAY",
                                         "grid 8 test package int 2",
                                         DataMember.BOOLEAN);
      p_oPackageDataMembers[6] = new DataMember("grid 8 test package char 2 DISPLAY",
                                         "grid 8 test package char 2",
                                         DataMember.CHAR);
      p_oPackageDataMembers[7] = new DataMember("grid 8 test package float 2 DISPLAY",
                                         "grid 8 test package float 2",
                                         DataMember.FLOAT);
      p_oPackageDataMembers[8] = new DataMember("grid 8 test package float 3 DISPLAY",
                                          "grid 8 test package float 3",
                                          DataMember.FLOAT);
      p_oPackageDataMembers[9] = new DataMember("grid 8 test package char 3 DISPLAY",
                                          "grid 8 test package char 3",
                                          DataMember.CHAR);
      p_oPackageDataMembers[10] = new DataMember("grid 8 test package char 4 DISPLAY",
                                          "grid 8 test package char 4",
                                          DataMember.CHAR);
      p_oPackageDataMembers[11] = new DataMember("grid 8 test package bool 3 DISPLAY",
                                          "grid 8 test package bool 3",
                                          DataMember.BOOLEAN);
      p_oPackageDataMembers[12] = new DataMember("grid 8 test package bool 4 DISPLAY",
                                          "grid 8 test package bool 4",
                                          DataMember.BOOLEAN);
      p_oPackageDataMembers[13] = new DataMember("grid 8 test package bool 5 DISPLAY",
                                          "grid 8 test package bool 5",
                                          DataMember.BOOLEAN);
      oNewGrid = new Grid(sGridName, p_oDataMembers, p_oPackageDataMembers, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 9";
      p_oPackageDataMembers = new DataMember[4];
      p_oPackageDataMembers[0] = new DataMember("grid 9 test package float 1 DISPLAY",
                                         "grid 9 test package float 1",
                                         DataMember.FLOAT);
      p_oPackageDataMembers[1] = new DataMember("grid 9 test package int 1 DISPLAY",
                                         "grid 9 test package int 1",
                                         DataMember.INTEGER);
      p_oPackageDataMembers[2] = new DataMember("grid 9 test package char 1 DISPLAY",
                                         "grid 9 test char 1",
                                         DataMember.CHAR);
      p_oPackageDataMembers[3] = new DataMember("grid 9 test package bool 1 DISPLAY",
                                         "grid 9 test bool 1",
                                         DataMember.BOOLEAN);
      oNewGrid = new Grid(sGridName, null, p_oPackageDataMembers, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);


      sGridName = "grid 10";
      p_oPackageDataMembers = new DataMember[14];
      p_oPackageDataMembers[0] = new DataMember("grid 10 test package float 1 DISPLAY",
                                         "grid 10 test package float 1",
                                         DataMember.FLOAT);
      p_oPackageDataMembers[1] = new DataMember("grid 10 test package int 1 DISPLAY",
                                         "grid 10 test package int 1",
                                         DataMember.INTEGER);
      p_oPackageDataMembers[2] = new DataMember("grid 10 test package char 1 DISPLAY",
                                         "grid 10 test package char 1",
                                         DataMember.CHAR);
      p_oPackageDataMembers[3] = new DataMember("grid 10 test package bool 1 DISPLAY",
                                         "grid 10 test package bool 1",
                                         DataMember.BOOLEAN);
      p_oPackageDataMembers[4] = new DataMember("grid 10 test package bool 2 DISPLAY",
                                         "grid 10 test package bool 2",
                                         DataMember.BOOLEAN);
      p_oPackageDataMembers[5] = new DataMember("grid 10 test package int 2 DISPLAY",
                                         "grid 10 test package int 2",
                                         DataMember.BOOLEAN);
      p_oPackageDataMembers[6] = new DataMember("grid 10 test package char 2 DISPLAY",
                                         "grid 10 test package char 2",
                                         DataMember.CHAR);
      p_oPackageDataMembers[7] = new DataMember("grid 10 test package float 2 DISPLAY",
                                         "grid 10 test package float 2",
                                         DataMember.FLOAT);
      p_oPackageDataMembers[8] = new DataMember("grid 10 test package float 3 DISPLAY",
                                          "grid 10 test package float 3",
                                          DataMember.FLOAT);
      p_oPackageDataMembers[9] = new DataMember("grid 10 test package char 3 DISPLAY",
                                          "grid 10 test package char 3",
                                          DataMember.CHAR);
      p_oPackageDataMembers[10] = new DataMember("grid 10 test package char 4 DISPLAY",
                                          "grid 10 test package char 4",
                                          DataMember.CHAR);
      p_oPackageDataMembers[11] = new DataMember("grid 10 test package bool 3 DISPLAY",
                                          "grid 10 test package bool 3",
                                          DataMember.BOOLEAN);
      p_oPackageDataMembers[12] = new DataMember("grid 10 test package bool 4 DISPLAY",
                                          "grid 10 test package bool 4",
                                          DataMember.BOOLEAN);
      p_oPackageDataMembers[13] = new DataMember("grid 10 test package bool 5 DISPLAY",
                                          "grid 10 test package bool 5",
                                          DataMember.BOOLEAN);
      oNewGrid = new Grid(sGridName, null, p_oPackageDataMembers, 2, 2);
      oNewGrid = oManager.AddGrid(oNewGrid);
  }

  public void ValidateData(TreePopulation oPop){;}
  public void DoSetup(TreePopulation oPop) {;}
} */
